package org.gvsig.fmap.dal;

import java.io.File;
import org.apache.commons.lang3.StringUtils;
import org.gvsig.expressionevaluator.SymbolTable;
import org.gvsig.fmap.dal.feature.FeatureStore;
import org.gvsig.tools.resourcesstorage.ResourcesStorage;
import org.gvsig.tools.util.LabeledValue;

/**
 *
 * @author jjdelcerro
 */
public interface DatabaseWorkspaceManager 
        extends LabeledValue<DatabaseWorkspaceManager>, SymbolTable 
    {
    
    public interface AddDatabaseWorkspaceListener {
        
        public String getName();
        
        public void onAddDatabaseWorkspace(DatabaseWorkspaceManager databaseWorkspace);
    }
    
    
    
    public static final String CONFIG_CAN_ANONYMOUS_USER_WRITE_IN_THE_TABLES = "CanAnonymousUserWriteInTheTables";

    public static final String TABLE_REPOSITORY_NAME = "gvsigd_repository";
    public static final String TABLE_RESOURCES_NAME = "gvsigd_resources";
    public static final String TABLE_CONFIGURATION_NAME = "gvsigd_config";
    
    public static final String FIELD_RESOURCES_NAME = "name";
    public static final String FIELD_RESOURCES_RESOURCE = "resource";

    public static final String FIELD_REPOSITORY_NAME = "name";
    public static final String FIELD_REPOSITORY_PARAMETERS = "parameters";

    public static final String FIELD_CONFIGURATION_NAME = "name";
    public static final String FIELD_CONFIGURATION_VALUE = "value";
    
    public static final int TABLE_RESOURCES = 0;
    public static final int TABLE_REPOSITORY = 1;
    public static final int TABLE_CONFIGURATION = 2;
    
    public static final String CONFIG_NAME_STORESREPOSITORYID = "StoresRepository.id";
    public static final String CONFIG_NAME_STORESREPOSITORYLABEL = "StoresRepository.label";

    /**
     * Check if the indicated name corresponds to one of the configuration tables of the workspace.
     * 
     * @param name to check.
     * @return true if the name is that of a configuration table.
     */
    public static boolean isInternalTable(String name) {
        if( StringUtils.isBlank(name) ) {
            return false;
        }
        String[] internalNames = new String[] {
            TABLE_REPOSITORY_NAME,
            TABLE_RESOURCES_NAME,
            TABLE_CONFIGURATION_NAME
        };
        for (String internalName : internalNames) {
            if( name.equalsIgnoreCase(internalName) ) {
                return true;
            }
        }
        return false;
    }

    /**
     * Returns the identifier of this workspace.
     * 
     * @return the id.
     */
    public String getId();

    /**
     * Returns the label of this workspace.
     * 
     * @return the label value.
     */
    public String getLabel();
    
    /**
     * Gets the value of a configuration variable associated with 
     * this work space.
     * 
     * @param name of the variable to consult 
     * @return the value of the variable 
     */
    public String get(String name);
    
    /**
     * Assigns the indicated value to the configuration variable.
     * 
     * @param name of the variable.
     * @param value value to set.
     * @return true if can assign the value.
     */
    public boolean set(String name, String value);
    
    /**
     * Gets the repository of data stores associated with this workspace.
     * 
     * @return the data stores repository.
     */
    public StoresRepository getStoresRepository();
    
    /**
     * Add a new data store to the workspace.
     * 
     * @param name of the data store.
     * @param parameters to open the data store.
     * @return true if ok.
     */
    public boolean writeStoresRepositoryEntry(String name, DataStoreParameters parameters);

    public boolean canAnonymousUserWriteInTheTables();

    /**
     * Check if the indicated configuration table exists in the workspace.
     * 
     * @param tableid
     * @return true if the table exists.
     */
    public boolean existsTable(int tableid);

    /**
     * Create the configuration table indicated in the workspace.
     * 
     * @param tableid identifier of the configuration table to create.
     */
    public void createTable(int tableid);
    
    /**
     * Remove the indicated configuration table from the workspace.
     * 
     * @param tableid identifier of the configuration table to remove.
     */
    public void dropTable(int tableid);

    /**
     * Gets the data store associated with the indicated configuration table.
     * 
     * @param tableid identifier of the configuration table to get.
     * @return the FeatureStore of the configuration table.
     */
    public FeatureStore getTable(int tableid);

    /**
     * Returns true if the connection associated with this object refers 
     * to a valid workspace.
     * At least the variable "StoresRepository.id" must be defined.
     * 
     * @return 
     */
    public boolean isValid();

    /**
     * Returns true if the connection associated with this object refers 
     * to a valid workspace with a stores-repository.
     * At least the repositories table must exist and the variable 
     * "StoresRepository.id" must be defined.
     * 
     * @return 
     */
    public boolean isValidStoresRepository();
    
    /**
     * If the workspace has an alternate resource storage defined, return it. 
     * If don't have it, return null.
     * 
     * @param tableName
     * @return the alternate resource storage.
     */
    public ResourcesStorage getAlternativeResourcesStorage(String tableName);
    
    /**
     * Create and initialize the tables associated with a gvSIG workspace.
     * 
     * @param id of the workspace
     * @param description of the workspace 
     */
    public void create(String id, String description);

    public void drop();
    
    public File getBaseFolder();
    
    public void setBaseFolder(File baseFolder);
    
    public ResourcesStorage getResourcesStorage();
        
}
