/* gvSIG. Geographic Information System of the Valencian Government
 *
 * Copyright (C) 2007-2016 Infrastructures and Transports Department
 * of the Valencian Government (CIT)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 *
 */
package org.gvsig.fmap.dal.store.h2;

import java.sql.Connection;
import java.sql.SQLException;
import java.text.MessageFormat;
import org.apache.commons.dbcp.BasicDataSource;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.lang3.StringUtils;
import org.gvsig.expressionevaluator.ExpressionBuilder.GeometrySupportType;
import org.gvsig.fmap.dal.exception.InitializeException;
import org.gvsig.fmap.dal.resource.exception.AccessResourceException;
import org.gvsig.fmap.dal.spi.DataServerExplorerProviderServices;
import org.gvsig.fmap.dal.store.h2.operations.H2SpatialOperationsFactory;
import org.gvsig.fmap.dal.store.jdbc.JDBCConnectionParameters;
import org.gvsig.fmap.dal.store.jdbc.JDBCNewStoreParameters;
import org.gvsig.fmap.dal.store.jdbc.JDBCServerExplorerParameters;
import org.gvsig.fmap.dal.store.jdbc.JDBCStoreParameters;
import org.gvsig.fmap.dal.store.jdbc.exception.JDBCDriverClassNotFoundException;
import org.gvsig.fmap.dal.store.jdbc2.JDBCServerExplorer;
import org.gvsig.fmap.dal.store.jdbc2.OperationsFactory;
import org.gvsig.fmap.dal.store.jdbc2.spi.JDBCHelperBase;
import org.gvsig.fmap.dal.store.jdbc2.spi.JDBCSQLBuilderBase;
import org.gvsig.fmap.dal.store.jdbc2.spi.JDBCServerExplorerBase;
import org.gvsig.fmap.dal.store.jdbc2.spi.SRSSolverBase;
import org.h2.tools.Server;
import org.h2gis.ext.H2GISExtension;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

@SuppressWarnings("UseSpecificCatch")
public class H2SpatialHelper extends JDBCHelperBase {

    static final Logger LOGGER = LoggerFactory.getLogger(H2SpatialHelper.class);

    public static final String H2SPATIAL_JDBC_DRIVER = "org.h2.Driver";
    
    public static String getConnectionURL(H2SpatialConnectionParameters params) {
        String connectionURL;
        String dbfilename = params.getFile().getAbsolutePath().replace("\\","/");
        if( dbfilename!=null && dbfilename.endsWith(".mv.db") ) {
            dbfilename = dbfilename.substring(0, dbfilename.length()-6);
        }
        StringBuilder commonParameters = new StringBuilder();
        commonParameters.append(";MODE=PostgreSQL");
        commonParameters.append(";SCHEMA=PUBLIC");
        commonParameters.append(";ALLOW_LITERALS=ALL");
        
//        Integer LOCK_TIMEOUT = (Integer) params.getDynValue("LOCK_TIMEOUT");
//        if( LOCK_TIMEOUT!=null ) {
//            commonParameters.append(";LOCK_TIMEOUT=").append(LOCK_TIMEOUT);
//        }
//        Integer MULTI_THREADED = (Integer) params.getDynValue("MULTI_THREADED");
//        if( MULTI_THREADED!=null ) {
//            commonParameters.append(";MULTI_THREADED=").append(MULTI_THREADED);
//        }
//        Integer CHACHE_SIZE = (Integer) params.getDynValue("CHACHE_SIZE");
//        if( LOCK_TIMEOUT!=null ) {
//            commonParameters.append(";CHACHE_SIZE=").append(CHACHE_SIZE);
//        }
//        Integer LOG = (Integer) params.getDynValue("LOG");
//        if( LOCK_TIMEOUT!=null ) {
//            commonParameters.append(";LOG=").append(LOG);
//        }
//        Integer LOCK_MODE = (Integer) params.getDynValue("LOCK_MODE");
//        if( LOCK_TIMEOUT!=null ) {
//            commonParameters.append(";LOCK_MODE=").append(LOCK_MODE);
//        }
        
        if( StringUtils.isEmpty(params.getHost()) ) {
            // Asumimos que es una conexion directa sobre el filesystem
            if( StringUtils.equalsIgnoreCase(FilenameUtils.getExtension(params.getFile().getName()),"zip") ) {
                connectionURL =  MessageFormat.format(
                    "jdbc:h2:zip:{0}!/{1}"+commonParameters.toString(),
                    dbfilename,
                    params.getDBName()
                );
            } else {
                connectionURL =  MessageFormat.format(
                    "jdbc:h2:file:{0}"+commonParameters.toString(),
                    dbfilename
                );
            }
        } else if( params.getPort() == null ) {
            connectionURL =  MessageFormat.format(
                "jdbc:h2:tcp://{0}/{1}"+commonParameters.toString(),
                params.getHost(),
                dbfilename
            );            
        } else {
            connectionURL =  MessageFormat.format(                
                "jdbc:h2:tcp://{0}:{1,number,#######}/{2}"+commonParameters.toString(),
                params.getHost(),
                params.getPort().intValue(),
                dbfilename
            );
        }
        LOGGER.debug("connectionURL: {}", connectionURL);
        return connectionURL;
    }

    private static class ConnectionProvider {

        private static boolean needRegisterDriver = true;

        private BasicDataSource dataSource = null;

        private final H2SpatialConnectionParameters connectionParameters;

        private static Server server = null;
        private static boolean startServer = true;

        public ConnectionProvider(H2SpatialConnectionParameters connectionParameters) {
            this.connectionParameters = connectionParameters;
        }
        
        private void startServer() {
        
            if( startServer && server == null ) {
                String port = "9123";
                try {
                    Server theServer;
                    String s = System.getProperty("H2Port");
                    if( s!=null ) {
                        try {
                            int n = Integer.parseInt(s);
                            port = String.valueOf(n);
                        } catch(Throwable th) {
                            // Ignore port number, use default.
                        }
                    }
                    if( System.getProperty("H2AllowOthers")!=null ) {
                        theServer = Server.createTcpServer("-tcpPort", port, "-tcpAllowOthers", "-ifExists");
                    } else {
                        theServer = Server.createTcpServer("-tcpPort", port, "-ifExists");
                    }
                    theServer.start();
                    server = theServer;
                    LOGGER.info("H2 Server started" );
                    LOGGER.info("  port  :"+ server.getPort());
                    LOGGER.info("  URL  :"+ server.getURL());
                    LOGGER.info("  status:"+ server.getStatus());
                } catch (SQLException ex) {
                    LOGGER.warn("H2 Server not started",ex);
                }
                // Tanto si consigue lanzar el server como si no, no lo vuelve a intentar
                startServer = false;
            }

        }

        @Override
        public String toString() {
            StringBuilder builder = new StringBuilder();
            builder.append(" url=").append(connectionParameters.getUrl());
            builder.append(" driver name=").append(connectionParameters.getJDBCDriverClassName());
            builder.append(" user=").append(connectionParameters.getUser());
            return builder.toString();
        }
        
        public Connection getConnection() throws SQLException {
            if (this.dataSource == null) {
                this.dataSource = this.createDataSource();               
            }
            Connection conn = this.dataSource.getConnection();
            try {
                conn.createStatement().execute("SELECT TOP 1 SRID FROM SPATIAL_REF_SYS");
            } catch(SQLException ex) {
                H2GISExtension.load(conn);
            }
            try {
                conn.createStatement().execute("CREATE SCHEMA IF NOT EXISTS PUBLIC;SET SCHEMA PUBLIC");
            } catch(SQLException ex) {
                // Ignore this error.
            }

            return conn;
        }

        private BasicDataSource createDataSource() throws SQLException {
            if (!this.isRegistered()) {
                this.registerDriver();
            }
            startServer();
            H2SpatialConnectionParameters params = connectionParameters;

            BasicDataSource ds = new BasicDataSource();
            ds.setDriverClassName(params.getJDBCDriverClassName());
            if( !StringUtils.isEmpty(params.getUser()) ) {
                ds.setUsername(params.getUser());
            }
            if( !StringUtils.isEmpty(params.getPassword()) ) {
                ds.setPassword(params.getPassword());
            }
            ds.setUrl(params.getUrl());

            ds.setMaxWait(60L * 1000);
            return ds;
        }

        private boolean isRegistered() {
            return needRegisterDriver;
        }

        public void registerDriver() throws SQLException {
            String className = this.connectionParameters.getJDBCDriverClassName();
            if (className == null) {
                return;
            }
            try {
                Class theClass = Class.forName(className);
                if (theClass == null) {
                    throw new JDBCDriverClassNotFoundException(H2SpatialLibrary.NAME, className);
                }
            } catch (Exception e) {
                throw new SQLException("Can't register JDBC driver '" + className + "'.", e);
            }
            needRegisterDriver = false;
        }

    }

    private ConnectionProvider connectionProvider = null;
   
    public H2SpatialHelper(JDBCConnectionParameters connectionParameters) {
        super(connectionParameters);
        this.srssolver = new SRSSolverBase(this);
    }

    @Override
    public synchronized Connection  getConnection() throws AccessResourceException {
        try {
            if (this.connectionProvider == null) {
                this.connectionProvider = new ConnectionProvider(this.getConnectionParameters());
            }
            Connection connection = this.connectionProvider.getConnection();
            if( LOGGER.isDebugEnabled() ) {
                LOGGER.debug("getConnection: connection = "+connection.hashCode()+ connectionProvider.toString());
            }
            return connection;
        } catch (SQLException ex) {
            throw new AccessResourceException(H2SpatialLibrary.NAME, ex);
        }
    }

    @Override
    public void closeConnection(Connection connection) {
        LOGGER.debug("closeConnection: connection = "+connection.hashCode());
        super.closeConnection(connection);
    }
    
    @Override
    public H2SpatialConnectionParameters getConnectionParameters() {
        return (H2SpatialConnectionParameters) super.getConnectionParameters();
    }
    
    @Override
    public String getConnectionURL() {
        return getConnectionURL(this.getConnectionParameters());
    }

    @Override
    protected String getResourceType() {
        return H2SpatialLibrary.NAME;
    }

    @Override
    public String getProviderName() {
        return H2SpatialLibrary.NAME;
    }

    @Override
    public JDBCSQLBuilderBase createSQLBuilder() {
        return new H2SpatialSQLBuilder(this);
    }
    
    @Override
    public OperationsFactory getOperations() {
        if (this.operationsFactory == null) {
            this.operationsFactory = new H2SpatialOperationsFactory(this);
        }
        return operationsFactory;
    }

    @Override
    public GeometrySupportType getGeometrySupportType() {
        return GeometrySupportType.WKB;
    }

    @Override
    public boolean hasSpatialFunctions() {
        return true;
    }

    @Override
    public boolean canWriteGeometry(int geometryType, int geometrySubtype) {
        return true;
    }

    @Override
    public String getQuoteForIdentifiers() {
        return "\"";
    }

    @Override
    public boolean allowAutomaticValues() {
        return true;
    }

    @Override
    public boolean supportOffsetInSelect() {
        return true;
    }

    @Override
    public String getQuoteForStrings() {
        return "'";
    }

    @Override
    public String getSourceId(JDBCStoreParameters parameters) {
        return parameters.getDBName() + "." + 
               parameters.getSchema()+ "." + 
               parameters.getTable();
    }

    @Override
    public JDBCNewStoreParameters createNewStoreParameters() {
        return new H2SpatialNewStoreParameters();
    }

    @Override
    public JDBCStoreParameters createOpenStoreParameters() {
        return new H2SpatialStoreParameters();
    }

    @Override
    public JDBCServerExplorerParameters createServerExplorerParameters() {
        return new H2SpatialExplorerParameters();
    }

    @Override
    public JDBCServerExplorer createServerExplorer(
            JDBCServerExplorerParameters parameters, 
            DataServerExplorerProviderServices providerServices
        ) throws InitializeException {
        
        JDBCServerExplorer explorer = new JDBCServerExplorerBase(
                parameters, 
                providerServices, 
                this
        );
        this.initialize(explorer, parameters, null);
        return explorer;
    }
}
