package org.gvsig.fmap.dal.store.h2;

import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Objects;
import org.apache.commons.lang3.tuple.Pair;
import org.gvsig.expressionevaluator.ExpressionBuilder.Parameter;
import org.gvsig.expressionevaluator.Formatter;
import org.gvsig.fmap.dal.DataTypes;
import org.gvsig.fmap.dal.feature.FeatureAttributeDescriptor;
import org.gvsig.fmap.dal.feature.FeatureType;
import org.gvsig.fmap.dal.feature.spi.FeatureProvider;
import org.gvsig.fmap.dal.store.h2.expressionbuilderformatter.H2SpatialFormatter;
import org.gvsig.fmap.dal.store.jdbc2.spi.JDBCSQLBuilderBase;
import org.gvsig.fmap.geom.Geometry;
import org.gvsig.fmap.geom.exception.CreateGeometryException;
import org.gvsig.tools.dispose.Disposable;

public class H2SpatialSQLBuilder extends JDBCSQLBuilderBase {

    protected Formatter formatter = null;
    
    public H2SpatialSQLBuilder(H2SpatialHelper helper) {
        super(helper);
        
        //
        // H2/H2GIS SQL functions reference list
        //
        // http://www.h2database.com/html/functions.html
        // http://www.h2gis.org/docs/1.3/functions/
        //
        // http://www.h2database.com/html/grammar.html
        //
        // http://www.h2database.com/html/datatypes.html
        //
        //

        
        this.defaultSchema = "PUBLIC";
        this.supportSchemas = true;
        this.allowAutomaticValues = true;
        this.geometrySupportType = this.helper.getGeometrySupportType();
        this.hasSpatialFunctions = this.helper.hasSpatialFunctions();

        this.STMT_DELETE_GEOMETRY_COLUMN_FROM_TABLE_schema_table = null;
        this.STMT_DELETE_GEOMETRY_COLUMN_FROM_TABLE_table = null;
         
        this.STMT_UPDATE_TABLE_STATISTICS_table = "ANALYZE SAMPLE_SIZE 0";

//        config.set(SQLConfig.ST_AsBinary, "NVL2({0}, ST_AsBinary({0}), null)");

        this.type_boolean = "BOOLEAN";
        this.type_byte = "TINYINT";
        this.type_bytearray = "BLOB";
        this.type_geometry = "GEOMETRY";
        this.type_char = "CHAR";
        this.type_date = "DATE";
        this.type_double = "DOUBLE"; 
        this.type_numeric_p = "DECIMAL({0,Number,#######})";
        this.type_numeric_ps = "DECIMAL({0,Number,#######},{1,Number,#######})";
        this.type_bigdecimal = "DOUBLE";
        this.type_float = "REAL";
        this.type_int = "INTEGER";
        this.type_long = "BIGINT";
        this.type_string = "VARCHAR";
        this.type_string_p = "VARCHAR({0,Number,#######})";
        this.type_time = "TIME";
        this.type_timestamp = "TIMESTAMP";
        this.type_version = "VARCHAR";
        this.type_URI = "VARCHAR";
        this.type_URL = "VARCHAR";
        this.type_FILE = "VARCHAR";
        this.type_FOLDER = "VARCHAR";        
    }

    @Override
    protected Formatter formatter() {
        if( this.formatter==null ) {
            this.formatter = new H2SpatialFormatter(this);
        }
        return this.formatter;
    }

    public class H2SpatialTableNameBuilderBase extends TableNameBuilderBase {

        @Override
        public boolean has_database() {
            return false;
        }
        
    }

    public class H2SpatialCreateIndexBuilder extends CreateIndexBuilderBase {
    
        @Override
        public List<String> toStrings(Formatter formatter) {
            StringBuilder builder = new StringBuilder();
            builder.append("CREATE ");
            if( this.isUnique ) {
                builder.append("UNIQUE ");
            }
            if( this.isSpatial ) {
                builder.append("SPATIAL ");
            }
            builder.append("INDEX ");
            if( this.ifNotExist ) {
                builder.append("IF NOT EXISTS ");
            }
            builder.append(as_identifier(this.indexName));
            builder.append(" ON ");
            builder.append(this.table.toString(formatter));
            builder.append(" ( ");
            boolean is_first_column = true;
            for( String column : this.columns) {
                if( is_first_column ) {
                    is_first_column = false;
                } else {
                    builder.append(", ");
                }
                builder.append(as_identifier(column));
            }
            builder.append(" )");
            
            List<String> sqls = new ArrayList<>();
            sqls.add(builder.toString());
            return sqls;
        }

    }
        
    protected class H2SpatialAlterTableBuilderBase extends AlterTableBuilderBase {
        @Override
        public List<String> toStrings(Formatter formatter) {
            List<String> sqls = new ArrayList<>();
            if( this.isEmpty() ) {
                return sqls;
            }
            for (String column : drops) {
                StringBuilder builder = new StringBuilder();
                builder.append("ALTER TABLE ");
                builder.append(this.table.toString(formatter));
                builder.append(" DROP COLUMN IF EXISTS ");
                builder.append(as_identifier(column)); 
                sqls.add(builder.toString());
            }
            for (ColumnDescriptor column : adds) {
                StringBuilder builder = new StringBuilder();
                builder.append("ALTER TABLE ");
                builder.append(this.table.toString(formatter));
                builder.append(" ADD COLUMN ");
                builder.append(as_identifier(column.getName())); 
                builder.append(" ");
                if( column.getType() == DataTypes.INT && column.isAutomatic() ) {
                    builder.append(" SERIAL");
                } else {
                    builder.append(
                        sqltype(
                            column.getType(), 
                            column.getPrecision(), 
                            column.getSize(),
                            column.getGeometryType(), 
                            column.getGeometrySubtype()
                        )
                    );
                }
                if (column.getDefaultValue() == null) {
                    if (column.allowNulls()) {
                        builder.append(" DEFAULT NULL");
                    }
                } else {
                    builder.append(" DEFAULT '");
                    builder.append(Objects.toString(column.getDefaultValue(),""));
                    builder.append("'");
                }
                if (column.allowNulls()) {
                    builder.append(" NULL");
                } else {
                    builder.append(" NOT NULL");
                }
                if (column.isPrimaryKey()) {
                    builder.append(" PRIMARY KEY");
                }
                sqls.add(builder.toString());
                
                if( column.isGeometry() ) {
                    String sql;
                    String constraint_name = "constraint_" + this.table().getName() + "_" + column.getName()+"_dim";
                    sql = MessageFormat.format(
                        "ALTER TABLE \"{0}\".\"{1}\" ADD CONSTRAINT IF NOT EXISTS \"{2}\" CHECK ST_CoordDim(\"{3}\") = {4}",
                        this.table().getSchema(),
                        this.table().getName(),
                        constraint_name,
                        column.getName(),
                        sqlgeometrynumdimension(column.getGeometryType(),column.getGeometrySubtype())
                    );
                    sqls.add(sql);
                }
            }
            for (ColumnDescriptor column : alters) {
                StringBuilder builder = new StringBuilder();
                builder.append("ALTER TABLE ");
                builder.append(this.table.toString(formatter));
                builder.append(" ALTER COLUMN ");
                builder.append(as_identifier(column.getName())); 
                builder.append(" ");
                builder.append(
                    sqltype(
                        column.getType(), 
                        column.getPrecision(), 
                        column.getSize(),
                        column.getGeometryType(), 
                        column.getGeometrySubtype()
                    )
                );
                if (column.getDefaultValue() == null) {
                    if (column.allowNulls()) {
                        builder.append(" DEFAULT NULL");
                    }
                } else {
                    builder.append(" DEFAULT '");
                    builder.append(column.getDefaultValue().toString());
                    builder.append("'");
                }
                if( column.isAutomatic() ) {
                    builder.append(" AUTO_INCREMENT");
                }
                sqls.add(builder.toString());
                if( column.isGeometry() ) {
                    String sql;
                    String constraint_name = "constraint_" + this.table().getName() + "_" + column.getName()+"_dim";
                    sql = MessageFormat.format(
                        "ALTER TABLE \"{0}\".\"{1}\" ADD CONSTRAINT IF NOT EXISTS \"{2}\" CHECK ST_CoordDim(\"{3}\") = {4}",
                        this.table().getSchema(),
                        this.table().getName(),
                        constraint_name,
                        column.getName(),
                        sqlgeometrynumdimension(column.getGeometryType(),column.getGeometrySubtype())
                    );
                    sqls.add(sql);
                }
            }
            for (Pair<String,String> pair : renames) {
                StringBuilder builder = new StringBuilder();
                builder.append("ALTER TABLE ");
                builder.append(this.table.toString(formatter));
                builder.append(" RENAME COLUMN ");
                builder.append(as_identifier(pair.getLeft())); 
                builder.append(" TO ");
                builder.append(as_identifier(pair.getRight())); 
                sqls.add(builder.toString());
            }
            return sqls;
        }

    }
    
    protected class H2SpatialCreateTableBuilder extends CreateTableBuilderBase {

        @Override
        public List<String> toStrings(Formatter formatter) {

            List<String> sqls = new ArrayList<>();
            StringBuilder builder = new StringBuilder();

            builder.append("CREATE TABLE ");
            builder.append(this.table.toString(formatter));
            builder.append(" (");
            boolean first = true;
            for (ColumnDescriptor column : columns) {

                if (first) {
                    first = false;
                } else {
                    builder.append(", ");
                }
                builder.append(as_identifier(column.getName()));
                builder.append(" ");
                builder.append(
                    sqltype(
                        column.getType(), 
                        column.getPrecision(), 
                        column.getSize(),
                        column.getGeometryType(), 
                        column.getGeometrySubtype()
                    )
                );
                if( !column.isGeometry() ) {
                    if (column.isPrimaryKey()) {
                        builder.append(" PRIMARY KEY");
                        if( column.isAutomatic() ) {
                            builder.append(" AUTO_INCREMENT");
                        }
                    } else {
                        if( column.isAutomatic() ) {
                            builder.append(" AUTO_INCREMENT");
                        }
                        if (column.getDefaultValue() == null) {
                            if (column.allowNulls()) {
                                builder.append(" DEFAULT NULL");
                            }
                        } else {
                            if( column.getType() == DataTypes.DATE ) {
                                builder.append(" DEFAULT ( TIMESTAMP '");
                                Date d = (Date) column.getDefaultValue();
                                builder.append(MessageFormat.format( "{0,date,yyyy-MM-dd HH:mm:ss.S}",d));
                                builder.append("' )");
                            } else {
                                builder.append(" DEFAULT '");
                                builder.append(Objects.toString(column.getDefaultValue(),""));
                                builder.append("'");
                            }
                        }
                    }
                }
                if (!column.allowNulls()) {
                    builder.append(" NOT NULL");
                }
            }
            builder.append(" )");
            sqls.add(builder.toString());
            for (ColumnDescriptor column : columns) {
                if( column.isGeometry() ) {
                    String sql;
                    String constraint_name = "constraint_" + this.table().getName() + "_" + column.getName()+"_dim";
                    sql = MessageFormat.format(
                        "ALTER TABLE {0} ADD CONSTRAINT IF NOT EXISTS \"{1}\" CHECK NVL2(\"{2}\", ST_CoordDim(\"{2}\") = {3}, TRUE)",
                        this.table().toString(),
                        constraint_name,
                        column.getName(),
                        sqlgeometrynumdimension(column.getGeometryType(),column.getGeometrySubtype())
                    );
                    if( column.getGeometrySRSId()!=null ) {
                        String sql2;
                        String constraint_name2 = "constraint_" + this.table().getName() + "_" + column.getName()+"_srid";
                        sql2 = MessageFormat.format(
                            "ALTER TABLE {0} ADD CONSTRAINT IF NOT EXISTS \"{1}\" CHECK NVL2(\"{2}\", ST_SRID(\"{2}\") = {3,number,#####}, TRUE)",
                            this.table().toString(),
                            constraint_name2, 
                            column.getName(), 
                            column.getGeometrySRSId() 
                        );
                        sqls.add(sql2);
                    }
                    sqls.add(sql);
                }
            }            
            
            return sqls;
        }
    }

    public class H2SpatialSelectBuilderBase extends SelectBuilderBase {
        
        @Override
        protected boolean isValid(StringBuilder message) {
            if( message == null ) {
                message = new StringBuilder();
            }
            if( this.has_offset() && !this.has_order_by() ) {
                // Algunos gestores de BBDD requieren que se especifique un
                // orden para poder usar OFFSET. Como eso parece buena idea para
                // asegurar que siempre tengamos los mismo resultados, lo exijimos
                // siempre.
                message.append("Can't use OFFSET without an ORDER BY.");
                return false;
            }
            return true;
        }        
        
        @Override
        public String toString(Formatter formatter) {
            StringBuilder builder = new StringBuilder();
            if( !isValid(builder) ) {
                throw new IllegalStateException(builder.toString());
            }
            builder.append("SELECT ");
            if( this.distinct ) {
                builder.append("DISTINCT ");
            }
            boolean first = true;
            for (SelectColumnBuilder column : columns) {
                if (first) {
                    first = false;
                } else {
                    builder.append(", ");
                }
                builder.append(column.toString(formatter));
            }

            if ( this.has_from() ) {
                builder.append(" FROM ");
                builder.append(this.from.toString(formatter));
            }
            if( this.has_group_by() ) {
                builder.append(" GROUP BY ");
                builder.append(this.groupColumn.get(0).toString(formatter));
                for (int i = 1; i < groupColumn.size(); i++) {
                    builder.append(", ");
                    builder.append(this.groupColumn.get(i).toString(formatter));
                }
            }
            if ( this.has_where() ) {
                builder.append(" WHERE ");
                builder.append(this.where.toString(formatter));
            }
            
            if( this.has_order_by() ) {
                builder.append(" ORDER BY ");
                first = true;
                for (OrderByBuilder item : this.order_by) {
                    if (first) {
                        first = false;
                    } else {
                        builder.append(", ");
                    }
                    builder.append(item.toString(formatter));                    
                }   
            }
            
            if ( this.has_limit() && this.has_offset() ) {
                builder.append(" LIMIT ");
                builder.append(this.limit);
                builder.append(" OFFSET ");
                builder.append(this.offset);
                
            } else if ( this.has_limit()) {
                builder.append(" LIMIT ");
                builder.append(this.limit);

            } else if ( this.has_offset() ) {
                builder.append(" LIMIT -1 OFFSET ");
                builder.append(this.offset);    
            }
            return builder.toString();

        }
    }

    @Override
    public Object sqlgeometrydimension(int type, int subtype) {
        //'XY' or 2: 2D points, identified by X and Y coordinates
        //'XYM': 2D points, identified by X and Y coordinates. Each point stores an M-value (measure) as well
        //'XYZ' or 3: 3D points, identified by X, Y and Z coordinates
        //'XYZM' or 4: 3D points, identified by X, Y and Z coordinates. Each point stores an M-value (measure) as well
        switch(subtype) {
            case Geometry.SUBTYPES.GEOM2D:
            default:
                return "XY";
            case Geometry.SUBTYPES.GEOM2DM:
                return "XYM";
            case Geometry.SUBTYPES.GEOM3D:
                return "XYZ";
            case Geometry.SUBTYPES.GEOM3DM:
                return "XYZM";
        }
    }

    @Override
    public String sqltype(int type, int p, int s, int geomtype, int geomSubtype) {
        if( type!=DataTypes.GEOMETRY ) {
            return super.sqltype(type, p, s, geomtype, geomSubtype);
        }
        //
        // https://github.com/orbisgis/h2gis/wiki/1.-Spatial-data#geometry-columns-view
        //
        switch(geomtype) {
        case Geometry.TYPES.POINT:
            return "POINT";
        case Geometry.TYPES.MULTIPOINT:
            return "MULTIPOINT";
        case Geometry.TYPES.LINE:
            return "LINESTRING";
        case Geometry.TYPES.MULTILINE:
            return "MULTILINESTRING";
        case Geometry.TYPES.POLYGON:
            return "POLYGON";
        case Geometry.TYPES.MULTIPOLYGON:
            return "MULTIPOLYGON";
        default:
            return "GEOMETRY";
        }
    }
    
    public Object sqlgeometrynumdimension(int type, int subtype) {
        int dimensions=2;
        switch(subtype) {
        case Geometry.SUBTYPES.GEOM3D:
            dimensions = 3;
            break;
        case Geometry.SUBTYPES.GEOM2D:
            dimensions = 2;
            break;
        case Geometry.SUBTYPES.GEOM2DM:
            dimensions = 2; // ??????
            break;
        case Geometry.SUBTYPES.GEOM3DM:
            dimensions = 3; // ??????
            break;
        }
        return dimensions;
    }
    
    @Override
    public H2SpatialHelper getHelper() {
        return (H2SpatialHelper) this.helper;
    }
   
    @Override
    public Disposable setParameters(PreparedStatement st, FeatureProvider feature) {
        try {
            FeatureType featureType = feature.getType();
            List<Object> values = new ArrayList<>();
            for (Parameter parameter : this.parameters()) {
                if (parameter.is_constant()) {
                    values.add(parameter.value());
                } else {
                    String name = parameter.name();
                    FeatureAttributeDescriptor descriptor = featureType.getAttributeDescriptor(name);
                    switch( descriptor.getType() ) {
                    case DataTypes.DATE:
                        Date value = (Date)(feature.get(name));
                        if( value == null ) {
                            values.add(null);
                        } else {
                            values.add(new java.sql.Date(value.getTime()));
                        }
                        break;
                    case DataTypes.GEOMETRY:
                        Geometry geom = this.forceGeometryType(
                            descriptor.getGeomType(),
                            (Geometry)(feature.get(name))
                        );
                        values.add(geom);
                        break;
                    default:
                        values.add(feature.get(name));
                        break;
                    }
                }
            }
            return this.setStatementParameters(st, values, this.geometry_support_type());
        } catch (SQLException | CreateGeometryException ex) {
            String f = "unknow";
            try {
                f = feature.toString();
            } catch (Exception ex2) {
                // Do nothing
            }
            throw new RuntimeException("Can't set parameters to prepared statement from the feature (" + f + ")", ex);
        }
    }

    @Override
    protected AlterTableBuilder createAlterTableBuilder() {
        return new H2SpatialAlterTableBuilderBase();
    }
            
    @Override
    public TableNameBuilder createTableNameBuilder() {
        return new H2SpatialTableNameBuilderBase();
    }
    
    @Override
    protected CreateTableBuilder createCreateTableBuilder() {
        return new H2SpatialCreateTableBuilder();
    }

    @Override
    protected SelectBuilder createSelectBuilder() {
        return new H2SpatialSelectBuilderBase();
    }

    @Override
    protected CreateIndexBuilder createCreateIndexBuilder() {
        return new H2SpatialCreateIndexBuilder();
    }

}
