
package org.gvsig.fmap.dal.store.h2.operations;

import java.sql.PreparedStatement;
import java.sql.SQLException;
import org.gvsig.expressionevaluator.ExpressionBuilder;
import org.gvsig.fmap.dal.DataTypes;
import org.gvsig.fmap.dal.exception.DataException;
import org.gvsig.fmap.dal.feature.FeatureAttributeDescriptor;
import org.gvsig.fmap.dal.feature.FeatureType;
import org.gvsig.fmap.dal.feature.exception.AlreadyEditingException;
import org.gvsig.fmap.dal.feature.spi.FeatureProvider;
import org.gvsig.fmap.dal.store.jdbc.exception.JDBCPreparingSQLException;
import org.gvsig.fmap.dal.store.jdbc2.JDBCHelper;
import org.gvsig.fmap.dal.store.jdbc2.JDBCUtils;
import org.gvsig.fmap.dal.store.jdbc2.OperationsFactory.TableReference;
import org.gvsig.fmap.dal.store.jdbc2.spi.operations.AppendOperation;
import org.gvsig.tools.dispose.Disposable;
import org.gvsig.tools.dispose.DisposeUtils;


@SuppressWarnings("UseSpecificCatch")
public class H2SpatialAppendOperation extends AppendOperation {

    public H2SpatialAppendOperation(
            JDBCHelper helper, 
            TableReference table, 
            FeatureType type
        ) {
        super(helper, table, type);
    }
    
    @Override
    public void begin() throws DataException {
        if (this.sqlbuilder != null) {
            throw new AlreadyEditingException(this.helper.getSourceId());
        }

        try {
            this.connection = this.helper.getConnectionWritable();
            
            this.sqlbuilder = this.helper.createSQLBuilder();
            this.expbuilder = this.sqlbuilder.expression();

            this.sqlbuilder.insert().table()
                .database(this.table.getDatabase())
                .schema(this.table.getSchema())
                .name(this.table.getTable());
            for (FeatureAttributeDescriptor attr : type) {
                if( attr.isAutomatic() || attr.isComputed() ) {
                    continue;
                }
                if (attr.getType() == DataTypes.GEOMETRY) {
                    this.sqlbuilder.insert().column().name(attr.getName()).with_value( 
                        this.expbuilder.parameter(attr.getName()).as_geometry_variable().srs( 
                                this.expbuilder.parameter().value(attr.getSRS()) 
                        ) 
                    );
                } else {
                    this.sqlbuilder.insert().column().name(attr.getName()).with_value(
                        this.expbuilder.parameter(attr.getName()).as_variable()
                    );
                }
            }

            PreparedStatement st;
            this.sql = this.sqlbuilder.insert().toString();
            this.preparedStatement = this.connection.prepareStatement(sql);
            this.connection.setAutoCommit(false);
            JDBCUtils.execute(this.connection, "SET LOG 1");
            JDBCUtils.execute(this.connection, "SET LOCK_MODE 1");
            JDBCUtils.execute(this.connection, "SET UNDO_LOG 0");
            
        } catch (SQLException ex) {
            throw new JDBCPreparingSQLException(this.sqlbuilder.toString(),ex);
        }

    }
    
    @Override
    protected void clean() {
        JDBCUtils.closeQuietly(this.preparedStatement);
        this.helper.closeConnection(this.connection);
        this.connection = null;
        this.preparedStatement = null;
        this.sqlbuilder = null;
        this.sql = null;        
    }
    
    @Override
    public void end() {
        try {
            this.connection.commit();
            JDBCUtils.execute(this.connection, "SET LOG 2");
            JDBCUtils.execute(this.connection, "SET LOCK_MODE 3");
            JDBCUtils.execute(this.connection, "SET UNDO_LOG 1");
        } catch (SQLException ex) {
            try {
                this.connection.rollback();
            } catch (SQLException ex1) {
            }
            throw new RuntimeException("Can't commit transaction", ex);
        } finally {
            clean();
        }
    }
    
    @Override
    public void abort() {        
        try {
            this.connection.rollback();
            JDBCUtils.execute(this.connection, "SET LOG 2");
            JDBCUtils.execute(this.connection, "SET LOCK_MODE 3");
            JDBCUtils.execute(this.connection, "SET UNDO_LOG 1");
        } catch (SQLException ex) {
        }
        clean();
    }
    
    @Override
    public void append(FeatureProvider feature) throws DataException {
        int n;
        Disposable paramsDisposer = null;
        try {
            paramsDisposer = this.sqlbuilder.setParameters(this.preparedStatement, feature);
            n = JDBCUtils.executeUpdate(this.preparedStatement,this.sql);
        } catch(Exception ex) {
            throw new RuntimeException("Can't insert feature.", ex);
        } finally {
            DisposeUtils.disposeQuietly(paramsDisposer);
        }
        if( n<1 ) {
            throw new RuntimeException("Can't insert feature (n="+n+").");
        }
    }
}
