
package org.gvsig.fmap.dal.store.h2.operations;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import org.cresques.cts.IProjection;
import org.gvsig.expressionevaluator.ExpressionBuilder;
import org.gvsig.fmap.dal.DataTypes;
import org.gvsig.fmap.dal.exception.DataException;
import org.gvsig.fmap.dal.feature.EditableFeatureAttributeDescriptor;
import org.gvsig.fmap.dal.feature.EditableFeatureType;
import org.gvsig.fmap.dal.resource.exception.AccessResourceException;
import org.gvsig.fmap.dal.store.jdbc2.JDBCHelper;
import org.gvsig.fmap.dal.store.jdbc2.JDBCUtils;
import org.gvsig.fmap.dal.store.jdbc2.OperationsFactory.TableReference;
import org.gvsig.fmap.dal.store.jdbc2.spi.JDBCSQLBuilderBase;
import org.gvsig.fmap.dal.store.jdbc2.spi.SRSSolver;
import org.gvsig.fmap.dal.store.jdbc2.spi.operations.FetchFeatureTypeOperation;
import org.gvsig.fmap.geom.Geometry;
import org.gvsig.fmap.geom.GeometryLocator;
import org.gvsig.fmap.geom.GeometryManager;
import org.gvsig.fmap.geom.type.GeometryType;

@SuppressWarnings("UseSpecificCatch")
public class H2SpatialFetchFeatureTypeOperation extends FetchFeatureTypeOperation {

    private static Map<String,GeometryType>h2spatialGeometryTypes = null;
    private Map<String,GeometryColumnInfo> geometry_column;
    
    private static class GeometryColumnInfo {
        
        public String columnName;
        public int geometryType;
        public String geometryTypeName;
        public int dimensions;
        public int srid;
        
        public GeometryColumnInfo(String columnName, int geometryType, String geometryTypeName, int dimensions, int srid) {
            this.columnName = columnName;
            this.geometryType = geometryType;
            this.geometryTypeName = geometryTypeName;
            this.dimensions = dimensions;
            this.srid = srid;
        }
    }
    
    
    public H2SpatialFetchFeatureTypeOperation(
            JDBCHelper helper
        ) {
        super(helper);
    }

    private GeometryType getGT(
            GeometryManager manager, 
            int type, 
            int subtype
        ) {
        try {
            return manager.getGeometryType(type, subtype);
        } catch (Exception ex) {
            return null;
        }
    }
    
    public H2SpatialFetchFeatureTypeOperation(
            JDBCHelper helper,
            EditableFeatureType featureType,
            TableReference table,
            List<String> primaryKeys,
            String defaultGeometryColumn,
            IProjection crs
        ) {
        super(helper, featureType, table, primaryKeys, defaultGeometryColumn, crs);
    }            

    private GeometryColumnInfo getGeometryColumnInfo(String name) {
        if( geometry_column==null ) {
            geometry_column = new HashMap<>();
            try {
                //
                // https://github.com/orbisgis/h2gis/wiki/1.-Spatial-data#geometry-columns-view
                //
                StringBuilder where = null;
                if( table.hasDatabase() ) {
                    if( where == null ) {
                        where = new StringBuilder();
                    } else {
                        where.append(" AND ");
                    }
                    where.append("UPPER(F_TABLE_CATALOG) = '");
                    where.append(table.getDatabase().toUpperCase());
                    where.append("'");
                }
                if( table.hasSchema()) {
                    if( where == null ) {
                        where = new StringBuilder();
                    } else {
                        where.append(" AND ");
                    }
                    where.append("UPPER(F_TABLE_SCHEMA) = '");
                    where.append(table.getSchema().toUpperCase());
                    where.append("'");
                }
                if( table.hasTable()) {
                    if( where == null ) {
                        where = new StringBuilder();
                    } else {
                        where.append(" AND ");
                    }
                    where.append("UPPER(F_TABLE_NAME) = '");
                    where.append(table.getTable().toUpperCase());
                    where.append("'");
                }            
                String sql = "SELECT F_GEOMETRY_COLUMN, GEOMETRY_TYPE, COORD_DIMENSION, SRID, TYPE FROM GEOMETRY_COLUMNS WHERE " + where;
                Statement st = this.getConnection().createStatement();
                ResultSet rs = JDBCUtils.executeQuery(st,sql);
                while( rs.next() ) {
                    geometry_column.put(
                        rs.getString("F_GEOMETRY_COLUMN"), 
                        new GeometryColumnInfo(
                            rs.getString("F_GEOMETRY_COLUMN"), 
                            rs.getInt("GEOMETRY_TYPE"), 
                            rs.getString("TYPE"), 
                            rs.getInt("COORD_DIMENSION"), 
                            rs.getInt("SRID")
                        )
                    );
                }
            } catch (SQLException | AccessResourceException ex) {
                LOGGER.warn("Can't read metadata from table '"+table+"'.",ex);
            }
        }
        return this.geometry_column.get(name);
    }
    
    @Override
    public void fetch(EditableFeatureType featureType, Connection conn, TableReference table, List<String> pks, String defaultGeometryColumn, IProjection crs) throws DataException {
        super.fetch(featureType, conn, table, pks, defaultGeometryColumn, crs);
    }
        
    @Override
    protected void fetchGeometryTypeAndSRS(
            EditableFeatureAttributeDescriptor attr,
            ResultSetMetaData rsMetadata,
            int colIndex
        ) {
        try {
            if( attr.getType()==DataTypes.GEOMETRY ) {
                GeometryColumnInfo column_info = this.getGeometryColumnInfo(attr.getName());
                String type = "GEOMETRY";
                if( column_info!=null ) {
                    // H2GIS solo soporte 2D y 3D no soporta Ms
                    if( column_info.dimensions==3 ) {
                        type = column_info.geometryTypeName+"Z";
                    } else {
                        type = column_info.geometryTypeName;
                    }
                    SRSSolver solver = this.helper.getSRSSolver();
                    attr.setSRS(
                        solver.getProjection(getConnection(), column_info.srid)
                    );
                }
                GeometryType gt = getGeometryTypeFromH2SpatialType(type);
                if( gt != null ) {
                    attr.setGeometryType(gt);
                }
            }
        } catch(Exception ex) {
            throw new RuntimeException("Can't fetch geometry type and SRS.", ex);
        }
    }

    private GeometryType getGeometryTypeFromH2SpatialType(String typeName) {
        if( h2spatialGeometryTypes==null ) {
            //
            // https://github.com/orbisgis/h2gis/wiki/1.-Spatial-data#geometry-columns-view
            //
            GeometryManager manager = GeometryLocator.getGeometryManager();
            h2spatialGeometryTypes = new HashMap<>();
            h2spatialGeometryTypes.put("POINT", getGT(manager, Geometry.TYPES.POINT,Geometry.SUBTYPES.GEOM2D));
            h2spatialGeometryTypes.put("POINTZ", getGT(manager, Geometry.TYPES.POINT,Geometry.SUBTYPES.GEOM3D));
            
            h2spatialGeometryTypes.put("LINESTRING", getGT(manager, Geometry.TYPES.LINE,Geometry.SUBTYPES.GEOM2D));
            h2spatialGeometryTypes.put("LINESTRINGZ", getGT(manager, Geometry.TYPES.LINE,Geometry.SUBTYPES.GEOM3D));
            
            h2spatialGeometryTypes.put("POLYGON", getGT(manager, Geometry.TYPES.POLYGON,Geometry.SUBTYPES.GEOM2D));
            h2spatialGeometryTypes.put("POLYGONZ", getGT(manager, Geometry.TYPES.POLYGON,Geometry.SUBTYPES.GEOM3D));

            h2spatialGeometryTypes.put("MULTIPOINT", getGT(manager, Geometry.TYPES.MULTIPOINT,Geometry.SUBTYPES.GEOM2D));
            h2spatialGeometryTypes.put("MULTIPOINTZ", getGT(manager, Geometry.TYPES.MULTIPOINT,Geometry.SUBTYPES.GEOM3D));

            h2spatialGeometryTypes.put("MULTILINESTRING", getGT(manager, Geometry.TYPES.MULTILINE,Geometry.SUBTYPES.GEOM2D));
            h2spatialGeometryTypes.put("MULTILINESTRINGZ", getGT(manager, Geometry.TYPES.MULTILINE,Geometry.SUBTYPES.GEOM3D));

            h2spatialGeometryTypes.put("MULTIPOLYGON", getGT(manager, Geometry.TYPES.MULTIPOLYGON,Geometry.SUBTYPES.GEOM2D));
            h2spatialGeometryTypes.put("MULTIPOLYGONZ", getGT(manager, Geometry.TYPES.MULTIPOLYGON,Geometry.SUBTYPES.GEOM3D));

            h2spatialGeometryTypes.put("GEOMETRY", getGT(manager, Geometry.TYPES.GEOMETRY,Geometry.SUBTYPES.GEOM2D));
            h2spatialGeometryTypes.put("GEOMETRYZ", getGT(manager, Geometry.TYPES.GEOMETRY,Geometry.SUBTYPES.GEOM3D));

            h2spatialGeometryTypes.put("GEOMCOLLECTION", getGT(manager, Geometry.TYPES.GEOMETRY,Geometry.SUBTYPES.GEOM2D));
            h2spatialGeometryTypes.put("GEOMCOLLECTIONZ", getGT(manager, Geometry.TYPES.GEOMETRY,Geometry.SUBTYPES.GEOM3D));
        }
        return h2spatialGeometryTypes.get(typeName);
    }

    @Override
    protected String getSQLToRetrievePrimaryKeysFromInformationSchema(
            String catalog,
            String schema,
            String table
        ) throws SQLException {
        JDBCSQLBuilderBase sqlbuilder = this.createSQLBuilder();
        ExpressionBuilder expbuilder = sqlbuilder.expression();

        sqlbuilder.select().column().name("COLUMN_LIST");
        sqlbuilder.select().column().name("CONSTRAINT_TYPE");
        sqlbuilder.select().from().table().schema("INFORMATION_SCHEMA").name("CONSTRAINTS");
        sqlbuilder.select().where().set(
                expbuilder.like(
                        expbuilder.column("TABLE_NAME"), 
                        expbuilder.constant(table)
                )
        );
        if (schema != null) {
            sqlbuilder.select().where().and(
                    expbuilder.like(
                            expbuilder.column("TABLE_SCHEMA"),
                            expbuilder.constant(schema)
                    )
            );
        }
        if (catalog != null) {
            sqlbuilder.select().where().and(
                    expbuilder.like(
                            expbuilder.column("CONSTRAINT_CATALOG"),
                            expbuilder.constant(catalog)
                    )
            );
        }
        sqlbuilder.select().where().and(
                expbuilder.eq(
                        expbuilder.column("CONSTRAINT_TYPE"),
                        expbuilder.constant("PRIMARY KEY")
                )
        );
        return sqlbuilder.toString();
    }
    
    
}
