package org.gvsig.fmap.dal.store.h2;

import junit.framework.TestCase;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import org.cresques.cts.IProjection;
import org.gvsig.expressionevaluator.ExpressionBuilder;
import org.gvsig.fmap.crs.CRSFactory;
import org.gvsig.fmap.dal.SQLBuilder;
import org.gvsig.fmap.dal.SQLBuilder.Privilege;
import org.gvsig.fmap.dal.feature.spi.SQLBuilderBase;
import org.gvsig.fmap.dal.store.jdbc2.OperationsFactory.TableReference;
import static org.gvsig.fmap.dal.store.jdbc2.spi.JDBCSQLBuilderBase.PROP_TABLE;
import org.gvsig.fmap.dal.store.jdbc2.spi.operations.OperationsFactoryBase;
import org.gvsig.fmap.geom.DataTypes;
import org.gvsig.fmap.geom.Geometry;
import org.gvsig.fmap.geom.GeometryLocator;
import org.gvsig.fmap.geom.GeometryManager;
import org.gvsig.fmap.geom.primitive.Polygon;
import org.gvsig.tools.library.impl.DefaultLibrariesInitializer;

public class H2SpatialSQLBuilderTest extends TestCase {
    
    public H2SpatialSQLBuilderTest(String testName) {
        super(testName);
    }

    @Override
    protected void setUp() throws Exception {
        super.setUp();
        new DefaultLibrariesInitializer().fullInitialize();
    }

    @Override
    protected void tearDown() throws Exception {
        super.tearDown();
    }

    private SQLBuilder createSQLBuilder() {
        return new H2SpatialSQLBuilder(new H2SpatialHelper(null));
    }
    
    public void testCalulateEnvelopeOfColumn() throws Exception {
        
        TableReference table = new OperationsFactoryBase.DefaultTableReference("master","dbo","test1", null);
        String columnName = "the_geom";
        
        SQLBuilder sqlbuilder = createSQLBuilder();
        ExpressionBuilder expbuilder = sqlbuilder.expression();
        
        sqlbuilder.select().column().value(
            expbuilder.as_geometry(
                expbuilder.ST_ExtentAggregate(
                        expbuilder.column(columnName)
                )
            )
        );
        //sqlbuilder.select().group_by(expbuilder.column(columnName));
        sqlbuilder.select().from().table()
                .database(table.getDatabase())
                .schema(table.getSchema())
                .name(table.getTable());
        sqlbuilder.select().from().subquery(table.getSubquery());

        sqlbuilder.select().where().and(        
            expbuilder.not_is_null(expbuilder.column(columnName))
        );
        
        System.out.println("# Test:: testCalulateEnvelopeOfColumn");
        System.out.println("# SQL:: " + sqlbuilder.toString());
        System.out.println("# Variables:: " + ArrayUtils.toString(sqlbuilder.variables_names()));
        System.out.println("# Parametros:: " + ArrayUtils.toString(sqlbuilder.parameters_names()));
        assertEquals(
                "SELECT ST_AsBinary(ST_Extent(\"the_geom\")) FROM \"dbo\".\"test1\" WHERE ( (\"the_geom\") IS NOT NULL )",
                sqlbuilder.toString()
        );
        assertEquals(
                "[the_geom]",
                ArrayUtils.toString(sqlbuilder.variables_names())
        );
        assertEquals(
                "[]",
                ArrayUtils.toString(sqlbuilder.parameters_names())
        );
    }
 
    public void testCalulateEnvelope() throws Exception {
        GeometryManager geometryManager = GeometryLocator.getGeometryManager();
        IProjection proj = CRSFactory.getCRS("EPSG:4326");
        
        Polygon limit = geometryManager.createPolygon(Geometry.SUBTYPES.GEOM2D);
        limit.addVertex(0, 0);
        limit.addVertex(0, 100);
        limit.addVertex(100, 100);
        limit.addVertex(100, 0);
        limit.addVertex(0, 0);
        
        SQLBuilder sqlbuilder = createSQLBuilder();
        ExpressionBuilder expbuilder = sqlbuilder.expression();
        
        sqlbuilder.select().column().value(
            expbuilder.as_geometry(
              expbuilder.ST_ExtentAggregate(
                expbuilder.column("the_geom")
              )
            )
        ).as("envelope");
        sqlbuilder.select().from().table().database("master").schema("dbo").name("test1");
        sqlbuilder.select().where().set(
            expbuilder.ST_Intersects(
                expbuilder.ST_Envelope(
                    expbuilder.column("the_geom")
                ),
                expbuilder.geometry(limit, proj)
            )
        );
        sqlbuilder.select().where().and(
                expbuilder.custom("x = 27")
        );
        
        System.out.println("# Test:: testCalulateEnvelope");
        System.out.println("# SQL:: " + sqlbuilder.toString());
        System.out.println("# Variables:: " + ArrayUtils.toString(sqlbuilder.variables_names()));
        System.out.println("# Parametros:: " + ArrayUtils.toString(sqlbuilder.parameters_names()));
        assertEquals(
                "SELECT ST_AsBinary(ST_Extent(\"the_geom\")) AS \"envelope\" FROM \"dbo\".\"test1\" WHERE (( ((ST_Envelope(\"the_geom\")) && (ST_GeomFromWKB((x'000000000300000001000000050000000000000000000000000000000000000000000000004059000000000000405900000000000040590000000000004059000000000000000000000000000000000000000000000000000000000000'), (4326)))) AND ST_Intersects((ST_Envelope(\"the_geom\")),(ST_GeomFromWKB((x'000000000300000001000000050000000000000000000000000000000000000000000000004059000000000000405900000000000040590000000000004059000000000000000000000000000000000000000000000000000000000000'), (4326))) )) AND x = 27)",
                sqlbuilder.toString()
        );
        assertEquals(
                "[the_geom]",
                ArrayUtils.toString(sqlbuilder.variables_names())
        );
        assertEquals(
                "[]",
                ArrayUtils.toString(sqlbuilder.parameters_names())
        );
    }

    public void testCount() throws Exception {
        SQLBuilder sqlbuilder = createSQLBuilder();
        ExpressionBuilder expbuilder = sqlbuilder.expression();
        
        sqlbuilder.select().column().value(sqlbuilder.count().all());
        sqlbuilder.select().from().table().database("master").schema("dbo").name("test1");
        sqlbuilder.select().from().subquery(null);
        sqlbuilder.select().where().set( expbuilder.custom("pp = 200"));

        System.out.println("# Test:: testCount");
        System.out.println("# SQL:: " + sqlbuilder.toString());
        System.out.println("# Variables:: " + ArrayUtils.toString(sqlbuilder.variables_names()));
        System.out.println("# Parametros:: " + ArrayUtils.toString(sqlbuilder.parameters_names()));

        //# Test:: testCount
        //# SQL:: SELECT COUNT(*) FROM "test1" WHERE pp = 200
        //# Variables:: []
        //# Parametros:: []

        assertEquals(
                "SELECT COUNT(*) FROM \"dbo\".\"test1\" WHERE pp = 200",
                sqlbuilder.toString()
        );
        assertEquals(
                "[]",
                ArrayUtils.toString(sqlbuilder.variables_names())
        );
        assertEquals(
                "[]",
                ArrayUtils.toString(sqlbuilder.parameters_names())
        );
    }
    
    public void testCreateTable() throws Exception {
        SQLBuilder sqlbuilder = createSQLBuilder();
        ExpressionBuilder expbuilder = sqlbuilder.expression();

        sqlbuilder.create_table().table().database("master").schema("dbo").name("test1");
        sqlbuilder.create_table().add_column(
                "name",
                DataTypes.STRING,
                45,
                0,
                false,
                false,
                true,
                false,
                null
        );
        sqlbuilder.create_table().add_column(
                "id",
                DataTypes.INT,
                0,
                0,
                true,
                false,
                false,
                true,
                0
        );
        sqlbuilder.create_table().add_column(
                "geom",
                DataTypes.GEOMETRY,
                0,
                0,
                false,
                false,
                true,
                false,
                null
        );

        
        // CREATE TABLE "test1" ("name" VARCHAR(45) DEFAULT NULL, "id" INTEGER PRIMARY KEY AUTO_INCREMENT DEFAULT '0' NOT NULL, "geom" GEOMETRY ); ALTER TABLE "test1" ADD CONSTRAINT IF NOT EXISTS "constraint_test1_geom_dim" CHECK ST_CoordDim("geom") = 2
        System.out.println("# Test:: testCreateTable");
        System.out.println("# SQL:: " + sqlbuilder.toString());
        System.out.println("# Variables:: " + ArrayUtils.toString(sqlbuilder.variables_names()));
        System.out.println("# Parametros:: " + ArrayUtils.toString(sqlbuilder.parameters_names()));
        assertEquals(
                "CREATE TABLE \"dbo\".\"test1\" (\"name\" VARCHAR(45) DEFAULT NULL, \"id\" INTEGER PRIMARY KEY AUTO_INCREMENT NOT NULL, \"geom\" GEOMETRY ); ALTER TABLE \"dbo\".\"test1\" ADD CONSTRAINT IF NOT EXISTS \"constraint_test1_geom_dim\" CHECK NVL2(\"geom\", ST_CoordDim(\"geom\") = 2, TRUE)",
                sqlbuilder.toString()
        );
        assertEquals(
                "[]",
                ArrayUtils.toString(sqlbuilder.variables_names())
        );
        assertEquals(
                "[]",
                ArrayUtils.toString(sqlbuilder.parameters_names())
        );
    }

    public void testDropTable() throws Exception {
        SQLBuilder sqlbuilder = createSQLBuilder();
        ExpressionBuilder expbuilder = sqlbuilder.expression();
        
        sqlbuilder.drop_table().table().database("master").schema("dbo").name("test1");

        // DROP TABLE "test1"
        
        System.out.println("# Test:: testDropTable");
        System.out.println("# SQL:: " + sqlbuilder.toString());
        System.out.println("# Variables:: " + ArrayUtils.toString(sqlbuilder.variables_names()));
        System.out.println("# Parametros:: " + ArrayUtils.toString(sqlbuilder.parameters_names()));
        assertEquals(
                "DROP TABLE \"dbo\".\"test1\"",
                sqlbuilder.toString()
        );
        assertEquals(
                "[]",
                ArrayUtils.toString(sqlbuilder.variables_names())
        );
        assertEquals(
                "[]",
                ArrayUtils.toString(sqlbuilder.parameters_names())
        );
    }
    
    public void testFetchFeatureProviderByReference() throws Exception {
        SQLBuilder sqlbuilder = new SQLBuilderBase();
        ExpressionBuilder expbuilder = sqlbuilder.expression();
        
        String value = "yoyo";
        sqlbuilder.select().column().name("name");
        sqlbuilder.select().column().name("id");
        sqlbuilder.select().column().name("geom").as_geometry();
        sqlbuilder.select().from().table().database("master").schema("dbo").name("test1");
        sqlbuilder.select().where().set(
            expbuilder.eq(
                expbuilder.column("name"),
                expbuilder.parameter(value).as_constant()
            )
        );
        sqlbuilder.select().limit(1);

        // SELECT "name", "id", ST_AsBinary("geom") FROM "master"."dbo"."test1" WHERE ( ("name") = (?) ) LIMIT 1

        System.out.println("# Test:: testFetchFeatureProviderByReference");
        System.out.println("# SQL:: " + sqlbuilder.toString());
        System.out.println("# Variables:: " + ArrayUtils.toString(sqlbuilder.variables_names()));
        System.out.println("# Parametros:: " + ArrayUtils.toString(sqlbuilder.parameters_names()));
        assertEquals(
                "SELECT \"name\", \"id\", ST_AsBinary(\"geom\") FROM \"master\".\"dbo\".\"test1\" WHERE ( (\"name\") = (?) ) LIMIT 1",
                sqlbuilder.toString()
        );
        assertEquals(
                "[geom, id, name]",
                ArrayUtils.toString(sqlbuilder.variables_names())
        );
        assertEquals(
                "['yoyo']",
                ArrayUtils.toString(sqlbuilder.parameters_names())
        );
    }
    
    public void testFetchFeatureType() throws Exception {
        SQLBuilder sqlbuilder = createSQLBuilder();
        ExpressionBuilder expbuilder = sqlbuilder.expression();

        sqlbuilder.select().column().all();
        sqlbuilder.select().from().table().database("master").schema("dbo").name("test1");
        sqlbuilder.select().limit(1);

        System.out.println("# Test:: testFetchFeatureType");
        System.out.println("# SQL:: " + sqlbuilder.toString());
        System.out.println("# Variables:: " + ArrayUtils.toString(sqlbuilder.variables_names()));
        System.out.println("# Parametros:: " + ArrayUtils.toString(sqlbuilder.parameters_names()));
        
        //# Test:: testFetchFeatureType
        //# SQL:: SELECT * FROM "test1" LIMIT 1
        //# Variables:: []
        //# Parametros:: []        
        
        assertEquals(
                "SELECT * FROM \"dbo\".\"test1\" LIMIT 1",
                sqlbuilder.toString()
        );
        assertEquals(
                "[]",
                ArrayUtils.toString(sqlbuilder.variables_names())
        );
        assertEquals(
                "[]",
                ArrayUtils.toString(sqlbuilder.parameters_names())
        );
    }
        
    public void testPerformDeletes() throws Exception {
        SQLBuilder sqlbuilder = createSQLBuilder();
        ExpressionBuilder expbuilder = sqlbuilder.expression();

        sqlbuilder.delete().table().database("master").schema("dbo").name("test1");
        sqlbuilder.delete().where().and(
            expbuilder.eq( 
                expbuilder.column("id1"),
                expbuilder.parameter("id1").as_variable()
            )
        );
        sqlbuilder.delete().where().and(
            expbuilder.eq( 
                expbuilder.column("id2"),
                expbuilder.parameter("id2").as_variable()
            )
        );

        // DELETE FROM "test1" WHERE ( ("id1") = (?) ) AND ( ("id2") = (?) )

        System.out.println("# Test:: testPerformDeletes");
        System.out.println("# SQL:: " + sqlbuilder.toString());
        System.out.println("# Variables:: " + ArrayUtils.toString(sqlbuilder.variables_names()));
        System.out.println("# Parametros:: " + ArrayUtils.toString(sqlbuilder.parameters_names()));
        assertEquals(
                "DELETE FROM \"dbo\".\"test1\" WHERE (( (\"id1\") = (?) ) AND ( (\"id2\") = (?) ))",
                sqlbuilder.toString()
        );
        assertEquals(
                "[id1, id2]",
                ArrayUtils.toString(sqlbuilder.variables_names())
        );
        assertEquals(
                "[\"id1\", \"id2\"]",
                ArrayUtils.toString(sqlbuilder.parameters_names())
        );
    }

    public void testPerformInserts1() throws Exception {
        IProjection proj = CRSFactory.getCRS("EPSG:4326");

        SQLBuilder sqlbuilder = createSQLBuilder();
        ExpressionBuilder expbuilder = sqlbuilder.expression();

        sqlbuilder.insert().table().database("master").schema("dbo").name("test1");
        sqlbuilder.insert().column().name("id").with_value(expbuilder.parameter("id"));
        sqlbuilder.insert().column().name("name").with_value(expbuilder.parameter("name"));
        sqlbuilder.insert().column().name("geom").with_value(expbuilder.parameter("geom").as_geometry_variable().srs(proj));
        
        System.out.println("# Test:: testPerformInserts1");
        System.out.println("# SQL:: " + sqlbuilder.toString());
        System.out.println("# Variables:: " + ArrayUtils.toString(sqlbuilder.variables_names()));
        System.out.println("# Parametros:: " + ArrayUtils.toString(sqlbuilder.parameters_names()));
        assertEquals(
                "INSERT INTO \"dbo\".\"test1\" ( \"id\", \"name\", \"geom\" ) VALUES ( ?, ?, ST_GeomFromWKB((?), (4326)) )",
                sqlbuilder.toString()
        );
        assertEquals(
                "[geom, id, name]",
                ArrayUtils.toString(sqlbuilder.variables_names())
        );
        assertEquals(
                "[\"id\", \"name\", \"geom\"]",
                ArrayUtils.toString(sqlbuilder.parameters_names())
        );
    }

    public void testPerformInserts2() throws Exception {
        IProjection proj = CRSFactory.getCRS("EPSG:4326");

        SQLBuilder sqlbuilder = createSQLBuilder();
        ExpressionBuilder expbuilder = sqlbuilder.expression();

        sqlbuilder.insert().table().database("master").schema("dbo").name("test1");
        sqlbuilder.insert().column().name("id").with_value(expbuilder.parameter("id"));
        sqlbuilder.insert().column().name("name").with_value(expbuilder.parameter("name"));
        sqlbuilder.insert().column().name("geom").with_value(expbuilder.parameter("geom").as_geometry_variable().srs(proj));
        
        System.out.println("# Test:: testPerformInserts2");
        System.out.println("# SQL:: " + sqlbuilder.toString());
        System.out.println("# Variables:: " + ArrayUtils.toString(sqlbuilder.variables_names()));
        System.out.println("# Parametros:: " + ArrayUtils.toString(sqlbuilder.parameters_names()));
        assertEquals(
                "INSERT INTO \"dbo\".\"test1\" ( \"id\", \"name\", \"geom\" ) VALUES ( ?, ?, ST_GeomFromWKB((?), (4326)) )",
                sqlbuilder.toString()
        );
        assertEquals(
                "[geom, id, name]",
                ArrayUtils.toString(sqlbuilder.variables_names())
        );
        assertEquals(
                "[\"id\", \"name\", \"geom\"]",
                ArrayUtils.toString(sqlbuilder.parameters_names())
        );
    }

    public void testPerformUpdates1() throws Exception {
        IProjection proj = CRSFactory.getCRS("EPSG:4326");

        SQLBuilder sqlbuilder = createSQLBuilder();
        ExpressionBuilder expbuilder = sqlbuilder.expression();

        sqlbuilder.update().table().database("master").schema("dbo").name("test1");
        sqlbuilder.update().where().and(
            expbuilder.eq(
                expbuilder.column("id"), 
                expbuilder.parameter("id").as_variable()
            )
        );
        sqlbuilder.update().column().name("name").with_value(expbuilder.parameter("name"));
        sqlbuilder.update().column().name("geom").with_value(
                expbuilder.parameter("geom").as_geometry_variable().srs(proj) 
        );

        System.out.println("# Test:: testPerformUpdates");
        System.out.println("# SQL:: " + sqlbuilder.toString());
        System.out.println("# Variables:: " + ArrayUtils.toString(sqlbuilder.variables_names()));
        System.out.println("# Parametros:: " + ArrayUtils.toString(sqlbuilder.parameters_names()));
        assertEquals(
                "UPDATE \"dbo\".\"test1\" SET \"name\" = ?, \"geom\" = ST_GeomFromWKB((?), (4326)) WHERE ( (\"id\") = (?) )",
                sqlbuilder.toString()
        );
        assertEquals(
                "[geom, id, name]",
                ArrayUtils.toString(sqlbuilder.variables_names())
        );
        assertEquals(
                "[\"name\", \"geom\", \"id\"]",
                ArrayUtils.toString(sqlbuilder.parameters_names())
        );
    }

    public void testPerformUpdates2() throws Exception {
        IProjection proj = CRSFactory.getCRS("EPSG:4326");

        SQLBuilder sqlbuilder = createSQLBuilder();
        ExpressionBuilder expbuilder = sqlbuilder.expression();

        sqlbuilder.update().table().database("master").schema("dbo").name("test1");
        sqlbuilder.update().where().and(
            expbuilder.eq(
                expbuilder.column("id"), 
                expbuilder.parameter("id").as_variable()
            )
        );
        sqlbuilder.update().column().name("name").with_value(expbuilder.parameter("name"));
        sqlbuilder.update().column().name("geom").with_value(
                expbuilder.parameter("geom").as_geometry_variable()
                        .srs(expbuilder.parameter().value(proj)) 
        );

        System.out.println("# Test:: testPerformUpdates");
        System.out.println("# SQL:: " + sqlbuilder.toString());
        System.out.println("# Variables:: " + ArrayUtils.toString(sqlbuilder.variables_names()));
        System.out.println("# Parametros:: " + ArrayUtils.toString(sqlbuilder.parameters_names()));
        assertEquals(
                "UPDATE \"dbo\".\"test1\" SET \"name\" = ?, \"geom\" = ST_GeomFromWKB((?), (?)) WHERE ( (\"id\") = (?) )",
                sqlbuilder.toString()
        );
        assertEquals(
                "[geom, id, name]",
                ArrayUtils.toString(sqlbuilder.variables_names())
        );
        assertEquals(
                "[\"name\", \"geom\", 4326, \"id\"]",
                ArrayUtils.toString(sqlbuilder.parameters_names())
        );
    }

    public void testGrant1() throws Exception {

        SQLBuilder sqlbuilder = createSQLBuilder();
        ExpressionBuilder expbuilder = sqlbuilder.expression();

        sqlbuilder.grant().table().database("master").schema("dbo").name("test1");
        sqlbuilder.grant().role("prueba").select().insert().update();
        sqlbuilder.grant().role("gis").all();
                
        
        System.out.println("# Test:: testGrant1");
        System.out.println("# SQL:: " + sqlbuilder.toString());
        System.out.println("# Variables:: " + ArrayUtils.toString(sqlbuilder.variables_names()));
        System.out.println("# Parametros:: " + ArrayUtils.toString(sqlbuilder.parameters_names()));
        assertEquals(
                "",
                sqlbuilder.toString()
        );
        assertEquals(
                "[]",
                ArrayUtils.toString(sqlbuilder.variables_names())
        );
        assertEquals(
                "[]",
                ArrayUtils.toString(sqlbuilder.parameters_names())
        );
    }

    public void testGrant2() throws Exception {

        SQLBuilder sqlbuilder = new SQLBuilderBase();
        ExpressionBuilder expbuilder = sqlbuilder.expression();

        sqlbuilder.grant().table().database("master").schema("dbo").name("test1");
        sqlbuilder.grant().role("prueba").privilege(Privilege.SELECT)
                .privilege(Privilege.INSERT)
                .privilege(Privilege.UPDATE);
        sqlbuilder.grant().role("gis").privilege(Privilege.ALL);
                
        
        System.out.println("# Test:: testGrant2");
        System.out.println("# SQL:: " + sqlbuilder.toString());
        System.out.println("# Variables:: " + ArrayUtils.toString(sqlbuilder.variables_names()));
        System.out.println("# Parametros:: " + ArrayUtils.toString(sqlbuilder.parameters_names()));
        assertEquals(
                "",
                sqlbuilder.toString()
        );
        assertEquals(
                "[]",
                ArrayUtils.toString(sqlbuilder.variables_names())
        );
        assertEquals(
                "[]",
                ArrayUtils.toString(sqlbuilder.parameters_names())
        );
    }


}
