package org.gvsig.fmap.dal.store.jdbc2;

import org.gvsig.fmap.dal.store.jdbc2.spi.JDBCSQLBuilderBase;
import java.sql.Connection;
import java.sql.ResultSet;
import java.util.List;
import org.gvsig.expressionevaluator.ExpressionBuilder.GeometrySupportType;
import org.gvsig.fmap.dal.SQLBuilder;
import org.gvsig.fmap.dal.exception.DataException;
import org.gvsig.fmap.dal.exception.InitializeException;
import org.gvsig.fmap.dal.feature.FeatureAttributeDescriptor;
import org.gvsig.fmap.dal.feature.FeatureQueryOrder;
import org.gvsig.fmap.dal.feature.FeatureType;
import org.gvsig.fmap.dal.feature.spi.FeatureProvider;
import org.gvsig.fmap.dal.resource.exception.AccessResourceException;
import org.gvsig.fmap.dal.resource.spi.ResourceProvider;
import org.gvsig.fmap.dal.spi.DataServerExplorerProviderServices;
import org.gvsig.fmap.dal.spi.DataStoreProviderServices;
import org.gvsig.fmap.dal.store.jdbc.JDBCConnectionParameters;
import org.gvsig.fmap.dal.store.jdbc.JDBCNewStoreParameters;
import org.gvsig.fmap.dal.store.jdbc.JDBCServerExplorerParameters;
import org.gvsig.fmap.dal.store.jdbc.JDBCStoreParameters;
import org.gvsig.fmap.dal.store.jdbc2.ResulSetControler.ResultSetEntry;
import org.gvsig.fmap.geom.Geometry;
import org.gvsig.fmap.geom.GeometryManager;
import org.gvsig.fmap.dal.store.jdbc2.spi.SRSSolver;
import org.gvsig.tools.evaluator.Evaluator;

public interface JDBCHelper extends AutoCloseable {

    /**
     * Return the name of the driver.
     * 
     * By default rerturn "JDBC".
     * 
     * @return 
     */
    public String getProviderName();
    
    /**
     * Indica como deben ser guardadas las geometrias en la BBDD. Pueden
     * guardarse en WKT, WKB o EWKB.
     *
     * @return
     */
    public GeometrySupportType getGeometrySupportType();

    /**
     * Devuelbe un SQLBuilder adaptado al proveedor. Por ejemplo, uno especifico
     * para PostgreSQL, MySQL, Oracle, MSSQLServer...
     *
     * @return
     */
    public JDBCSQLBuilderBase createSQLBuilder();

    /**
     * Devueble las comillas que han de usaese para los identificadores.
     *
     * @return
     */
    public String getQuoteForIdentifiers();

    /**
     * Devueble las comillas que han de usaese en las constantes de cadena.
     *
     * @return
     */
    public String getQuoteForStrings();

    /**
     * Indica si la BBDD soporta valores automaticos, tipo serial.
     *
     * @return
     */
    public boolean allowAutomaticValues();

    /**
     * Indica si la BBDD soporta el uso de OFFSET en la sentencia select.
     *
     * @return
     */
    public boolean supportOffsetInSelect();

    /**
     * Indica si se especifico un subquery al abrir el proveedor.
     *
     * @return
     */
    public boolean useSubquery();

    /**
     * Indica si la BBDD tiene soporte espacial.
     *
     * @return
     */
    public boolean hasSpatialFunctions();

    public boolean supportFilter(FeatureType type, Evaluator evaluator);
    
    public boolean supportOrder(FeatureType type, FeatureQueryOrder order);
    
    public boolean allowNestedOperations();
    /**
     * Indica si podemos escribir el tipo de geometria indicado.
     *
     * @param geometryType
     * @param geometrySubtype
     * @return
     */
    public boolean canWriteGeometry(
            int geometryType, 
            int geometrySubtype
    );

    public Connection getConnection() throws AccessResourceException;

    public Connection getConnectionWritable() throws AccessResourceException;

    public String getConnectionURL();
    
    public JDBCConnectionParameters getConnectionParameters();

    public void closeConnection(Connection connection);

    public void closeConnectionQuietly(Connection connection);

    public GeometryManager getGeometryManager();

    public ResulSetControler getResulSetControler();

    public String getSourceId();

    public ResourceProvider getResource();

    public void dispose();

    public void fetchFeature(
            FeatureProvider feature, 
            ResultSet rs,
            FeatureAttributeDescriptor[] columns,
            String[] extraValueNames
    ) throws DataException;
    
    public void fetchFeature(
            FeatureProvider feature, 
            ResultSetEntry rs
    ) throws DataException;

    public Geometry getGeometryFromColumn(
            ResultSet rs, 
            int index
    ) throws DataException;

    public Geometry getGeometryFromColumn(
            ResultSetEntry rs, 
            int index
    ) throws DataException;

    public OperationsFactory getOperations();

    public FeatureProvider createFeature(
            FeatureType featureType
    ) throws DataException;

    public JDBCStoreProvider createProvider(
            JDBCStoreParameters parameters,
            DataStoreProviderServices providerServices
    ) throws InitializeException;

    public JDBCServerExplorer createServerExplorer(
            JDBCServerExplorerParameters parameters, 
            DataServerExplorerProviderServices providerServices
    ) throws InitializeException;

    public SRSSolver getSRSSolver();
    
    public JDBCNewStoreParameters createNewStoreParameters();

    public JDBCStoreParameters createOpenStoreParameters();
    
    public JDBCServerExplorerParameters createServerExplorerParameters();

    public String getSourceId(JDBCStoreParameters parameters);

    public boolean isThreadSafe();
    
    public String[] replaceForeingValueFunction(SQLBuilder sqlbuilder, FeatureType type);
}
