package org.gvsig.fmap.dal.store.jdbc2;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Iterator;
import java.util.List;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

@SuppressWarnings("UseSpecificCatch")
public class JDBCUtils {

    private static final Logger LOGGER = LoggerFactory.getLogger(JDBCUtils.class);
    
    private JDBCUtils() {

    }

    public static ResultSet executeQuery(Statement st, String sql) throws SQLException {
        LOGGER.debug("execute query SQL: " + sql);
        try {
            ResultSet rs = st.executeQuery(sql);
            return rs;
        } catch(Exception ex) {
            LOGGER.warn("execute SQL: " + sql, ex);
            throw ex;
        }
    }

    public static void execute(Statement st, String sql) throws SQLException {
        LOGGER.debug("execute SQL: " + sql);
        try {
            st.execute(sql);
        } catch(Exception ex) {
            LOGGER.warn("execute SQL: " + sql, ex);
            throw ex;
        }
    }

    public static void execute(Connection connection, String sql) throws SQLException {
        LOGGER.debug("execute SQL: " + sql);
        Statement st = connection.createStatement();
        try {
            st.execute(sql);
        } catch(Exception ex) {
            LOGGER.warn("execute SQL: " + sql, ex);
            throw ex;
        }
    }

    public static void execute(Connection connection, List<String> sqls) throws SQLException {
        Statement st = null;
        String sql = null;
        try {
            st = connection.createStatement();
            Iterator<String> it = sqls.iterator();
            while( it.hasNext() ) {
                sql = it.next();
                LOGGER.debug("execute SQL: " + sql);
                st.execute(sql);
            }
        } catch (SQLException ex) {
            LOGGER.warn("execute SQL: " + sql, ex);
            throw ex;
        } finally {
            JDBCUtils.closeQuietly(st);
        }       
    }

    public static ResultSet executeQuery(PreparedStatement st, String sql) throws SQLException {
        LOGGER.debug("execute query SQL: " + sql);
        try {
            ResultSet rs = st.executeQuery();
            return rs;
        } catch(Exception ex) {
            LOGGER.warn("execute SQL: " + sql, ex);
            throw ex;
        }
    }

    public static void execute(PreparedStatement st, String sql) throws SQLException {
        LOGGER.debug("execute SQL: " + sql);
        try {
            st.execute();
        } catch(Exception ex) {
            LOGGER.warn("execute SQL: " + sql, ex);
            throw ex;
        }
    }

    public static int executeUpdate(PreparedStatement st, String sql) throws SQLException {
        LOGGER.debug("execute update SQL: "+ sql);
        try {
            return st.executeUpdate();
        } catch(Exception ex) {
            LOGGER.warn("execute SQL: " + sql, ex);
            throw ex;
        }
    }

    public static void closeQuietly(Statement obj) {
        if (obj == null) {
            return;
        }
        try {
            obj.close();
        } catch (Exception e) {
            LOGGER.warn("Problems closing " + obj.getClass().getSimpleName() + " '" + obj.toString() + "'.", e);
        }
    }

    public static void closeQuietly(ResultSet obj) {
        if (obj == null) {
            return;
        }
        try {
            obj.close();
        } catch (Exception e) {
            LOGGER.warn("Problems closing " + obj.getClass().getSimpleName() + " '" + obj.toString() + "'.", e);
        }
    }

    public static void closeQuietly(Connection conn) {
        if (conn == null) {
            return;
        }
        LOGGER.debug("Closing connection "+ conn.hashCode() + " ("+conn.toString()+")");
        try {
            conn.close();
        } catch (Exception e) {
            LOGGER.warn("Problems closing " + conn.getClass().getSimpleName() + " '" + conn.toString() + "'.", e);
        }
    }
    
    public static void closeQuietly(AutoCloseable obj) {
        if (obj == null) {
            return;
        }
        try {
            LOGGER.debug("Closing " + obj.getClass().getSimpleName() + " "+ obj.hashCode() + " ("+obj.toString()+")");
            obj.close();
        } catch (Exception e) {
            LOGGER.warn("Problems closing " + obj.getClass().getSimpleName() + " '" + obj.toString() + "'.", e);
        }
    }
}
