/**
 * gvSIG. Desktop Geographic Information System.
 *
 * Copyright (C) 2007-2013 gvSIG Association.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * For any additional information, do not hesitate to contact us at info AT
 * gvsig.com, or visit our website www.gvsig.com.
 */
package org.gvsig.fmap.dal.store.jdbc2.impl;

import java.util.ArrayList;
import java.util.List;
import java.util.NoSuchElementException;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang3.BooleanUtils;
import org.gvsig.fmap.dal.exception.DataException;
import org.gvsig.fmap.dal.feature.FeatureQuery;
import org.gvsig.fmap.dal.feature.FeatureType;

import org.gvsig.fmap.dal.feature.spi.AbstractFeatureSetProvider;
import org.gvsig.fmap.dal.feature.spi.AbstractFeatureStoreProvider;
import org.gvsig.fmap.dal.store.jdbc.JDBCStoreParameters;
import org.gvsig.fmap.dal.store.jdbc2.JDBCHelper;
import org.gvsig.fmap.dal.store.jdbc2.JDBCStoreProvider;
import org.gvsig.fmap.dal.store.jdbc2.JDBCUtils;
import org.gvsig.fmap.dal.store.jdbc2.OperationsFactory;
import org.gvsig.fmap.dal.store.jdbc2.ResulSetControler;
import org.gvsig.fmap.dal.store.jdbc2.ResulSetControler.ResultSetEntry;
import org.gvsig.fmap.dal.store.jdbc2.spi.operations.TableIsEmptyOperation;
import org.gvsig.fmap.dal.store.jdbc2.spi.operations.ResultSetForSetProviderOperation;
import org.gvsig.fmap.dal.store.jdbc2.spi.operations.CountOperation;
import org.gvsig.tools.evaluator.Evaluator;
import org.gvsig.tools.exception.BaseException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class JDBCSetProvider extends AbstractFeatureSetProvider {

    final static protected Logger LOGGER = LoggerFactory.getLogger(JDBCSetProvider.class);

    private class EmptyJDBCIterator extends JDBCIterator {

        protected EmptyJDBCIterator(JDBCStoreProvider store) throws DataException {
            super(store, null, null, null);
        }

        @Override
        protected boolean internalHasNext() {
            return false;
        }

        @Override
        protected Object internalNext() {
            throw new NoSuchElementException();
        }

        @Override
        protected void doDispose() throws BaseException {
            // nothing to do
        }

    }

    
    protected Long size = null;
    protected Boolean isEmpty = null;
    protected List<ResultSetEntry> resultSets;
    private int defaultFetchSize = 1000;

    public JDBCSetProvider(
            AbstractFeatureStoreProvider store,
            JDBCHelper helper,
            FeatureQuery query,
            FeatureType featureType
    ) throws DataException {
        super(store, query, featureType);
        this.resultSets = new ArrayList<>();
    }

    /**
     * @return the defaultFetchSize
     */
    public int getDefaultFetchSize() {
        return defaultFetchSize;
    }

    /**
     * @param defaultFetchSize the defaultFetchSize to set
     */
    public void setDefaultFetchSize(int defaultFetchSize) {
        this.defaultFetchSize = defaultFetchSize;
    }

//    @Override
//    protected JDBCStoreProvider getStore() {
//        return (JDBCStoreProvider) super.getStore();
//    }

    protected JDBCStoreProvider getJDBCStore() {
//        Corregido el problema en AbstractFeatureSetProvider, de que getStore
//        debe ser un FeatureStoreProvider y no un AbstractFeatureStoreProvider
//        este metodo debe desaparecer y quedar el getStore de arriba que esta
//          comentarizado
        return (JDBCStoreProvider) super.getStore();
    }    
    
    protected JDBCHelper getHelper() {
        return this.getJDBCStore().getHelper();
    }
    
    protected OperationsFactory getOperations() {
        return this.getHelper().getOperations();
    }

    @Override
    public boolean canIterateFromIndex() {
        return this.getHelper().supportOffsetInSelect();
    }
    
    @Override
    public boolean canFilter() {
        return this.getHelper().supportFilter(this.getFeatureType(), getQuery().getFilter());
    }

    @Override
    public boolean canOrder() {
        return this.getHelper().supportOrder(this.getFeatureType(), getQuery().getOrder());
    }
    
    public ResulSetControler getResulSetControler() {
        return this.getHelper().getResulSetControler();
    }

    @Override
    protected void doDispose() throws BaseException {
        if( ! CollectionUtils.isEmpty(this.resultSets) ) {
            for( ResultSetEntry resulset : this.resultSets ) {
               JDBCUtils.closeQuietly(resulset);
            }
        }
        size = null;
        isEmpty = null;
    }

    protected int getFetchSize() {
        long pageSize = -1;
        if (getQuery() != null) {
            pageSize = getQuery().getPageSize();
            pageSize = pageSize > Integer.MAX_VALUE ? Integer.MAX_VALUE : pageSize;
        }
        return (pageSize > 0 ? (int) pageSize : defaultFetchSize);
    }

    @Override
    public long getSize() throws DataException {
        if (size == null) {
            JDBCStoreParameters params = this.getJDBCStore().getParameters();
            String filtersql = null;
            Evaluator filter = this.getQuery().getFilter();
            if( filter != null ) {
                filtersql = filter.getSQL();
            }
            CountOperation selectCount = this.getOperations().createCount(
                    this.getFeatureType(),
                    this.getOperations().createTableReference(params),
                    params.getBaseFilter(), 
                    filtersql
            );
            size = (Long) selectCount.perform();              
        }
        return size;
    }

    @Override
    public boolean isEmpty() throws DataException {
        if (this.isEmpty == null) {
            if (this.size == null) {
                JDBCStoreParameters params = this.getJDBCStore().getParameters();
                String filtersql = null;
                Evaluator filter = this.getQuery().getFilter();
                if( filter != null ) {
                    filtersql = filter.getSQL();
                }
                TableIsEmptyOperation isEmpty_ = this.getOperations().createTableIsEmpty(
                        this.getFeatureType(),
                        this.getOperations().createTableReference(params),
                        params.getBaseFilter(), 
                        filtersql
                );
                this.isEmpty = (Boolean) isEmpty_.perform();   
            } else {
                this.isEmpty = (this.size < 1);
            }
        }
        return isEmpty;
    }


    protected JDBCIterator createFastIterator(long index) throws DataException {
        return createFastIterator(index, 0);
    }
    
    @Override
    protected JDBCIterator createFastIterator(long index, long elements) throws DataException {
        if( BooleanUtils.isTrue(isEmpty) ) {
            return new EmptyJDBCIterator(this.getJDBCStore());
        }
        
        JDBCStoreParameters params = this.getJDBCStore().getParameters();
        FeatureType storeType = this.getStore()
                .getStoreServices()
		.getProviderFeatureType(this.getFeatureType().getId());
        ResultSetForSetProviderOperation createResultSet 
            = getOperations().createResultSetForSetProvider(
                    this.getOperations().createTableReference(params),
                    params.getBaseFilter(), 
                    params.getBaseOrder(), 
                    this.getQuery(), 
                    storeType, 
                    this.getFeatureType(), 
                    elements, 
                    index, 
                    this.getDefaultFetchSize()
            );

        ResultSetEntry resultSetEntry = (ResultSetEntry) createResultSet.perform();

        this.resultSets.add(resultSetEntry);
        return new JDBCFastIterator(
                this.getJDBCStore(), 
                this, 
                this.getFeatureType(),
                resultSetEntry
        );
    }

    @Override
    protected JDBCIterator createIterator(long index) throws DataException {
        return createIterator(index, -1);
    }
    
    @Override
    protected JDBCIterator createIterator(long index, long elements) throws DataException {
        if( BooleanUtils.isTrue(isEmpty) ) {
            return new EmptyJDBCIterator(this.getJDBCStore());
        }
        JDBCStoreParameters params = this.getJDBCStore().getParameters();
        FeatureType storeType = this.getStore()
                .getStoreServices()
		.getProviderFeatureType(this.getFeatureType().getId());
        ResultSetForSetProviderOperation createResultSet 
            = getOperations().createResultSetForSetProvider(
                    this.getOperations().createTableReference(params),
                    params.getBaseFilter(), 
                    params.getBaseOrder(), 
                    this.getQuery(), 
                    storeType, 
                    this.getFeatureType(), 
                    elements, 
                    index, 
                    this.getDefaultFetchSize()
            );

        ResultSetEntry resultSetEntry = (ResultSetEntry) createResultSet.perform();
        this.resultSets.add(resultSetEntry);
        return new JDBCIterator(
                this.getJDBCStore(), 
                this, 
                this.getFeatureType(),
                resultSetEntry
        );
    }

}
