package org.gvsig.fmap.dal.store.jdbc2.spi;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import org.cresques.cts.IProjection;
import org.gvsig.expressionevaluator.ExpressionBuilder.GeometrySupportType;
import org.gvsig.expressionevaluator.ExpressionBuilder.Parameter;
import org.gvsig.fmap.dal.feature.FeatureAttributeDescriptor;
import org.gvsig.fmap.dal.feature.FeatureReference;
import org.gvsig.fmap.dal.feature.FeatureType;
import org.gvsig.fmap.dal.feature.spi.SQLBuilderBase;
import org.gvsig.fmap.dal.feature.spi.FeatureProvider;
import org.gvsig.fmap.dal.feature.spi.FeatureReferenceProviderServices;
import org.gvsig.fmap.dal.store.jdbc2.JDBCHelper;
import org.gvsig.fmap.geom.Geometry;
import org.gvsig.fmap.geom.GeometryLocator;
import org.gvsig.fmap.geom.GeometryManager;
import org.gvsig.fmap.geom.aggregate.MultiLine;
import org.gvsig.fmap.geom.aggregate.MultiPoint;
import org.gvsig.fmap.geom.aggregate.MultiPolygon;
import org.gvsig.fmap.geom.exception.CreateGeometryException;
import org.gvsig.fmap.geom.primitive.Primitive;
import org.gvsig.fmap.geom.type.GeometryType;
import org.gvsig.tools.dispose.Disposable;

@SuppressWarnings("UseSpecificCatch")
public class JDBCSQLBuilderBase extends SQLBuilderBase {

    public static final String PROP_FEATURE_TYPE = "FeatureType";
    public static final String PROP_TABLE = "Table";
    
    private GeometryManager geometryManager = null;
    protected final JDBCHelper helper;
    
    public JDBCSQLBuilderBase(JDBCHelper helper) {
        super();
        this.helper = helper;
    }
    
    public JDBCHelper getHelper() {
        return helper;
    }
    
    protected GeometryManager getGeometryManager() {
        if (this.geometryManager == null) {
            this.geometryManager = GeometryLocator.getGeometryManager();
        }
        return this.geometryManager;
    }
    
    @Override
    public Object srs_id(IProjection projection) {
        Connection conn = null;
        try {
            conn = this.helper.getConnection();
            SRSSolver solver = this.helper.getSRSSolver();
            Object srscode = solver.getDatabaseCode(conn, projection);
//            logger.debug("database code srs {}, type {}, srssolver {}.", 
//                new Object[] { 
//                    srscode, 
//                    srscode==null? "null":srscode.getClass().getSimpleName(), 
//                    solver
//                }
//            );
            return srscode;
        } catch (Exception ex) {
            throw new RuntimeException("Can't locate database code for SRS '"+projection.getAbrev()+"'.");
        } finally {
            this.helper.closeConnectionQuietly(conn);
        }
    }
    
    public void setParameters(PreparedStatement st) {
        try {
            int columnIndex = 1;
            for (Parameter parameter : this.parameters()) {
                st.setObject(columnIndex++, parameter.value());
            }
        } catch (Exception ex) {
            String p = "unknow";
            try {
                p =  this.parameters().toString();
            } catch (Exception ex2) {
                // Do nothing
            }
            throw new RuntimeException("Can't set parameters to prepared statement from parameters (" + p + ")", ex);
        }
    }

    public Disposable setParameters(PreparedStatement st, FeatureProvider feature) {
        try {
            FeatureType type = feature.getType();
            List<Object> values = new ArrayList<>();
            Object value;
            for (Parameter parameter : this.parameters()) {
                if (parameter.is_constant()) {
                    value = parameter.value();
                    values.add(value);
                } else {
                    String name = parameter.name();
                    value = feature.get(name);
                    FeatureAttributeDescriptor attrDesc = type.getAttributeDescriptor(name);
                    switch( attrDesc.getType() ) {
                    case org.gvsig.fmap.dal.DataTypes.DATE:
                        if( value == null ) {
                            values.add(null);
                        } else {
                            values.add(new java.sql.Date(((Date)value).getTime()));
                        }
                        break;
                    case org.gvsig.fmap.dal.DataTypes.GEOMETRY:
                        Geometry geom = this.forceGeometryType(
                            attrDesc.getGeomType(),
                            (Geometry)value
                        );
                        values.add(geom);
                        break;
                    default:
                        values.add(value);
                        break;
                    }
                }
            }
            return this.setStatementParameters(st, values, this.geometry_support_type());
        } catch (Exception ex) {
            String f = "unknow";
            try {
                f = feature.toString();
            } catch (Exception ex2) {
                // Do nothing
            }
            throw new RuntimeException("Can't set parameters to prepared statement from the feature (" + f + ")", ex);
        }
    }
    
    protected Geometry forceGeometryType(GeometryType geomtype, Geometry geom) throws CreateGeometryException {
        if( geom == null ) {
            return null;
        }
        switch( geomtype.getType() ) {
        case Geometry.TYPES.MULTIPOLYGON:
            if( geom.getType()==Geometry.TYPES.POLYGON ) {
                MultiPolygon x = getGeometryManager().createMultiPolygon(geomtype.getSubType());
                x.addPrimitive((Primitive) geom);
                geom = x;
            }
            break;
        case Geometry.TYPES.MULTILINE:
            if( geom.getType()==Geometry.TYPES.LINE ) {
                MultiLine x = getGeometryManager().createMultiLine(geomtype.getSubType());
                x.addPrimitive((Primitive) geom);
                geom = x;
            }
            break;
        case Geometry.TYPES.MULTIPOINT:
            if( geom.getType()==Geometry.TYPES.POINT ) {
                MultiLine x = getGeometryManager().createMultiLine(geomtype.getSubType());
                x.addPrimitive((Primitive) geom);
                geom = x;
            }
            break;
        case Geometry.TYPES.POLYGON:
            if( geom.getType()==Geometry.TYPES.MULTIPOLYGON ) {
                MultiPolygon x = (MultiPolygon) geom;
                if( x.getPrimitivesNumber()==1 ) {
                    geom = x.getPrimitiveAt(0);
                }
            }
            break;
        case Geometry.TYPES.LINE:
            if( geom.getType()==Geometry.TYPES.MULTILINE ) {
                MultiLine x = (MultiLine) geom;
                if( x.getPrimitivesNumber()==1 ) {
                    geom = x.getPrimitiveAt(0);
                }
            }
            break;
        case Geometry.TYPES.POINT:
            if( geom.getType()==Geometry.TYPES.MULTIPOINT ) {
                MultiPoint x = (MultiPoint) geom;
                if( x.getPrimitivesNumber()==1 ) {
                    geom = x.getPrimitiveAt(0);
                }
            }
        }
        return geom;
    }
    
    public Disposable setParameters(PreparedStatement st, FeatureReference reference) {
        try {
            
            List<Object> values = new ArrayList<>();
            for (Parameter parameter : this.parameters()) {
                if (parameter.is_constant()) {
                    values.add(parameter.value());
                } else {
                    String name = parameter.name();
                    values.add(((FeatureReferenceProviderServices)reference).getKeyValue(name));
                }
            }
            return this.setStatementParameters(st, values, this.geometry_support_type());
        } catch (Exception ex) {
            String f = "unknow";
            try {
                f = reference.toString();
            } catch (Exception ex2) {
                // Do nothing
            }
            throw new RuntimeException("Can't set parameters to prepared statement from the feature (" + f + ")", ex);
        }
    }

    public Disposable setStatementParameters(
        PreparedStatement st, 
        List values, 
        GeometrySupportType geometrySupportType) throws SQLException {
        
        if (values == null) {
            return new Disposable() {
                @Override
                public void dispose() {
                }
            };
        }
        if( LOGGER.isDebugEnabled() ) {
            StringBuilder debug = new StringBuilder();
            debug.append("st.set(");
            try {
                byte[] bytes;
                int columnIndex = 1;
                for (Object value : values) {
                    if (value instanceof Geometry) {
                        switch(geometrySupportType) {
                            case WKT:
                                value = ((Geometry) value).convertToWKT();
                                debug.append("/*");
                                debug.append(columnIndex);
                                debug.append("*/ ");
                                debug.append(as_string(value));
                                debug.append(", ");
                                break;
                            case NATIVE:
                            case WKB: 
                                bytes = ((Geometry) value).convertToWKB();
                                debug.append("/*");
                                debug.append(columnIndex);
                                debug.append("*/ ");
                                debug.append(as_string(bytes));
                                debug.append(", ");
                                break;
                            case EWKB:
                                bytes = ((Geometry) value).convertToEWKB();
                                debug.append("/*");
                                debug.append(columnIndex);
                                debug.append("*/ ");
                                debug.append(as_string(bytes));
                                debug.append(", ");
                                break;
                        }
                    } else {
                        debug.append("/*");
                        debug.append(columnIndex);
                        debug.append("*/ ");
                        if( value instanceof String ) {
                            debug.append(as_string(value));
                        } else if( value instanceof Boolean ) {
                            debug.append( ((Boolean)value)? constant_true:constant_false );
                        } else {
                            debug.append(value);
                        }
                        debug.append(", ");
                    }
                    columnIndex++;
                }
                debug.append(")");
                LOGGER.debug(debug.toString());
            } catch(Exception ex) {
            }        
        }
        try {
            byte[] bytes;
            int columnIndex = 1;
            for (Object value : values) {
                if (value instanceof Geometry) {
                    switch(geometrySupportType) {
                        case WKT:
                            value = ((Geometry) value).convertToWKT();
                            st.setObject(columnIndex, value);
                            break;
                        case NATIVE:
                        case WKB: 
                            bytes = ((Geometry) value).convertToWKB();
                            st.setBytes(columnIndex, bytes);
                            break;
                        case EWKB:
                            bytes = ((Geometry) value).convertToEWKB();
                            st.setBytes(columnIndex, bytes);
                            break;
                    }
                } else {
                    st.setObject(columnIndex, value);
                }
                columnIndex++;
            }
            return new Disposable() {
                @Override
                public void dispose() {
                }
            };
        } catch(Exception ex) {
            throw new SQLException("Can't set values for the prepared statement.", ex);
        }        
    }
}
