package org.gvsig.fmap.dal.store.jdbc2.spi.operations;

import java.sql.Connection;
import java.sql.SQLException;
import org.gvsig.fmap.dal.resource.exception.AccessResourceException;
import org.gvsig.fmap.dal.store.jdbc.exception.JDBCTransactionRollbackException;
import org.gvsig.fmap.dal.store.jdbc2.JDBCHelper;
import org.gvsig.fmap.dal.store.jdbc2.spi.JDBCSQLBuilderBase;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

@SuppressWarnings("UseSpecificCatch")
public abstract class AbstractConnectionOperation implements ConnectionOperation {

    final static protected Logger LOGGER = LoggerFactory.getLogger(AbstractConnectionOperation.class);

    final protected JDBCHelper helper;
    
    protected Connection conn = null;

    public AbstractConnectionOperation(JDBCHelper helper) {
        this.helper = helper;
    }

    @Override
    public boolean continueTransactionAllowed() {
        return false;
    }

    @Override
    public boolean needTransaction() {
        return true;
    }

    protected JDBCSQLBuilderBase createSQLBuilder() {
        return this.helper.createSQLBuilder();
    }

    @Override
    public Object perform()  {
        try {
            if( this.helper.isThreadSafe() ) {
                return this.perform_operation();
            } 
            synchronized(AbstractConnectionOperation.class) {
                return this.perform_operation();
            }
        } catch (RuntimeException ex) {
            throw ex;
        } catch (Exception ex) {
            throw new RuntimeException("Can't perform operation '"+this.getClass().getSimpleName()+"'.",ex);
        }
    }

    protected Connection getConnection() throws AccessResourceException {
        if( this.conn == null ) {
            this.conn = this.helper.getConnection();
        }
        return this.conn;
    }
    
    protected Object perform_operation() throws Exception {
        Object result = null;
        LOGGER.debug("preparing execution of "+this.getClass().getSimpleName()+".");
        if( needTransaction() ) {
            try {
                if( this.helper.allowNestedOperations() ) {
                    boolean commitOnFinish = true;
                    if (!this.getConnection().getAutoCommit()) {
                        // Casi seguro que es una operacion anidada, asi que no
                        // terminamos la transaccion.
                        commitOnFinish = false;
                    } 
                    this.getConnection().setAutoCommit(false);
                    try {
                        LOGGER.debug("Excuting operation {}.", this.getClass().getSimpleName());
                        result = perform(this.getConnection());
                    } catch (Exception ex) {
                        try {
                            this.getConnection().rollback();
                        } catch (Exception e1) {
                            throw new JDBCTransactionRollbackException(e1, ex);
                        }
                        throw new RuntimeException("Can't perform operation '"+this.getClass().getSimpleName()+"'.",ex);
                    }
                    if( commitOnFinish ) {
                        this.getConnection().commit();
                        this.getConnection().setAutoCommit(true);
                    }
                } else {
                    if (!this.getConnection().getAutoCommit()) {
                        // Casi seguro que es una transaccion anidada, si no
                        // esta permitido, petamos.
                        if (!continueTransactionAllowed()) {
                            throw new SQLException("nested operations not allowed.");
                        }
                    }
                    this.getConnection().setAutoCommit(false);
                    try {
                        LOGGER.debug("Excuting operation {}.", this.getClass().getSimpleName());
                        result = perform(this.getConnection());
                    } catch (Exception ex) {
                        try {
                            this.getConnection().rollback();
                        } catch (Exception e1) {
                            throw new JDBCTransactionRollbackException(e1, ex);
                        }
                        throw new RuntimeException("Can't perform operation '"+this.getClass().getSimpleName()+"'.",ex);
                    }
                    this.getConnection().commit();
                }
            } finally {
                helper.closeConnection(this.getConnection());
                conn = null;
            }
        } else {
            try {
                LOGGER.debug("Excuting operation {}.", this.getClass().getSimpleName());
                result = perform(this.getConnection());
            } catch (Exception ex) {
                throw new RuntimeException("Can't perform operation '"+this.getClass().getSimpleName()+"'.",ex);
            } finally {
                helper.closeConnection(this.getConnection());
                conn = null;
            }
        }
        return result;
    }
}
