
package org.gvsig.fmap.dal.store.jdbc2.spi.operations;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import org.gvsig.expressionevaluator.ExpressionBuilder;
import org.gvsig.fmap.dal.DataTypes;
import org.gvsig.fmap.dal.exception.DataException;
import org.gvsig.fmap.dal.feature.FeatureAttributeDescriptor;
import org.gvsig.fmap.dal.feature.FeatureType;
import org.gvsig.fmap.dal.feature.exception.AlreadyEditingException;
import org.gvsig.fmap.dal.feature.spi.FeatureProvider;
import org.gvsig.fmap.dal.store.jdbc.exception.JDBCPreparingSQLException;
import org.gvsig.fmap.dal.store.jdbc2.JDBCHelper;
import org.gvsig.fmap.dal.store.jdbc2.JDBCUtils;
import org.gvsig.fmap.dal.store.jdbc2.OperationsFactory.TableReference;
import org.gvsig.fmap.dal.store.jdbc2.spi.JDBCSQLBuilderBase;
import org.gvsig.tools.dispose.Disposable;
import org.gvsig.tools.dispose.DisposeUtils;


public class AppendOperation {

    protected Connection connection = null;
    protected final JDBCHelper helper;
    protected final TableReference table;
    protected final FeatureType type;
    
    protected JDBCSQLBuilderBase sqlbuilder = null;
    protected ExpressionBuilder expbuilder;

    protected PreparedStatement preparedStatement;
    protected String sql;
    
    public AppendOperation(
            JDBCHelper helper, 
            TableReference table, 
            FeatureType type
        ) {
        this.helper = helper;
        this.table = table;
        this.type = type;
    }
    
    public void begin() throws DataException {
        if (this.sqlbuilder != null) {
            throw new AlreadyEditingException(this.helper.getSourceId());
        }

        try {
            this.connection = this.helper.getConnectionWritable();
            
            this.sqlbuilder = this.helper.createSQLBuilder();
            this.expbuilder = this.sqlbuilder.expression();

            this.sqlbuilder.insert().table()
                    .database(this.table.getDatabase())
                    .schema(this.table.getSchema())
                    .name(this.table.getTable()
            );
            for (FeatureAttributeDescriptor attr : type) {
                if( attr.isAutomatic() || attr.isComputed() ) {
                    continue;
                }
                if (attr.getType() == DataTypes.GEOMETRY) {
                    this.sqlbuilder.insert().column().name(attr.getName()).with_value( 
                        expbuilder.parameter(attr.getName()).as_geometry_variable().srs( 
                                expbuilder.parameter().value(attr.getSRS()) 
                        ) 
                    );
                } else {
                    this.sqlbuilder.insert().column().name(attr.getName()).with_value(
                        expbuilder.parameter(attr.getName()).as_variable()
                    );
                }
            }

            this.sql = this.sqlbuilder.insert().toString();
            this.preparedStatement = this.connection.prepareStatement(sql);
            this.connection.setAutoCommit(false);
            
        } catch (SQLException ex) {
            throw new JDBCPreparingSQLException(this.sqlbuilder.toString(),ex);
        }

    }
    
    protected void clean() {
        JDBCUtils.closeQuietly(this.preparedStatement);
        this.helper.closeConnection(this.connection);
        this.connection = null;
        this.preparedStatement = null;
        this.sqlbuilder = null;
        this.sql = null;        
    }
    
    public void end() {
        try {
            this.connection.commit();
        } catch (SQLException ex) {
            try {
                this.connection.rollback();
            } catch (SQLException ex1) {
            }
            throw new RuntimeException("Can't commit transaction", ex);
        } finally {
            clean();
        }
    }
    
    public void abort() {        
        try {
            this.connection.rollback();
        } catch (SQLException ex) {
        }
        clean();
    }
    
    public void append(FeatureProvider feature) throws DataException {
        int n;
        Disposable paramsDisposer = null;
        try {
            paramsDisposer = this.sqlbuilder.setParameters(this.preparedStatement, feature);
            n = JDBCUtils.executeUpdate(this.preparedStatement,this.sql);
        } catch(Exception ex) {
            throw new RuntimeException("Can't insert feature.", ex);
        } finally {
            DisposeUtils.disposeQuietly(paramsDisposer);
        }
        if( n<1 ) {
            throw new RuntimeException("Can't insert feature (n="+n+").");
        }
    }
}
