package org.gvsig.fmap.dal.store.jdbc2.spi.operations;

import java.util.Iterator;
import java.util.List;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.Pair;
import org.cresques.cts.IProjection;
import org.gvsig.fmap.dal.exception.DataException;
import org.gvsig.fmap.dal.feature.EditableFeatureType;
import org.gvsig.fmap.dal.feature.FeatureQuery;
import org.gvsig.fmap.dal.feature.FeatureType;
import org.gvsig.fmap.dal.SQLBuilder;
import org.gvsig.fmap.dal.feature.spi.FeatureReferenceProviderServices;
import org.gvsig.fmap.dal.store.jdbc.JDBCNewStoreParameters;
import org.gvsig.fmap.dal.store.jdbc.JDBCStoreParameters;
import org.gvsig.fmap.dal.store.jdbc2.JDBCHelper;
import org.gvsig.fmap.dal.store.jdbc2.OperationsFactory;
import org.gvsig.fmap.geom.primitive.Envelope;

public class OperationsFactoryBase implements OperationsFactory {

    public static class DefaultTableReference implements TableReference {

        private final String database;
        private final String schema;
        private final String table;
        private final String subquery;

        public DefaultTableReference(String database, String schema, String table, String subquery) {
            this.database = StringUtils.defaultIfBlank(database, null);
            this.schema = StringUtils.defaultIfBlank(schema, null);
            this.table = StringUtils.defaultIfBlank(table, null);
            this.subquery = StringUtils.defaultIfBlank(subquery, null);
        }
        
        @Override
        public String getDatabase() {
            return this.database;
        }

        @Override
        public String getSchema() {
            return this.schema;
        }

        @Override
        public String getTable() {
            return this.table;
        }

        @Override
        public String getSubquery() {
            return this.subquery;
        }

        @Override
        public boolean hasDatabase() {
            return this.database != null;
        }
        
        @Override
        public boolean hasSchema() {
            return this.schema != null;
        }
        
        @Override
        public boolean hasTable() {
            return this.table != null;
        }
        
        @Override
        public boolean hasSubquery() {
            return this.subquery != null;
        }

        @Override
        public String toString() {
            StringBuilder builder = new StringBuilder();
            if( this.hasDatabase() ) {
                builder.append(this.database);
                builder.append(".");
            }
            if( this.hasSchema()) {
                builder.append(this.schema);
                builder.append(".");
            }
            builder.append(this.table);
            if( this.hasSubquery() ) {
                builder.append("[");
                builder.append(this.subquery);
                builder.append("]");
            }
            return builder.toString();
        }
        
        
    }
    
    protected final JDBCHelper helper;

    public OperationsFactoryBase(JDBCHelper helper) {
        this.helper = helper;
    }

    @Override
    public TableReference createTableReference(JDBCStoreParameters params) {
        TableReference t = new DefaultTableReference(
                params.getDBName(),
                params.getSchema(),
                params.getTable(),
                params.getSQL()
        );
        return t;
    }

    @Override
    public TableReference createTableReference(JDBCNewStoreParameters params) {
        TableReference t = new DefaultTableReference(
                params.getDBName(),
                params.getSchema(),
                params.getTable(),
                null
        );
        return t;
    }

    @Override
    public TableReference createTableReference(String database, String schema, String table, String subquery) {
        TableReference t = new DefaultTableReference(database, schema, table, subquery);
        return t;
    }

    
    @Override
    public FetchFeatureTypeOperation createFetchFeatureType(
            EditableFeatureType type, 
            TableReference table, 
            List<String> primaryKeys, 
            String defaultGeometryField, 
            IProjection crs
        ) {
        return new FetchFeatureTypeOperation(helper, type, table, 
                primaryKeys, defaultGeometryField, crs);
    }
    
    @Override
    public FetchFeatureProviderByReferenceOperation createFetchFeatureProviderByReference(
            FeatureReferenceProviderServices reference, 
            FeatureType featureType, 
            TableReference table
        ) {
        return new FetchFeatureProviderByReferenceOperation(helper, reference, 
                featureType, table);
    }

    @Override
    public CalculateEnvelopeOfColumnOperation createCalculateEnvelopeOfColumn(
            FeatureType featureType,
            TableReference table, 
            String columnName, 
            String baseFilter, 
            Envelope workingArea, 
            IProjection crs
        ) {
        return new CalculateEnvelopeOfColumnOperation(helper, 
                featureType, table, columnName, baseFilter, workingArea, crs);
    }

    @Override
    public PerformChangesOperation createPerformChanges(
            TableReference table, 
            FeatureType type, 
            Iterator deleteds, 
            Iterator inserteds, 
            Iterator updateds, 
            Iterator featureTypesChanged
        ) {
        return new PerformChangesOperation(helper, table, type, 
                deleteds, inserteds, updateds, featureTypesChanged);
    }

    @Override
    public AppendOperation createAppend(
            TableReference table, 
            FeatureType type
        ) {
        return new AppendOperation(helper, table, type);
    }

    @Override
    public CountOperation createCount(
            FeatureType featureType,
            TableReference table, 
            String baseFilter, 
            String filter
        ) {
        return new CountOperation(helper, featureType, table, baseFilter, filter);
    }

    @Override
    public TableIsEmptyOperation createTableIsEmpty(
            FeatureType featureType,
            TableReference table,
            String baseFilter, 
            String filter
        ) {
        return new TableIsEmptyOperation(helper, featureType, table, baseFilter, filter);
    }

    @Override
    public ResultSetForSetProviderOperation createResultSetForSetProvider(
            TableReference table, 
            String baseFilter, 
            String baseOrder, 
            FeatureQuery query, 
            FeatureType storeType, 
            FeatureType setType, 
            long limit, 
            long offset, 
            int fetchSize
        ) {
        return new ResultSetForSetProviderOperation(helper, table, 
                baseFilter, baseOrder, query, storeType, setType, 
                limit, offset, fetchSize);
    }

    @Override
    public ListTablesOperation createListTables(
            int mode,
            JDBCStoreParameters baseParameters, 
            boolean informationTables
        ) {
        return new ListTablesOperation(
                helper, mode, baseParameters, informationTables );
    }
    
    @Override
    public DropTableOperation createDropTable( 
            TableReference tableName            
        ) {
        return new DropTableOperation(helper, tableName);
    }
    
    @Override
    public CreateTableOperation createTable(
            TableReference table, 
            FeatureType type, 
            List<Pair<String, SQLBuilder.Privilege>> userAndPrivileges, 
            List<String> additionalSQLs
        ) throws DataException {
        return new CreateTableOperation(
                helper, table, type, 
                userAndPrivileges, additionalSQLs
        );
    }

    @Override
    public CanCreateTablesOperation createCanCreateTables() {
        return new CanCreateTablesOperation(helper);
    }

    @Override
    public UpdateTableStatisticsOperation createUpdateTableStatistics(
            TableReference table
        ) {
        return new UpdateTableStatisticsOperation(helper, table);
    }
    
    @Override
    public CanModifyTableOperation createCanModifyTableOperation(
            TableReference table
        ) {
        return new CanModifyTableOperation(helper, table);
    }
    
    @Override
    public ExecuteOperation createExecute(String sql) {
        return new ExecuteOperation(helper, sql);
    }
    
}
