package org.gvsig.fmap.dal.store.jdbc2.spi.operations;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.gvsig.expressionevaluator.ExpressionBuilder;
import org.gvsig.expressionevaluator.ExpressionBuilder.Variable;
import org.gvsig.fmap.dal.DataTypes;
import org.gvsig.fmap.dal.exception.DataException;
import org.gvsig.fmap.dal.feature.FeatureAttributeDescriptor;
import org.gvsig.fmap.dal.feature.FeatureReference;
import org.gvsig.fmap.dal.feature.FeatureType;
import org.gvsig.fmap.dal.feature.spi.FeatureProvider;
import org.gvsig.fmap.dal.feature.spi.FeatureReferenceProviderServices;
import org.gvsig.fmap.dal.feature.spi.FeatureStoreProvider;
import org.gvsig.fmap.dal.store.jdbc2.JDBCHelper;
import org.gvsig.fmap.dal.store.jdbc.exception.JDBCExecuteSQLException;
import org.gvsig.fmap.dal.store.jdbc.exception.JDBCSQLException;
import org.gvsig.fmap.dal.store.jdbc.exception.JDBCUpdateWithoutChangesException;
import org.gvsig.fmap.dal.store.jdbc2.spi.JDBCSQLBuilderBase;
import org.gvsig.fmap.dal.store.jdbc2.JDBCUtils;
import org.gvsig.fmap.dal.store.jdbc2.OperationsFactory.TableReference;
import static org.gvsig.fmap.dal.store.jdbc2.spi.JDBCSQLBuilderBase.PROP_FEATURE_TYPE;
import static org.gvsig.fmap.dal.store.jdbc2.spi.JDBCSQLBuilderBase.PROP_TABLE;
import org.gvsig.tools.dispose.Disposable;
import org.gvsig.tools.dispose.DisposeUtils;

public class PerformChangesOperation extends AbstractConnectionWritableOperation {

    protected TableReference table;
    protected FeatureType featureType;
    protected Iterator<FeatureReferenceProviderServices> deleteds;
    protected Iterator<FeatureStoreProvider.FeatureTypeChanged> featureTypesChanged;
    protected Iterator<FeatureProvider> updateds;
    protected Iterator<FeatureProvider> inserteds;

    protected boolean typeChanged = false;
    
    public PerformChangesOperation(JDBCHelper helper) {
        this(helper, null, null, null, null, null, null);
    }

    public PerformChangesOperation(JDBCHelper helper,
            TableReference table,
            FeatureType featureType,
            Iterator<FeatureReferenceProviderServices> deleteds,
            Iterator<FeatureProvider> inserteds,
            Iterator<FeatureProvider> updateds,
            Iterator<FeatureStoreProvider.FeatureTypeChanged> featureTypesChanged) {
        super(helper);
        this.deleteds = deleteds;
        this.inserteds = inserteds;
        this.updateds = updateds;
        this.table = table;
        this.featureType = featureType;
        this.featureTypesChanged = featureTypesChanged;
    }

    public boolean isTypeChanged() {
        return typeChanged;
    }

    @Override
    public Object perform(Connection conn) throws DataException {
        if (featureTypesChanged.hasNext()) {
            FeatureStoreProvider.FeatureTypeChanged item = featureTypesChanged.next();
            this.performUpdateTable(conn, table, item.getSource(), item.getTarget());
            typeChanged = true;
        } else {
            typeChanged = false;
        }
        if (deleteds.hasNext()) {
            performDeletes(conn, table, featureType, deleteds);
        }
        if (updateds.hasNext()) {
            performUpdates(conn, table, featureType, updateds);
        }
        if (inserteds.hasNext()) {
            performInserts(conn, table, featureType, inserteds);
        }
        return true;
    }
    
    protected JDBCSQLBuilderBase buildDeleteSQL(
            TableReference table,
            FeatureType type
        ) {
        JDBCSQLBuilderBase sqlbuilder = createSQLBuilder();
        ExpressionBuilder expbuilder = sqlbuilder.expression();

        sqlbuilder.delete().table()
                .database(this.table.getDatabase())
                .schema(this.table.getSchema())
                .name(this.table.getTable());
        for (FeatureAttributeDescriptor attr : type) {
            if (attr.isPrimaryKey()) {
                sqlbuilder.delete().where().and(
                        expbuilder.eq(
                                expbuilder.column(attr.getName()),
                                expbuilder.parameter(attr.getName()).as_variable()
                        )
                );
            }
        }
        if (!sqlbuilder.delete().has_where() ) {
            throw new RuntimeException("Operation requires missing pk");
        }
        return sqlbuilder;
    }

    public void performDeletes(Connection conn,
            TableReference table,
            FeatureType type,
            Iterator<FeatureReferenceProviderServices> deleteds) throws DataException {

        JDBCSQLBuilderBase sqlbuilder = buildDeleteSQL(table, type);

        PreparedStatement st = null;
        Disposable paramsDisposer = null;
        sqlbuilder.setProperties(
                Variable.class, 
                PROP_TABLE, table
        );
        String sql = sqlbuilder.delete().toString();
        try {
            st = conn.prepareStatement(sql);
            while (deleteds.hasNext()) {
                FeatureReference reference = (FeatureReference) deleteds.next();
                paramsDisposer = sqlbuilder.setParameters(st, reference);
                int nAffected = JDBCUtils.executeUpdate(st,sql);
                if (nAffected == 0) {
                    throw new JDBCUpdateWithoutChangesException(
                            sqlbuilder.delete().toString(),
                            null
                    );
                }
                if (nAffected > 1) {
                    LOGGER.warn("Remove statement affectst to {} rows ( {} )",
                            nAffected, sql
                    );
                }
            }
        } catch (SQLException e) {
            throw new JDBCSQLException(e);
        } finally {
            JDBCUtils.closeQuietly(st);
            DisposeUtils.disposeQuietly(paramsDisposer);
        }
    }
    
    protected JDBCSQLBuilderBase buildInsertSQL(
            TableReference table,
            FeatureType type
        ) {
        JDBCSQLBuilderBase sqlbuilder = createSQLBuilder();
        ExpressionBuilder expbuilder = sqlbuilder.expression();

        sqlbuilder.insert().table()
                .database(this.table.getDatabase())
                .schema(this.table.getSchema())
                .name(this.table.getTable());
        for (FeatureAttributeDescriptor attr : type) {
            if( attr.isAutomatic() || attr.isComputed() ) {
                continue;
            }
            if (attr.getType() == DataTypes.GEOMETRY) {
                sqlbuilder.insert().column().name(attr.getName()).with_value(
                    expbuilder.parameter(attr.getName()).as_geometry_variable()
                        .srs(
                            expbuilder.parameter().value(
                                attr.getSRS()).as_constant()
                        )
                );
            } else {
                sqlbuilder.insert().column().name(attr.getName()).with_value(
                        expbuilder.parameter(attr.getName())
                );
            }
        }
        return sqlbuilder;
    }

    public void performInserts(Connection conn,
            TableReference table,
            FeatureType type,
            Iterator<FeatureProvider> inserteds) throws DataException {
        JDBCSQLBuilderBase sqlbuilder = buildInsertSQL(table, type);

        PreparedStatement st;
        Disposable paramsDisposer;
        sqlbuilder.setProperties(
                Variable.class, 
                PROP_FEATURE_TYPE, type,
                PROP_TABLE, table
        );
        String sql = sqlbuilder.insert().toString();
        try {
            st = conn.prepareStatement(sql);
            while (inserteds.hasNext()) {
                FeatureProvider feature = inserteds.next();
                paramsDisposer = sqlbuilder.setParameters(st, feature);
                try {
                    if (JDBCUtils.executeUpdate(st,sql) == 0) {
                        throw new JDBCExecuteSQLException(
                                sqlbuilder.insert().toString(),
                                null
                        );
                    }
                } finally {
                    DisposeUtils.disposeQuietly(paramsDisposer);
                }
            }
        } catch (JDBCExecuteSQLException ex) {
            throw ex;
        } catch (Exception ex) {
            throw new JDBCExecuteSQLException(sql,ex);
        }
    }

    protected JDBCSQLBuilderBase buildUpdateSQL(
            TableReference table,
            FeatureType type
        ) {
        JDBCSQLBuilderBase sqlbuilder = createSQLBuilder();
        ExpressionBuilder expbuilder = sqlbuilder.expression();

        sqlbuilder.update().table()
                .database(this.table.getDatabase())
                .schema(this.table.getSchema())
                .name(this.table.getTable());
        for (FeatureAttributeDescriptor attr : type) {
            if (attr.isPrimaryKey()) {
                sqlbuilder.update().where().and(
                        expbuilder.eq(
                                expbuilder.column(attr.getName()),
                                expbuilder.parameter(attr.getName()).as_variable()
                        )
                );
            } 
            if ( attr.isAutomatic() || attr.isReadOnly() || attr.isComputed() ) {
                continue;
            }
            if (attr.getType() == DataTypes.GEOMETRY) {
                sqlbuilder.update().column().name(attr.getName()).with_value(
                    expbuilder.parameter(attr.getName()).as_geometry_variable().srs(
                        expbuilder.parameter().value(attr.getSRS()).as_constant()
                    )
                );
            } else {
                sqlbuilder.update().column().name(attr.getName()).with_value(
                    expbuilder.parameter(attr.getName()).as_variable()
                );
            }
        }
        if (!sqlbuilder.update().has_where() ) {
            throw new RuntimeException("Operation requires missing pk");
        }
        return sqlbuilder;
    }
    
    public void performUpdates(Connection conn,
            TableReference table,
            FeatureType type,
            Iterator<FeatureProvider> updateds) throws DataException {

        JDBCSQLBuilderBase sqlbuilder = buildUpdateSQL(table, type);
        
        PreparedStatement st = null;
        Disposable paramsDisposer = null;
        sqlbuilder.setProperties(
                Variable.class, 
                PROP_FEATURE_TYPE, type,
                PROP_TABLE, table
        );
        String sql = sqlbuilder.update().toString();
        try {
            st = conn.prepareStatement(sql);
            while (updateds.hasNext()) {
                FeatureProvider featureProvider = (FeatureProvider) updateds.next();
                paramsDisposer = sqlbuilder.setParameters(st, featureProvider);
                if (JDBCUtils.executeUpdate(st,sql) == 0) {
                    throw new JDBCUpdateWithoutChangesException(sql,null);
                }
            }
        } catch (SQLException e) {
            throw new JDBCSQLException(e);
        } finally {
            JDBCUtils.closeQuietly(st);
            DisposeUtils.disposeQuietly(paramsDisposer);
        }
    }

    protected List<String> buildUpdateTableSQL(
            TableReference table,
            FeatureType original,
            FeatureType target
        ) {
        JDBCSQLBuilderBase sqlbuilder = createSQLBuilder();
        sqlbuilder.alter_table().table()
                .database(this.table.getDatabase())
                .schema(this.table.getSchema())
                .name(this.table.getTable());

        for (FeatureAttributeDescriptor attrOrgiginal : original) {
            FeatureAttributeDescriptor attrTarget = target.getAttributeDescriptor(
                    attrOrgiginal.getName()
            );
            if (attrTarget == null) {
                if( attrOrgiginal.isComputed() ) {
                    continue;
                }
                sqlbuilder.alter_table().drop_column(attrOrgiginal.getName());
            } else if( !this.areEquals(attrOrgiginal, attrTarget) ) {
                if( attrTarget.isComputed() ) {
                    continue;
                }
                if( attrTarget.getType()==DataTypes.GEOMETRY ) {
                    sqlbuilder.alter_table().alter_geometry_column(
                            attrTarget.getName(),
                            attrTarget.getGeomType().getType(),
                            attrTarget.getGeomType().getSubType(),
                            attrTarget.getSRS(),
                            attrTarget.isIndexed(),
                            attrTarget.allowNull()
                    );
                } else {
                    sqlbuilder.alter_table().alter_column(
                            attrTarget.getName(),
                            attrTarget.getType(),
                            attrTarget.getPrecision(),
                            attrTarget.getSize(),
                            attrTarget.isPrimaryKey(),
                            attrTarget.isIndexed(),
                            attrTarget.allowNull(),
                            attrTarget.isAutomatic(),
                            attrTarget.getDefaultValue()
                    );
                }
            }
        }
        for (FeatureAttributeDescriptor attrTarget : target) {
            if( attrTarget.isComputed() ) {
                continue;
            }
            if (original.getAttributeDescriptor(attrTarget.getName()) == null) {
                if( attrTarget.getType()==DataTypes.GEOMETRY ) {
                    sqlbuilder.alter_table().add_geometry_column(
                            attrTarget.getName(),
                            attrTarget.getGeomType().getType(),
                            attrTarget.getGeomType().getSubType(),
                            attrTarget.getSRS(),
                            attrTarget.isIndexed(),
                            attrTarget.allowNull()
                    );
                } else {
                    sqlbuilder.alter_table().add_column(
                            attrTarget.getName(),
                            attrTarget.getType(),
                            attrTarget.getPrecision(),
                            attrTarget.getSize(),
                            attrTarget.isPrimaryKey(),
                            attrTarget.isIndexed(),
                            attrTarget.allowNull(),
                            attrTarget.isAutomatic(),
                            attrTarget.getDefaultValue()
                    );
                }
            }
        }
        sqlbuilder.setProperties(
                Variable.class, 
                PROP_TABLE, table
        );
        return sqlbuilder.alter_table().toStrings();
    }
    
    protected boolean areEquals(FeatureAttributeDescriptor attr1, FeatureAttributeDescriptor attr2) {
        // No interesa si son o no iguales en general, solo si son iguales en lo 
        // que a los atributos usados para crear la columna de la tabla se refiere.
        if( !StringUtils.equals(attr1.getName(), attr2.getName()) ) {
            return false;
        }
        if( attr1.getType() != attr2.getType() ) {
            return false;
        }
        if( attr1.getPrecision() != attr2.getPrecision() ) {
            return false;
        }
        if( attr1.getSize() != attr2.getSize() ) {
            return false;
        }
        if( attr1.isPrimaryKey() != attr2.isPrimaryKey() ) {
            return false;
        }        
//        if( attr1.isIndexed() != attr2.isIndexed() ) {
//            return false;
//        }
        if( attr1.allowNull() != attr2.allowNull() ) {
            return false;
        }
        if( attr1.isAutomatic() != attr2.isAutomatic() ) {
            return false;
        }
        if( attr1.getDefaultValue() != attr2.getDefaultValue() ) {
            if( attr1.getDefaultValue()==null || attr2.getDefaultValue()==null) {
                return false;
            }
            if( !attr1.getDefaultValue().equals(attr2.getDefaultValue()) ) {
                return false;
            }
        }
        return true;
    }

    protected List<String> buildCreateIndexSQL(
            TableReference table,
            FeatureType original,
            FeatureType target
        ) {
        ArrayList<String> sqls = new ArrayList<>();
        
        for (FeatureAttributeDescriptor attrTarget : target) {
            boolean createIndex = false;
            if( attrTarget.isIndexed() ) {
                FeatureAttributeDescriptor attrOriginal = original.getAttributeDescriptor(attrTarget.getName());
                if ( attrOriginal == null) {
                    createIndex = true;
                } else {
                    if( attrOriginal.isIndexed() ) {
                        createIndex = false;
                    } else {
                        createIndex = true;
                    }
                }
            }
            if( createIndex ) {
                JDBCSQLBuilderBase sqlbuilder = createSQLBuilder();
                if( attrTarget.getType()==DataTypes.GEOMETRY ) {
                    sqlbuilder.create_index().spatial();
                }
                sqlbuilder.create_index().if_not_exist();
                sqlbuilder.create_index().name("idx_" + table + "_" + attrTarget.getName());
                sqlbuilder.create_index().column(attrTarget.getName());
                sqlbuilder.create_index().table()
                    .database(this.table.getDatabase())
                    .schema(this.table.getSchema())
                    .name(this.table.getTable());
                sqlbuilder.setProperties(
                        Variable.class, 
                        PROP_TABLE, table
                );
                sqls.addAll(sqlbuilder.create_index().toStrings());
            }
        }
        return sqls;
    }
    
    public void performUpdateTable(Connection conn,
            TableReference table,
            FeatureType original,
            FeatureType target) throws DataException {
        
        ArrayList<String> sqls = new ArrayList<>();
        
        sqls.addAll(buildUpdateTableSQL(table, original, target));
        sqls.addAll(buildCreateIndexSQL(table, original, target));

        if( !CollectionUtils.isEmpty(sqls) ) {
            Statement st = null;
            try {
                st = conn.createStatement();
                for (String sql : sqls) {
                    if( !StringUtils.isBlank(sql) ) {
                        JDBCUtils.execute(st, sql);
                    }
                }
            } catch (SQLException e) {
                throw new JDBCSQLException(e);
            } finally {
                JDBCUtils.closeQuietly(st);
            }
        }
    }
}
