/* gvSIG. Geographic Information System of the Valencian Government
 *
 * Copyright (C) 2007-2016 Infrastructures and Transports Department
 * of the Valencian Government (CIT)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 *
 */
package org.gvsig.fmap.dal.store.mdb;

import java.sql.Connection;
import java.sql.SQLException;
import java.text.MessageFormat;
import org.apache.commons.dbcp.BasicDataSource;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.lang3.StringUtils;
import org.gvsig.fmap.dal.SQLBuilder;
import org.gvsig.fmap.dal.resource.exception.AccessResourceException;
import org.gvsig.fmap.dal.store.mdb.operations.MdbOperationsFactory;
import org.gvsig.fmap.dal.store.jdbc.JDBCConnectionParameters;
import org.gvsig.fmap.dal.store.jdbc.JDBCNewStoreParameters;
import org.gvsig.fmap.dal.store.jdbc.JDBCServerExplorerParameters;
import org.gvsig.fmap.dal.store.jdbc.JDBCStoreParameters;
import org.gvsig.fmap.dal.store.jdbc.exception.JDBCDriverClassNotFoundException;
import org.gvsig.fmap.dal.store.jdbc2.OperationsFactory;
import org.gvsig.fmap.dal.store.jdbc2.spi.JDBCHelperBase;
import org.gvsig.fmap.dal.store.jdbc2.spi.JDBCSQLBuilderBase;
import org.gvsig.fmap.dal.store.jdbc2.spi.SRSSolverBase;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class MdbHelper extends JDBCHelperBase {

    static final Logger logger = LoggerFactory.getLogger(MdbHelper.class);

    public static final String MdbJDBCDriver = "org.h2.Driver";

    public static String getConnectionURL(MdbConnectionParameters params) {
        String connectionURL;
        if( StringUtils.isEmpty(params.getHost()) ) {
            // Asumimos que es una conexion directa sobre el filesystem
            String dbname = params.getFile().getAbsolutePath().replace("\\","/");
            if( StringUtils.equalsIgnoreCase(FilenameUtils.getExtension(params.getFile().getName()),"zip") ) {
                connectionURL =  MessageFormat.format(
                    "jdbc:h2:zip:{0}!/{1};MODE=PostgreSQL",
                    dbname,
                    params.getDBName()
                );
            } else {
                if( dbname.endsWith(".mv.db") ) {
                    dbname = dbname.substring(0, dbname.length()-6);
                }
                connectionURL =  MessageFormat.format(
                    "jdbc:h2:file:{0};MODE=PostgreSQL",
                    dbname
                );
            }
        } else if( params.getPort() == null ) {
            connectionURL =  MessageFormat.format(
                "jdbc:h2:tcp://{0}/{1};MODE=PostgreSQL",
                params.getHost(),
                params.getDBName()
            );            
        } else {
            connectionURL =  MessageFormat.format(
                "jdbc:h2:tcp://{0}:{1,number,#######}/{2};MODE=PostgreSQL",
                params.getHost(),
                params.getPort().intValue(),
                params.getDBName()
            );
        }
        logger.debug("connectionURL: {}", connectionURL);
        return connectionURL;
    }

    private static class ConnectionProvider {

        private static boolean needRegisterDriver = true;

        private BasicDataSource dataSource = null;

        private final MdbConnectionParameters connectionParameters;

        public ConnectionProvider(MdbConnectionParameters connectionParameters) {
            this.connectionParameters = connectionParameters;
        }

        @Override
        public String toString() {
            StringBuilder builder = new StringBuilder();
            builder.append(" url=").append(connectionParameters.getUrl());
            builder.append(" driver name=").append(connectionParameters.getJDBCDriverClassName());
            builder.append(" user=").append(connectionParameters.getUser());
            return builder.toString();
        }
        
        public Connection getConnection() throws SQLException {
            if (this.dataSource == null) {
                this.dataSource = this.createDataSource();               
            }
            Connection conn = this.dataSource.getConnection();
            try {
                conn.createStatement().execute("SELECT TOP 1 SRID FROM SPATIAL_REF_SYS");
            } catch(SQLException ex) {
                H2GISExtension.load(conn);
            }
            return conn;
        }

        private BasicDataSource createDataSource() throws SQLException {
            if (!this.isRegistered()) {
                this.registerDriver();
            }
            MdbConnectionParameters params = connectionParameters;

            BasicDataSource ds = new BasicDataSource();
            ds.setDriverClassName(params.getJDBCDriverClassName());
            if( !StringUtils.isEmpty(params.getUser()) ) {
                ds.setUsername(params.getUser());
            }
            if( !StringUtils.isEmpty(params.getPassword()) ) {
                ds.setPassword(params.getPassword());
            }
            ds.setUrl(params.getUrl());

            ds.setMaxWait(60L * 1000);
            return ds;
        }

        private boolean isRegistered() {
            return needRegisterDriver;
        }

        public void registerDriver() throws SQLException {
            String className = this.connectionParameters.getJDBCDriverClassName();
            if (className == null) {
                return;
            }
            try {
                Class theClass = Class.forName(className);
                if (theClass == null) {
                    throw new JDBCDriverClassNotFoundException(MdbLibrary.NAME, className);
                }
            } catch (Exception e) {
                throw new SQLException("Can't register JDBC driver '" + className + "'.", e);
            }
            needRegisterDriver = false;
        }

    }

//    private static class ConnectionProvider {
//
//        private static boolean needRegisterDriver = true;
//
//
//        private final MdbConnectionParameters connectionParameters;
//
//        public ConnectionProvider(MdbConnectionParameters connectionParameters) {
//            this.connectionParameters = connectionParameters;
//        }
//
//        public Connection getConnection() throws SQLException {
//            if (!this.isRegistered()) {
//                this.registerDriver();
//            }
//            Connection conn;
//            if( !StringUtils.isEmpty(connectionParameters.getUser()) ) {
//                conn = DriverManager.getConnection(
//                    connectionParameters.getUrl(),
//                    connectionParameters.getUser(),
//                    connectionParameters.getPassword()
//                );
//            } else {
//                conn = DriverManager.getConnection(connectionParameters.getUrl());
//            }
//            try {
//                conn.createStatement().execute("SELECT TOP 1 SRID FROM SPATIAL_REF_SYS");
//            } catch(SQLException ex) {
//                H2GISExtension.load(conn);
//            }
//            return conn;
//        }
//
//        private boolean isRegistered() {
//            return needRegisterDriver;
//        }
//
//        public void registerDriver() throws SQLException {
//            String className = this.connectionParameters.getJDBCDriverClassName();
//            if (className == null) {
//                return;
//            }
//            try {
//                Class theClass = Class.forName(className);
//                if (theClass == null) {
//                    throw new JDBCDriverClassNotFoundException(MdbLibrary.NAME, className);
//                }
//            } catch (Exception e) {
//                throw new SQLException("Can't register JDBC driver '" + className + "'.", e);
//            }
//            needRegisterDriver = false;
//        }
//
//    }

    private ConnectionProvider connectionProvider = null;
   
    public MdbHelper(JDBCConnectionParameters connectionParameters) {
        super(connectionParameters);
        this.srssolver = new SRSSolverBase(this);
    }

    @Override
    public synchronized Connection  getConnection() throws AccessResourceException {
        try {
            if (this.connectionProvider == null) {
                this.connectionProvider = new ConnectionProvider(this.getConnectionParameters());
            }
            Connection connection = this.connectionProvider.getConnection();
            if( logger.isDebugEnabled() ) {
                logger.debug("getConnection: connection = "+connection.hashCode()+ connectionProvider.toString());
            }
            return connection;
        } catch (SQLException ex) {
            throw new AccessResourceException(MdbLibrary.NAME, ex);
        }
    }

    @Override
    public void closeConnection(Connection connection) {
        logger.debug("closeConnection: connection = "+connection.hashCode());
        super.closeConnection(connection);
    }
    
    @Override
    public MdbConnectionParameters getConnectionParameters() {
        return (MdbConnectionParameters) super.getConnectionParameters();
    }
    
    @Override
    public String getConnectionURL() {
        return getConnectionURL(this.getConnectionParameters());
    }

    @Override
    protected String getResourceType() {
        return MdbLibrary.NAME;
    }

    @Override
    public String getProviderName() {
        return MdbLibrary.NAME;
    }

    @Override
    public JDBCSQLBuilderBase createSQLBuilder() {
        return new MdbSQLBuilder(this);
    }
    
    @Override
    public OperationsFactory getOperations() {
        if (this.operationsFactory == null) {
            this.operationsFactory = new MdbOperationsFactory(this);
        }
        return operationsFactory;
    }

    @Override
    public SQLBuilder.GeometrySupportType getGeometrySupportType() {
        return SQLBuilder.GeometrySupportType.WKB;
    }

    @Override
    public boolean hasSpatialFunctions() {
        return true;
    }

    @Override
    public boolean canWriteGeometry(int geometryType, int geometrySubtype) {
        return true;
    }

    @Override
    public String getQuoteForIdentifiers() {
        return "\"";
    }

    @Override
    public boolean allowAutomaticValues() {
        return true;
    }

    @Override
    public boolean supportOffsetInSelect() {
        return true;
    }

    @Override
    public String getQuoteForStrings() {
        return "'";
    }

    @Override
    public String getSourceId(JDBCStoreParameters parameters) {
        return parameters.getDBName() + "." + 
               parameters.getSchema()+ "." + 
               parameters.getTable();
    }

    @Override
    public JDBCNewStoreParameters createNewStoreParameters() {
        return new MdbNewStoreParameters();
    }

    @Override
    public JDBCStoreParameters createOpenStoreParameters() {
        return new MdbStoreParameters();
    }

    @Override
    public JDBCServerExplorerParameters createServerExplorerParameters() {
        return new MdbExplorerParameters();
    }

}
