package org.gvsig.fmap.dal.store.mdb;

import java.sql.PreparedStatement;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import org.apache.commons.lang3.tuple.Pair;
import org.gvsig.fmap.dal.DataTypes;
import org.gvsig.fmap.dal.feature.FeatureAttributeDescriptor;
import org.gvsig.fmap.dal.feature.FeatureType;
import org.gvsig.fmap.dal.feature.spi.FeatureProvider;
import org.gvsig.fmap.dal.store.jdbc2.spi.JDBCSQLBuilderBase;
import org.gvsig.fmap.geom.Geometry;
import org.gvsig.tools.dispose.Disposable;

public class MdbSQLBuilder extends JDBCSQLBuilderBase {

    public MdbSQLBuilder(MdbHelper helper) {
        super(helper);
        
        //
        // H2/H2GIS SQL functions reference list
        //
        // http://www.h2database.com/html/functions.html
        // http://www.h2gis.org/docs/1.3/functions/
        //
        // http://www.h2database.com/html/grammar.html
        //
        // http://www.h2database.com/html/datatypes.html
        //
        //
        
        config.set(SQLConfig.default_schema, "");
        config.set(SQLConfig.support_schemas, false);
        config.set(SQLConfig.allowAutomaticValues, true);
        config.set(SQLConfig.geometry_type_support, this.helper.getGeometrySupportType());
        config.set(SQLConfig.has_spatial_functions, this.helper.hasSpatialFunctions());

        config.remove_functionality(SQLConfig.DELETE_GEOMETRY_COLUMN_FROM_TABLE_schema_table);
        config.remove_functionality(SQLConfig.DELETE_GEOMETRY_COLUMN_FROM_TABLE_table);
         
//        config.set(SQLConfig.UPDATE_TABLE_STATISTICS_table,"ANALYZE TABLE {0} SAMPLE_SIZE 0");
        config.set(SQLConfig.UPDATE_TABLE_STATISTICS_table,"ANALYZE SAMPLE_SIZE 0");

        config.set(SQLConfig.ST_GeomFromEWKB, "GeomFromWKB({0}, {1})");
        config.set(SQLConfig.ST_AsEWKB, "AsWKB(({0}))");        
        config.set(SQLConfig.ST_Intersects, "( (({0}) && ({1})) AND ST_Intersects(({0}),({1}) ))");  
        config.set(SQLConfig.ST_AsBinary, "NVL2({0}, ST_AsBinary({0}), null)");
        
        config.set(SQLConfig.lcase, "LOWER({0})");
        config.set(SQLConfig.ucase, "UPPER({0})");
        config.set(SQLConfig.operator_ILIKE, "({0}) ILIKE ({1})");
        config.set(SQLConfig.isNull, "( ({0}) IS NULL )");
        config.set(SQLConfig.notIsNull, "( ({0}) IS NOT NULL )");

        config.set(SQLConfig.type_boolean, "BOOLEAN");
        config.set(SQLConfig.type_byte, "TINYINT");
        config.set(SQLConfig.type_bytearray, "TINYINT");
        config.set(SQLConfig.type_geometry, "GEOMETRY");
        config.set(SQLConfig.type_char, "CHAR");
        config.set(SQLConfig.type_date, "DATE");
        config.set(SQLConfig.type_double, "DOUBLE"); 
        config.set(SQLConfig.type_numeric_p, "DECIMAL({0})");
        config.set(SQLConfig.type_numeric_ps, "DECIMAL({0},{1})");
        config.set(SQLConfig.type_bigdecimal, "DOUBLE");
        config.set(SQLConfig.type_float, "REAL");
        config.set(SQLConfig.type_int, "INTEGER");
        config.set(SQLConfig.type_long, "BIGINT");
        config.set(SQLConfig.type_string, "VARCHAR");
        config.set(SQLConfig.type_string_p, "VARCHAR({0})");
        config.set(SQLConfig.type_time, "TIME");
        config.set(SQLConfig.type_timestamp, "TIMESTAMP");
        config.set(SQLConfig.type_version, "VARCHAR");
        config.set(SQLConfig.type_URI, "VARCHAR");
        config.set(SQLConfig.type_URL, "VARCHAR");
        config.set(SQLConfig.type_FILE, "VARCHAR");
        config.set(SQLConfig.type_FOLDER, "VARCHAR");        
    }

    public class MdbTableNameBuilderBase extends TableNameBuilderBase {

        @Override
        public boolean has_schema() {
            return false;
        }

        @Override
        public boolean has_database() {
            return false;
        }
        
    }

    public class MdbCreateIndexBuilder extends CreateIndexBuilderBase {
    
        @Override
        public List<String> toStrings() {
            StringBuilder builder = new StringBuilder();
            builder.append("CREATE ");
            if( this.isUnique ) {
                builder.append("UNIQUE ");
            }
            if( this.isSpatial ) {
                builder.append("SPATIAL ");
            }
            builder.append("INDEX ");
            if( this.ifNotExist ) {
                builder.append("IF NOT EXISTS ");
            }
            builder.append(identifier(this.indexName));
            builder.append(" ON ");
            builder.append(this.table.toString());
            builder.append(" ( ");
            boolean is_first_column = true;
            for( String column : this.columns) {
                if( is_first_column ) {
                    is_first_column = false;
                } else {
                    builder.append(", ");
                }
                builder.append(identifier(column));
            }
            builder.append(" )");
            
            List<String> sqls = new ArrayList<>();
            sqls.add(builder.toString());
            return sqls;
        }

    }
        
    protected class MdbAlterTableBuilderBase extends AlterTableBuilderBase {
        @Override
        public List<String> toStrings() {
            List<String> sqls = new ArrayList<>();
            if( this.isEmpty() ) {
                return sqls;
            }
            for (String column : drops) {
                StringBuilder builder = new StringBuilder();
                builder.append("ALTER TABLE ");
                builder.append(this.table.toString());
                builder.append(" DROP COLUMN IF EXISTS ");
                builder.append(identifier(column)); 
                sqls.add(builder.toString());
            }
            for (ColumnDescriptor column : adds) {
                StringBuilder builder = new StringBuilder();
                builder.append("ALTER TABLE ");
                builder.append(this.table.toString());
                builder.append(" ADD COLUMN ");
                builder.append(identifier(column.getName())); 
                builder.append(" ");
                if( column.getType() == DataTypes.INT && column.isAutomatic() ) {
                    builder.append(" SERIAL");
                } else {
                    builder.append(
                        sqltype(
                            column.getType(), 
                            column.getPrecision(), 
                            column.getSize(),
                            column.getGeometryType(), 
                            column.getGeometrySubtype()
                        )
                    );
                }
                if (column.getDefaultValue() == null) {
                    if (column.allowNulls()) {
                        builder.append(" DEFAULT NULL");
                    }
                } else {
                    builder.append(" DEFAULT '");
                    builder.append(column.getDefaultValue().toString());
                    builder.append("'");
                }
                if (column.allowNulls()) {
                    builder.append(" NULL");
                } else {
                    builder.append(" NOT NULL");
                }
                if (column.isPrimaryKey()) {
                    builder.append(" PRIMARY KEY");
                }
                sqls.add(builder.toString());
                
                if( column.isGeometry() ) {
                    String sql;
                    String constraint_name = "constraint_" + this.table().getName() + "_" + column.getName()+"_dim";
                    sql = MessageFormat.format(
                        "ALTER TABLE \"{0}\" ADD CONSTRAINT IF NOT EXISTS \"{1}\" CHECK ST_CoordDim(\"{2}\") = {3}",
                        this.table().getName(),
                        constraint_name,
                        column.getName(),
                        sqlgeometrynumdimension(column.getGeometryType(),column.getGeometrySubtype())
                    );
                    sqls.add(sql);
                }
            }
            for (ColumnDescriptor column : alters) {
                StringBuilder builder = new StringBuilder();
                builder.append("ALTER TABLE ");
                builder.append(this.table.toString());
                builder.append(" ALTER COLUMN ");
                builder.append(identifier(column.getName())); 
                builder.append(" ");
                builder.append(
                    sqltype(
                        column.getType(), 
                        column.getPrecision(), 
                        column.getSize(),
                        column.getGeometryType(), 
                        column.getGeometrySubtype()
                    )
                );
                if (column.getDefaultValue() == null) {
                    if (column.allowNulls()) {
                        builder.append(" DEFAULT NULL");
                    }
                } else {
                    builder.append(" DEFAULT '");
                    builder.append(column.getDefaultValue().toString());
                    builder.append("'");
                }
                if( column.isAutomatic() ) {
                    builder.append(" AUTO_INCREMENT");
                }
                sqls.add(builder.toString());
                if( column.isGeometry() ) {
                    String sql;
                    String constraint_name = "constraint_" + this.table().getName() + "_" + column.getName()+"_dim";
                    sql = MessageFormat.format(
                        "ALTER TABLE \"{0}\" ADD CONSTRAINT IF NOT EXISTS \"{1}\" CHECK ST_CoordDim(\"{2}\") = {3}",
                        this.table().getName(),
                        constraint_name,
                        column.getName(),
                        sqlgeometrynumdimension(column.getGeometryType(),column.getGeometrySubtype())
                    );
                    sqls.add(sql);
                }
            }
            for (Pair<String,String> pair : renames) {
                StringBuilder builder = new StringBuilder();
                builder.append("ALTER TABLE ");
                builder.append(this.table.toString());
                builder.append(" RENAME COLUMN ");
                builder.append(identifier(pair.getLeft())); 
                builder.append(" TO ");
                builder.append(identifier(pair.getRight())); 
                sqls.add(builder.toString());
            }
            return sqls;
        }

    }
    
    protected class MdbCreateTableBuilder extends CreateTableBuilderBase {

        @Override
        public List<String> toStrings() {

            List<String> sqls = new ArrayList<>();
            StringBuilder builder = new StringBuilder();

            builder.append("CREATE TABLE ");
            builder.append(this.table.toString());
            builder.append(" (");
            boolean first = true;
            for (ColumnDescriptor column : columns) {

                if (first) {
                    first = false;
                } else {
                    builder.append(", ");
                }
                builder.append(identifier(column.getName()));
                builder.append(" ");
                builder.append(
                    sqltype(
                        column.getType(), 
                        column.getPrecision(), 
                        column.getSize(),
                        column.getGeometryType(), 
                        column.getGeometrySubtype()
                    )
                );
                if( !column.isGeometry() ) {
                    if (column.isPrimaryKey()) {
                        builder.append(" PRIMARY KEY");
                    }
                    if( column.isAutomatic() ) {
                        builder.append(" AUTO_INCREMENT");
                    }
                    if (column.getDefaultValue() == null) {
                        if (column.allowNulls()) {
                            builder.append(" DEFAULT NULL");
                        }
                    } else {
                        if( column.getType() == DataTypes.DATE ) {
                            builder.append(" DEFAULT ( TIMESTAMP '");
                            Date d = (Date) column.getDefaultValue();
                            builder.append(MessageFormat.format( "{0,date,yyyy-MM-dd HH:mm:ss.S}",d));
                            builder.append("' )");
                        } else {
                            builder.append(" DEFAULT '");
                            builder.append(column.getDefaultValue().toString());
                            builder.append("'");
                        }
                    }
                }
                if (!column.allowNulls()) {
                    builder.append(" NOT NULL");
                }
            }
            builder.append(" )");
            sqls.add(builder.toString());
            for (ColumnDescriptor column : columns) {
                if( column.isGeometry() ) {
                    String sql;
                    String constraint_name = "constraint_" + this.table().getName() + "_" + column.getName()+"_dim";
                    sql = MessageFormat.format(
                        "ALTER TABLE \"{0}\" ADD CONSTRAINT IF NOT EXISTS \"{1}\" CHECK ST_CoordDim(\"{2}\") = {3}",
                        this.table().getName(),
                        constraint_name,
                        column.getName(),
                        sqlgeometrynumdimension(column.getGeometryType(),column.getGeometrySubtype())
                    );
                    if( column.getGeometrySRSId()!=null ) {
                        String sql2;
                        String constraint_name2 = "constraint_" + this.table().getName() + "_" + column.getName()+"_srid";
                        sql2 = MessageFormat.format(
                            "ALTER TABLE \"{0}\" ADD CONSTRAINT IF NOT EXISTS \"{1}\" CHECK ST_SRID(\"{2}\") = {3,number,#####}",
                            this.table().getName(),
                            constraint_name2,
                            column.getName(),
                            column.getGeometrySRSId()
                        );
                        sqls.add(sql2);
                    }
                    sqls.add(sql);
                }
            }            
            
            return sqls;
        }
    }

    public class MdbSelectBuilderBase extends SelectBuilderBase {
        
        @Override
        protected boolean isValid(StringBuilder message) {
            if( message == null ) {
                message = new StringBuilder();
            }
            if( this.has_offset() && !this.has_order_by() ) {
                // Algunos gestores de BBDD requieren que se especifique un
                // orden para poder usar OFFSET. Como eso parece buena idea para
                // asegurar que siempre tengamos los mismo resultados, lo exijimos
                // siempre.
                message.append("Can't use OFFSET without an ORDER BY.");
                return false;
            }
            return true;
        }        
        
        @Override
        public String toString() {
            StringBuilder builder = new StringBuilder();
            if( !isValid(builder) ) {
                throw new IllegalStateException(builder.toString());
            }
            builder.append("SELECT ");
            if( this.distinct ) {
                builder.append("DISTINCT ");
            }
            boolean first = true;
            for (SelectColumnBuilder column : columns) {
                if (first) {
                    first = false;
                } else {
                    builder.append(", ");
                }
                builder.append(column.toString());
            }

            if ( this.has_from() ) {
                builder.append(" FROM ");
                builder.append(this.from.toString());
            }
            if ( this.has_where() ) {
                builder.append(" WHERE ");
                builder.append(this.where.toString());
            }
            
            if( this.has_order_by() ) {
                builder.append(" ORDER BY ");
                first = true;
                for (OrderByBuilder item : this.order_by) {
                    if (first) {
                        first = false;
                    } else {
                        builder.append(", ");
                    }
                    builder.append(item.toString());                    
                }   
            }
            
            if ( this.has_limit() && this.has_offset() ) {
                builder.append(" LIMIT ");
                builder.append(this.limit);
                builder.append(" OFFSET ");
                builder.append(this.offset);
                
            } else if ( this.has_limit()) {
                builder.append(" LIMIT ");
                builder.append(this.limit);

            } else if ( this.has_offset() ) {
                builder.append(" LIMIT -1 OFFSET ");
                builder.append(this.offset);    
            }
            return builder.toString();

        }
    }

    @Override
    public String bytearray(byte[] data) {
        // Mdb usa un formato diferencte para especificar un array de 
        // bytes. En lugar de 0x... usa x'...' .
        StringBuilder builder = new StringBuilder();
        builder.append("x'");
        for (byte abyte : data) {
            int v = abyte & 0xff;
            builder.append(String.format("%02x", v));
        }
        builder.append("'");
        return builder.toString();
    }

    @Override
    public Object sqlgeometrydimension(int type, int subtype) {
        //'XY' or 2: 2D points, identified by X and Y coordinates
        //'XYM': 2D points, identified by X and Y coordinates. Each point stores an M-value (measure) as well
        //'XYZ' or 3: 3D points, identified by X, Y and Z coordinates
        //'XYZM' or 4: 3D points, identified by X, Y and Z coordinates. Each point stores an M-value (measure) as well
        switch(subtype) {
            case Geometry.SUBTYPES.GEOM2D:
            default:
                return "XY";
            case Geometry.SUBTYPES.GEOM2DM:
                return "XYM";
            case Geometry.SUBTYPES.GEOM3D:
                return "XYZ";
            case Geometry.SUBTYPES.GEOM3DM:
                return "XYZM";
        }
    }

    @Override
    public String sqltype(int type, int p, int s, int geomtype, int geomSubtype) {
        if( type!=DataTypes.GEOMETRY ) {
            return super.sqltype(type, p, s, geomtype, geomSubtype);
        }
        //
        // https://github.com/orbisgis/h2gis/wiki/1.-Spatial-data#geometry-columns-view
        //
        switch(geomtype) {
        case Geometry.TYPES.POINT:
            return "POINT";
        case Geometry.TYPES.MULTIPOINT:
            return "MULTIPOINT";
        case Geometry.TYPES.LINE:
            return "LINESTRING";
        case Geometry.TYPES.MULTILINE:
            return "MULTILINESTRING";
        case Geometry.TYPES.POLYGON:
            return "POLYGON";
        case Geometry.TYPES.MULTIPOLYGON:
            return "MULTIPOLYGON";
        default:
            return "GEOMETRY";
        }
    }
    
    public Object sqlgeometrynumdimension(int type, int subtype) {
        int dimensions=2;
        switch(subtype) {
        case Geometry.SUBTYPES.GEOM3D:
            dimensions = 3;
            break;
        case Geometry.SUBTYPES.GEOM2D:
            dimensions = 2;
            break;
        case Geometry.SUBTYPES.GEOM2DM:
            dimensions = 2; // ??????
            break;
        case Geometry.SUBTYPES.GEOM3DM:
            dimensions = 3; // ??????
            break;
        }
        return dimensions;
    }
    
    public MdbHelper getHelper() {
        return (MdbHelper) this.helper;
    }
   
    @Override
    public Disposable setParameters(PreparedStatement st, FeatureProvider feature) {
        try {
            FeatureType featureType = feature.getType();
            List<Object> values = new ArrayList<>();
            for (Parameter parameter : this.getParameters()) {
                if (parameter.is_constant()) {
                    values.add(parameter.getValue());
                } else {
                    String name = parameter.getName();
                    FeatureAttributeDescriptor descriptor = featureType.getAttributeDescriptor(name);
                    switch( descriptor.getType() ) {
                    case DataTypes.DATE:
                        Date value = (Date)(feature.get(name));
                        if( value == null ) {
                            values.add(null);
                        } else {
                            values.add(value.getTime());
                        }
                        break;
                    case DataTypes.GEOMETRY:
                        Geometry geom = this.forceGeometryType(
                            descriptor.getGeomType(),
                            (Geometry)(feature.get(name))
                        );
                        values.add(geom);
                        break;
                    default:
                        values.add(feature.get(name));
                        break;
                    }
                }
            }
            return this.setStatementParameters(st, values, this.geometry_support_type());
        } catch (Exception ex) {
            String f = "unknow";
            try {
                f = feature.toString();
            } catch (Exception ex2) {
                // Do nothing
            }
            throw new RuntimeException("Can't set parameters to prepared statement from the feature (" + f + ")", ex);
        }
    }

    @Override
    protected AlterTableBuilder createAlterTableBuilder() {
        return new MdbAlterTableBuilderBase();
    }
            
    @Override
    protected TableNameBuilder createTableNameBuilder() {
        return new MdbTableNameBuilderBase();
    }
    
    @Override
    protected CreateTableBuilder createCreateTableBuilder() {
        return new MdbCreateTableBuilder();
    }

    @Override
    protected SelectBuilder createSelectBuilder() {
        return new MdbSelectBuilderBase();
    }

    @Override
    protected CreateIndexBuilder createCreateIndexBuilder() {
        return new MdbCreateIndexBuilder();
    }

}
