
package org.gvsig.fmap.dal.store.mdb.operations;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import org.apache.commons.lang3.StringUtils;
import org.cresques.cts.IProjection;
import org.gvsig.fmap.dal.DataTypes;
import org.gvsig.fmap.dal.exception.DataException;
import org.gvsig.fmap.dal.feature.EditableFeatureAttributeDescriptor;
import org.gvsig.fmap.dal.feature.EditableFeatureType;
import org.gvsig.fmap.dal.store.jdbc2.JDBCHelper;
import org.gvsig.fmap.dal.store.jdbc2.JDBCUtils;
import org.gvsig.fmap.dal.store.jdbc2.spi.SRSSolver;
import org.gvsig.fmap.dal.store.jdbc2.spi.operations.FetchFeatureTypeOperation;
import org.gvsig.fmap.geom.Geometry;
import org.gvsig.fmap.geom.GeometryLocator;
import org.gvsig.fmap.geom.GeometryManager;
import org.gvsig.fmap.geom.type.GeometryType;

public class MdbFetchFeatureTypeOperation extends FetchFeatureTypeOperation {

    private static Map<String,GeometryType>h2spatialGeometryTypes = null;
    private Map<String,GeometryColumnInfo> geometry_column;
    
    private static class GeometryColumnInfo {
        
        public String columnName;
        public int geometryType;
        public String geometryTypeName;
        public int dimensions;
        public int srid;
        
        public GeometryColumnInfo(String columnName, int geometryType, String geometryTypeName, int dimensions, int srid) {
            this.columnName = columnName;
            this.geometryType = geometryType;
            this.geometryTypeName = geometryTypeName;
            this.dimensions = dimensions;
            this.srid = srid;
        }
    }
    
    
    public MdbFetchFeatureTypeOperation(
            JDBCHelper helper
        ) {
        super(helper);
    }

    private GeometryType getGT(
            GeometryManager manager, 
            int type, 
            int subtype
        ) {
        try {
            return manager.getGeometryType(type, subtype);
        } catch (Exception ex) {
            return null;
        }
    }
    
    public MdbFetchFeatureTypeOperation(
            JDBCHelper helper,
            EditableFeatureType featureType,
            String dbname,
            String schema,
            String table,
            List<String> primaryKeys,
            String defaultGeometryColumn,
            IProjection crs
        ) {
        super(helper, featureType, dbname, schema, table, primaryKeys, defaultGeometryColumn, crs);
    }            

    @Override
    public void fetch(EditableFeatureType featureType, Connection conn, String dbname, String schema, String table, List<String> pks, String defaultGeometryColumn, IProjection crs) throws DataException {
        geometry_column = new HashMap<>();
        try {
            //
            // https://github.com/orbisgis/h2gis/wiki/1.-Spatial-data#geometry-columns-view
            //
            StringBuilder where = null;
            if( !StringUtils.isEmpty(dbname) ) {
                if( where == null ) {
                    where = new StringBuilder();
                } else {
                    where.append(" AND ");
                }
                where.append("UPPER(F_TABLE_CATALOG) = '");
                where.append(dbname.toUpperCase());
                where.append("'");
            }
            if( !StringUtils.isEmpty(schema) ) {
                if( where == null ) {
                    where = new StringBuilder();
                } else {
                    where.append(" AND ");
                }
                where.append("UPPER(F_TABLE_SCHEMA) = '");
                where.append(schema.toUpperCase());
                where.append("'");
            }
            if( !StringUtils.isEmpty(table) ) {
                if( where == null ) {
                    where = new StringBuilder();
                } else {
                    where.append(" AND ");
                }
                where.append("UPPER(F_TABLE_NAME) = '");
                where.append(table.toUpperCase());
                where.append("'");
            }            
            String sql = "SELECT F_GEOMETRY_COLUMN, GEOMETRY_TYPE, COORD_DIMENSION, SRID, TYPE FROM GEOMETRY_COLUMNS WHERE " + where;
            Statement st = this.getConnection().createStatement();
            ResultSet rs = JDBCUtils.executeQuery(st,sql);
            while( rs.next() ) {
                geometry_column.put(
                    rs.getString("F_GEOMETRY_COLUMN"), 
                    new GeometryColumnInfo(
                        rs.getString("F_GEOMETRY_COLUMN"), 
                        rs.getInt("GEOMETRY_TYPE"), 
                        rs.getString("TYPE"), 
                        rs.getInt("COORD_DIMENSION"), 
                        rs.getInt("SRID")
                    )
                );
            }
        } catch (SQLException ex) {
            logger.warn("Can't read metadata from table '"+table+"'.",ex);
        }
        super.fetch(featureType, conn, dbname, schema, table, pks, defaultGeometryColumn, crs);
    }
        
    @Override
    protected void fetchGeometryTypeAndSRS(
            EditableFeatureAttributeDescriptor attr,
            ResultSetMetaData rsMetadata,
            int colIndex
        ) {
        try {
            if( attr.getType()==DataTypes.GEOMETRY ) {
                GeometryColumnInfo column_info = this.geometry_column.get(attr.getName());
                String type = "GEOMETRY";
                if( column_info!=null ) {
                    // H2GIS solo soporte 2D y 3D no soporta Ms
                    if( column_info.dimensions==3 ) {
                        type = column_info.geometryTypeName+"Z";
                    } else {
                        type = column_info.geometryTypeName;
                    }
                    SRSSolver solver = this.helper.getSRSSolver();
                    attr.setSRS(
                        solver.getProjection(getConnection(), column_info.srid)
                    );
                }
                GeometryType gt = getGeometryTypeFromMdbType(type);
                if( gt != null ) {
                    attr.setGeometryType(gt);
                }
            }
        } catch(Exception ex) {
            throw new RuntimeException("Can't fetch geometry type and SRS.", ex);
        }
    }

    private GeometryType getGeometryTypeFromMdbType(String typeName) {
        if( h2spatialGeometryTypes==null ) {
            //
            // https://github.com/orbisgis/h2gis/wiki/1.-Spatial-data#geometry-columns-view
            //
            GeometryManager manager = GeometryLocator.getGeometryManager();
            h2spatialGeometryTypes = new HashMap<>();
            h2spatialGeometryTypes.put("POINT", getGT(manager, Geometry.TYPES.POINT,Geometry.SUBTYPES.GEOM2D));
            h2spatialGeometryTypes.put("POINTZ", getGT(manager, Geometry.TYPES.POINT,Geometry.SUBTYPES.GEOM3D));
            
            h2spatialGeometryTypes.put("LINESTRING", getGT(manager, Geometry.TYPES.LINE,Geometry.SUBTYPES.GEOM2D));
            h2spatialGeometryTypes.put("LINESTRINGZ", getGT(manager, Geometry.TYPES.LINE,Geometry.SUBTYPES.GEOM3D));
            
            h2spatialGeometryTypes.put("POLYGON", getGT(manager, Geometry.TYPES.POLYGON,Geometry.SUBTYPES.GEOM2D));
            h2spatialGeometryTypes.put("POLYGONZ", getGT(manager, Geometry.TYPES.POLYGON,Geometry.SUBTYPES.GEOM3D));

            h2spatialGeometryTypes.put("MULTIPOINT", getGT(manager, Geometry.TYPES.MULTIPOINT,Geometry.SUBTYPES.GEOM2D));
            h2spatialGeometryTypes.put("MULTIPOINTZ", getGT(manager, Geometry.TYPES.MULTIPOINT,Geometry.SUBTYPES.GEOM3D));

            h2spatialGeometryTypes.put("MULTILINESTRING", getGT(manager, Geometry.TYPES.MULTILINE,Geometry.SUBTYPES.GEOM2D));
            h2spatialGeometryTypes.put("MULTILINESTRINGZ", getGT(manager, Geometry.TYPES.MULTILINE,Geometry.SUBTYPES.GEOM3D));

            h2spatialGeometryTypes.put("MULTIPOLYGON", getGT(manager, Geometry.TYPES.MULTIPOLYGON,Geometry.SUBTYPES.GEOM2D));
            h2spatialGeometryTypes.put("MULTIPOLYGONZ", getGT(manager, Geometry.TYPES.MULTIPOLYGON,Geometry.SUBTYPES.GEOM3D));

            h2spatialGeometryTypes.put("GEOMETRY", getGT(manager, Geometry.TYPES.GEOMETRY,Geometry.SUBTYPES.GEOM2D));
            h2spatialGeometryTypes.put("GEOMETRYZ", getGT(manager, Geometry.TYPES.GEOMETRY,Geometry.SUBTYPES.GEOM3D));

            h2spatialGeometryTypes.put("GEOMCOLLECTION", getGT(manager, Geometry.TYPES.GEOMETRY,Geometry.SUBTYPES.GEOM2D));
            h2spatialGeometryTypes.put("GEOMCOLLECTIONZ", getGT(manager, Geometry.TYPES.GEOMETRY,Geometry.SUBTYPES.GEOM3D));
        }
        return h2spatialGeometryTypes.get(typeName);
    }
    
}
