package org.gvsig.fmap.dal.store.csv;

import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import org.gvsig.fmap.dal.DataTypes;
import org.gvsig.tools.ToolsLocator;
import org.gvsig.tools.dataTypes.DataTypesManager;

/**
 *
 * @author jjdelcerro
 */
public class AutomaticDetectionOfTypes {

    public interface Rows {

        public List<String> nextRowValues();
    }

    private static class PossibleDataType {

        public boolean possibleInt = true;
        public boolean possibleFloat = true;
        public boolean possibleDouble = true;
        public boolean possibleLong = true;
        public boolean possibleURL = true;
        public boolean possibleDate = true;
        public boolean possibleGeometry = true;
    }

    private final String filename;

    public AutomaticDetectionOfTypes() {
        this("(unknown)");
    }

    public AutomaticDetectionOfTypes(String filename) {
        this.filename = filename;
    }

    private String getFullFileName() {
        return this.filename;
    }

    @SuppressWarnings({"UseSpecificCatch", "ResultOfObjectAllocationIgnored"})
    public int[] detect(int columns, Rows rows, boolean isFirstLineHeader, Locale locale) throws IOException {
        List<PossibleDataType> possibleDataTypes;
        int[] types = null;

        int lineno = 0;
        try {
            if (isFirstLineHeader) {
                rows.nextRowValues();
                lineno++;
            }
            possibleDataTypes = new ArrayList<>(columns);
            for (int i = 0; i < columns; i++) {
                possibleDataTypes.add(new PossibleDataType());
            }
            if (locale == null) {
                locale = Locale.getDefault();
            }
            DataTypesManager typeManager = ToolsLocator.getDataTypesManager();
            DataTypesManager.CoercionWithLocale toDouble = (DataTypesManager.CoercionWithLocale) typeManager.getCoercion(DataTypes.DOUBLE);
            DataTypesManager.CoercionWithLocale toFloat = (DataTypesManager.CoercionWithLocale) typeManager.getCoercion(DataTypes.FLOAT);
            DataTypesManager.CoercionWithLocale toDate = (DataTypesManager.CoercionWithLocale) typeManager.getCoercion(DataTypes.DATE);
            DataTypesManager.CoercionWithLocale toInt = (DataTypesManager.CoercionWithLocale) typeManager.getCoercion(DataTypes.INT);
            DataTypesManager.CoercionWithLocale toLong = (DataTypesManager.CoercionWithLocale) typeManager.getCoercion(DataTypes.LONG);
            DataTypesManager.Coercion toGeom = typeManager.getCoercion(DataTypes.GEOMETRY);

            List<String> row = rows.nextRowValues();
            lineno++;

            while (row != null) {
                for (int i = 0; i < row.size(); i++) {
                    while( possibleDataTypes.size()<row.size() ) {
                        possibleDataTypes.add(new PossibleDataType());
                    }
                    String rawvalue = row.get(i);
                    PossibleDataType possibleDataType = possibleDataTypes.get(i);
                    if (possibleDataType.possibleDouble) {
                        try {
                            toDouble.coerce(rawvalue, locale);
                            possibleDataType.possibleDouble = true;
                        } catch (Exception ex) {
                            possibleDataType.possibleDouble = false;
                        }
                    }
                    if (possibleDataType.possibleFloat) {
                        try {
                            toFloat.coerce(rawvalue, locale);
                            possibleDataType.possibleFloat = true;
                        } catch (Exception ex) {
                            possibleDataType.possibleFloat = false;
                        }
                    }
                    if (possibleDataType.possibleLong) {
                        possibleDataType.possibleLong = isValidLong(rawvalue);
                    }
                    if (possibleDataType.possibleInt) {
                        possibleDataType.possibleInt = isValidInteger(rawvalue);
                    }
                    if (possibleDataType.possibleDate) {
                        try {
                            toDate.coerce(rawvalue, locale);
                            possibleDataType.possibleDate = true;
                        } catch (Exception ex) {
                            possibleDataType.possibleDate = false;
                        }
                    }
                    if (possibleDataType.possibleURL) {
                        try {
                            new URL((String) rawvalue);
                            possibleDataType.possibleURL = true;
                        } catch (Exception ex) {
                            possibleDataType.possibleURL = false;
                        }
                    }
                    if (possibleDataType.possibleGeometry) {
                        try {
                            toGeom.coerce((String) rawvalue);
                            possibleDataType.possibleGeometry = true;
                        } catch (Exception ex) {
                            possibleDataType.possibleGeometry = false;
                        }
                    }
                }
                row = rows.nextRowValues();
            }
            int n = 0;
            types = new int[possibleDataTypes.size()];
            for (PossibleDataType possibleDataType : possibleDataTypes) {
                if (possibleDataType.possibleInt) {
                    types[n++] = DataTypes.INT;
                    continue;
                }
                if (possibleDataType.possibleLong) {
                    types[n++] = DataTypes.LONG;
                    continue;
                }
                if (possibleDataType.possibleFloat) {
                    // Forzamos los float a double para evitar perder precision
                    types[n++] = DataTypes.DOUBLE;
                    continue;
                }
                if (possibleDataType.possibleDouble) {
                    types[n++] = DataTypes.DOUBLE;
                    continue;
                }
                if (possibleDataType.possibleURL) {
                    types[n++] = DataTypes.URL;
                    continue;
                }
                if (possibleDataType.possibleDate) {
                    types[n++] = DataTypes.DATE;
                    continue;
                }
                if (possibleDataType.possibleGeometry) {
                    types[n++] = DataTypes.GEOMETRY;
                    continue;
                }
                types[n++] = DataTypes.STRING;
            }
        } catch (Exception ex) {
            throw new RuntimeException("Problems reading file '" + this.getFullFileName() + "' near line " + lineno + ".", ex);
        }
        return types;
    }

    @SuppressWarnings("UseSpecificCatch")
    private boolean isValidLong(String s) {
        if (s == null) {
            return true;
        }
        s = s.trim().toLowerCase();
        if (s.isEmpty()) {
            return true;
        }
        try {
            if (s.startsWith("0x")) {
                Long.valueOf(s.substring(2), 16);
            } else {
                Long.valueOf(s);
            }
            return true;
        } catch (Exception ex) {
            return false;
        }
    }

    @SuppressWarnings("UseSpecificCatch")
    private boolean isValidInteger(String s) {
        if (s == null) {
            return true;
        }
        s = s.trim().toLowerCase();
        if (s.isEmpty()) {
            return true;
        }
        try {
            if (s.startsWith("0x")) {
                Integer.valueOf(s.substring(2), 16);
            } else {
                Integer.valueOf(s);
            }
            return true;
        } catch (Exception ex) {
            return false;
        }
    }

}
