/**
 * gvSIG. Desktop Geographic Information System.
 *
 * Copyright (C) 2007-2016 gvSIG Association.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 *
 * For any additional information, do not hesitate to contact us
 * at info AT gvsig.com, or visit our website www.gvsig.com.
 */
package org.gvsig.fmap.dal.store.dbf.utils;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.UnsupportedEncodingException;

import org.apache.commons.io.FilenameUtils;
import org.apache.commons.lang3.StringUtils;
import org.gvsig.utils.StringUtilities;


/**
 * Reads and writes codepage (.cpg) files which are sometime
 * used to externally specify the charset encoding of a DBF file.
 * 
 * @author Cesar Martinez Izquierdo <cmartinez@scolab.es>
 *
 */
public class DbaseCodepage {
	private File codePageFile;
	private String charset = null;
	
	/**
	 * Define the valid dbf Language driver IDs (equivalent to MSDOS code pages), which are used on the byte 29
	 * of the DBF header to define the DBF codepage.
	 * 
	 * The equivalences of these charsets using Java NIO charset names are defined on the {@link #ldidJava} array
	 * (so 0x01 is equivalent to IBM437, 0x02 to IBM850, etc)
	 * 
	 * See some other equivalences in:
	 *  https://github.com/infused/dbf/blob/master/docs/supported_encodings.csv
	 *  https://github.com/olemb/dbfread/blob/master/dbfread/codepages.py
	 *  https://joinup.ec.europa.eu/svn/gvsig-desktop/trunk/libraries/libFMap/src/com/iver/cit/gvsig/fmap/drivers/dbf/DbfEncodings.java
	 */
	public static final short[] dbfLdid= {
			0x01, 0x02, 0x03, 0x04,
			0x08, 0x09, 0x0a, 0x0b,
			0x0d, 0x0e, 0x0f, 0x10,
			0x11, 0x12, 0x13, 0x14,
			0x15, 0x16, 0x17, 0x18,
			0x19, 0x1a, 0x1b, 0x1c,
			0x1d, 0x1f, 0x22, 0x23,
			0x24, 0x25, 0x26, 0x37,
			0x40, 0x4d, 0x4e, 0x4f,
			0x50, 0x57, 0x58, 0x59,
			0x64, 0x65, 0x66, 0x67,
			0x68, 0x69, 0x6a, 0x6b,
			0x6c, 0x78, 0x79, 0x7a,
			0x7b, 0x7c, 0x7d, 0x7d, 
			0x86, 0x87,	0x88, 0xc8,
			0xc9, 0xca,	0xcb, 0xcc};
	
	/**
	 * Equivalent Java charset names to the code pages defined in {@link #dbfLdid}, using Java NIO Charset names
	 * (which differ from JAVA IO names, see https://docs.oracle.com/javase/8/docs/technotes/guides/intl/encoding.doc.html)
	 */
	public static final String[] ldidJava = new String[] {
			"IBM437", "IBM850", "windows-1252", "x-MacRoman",
			"IBM865", "IBM437", "IBM850", "IBM437",
			"IBM437", "IBM850", "IBM437", "IBM850",
			"IBM437", "IBM850", "x-IBM943", "IBM850",
			"IBM437", "IBM850", "IBM865", "IBM437",
			"IBM437", "IBM850", "IBM437", "IBM863",
			"IBM850", "IBM852", "IBM852", "IBM852",
			"IBM860", "IBM850", "IBM866", "IBM850",
			"IBM852", "x-mswin-936", "x-IBM949", "IBM950",
			"x-IBM874", "windows-1252", "windows-1252", "windows-1252",
			"IBM852", "IBM866", "IBM865", "IBM861",
			// 0x68 and 0x69 are unofficial "Codepage 895 Kamenicky (Czech) MS-DOS" and "Codepage 620  Mazovia (Polish) MS-DOS",
			// but there is no Java equivalent
			// so we use CP437 which is the closest charset for the latin characters part
			"IBM437", "IBM437", "x-IBM737", "IBM857",
			"IBM863", "x-IBM950", "x-IBM949", "x-mswin-936",
			"x-IBM942", "x-IBM874", "windows-1255", "windows-1256",
			"x-IBM737", "IBM852", "IBM857", "windows-1250",
			"windows-1251", "windows-1254",	"windows-1253", "windows-1257"};
	
	/**
	 * Defines the dbfFile and constructs the name of the .cpg
	 * file based on it
	 * @param dbfFile
	 */
	public DbaseCodepage(File dbfFile) {
		String cpfileName = getCpgFileName(dbfFile.getAbsolutePath()); 
		this.codePageFile = new File(cpfileName);
	}
	
	public static String getCpgFileName(String dbfFile) {
		return FilenameUtils.removeExtension(dbfFile) + ".cpg";
	}
	
	public String getCharset() {
		return this.charset;
	}
	
	public void setCharset(String charset) {
		this.charset = charset;
	}
	
	public static String cpg2java(String codePage) {
		if (codePage.equals("UTF8")) {
			return "UTF-8";
		}
		if (codePage.equals("SJIS")) {
			return "Shift_JIS";
		}
		
		if (StringUtils.isNumeric(codePage)) {
			if (codePage.startsWith("8859") && codePage.length()>4) {
				return "ISO-8859-" + codePage.substring(4); 
			}
			if (codePage.startsWith("125") && codePage.length()==4) {
				return "windows-"+codePage;
			}
			if (codePage.length()==3) {
				return "IBM-"+codePage;
			}
			for (int i=0; i<ldidJava.length; i++) {
				if (ldidJava[i].contains(codePage)) {
					return codePage;
				}
			}
		}
		if (codePage=="65001") {
			return "UTF-8";
		}
		return codePage;
	}
	
	public static String java2cpg(String charset) {
		if (charset.startsWith("windows-") ||
				charset.startsWith("ISO-8859") ||
				charset.startsWith("IBM-") ||
				charset.startsWith("x-IBM") ||
				charset.startsWith("x-mswin-")) {
			return charset.replaceAll("[^\\d]", "");
		}
		if (charset.equals("Shift_JIS")) {
			return "SJIS";
		}
		// For the rest of the charsets, we'll directly write the Java NIO Charset
		// Probably they will only be recognized by gvSIG, but it's better than nothing
	    return charset;
	}
	
	/**
	 * Reads the charset from the codepage file, and updates the internal
	 * charset definition
	 * @return
	 */
	public String read() {
		BufferedReader bufReader = null;
		try {
			InputStreamReader reader = new InputStreamReader(new FileInputStream(codePageFile), "ISO-8859-1");
			bufReader = new BufferedReader(reader);
			String cp = bufReader.readLine();
			String charset = cpg2java(cp);
			this.charset = charset;
		} catch (FileNotFoundException e) {
		} catch (IOException e) {
		} finally {
			if (bufReader!=null) {
				try {
					bufReader.close();
				}
				catch (Exception ex) {}
			}
		}
		return getCharset();
	}
	
	/**
	 * Writes the specified charset to the codepage file, and updates
	 * the internal charset definition
	 * 
	 * @param charset
	 */
	
	public void write(String charset) {
		this.charset = charset;
		OutputStreamWriter writer = null;
		try {
			writer = new OutputStreamWriter(new FileOutputStream(codePageFile), "ISO-8859-1");
			writer.write(java2cpg(charset));
		} catch (UnsupportedEncodingException e) {
		} catch (FileNotFoundException e) {
		} catch (IOException e) {
		} finally {
			if (writer!=null) {
				try {
					writer.close();
				} catch (IOException e) {
				}
			}
		}
	}
	
	/**
	 * Gets the dbf Language driver IDs (code pages) corresponding
	 * to the provided charset name
	 *  
	 * @return The LDID code, or 0x00 if no equivalent LDID code
	 * was found for the provided charsetName
	 */
	public static int getLdid(String charsetName) {
		for (int i=0; i<DbaseCodepage.ldidJava.length; i++) {
			if (DbaseCodepage.ldidJava[i].equals(charsetName)) {
				return DbaseCodepage.dbfLdid[i];
			}
		}
		// default
		return 0x00;
	}
	
	/**
	 * Gets the Java NIO charset name equivalent to the provided ldid
	 * value. Gets null if the provided ldid is not recognised as
	 * a valid code
	 * 
	 * @param ldid
	 * @return
	 */
	public static String getCharsetName(int ldid) {
		if (ldid!=0) {
			for (int i=0; i<DbaseCodepage.dbfLdid.length; i++) {
				if (DbaseCodepage.dbfLdid[i]==ldid) {
					return DbaseCodepage.ldidJava[i];
				}
			}
		}
		return null;
	}

}
