/**
 * gvSIG. Desktop Geographic Information System.
 *
 * Copyright (C) 2007-2013 gvSIG Association.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * For any additional information, do not hesitate to contact us at info AT
 * gvsig.com, or visit our website www.gvsig.com.
 */
package org.gvsig.fmap.dal.feature.impl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.gvsig.fmap.dal.DataStoreNotification;
import org.gvsig.fmap.dal.exception.DataException;
import org.gvsig.fmap.dal.exception.DataRuntimeException;
import org.gvsig.fmap.dal.feature.EditableFeature;
import org.gvsig.fmap.dal.feature.Feature;
import org.gvsig.fmap.dal.feature.FeatureQuery;
import org.gvsig.fmap.dal.feature.FeatureReference;
import org.gvsig.fmap.dal.feature.FeatureSelection;
import org.gvsig.fmap.dal.feature.FeatureSet;
import org.gvsig.fmap.dal.feature.FeatureStore;
import org.gvsig.fmap.dal.feature.FeatureStoreNotification;
import org.gvsig.fmap.dal.feature.FeatureType;
import org.gvsig.fmap.dal.feature.exception.ReversedSelectionIteratorException;
import org.gvsig.fmap.dal.feature.impl.DefaultFeatureReferenceSelection.SelectionData;
import org.gvsig.fmap.dal.feature.impl.featureset.AbstractFeatureSet;
import org.gvsig.fmap.dal.feature.impl.featureset.DefaultFeatureSet;
import org.gvsig.fmap.dal.feature.impl.undo.FeatureCommandsStack;
import org.gvsig.tools.ToolsLocator;
import org.gvsig.tools.dispose.DisposableIterator;
import org.gvsig.tools.dispose.DisposeUtils;
import org.gvsig.tools.dynobject.DynStruct;
import org.gvsig.tools.exception.BaseException;
import org.gvsig.tools.observer.Observable;
import org.gvsig.tools.observer.Observer;
import org.gvsig.tools.persistence.PersistenceManager;
import org.gvsig.tools.persistence.PersistentState;
import org.gvsig.tools.persistence.exception.PersistenceException;
import org.gvsig.tools.visitor.Visitor;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Default implementation of the FeatureSelection interface. Internally, only
 * FeatureReference values are stored.
 *
 * This implementation performs better if used with the selection related
 * methods: select, deselect and isSelected ones.
 *
 * @author <a href="mailto:cordin@disid.com">Csar Ordiana</a>
 */
public class DefaultFeatureSelection extends AbstractFeatureSet
        implements FeatureSelection {

    public class RemoveFromFeatureSelectionException extends DataRuntimeException {

        /**
         *
         */
        private static final long serialVersionUID = 2636692469445838928L;
        private final static String MESSAGE_FORMAT = "Can't remove feature from reversed selection.";
        private final static String MESSAGE_KEY = "_RemoveFromFeatureSelectionException";

        public RemoveFromFeatureSelectionException(Throwable cause) {
            super(MESSAGE_FORMAT, cause, MESSAGE_KEY, serialVersionUID);
            //setValue("store", store);
        }
    }

    /**
     * Facade over a Iterator of FeatureReferences, to return Features instead.
     *
     * @author <a href="mailto:cordin@disid.com">Csar Ordiana</a>
     */
    private class FeatureIteratorFacade implements DisposableIterator {

        private final Logger LOGGER = LoggerFactory
                .getLogger(FeatureIteratorFacade.class);

        private java.util.Iterator refIterator;

        private FeatureStore featureStore;
        private Feature currentFeature = null;

        public FeatureIteratorFacade(java.util.Iterator iter,
                FeatureStore featureStore) {
            this.refIterator = iter;
            this.featureStore = featureStore;
        }

        @Override
        public boolean hasNext() {
            return refIterator.hasNext();
        }

        @Override
        public Object next() {
            FeatureReference ref = nextFeatureReference();
            try {
                currentFeature = featureStore.getFeatureByReference(ref);
                return currentFeature;
            } catch (DataException ex) {
                LOGGER.error(
                        "Error loading the Feature with FeatureReference: "
                        + ref, ex);
                return null;
            }
        }

        /**
         * Returns the next FeatureReference.
         *
         * @return the next FeatureReference
         */
        public FeatureReference nextFeatureReference() {
            return (FeatureReference) refIterator.next();
        }

        @Override
        public void remove() {
            try {
                featureStore.delete(currentFeature);
                refIterator.remove();
            } catch (DataException e) {
                throw new RemoveFromFeatureSelectionException(e);
            }
        }

        public class RemoveFromFeatureSelectionException extends DataRuntimeException {

            /**
             *
             */
            private static final long serialVersionUID = 2636692469445838928L;
            private final static String MESSAGE_FORMAT = "Can't remove feature from selection.";
            private final static String MESSAGE_KEY = "_RemoveFromFeatureSelectionException";

            public RemoveFromFeatureSelectionException(Throwable cause) {
                super(MESSAGE_FORMAT, cause, MESSAGE_KEY, serialVersionUID);
                //setValue("store", store);
            }
        }

        @Override
        public void dispose() {
            if (refIterator instanceof DisposableIterator) {
                ((DisposableIterator) refIterator).dispose();
            }
            refIterator = null;
            featureStore = null;
        }
    }

    /**
     * Facade over a Iterator of FeatureReferences, to return Features instead,
     * when the Selection is reversed
     *
     * @author <a href="mailto:cordin@disid.com">Csar Ordiana</a>
     */
    private class ReversedFeatureIteratorFacade implements DisposableIterator {

        private SelectionData selectionData;

        private DisposableIterator iterator;

        private Feature nextFeature = null;
        private Feature currentFeature = null;

        private FeatureSet featureSet;

        public ReversedFeatureIteratorFacade(SelectionData selectionData,
                FeatureStore featureStore, boolean fastIterator) {
            this.selectionData = selectionData;

            // Load a Set with all the store features
            try {
                featureSet = new IgnoreInsertAndUpdateFeatureSet(
                        (DefaultFeatureStore) featureStore, 
                        new DefaultFeatureQuery(featureStore.getDefaultFeatureType())
                );
                iterator = featureSet.fastIterator();
            } catch (DataException ex) {
                throw new ReversedSelectionIteratorException(ex);
            }

            // Filter the features not selected and position in the next
            // selected feature
            positionInNextElement();
        }

        @Override
        public boolean hasNext() {
            return nextFeature != null;
        }

        @Override
        public Object next() {
            featureIterators.remove(currentFeature);
            currentFeature = nextFeature.getCopy();
            featureIterators.put(currentFeature, this);
            positionInNextElement();
            return currentFeature;
        }

        @Override
        public void remove() {
            try {
                featureSet.delete(currentFeature);
            } catch (DataException e) {
                throw new RemoveFromFeatureSelectionException(e);

            }
        }

        private void positionInNextElement() {
            nextFeature = null;
            while (iterator.hasNext()) {
                nextFeature = (Feature) iterator.next();
                if (selectionData.contains(nextFeature.getReference())) {
                    nextFeature = null;
                } else {
                    break;
                }
            }
        }

        @Override
        public void dispose() {
            this.featureSet.dispose();
            this.iterator.dispose();
            this.selectionData = null;
            this.nextFeature = null;
        }
    }

    private Map featureTypeCounts = new HashMap(1);
    private final Map<Feature, Iterator> featureIterators = new HashMap<>();
    private final DefaultFeatureReferenceSelection featureReferenceSelection;

    /**
     * Creates a DefaultFeatureSelection, with a FeatureStore.
     *
     * @param featureStore the FeatureStore to load Features from
     * @throws DataException if there is an error while getting the total number
     * of Features of the Store.
     * @see AbstractSetBasedDataSelection#DefaultSelection(int)
     */
    public DefaultFeatureSelection(DefaultFeatureStore featureStore)
            throws DataException {
        this.featureReferenceSelection = new DefaultFeatureReferenceSelection(featureStore);
    }

    /**
     * Creates a new Selection with the total size of Features from which the
     * selection will be performed.
     *
     * @param featureStore the FeatureStore of the selected FeatureReferences
     * @param helper to get some information of the Store
     * @throws DataException if there is an error while getting the total number
     * of Features of the Store.
     */
    public DefaultFeatureSelection(FeatureStore featureStore,
            FeatureSelectionHelper helper) throws DataException {
        this.featureReferenceSelection = new DefaultFeatureReferenceSelection(featureStore, helper);
    }

    /**
     * Constructor used by the persistence manager. Don't use directly. After to
     * invoke this method, the persistence manager calls the the method
     * {@link #loadFromState(PersistentState)} to set the values of the internal
     * attributes that this class needs to work.
     */
    public DefaultFeatureSelection() {
        this.featureReferenceSelection = new DefaultFeatureReferenceSelection();
    }

    @Override
    public FeatureStore getFeatureStore() {
        return this.featureReferenceSelection.getFeatureStore();
    }

    private void notifyObservers(String notificationType) {
        this.featureReferenceSelection.notifyObservers(notificationType);
    }

    public FeatureCommandsStack getCommands() {
        return this.featureReferenceSelection.getCommands();
    }

    @Override
    public void enableNotifications() {
        this.featureReferenceSelection.enableNotifications();
    }

    @Override
    public void disableNotifications() {
        this.featureReferenceSelection.disableNotifications();
    }

    public boolean isReversed() {
        return this.featureReferenceSelection.isReversed();
    }

    @Override
    public long getSelectedCount() {
        return this.featureReferenceSelection.getSelectedCount();
    }

    public DefaultFeatureReferenceSelection.SelectionData getData() {
        return this.featureReferenceSelection.getData();
    }

    @Override
    public boolean select(FeatureReference reference) {
        return this.featureReferenceSelection.select(reference);
    }

    public boolean select(FeatureReference reference, boolean undoable) {
        return this.featureReferenceSelection.select(reference, undoable);
    }

    @Override
    public boolean deselect(FeatureReference reference) {
        return this.featureReferenceSelection.deselect(reference);
    }

    public boolean deselect(FeatureReference reference, boolean undoable) {
        return this.featureReferenceSelection.deselect(reference, undoable);
    }

    @Override
    public Iterator referenceIterator() {
        return this.featureReferenceSelection.referenceIterator();
    }

    @Override
    public void selectAll() throws DataException {
        this.featureReferenceSelection.selectAll();
    }

    @Override
    public void deselectAll() throws DataException {
        this.featureReferenceSelection.deselectAll();
    }
    
    public void deselectAll(boolean undoable) throws DataException {
        this.featureReferenceSelection.deselectAll(undoable);
    }
    
    @Override
    public boolean isSelected(FeatureReference reference) {
        return this.featureReferenceSelection.isSelected(reference);
    }

    @Override
    public void reverse() {
        this.featureReferenceSelection.reverse();
    }

    @Override
    public void dispose() {
        this.featureReferenceSelection.dispose();
    }

    @Override
    public void update(Observable o, Object o1) {
        this.featureReferenceSelection.update(o, o1);
    }

    @Override
    public void addObserver(Observer obsrvr) {
        this.featureReferenceSelection.addObserver(obsrvr);
    }

    @Override
    public void deleteObserver(Observer obsrvr) {
        this.featureReferenceSelection.deleteObserver(obsrvr);
    }

    @Override
    public void deleteObservers() {
        this.featureReferenceSelection.deleteObservers();
    }

    @Override
    public void beginComplexNotification() {
        this.featureReferenceSelection.beginComplexNotification();
    }

    @Override
    public void endComplexNotification() {
        this.featureReferenceSelection.endComplexNotification();
    }

    @Override
    public void saveToState(PersistentState ps) throws PersistenceException {
        this.featureReferenceSelection.saveToState(ps);
    }

    @Override
    public boolean select(Feature feature) {
        return select(feature, true);
    }

    /**
     * @param feature
     * @return
     * @see #select(Feature)
     * @param undoable if the action must be undoable
     */
    public boolean select(Feature feature, boolean undoable) {
        // TODO: should we check if the feature is from the same FeatureStore??
        if (feature == null) {
            return false;
        }

        // LOGGER.debug("Selected feature: {}", feature);
        if (isReversed()) {
            removeFeatureTypeCount(feature.getType());
        } else {
            addFeatureTypeCount(feature.getType());
        }
        return select(feature.getReference(), undoable);
    }

    @Override
    public boolean select(FeatureSet features) throws DataException {
        return select(features, true);
    }

    /**
     * @param features
     * @return
     * @throws org.gvsig.fmap.dal.exception.DataException
     * @see #select(FeatureSet)
     * @param undoable if the action must be undoable
     */
    public boolean select(FeatureSet features, boolean undoable)
            throws DataException {
        boolean change = false;
        boolean inComplex = false;
        if (undoable && this.featureReferenceSelection.getFeatureStore().isEditing()
                && !this.featureReferenceSelection.getCommands().inComplex()) {

            this.featureReferenceSelection.getCommands().startComplex("_selectionSelectFeatureSet");
            inComplex = this.featureReferenceSelection.getCommands().inComplex();
        }

        disableNotifications();
        DisposableIterator iter = null;
        try {
            for (iter = features.fastIterator(); iter.hasNext();) {
                change |= select((Feature) iter.next(), undoable);
            }
        } finally {
            DisposeUtils.disposeQuietly(iter);
        }
        enableNotifications();
        if (undoable && getFeatureStore().isEditing() && inComplex) {
            getCommands().endComplex();
        }
        if (change) {
            notifyObservers(DataStoreNotification.SELECTION_CHANGE);
        }
        return change;
    }

    @Override
    public boolean deselect(Feature feature) {
        return deselect(feature, true);
    }

    /**
     * @param feature
     * @return
     * @see #deselect(Feature)
     * @param undoable if the action must be undoable
     */
    public boolean deselect(Feature feature, boolean undoable) {
        if (feature == null) {
            return false;
        }

        LOG.debug("Deselected feature: {}", feature);

        if (isReversed()) {
            addFeatureTypeCount(feature.getType());
        } else {
            removeFeatureTypeCount(feature.getType());
        }
        return deselect(feature.getReference(), undoable);
    }

    @Override
    public boolean deselect(FeatureSet features) throws DataException {
        return deselect(features, true);
    }

    /**
     * @param features
     * @return
     * @throws org.gvsig.fmap.dal.exception.DataException
     * @see #deselect(FeatureSet)
     * @param undoable if the action must be undoable
     */
    public boolean deselect(FeatureSet features, boolean undoable)
            throws DataException {
        boolean change = false;
        if (undoable && getFeatureStore().isEditing()) {
            getCommands().startComplex("_selectionDeselectFeatureSet");
        }
        disableNotifications();
        DisposableIterator iter = null;
        try {
            for (iter = features.fastIterator(); iter.hasNext();) {
                change |= deselect((Feature) iter.next(), undoable);
            }
        } finally {
            DisposeUtils.disposeQuietly(iter);
        }
        enableNotifications();
        if (undoable && getFeatureStore().isEditing()) {
            getCommands().endComplex();
        }
        if (change) {
            notifyObservers(DataStoreNotification.SELECTION_CHANGE);
        }
        return change;
    }

    @Override
    public boolean isSelected(Feature feature) {
        if (feature == null) {
            return false;
        }
        if( this.featureReferenceSelection.isEmpty() ) {
            return false;
        }
        // Use the selection data size as a small optimization for the most
        // common case, when nothing is selected and every feature is checked
        // while drawing or painting the table document.
        if (getData().isReversed()) {
            return getData().getSize() == 0
                    || !getData().contains(feature.getReference());
        } else {
            return getData().getSize() > 0
                    && getData().contains(feature.getReference());
        }
    }

    @Override
    public FeatureType getDefaultFeatureType() {
        try {
            return getFeatureStore().getDefaultFeatureType();
        } catch (DataException ex) {
            LOG.error("Error getting the default feature type "
                    + "of the FeatureStore: " + getFeatureStore(), ex);
        }
        return null;
    }

    @Override
    public List getFeatureTypes() {
        // Go through the map of FeatureTypes, and return only the ones that
        // have at least a Feature.
        List types = new ArrayList();
        for (java.util.Iterator iterator = featureTypeCounts.entrySet()
                .iterator(); iterator.hasNext();) {
            Map.Entry entry = (Entry) iterator.next();
            FeatureType type = (FeatureType) entry.getKey();
            Long count = (Long) entry.getValue();

            if (count > 0) {
                types.add(type);
            }
        }

        return types;
    }

    @Override
    public long getSize() throws DataException {
        return getSelectedCount();
    }

    /**
     * Returns the list of selected values, or the deselected ones if the
     * selection has been reversed.
     *
     * WARN: not very good performance implementation.
     */
    @Override
	public DisposableIterator iterator(long index) {
		return iterator(index, 0, false);
	}

    @Override
        public DisposableIterator iterator(long index, long elements) {
		return iterator(index, elements, false);
        }

    /**
     * Returns the list of selected values, or the deselected ones if the
     * selection has been reversed.
     *
     * WARN: not really a fast implementation.
     */
    @Override
	public DisposableIterator fastIterator(long index) {
            return fastIterator(index, 0);
	}

    @Override
    public DisposableIterator fastIterator(long index, long elements) {
        return iterator(index, elements, true);
    }


    protected void clearFeatureReferences() {
        this.featureReferenceSelection.clearFeatureReferences();
        featureTypeCounts.clear();
    }

    /**
     * Creates an iterator for the Selection.
     */
    private DisposableIterator iterator(long index, long elements, boolean fastIterator) {
        if (isReversed()) {
            DisposableIterator iter = new ReversedFeatureIteratorFacade(
                    getData(), getFeatureStore(), fastIterator);
            for (long l = 0; l < index && iter.hasNext(); l++) {
                iter.next();
            }
            return iter;

        } else {
            // TODO: maybe we could add a new referenceIterator(int index)
            // method that could be implemented in a more performant way

            java.util.Iterator iter = getData().getSelected().iterator();
            for (long l = 0; l < index && iter.hasNext(); l++) {
                iter.next();
            }
            return new FeatureIteratorFacade(iter, getFeatureStore());
        }
    }

    private Long removeFeatureTypeCount(FeatureType featureType) {
        Long count = (Long) featureTypeCounts.get(featureType);
        if (count == null) {
            count = new Long(-1);
        } else {
            count = count - 1;
        }
        featureTypeCounts.put(featureType, count);
        return count;
    }

    private Long addFeatureTypeCount(FeatureType featureType) {
        Long count = (Long) featureTypeCounts.get(featureType);
        if (count == null) {
            count = new Long(1);
        } else {
            count = count + 1;
        }
        featureTypeCounts.put(featureType, count);
        return count;
    }

    @Override
    public void delete(Feature feature) throws DataException {
        Iterator it = this.featureIterators.get(feature);
        if (it != null) {
            it.remove();
            return;
        }
        feature.getStore().delete(feature);
    }

    @Override
    public void insert(EditableFeature feature) throws DataException {
        feature.getStore().insert(feature);
    }

    @Override
    public void update(EditableFeature feature) throws DataException {
        feature.getStore().update(feature);
    }
    
    
    @Override
    public void commitChanges() throws DataException {
    }

    /*
     * (non-Javadoc)
     *
     * @seeorg.gvsig.fmap.dal.feature.impl.DefaultFeatureReferenceSelection#
     * loadFromState(org.gvsig.tools.persistence.PersistentState)
     */
    @Override
    public void loadFromState(PersistentState state)
            throws PersistenceException {
        this.featureReferenceSelection.loadFromState(state);
    }

    protected void doDispose() throws BaseException {
        this.featureReferenceSelection.doDispose();
        featureTypeCounts.clear();
    }

    public static void registerPersistent() {
        PersistenceManager manager = ToolsLocator.getPersistenceManager();
        DynStruct definition = manager.addDefinition(
                DefaultFeatureSelection.class, "DefaultFeatureSelection",
                "DefaultFeatureSelection Persistent definition", null, null);

        definition.extend(manager.getDefinition(DefaultFeatureReferenceSelection.DYNCLASS_PERSISTENT_NAME));
        definition.addDynFieldMap("featureTypeCounts")
                .setClassOfItems(Long.class).setMandatory(false);

    }

    @Override
    public Object clone() throws CloneNotSupportedException {
        DefaultFeatureSelection clone = (DefaultFeatureSelection) super.clone();
        clone.featureTypeCounts = new HashMap(featureTypeCounts);
        return clone;
    }

    @Override
    protected void doAccept(Visitor visitor, long firstValueIndex, long elements) throws BaseException {
        if( this.featureReferenceSelection.isEmpty() ) {
            return;
        }
        DisposableIterator iterator = fastIterator(firstValueIndex, elements);
        if (iterator != null) {
            try {
                while (iterator.hasNext()) {
                    Feature feature = (Feature) iterator.next();
                    visitor.visit(feature);
                }
            } finally {
                iterator.dispose();
            }
        }
    }

    private static class IgnoreInsertAndUpdateFeatureSet extends DefaultFeatureSet {

        public IgnoreInsertAndUpdateFeatureSet(DefaultFeatureStore store, FeatureQuery query) throws DataException {
            super(store, query);
        }

        public void update(Observable obsevable, Object notification) {
            String type = ((FeatureStoreNotification) notification).getType();
            if (type.equalsIgnoreCase(FeatureStoreNotification.AFTER_INSERT)
                    || type.equalsIgnoreCase(FeatureStoreNotification.AFTER_UPDATE)) {
                return;
            }
            super.update(obsevable, notification);
        }

    }

    @Override
    public boolean isAvailable() {
        return this.featureReferenceSelection.isAvailable();
    }

}
