/* gvSIG. Desktop Geographic Information System.
 *
 * Copyright  2007-2017 gvSIG Association
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 *
 * For any additional information, do not hesitate to contact us
 * at info AT gvsig.com, or visit our website www.gvsig.com.
 */
package org.gvsig.fmap.dal.feature.impl;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import org.gvsig.fmap.dal.feature.Feature;
import org.gvsig.fmap.dal.feature.FeatureReference;
import org.gvsig.fmap.dal.feature.FeatureStore;
import org.gvsig.fmap.dal.feature.spi.FeatureReferenceProviderServices;
import org.gvsig.fmap.geom.Geometry;
import org.gvsig.fmap.geom.SpatialIndex;
import org.gvsig.fmap.geom.SpatialIndexFactory;
import org.gvsig.fmap.geom.primitive.Envelope;
import org.gvsig.tools.exception.BaseException;
import org.gvsig.tools.service.Manager;
import org.gvsig.tools.visitor.VisitCanceledException;
import org.gvsig.tools.visitor.Visitor;

/**
 * @author fdiaz
 *
 */
public class WrappedSpatialIndex implements SpatialIndex {

    private final SpatialIndex index;
    private final FeatureStore store;

    protected class Oid2FeatureReferenceVisitor implements Visitor {

        private final Visitor wrapedVisitor;

        public Oid2FeatureReferenceVisitor(Visitor wrapedVisitor) {
            this.wrapedVisitor = wrapedVisitor;
        }

        @Override
        public void visit(Object oid) throws VisitCanceledException, BaseException {
            DefaultFeatureReference ref = new DefaultFeatureReference(store, oid);
            wrapedVisitor.visit(ref);
        }

    }

    protected class Oid2FeatureReferenceIterator implements Iterator<FeatureReference> {

        private final Iterator wrapedIterator;

        public Oid2FeatureReferenceIterator(Iterator wrapedIterator) {
            this.wrapedIterator = wrapedIterator;
        }

        @Override
        public boolean hasNext() {
            return this.wrapedIterator.hasNext();
        }

        @Override
        public FeatureReference next() {
            Object oid = this.wrapedIterator.next();
            DefaultFeatureReference ref = new DefaultFeatureReference(store, oid);
            return ref;
        }

        @Override
        public void remove() {
            this.wrapedIterator.remove();
        }
    }

    public WrappedSpatialIndex(SpatialIndex index, FeatureStore store) {
        this.index = index;
        this.store = store;
    }

    @Override
    public SpatialIndexFactory getFactory() {
        return this.index.getFactory();
    }

    @Override
    public void open() {
        this.index.open();
    }

    @Override
    public void close() {
        this.index.close();
    }

    @Override
    public void query(Envelope envelope, Visitor visitor) {
        this.index.query(envelope, new Oid2FeatureReferenceVisitor(visitor));
    }

    @Override
    public void query(Geometry geom, Visitor visitor) {
        this.index.query(geom, new Oid2FeatureReferenceVisitor(visitor));
    }

    @Override
    public Iterator<FeatureReference> query(Envelope envelope, long limit) {
        return new Oid2FeatureReferenceIterator(this.index.query(envelope, limit));
    }

    @Override
    public Iterator<FeatureReference> query(Envelope envelope) {
        return new Oid2FeatureReferenceIterator(this.index.query(envelope));
    }

    @Override
    public Iterator<FeatureReference> query(Geometry geom, long limit) {
        return new Oid2FeatureReferenceIterator(this.index.query(geom, limit));
    }

    @Override
    public Iterator<FeatureReference> query(Geometry geom) {
        return new Oid2FeatureReferenceIterator(this.index.query(geom));
    }

    @Override
    public Iterator<FeatureReference> queryNearest(Envelope envelope, long limit) {
        return new Oid2FeatureReferenceIterator(this.index.queryNearest(envelope, limit));
    }

    @Override
    public Iterator <FeatureReference>queryNearest(Envelope envelope) {
        return new Oid2FeatureReferenceIterator(this.index.queryNearest(envelope));
    }

    @Override
    public Iterator<FeatureReference> queryNearest(Geometry geom, long limit) {
        return new Oid2FeatureReferenceIterator(this.index.queryNearest(geom, limit));
    }

    @Override
    public Iterator<FeatureReference> queryNearest(Geometry geom) {
        return new Oid2FeatureReferenceIterator(this.index.queryNearest(geom));
    }

    @Override
    public Iterator<FeatureReference> queryAll() {
        return new Oid2FeatureReferenceIterator(this.index.queryAll());
    }

    @Override
    public List<FeatureReference> queryAsList(Envelope envelope) {
        return asList( this.query(envelope));
    }

    @Override
    public List<FeatureReference> queryAsList(Geometry geom) {
        return asList( this.query(geom));
    }

    @Override
    public List<FeatureReference> queryAllAsList() {
        return asList( this.queryAll());
    }

    @Override
    public void insert(Envelope envelope, Object data) {
        this.index.insert(envelope, getOID(data));
    }

    @Override
    public void insert(Geometry geom, Object data) {
        this.index.insert(geom, getOID(data));
    }

    @Override
    public void insert(Geometry geom) {
        this.index.insert(geom);
    }

    @Override
    public boolean remove(Envelope envelope, Object data) {
        return this.index.remove(envelope, getOID(data));
    }

    @Override
    public boolean remove(Geometry geom, Object data) {
        return this.index.remove(geom, getOID(data));
    }

    @Override
    public boolean remove(Geometry geom) {
        return this.index.remove(geom);
    }

    @Override
    public void removeAll() {
        this.index.removeAll();
    }

    @Override
    public long size() {
        return this.index.size();
    }

    @Override
    public void flush() {
        this.index.flush();
    }

    @Override
    public Manager getManager() {
        return this.index.getManager();
    }

    protected Object getOID(Object obj) {
        if( obj == null ) {
            return null;
        }
        if( obj instanceof FeatureReferenceProviderServices ) {
            return ((FeatureReferenceProviderServices)obj).getOID();
        }
        if( obj instanceof Feature ) {
            FeatureReference f = ((Feature)obj).getReference();
            return ((FeatureReferenceProviderServices)f).getOID();
        }
        if( obj instanceof Number ) {
            return obj;
        }
        throw new IllegalArgumentException("Can't get OID from object of type '"+obj.getClass().getName()+"'.");
    }

    protected List<FeatureReference> asList(Iterator it) {
        List l = new ArrayList();
        while (it.hasNext()) {
            l.add(it.next());
        }
        return l;
    }
}