/**
 * gvSIG. Desktop Geographic Information System.
 *
 * Copyright (C) 2007-2013 gvSIG Association.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 *
 * For any additional information, do not hesitate to contact us
 * at info AT gvsig.com, or visit our website www.gvsig.com.
 */
package org.gvsig.fmap.dal.feature.impl.featureset;

import org.gvsig.fmap.dal.DataStore;
import org.gvsig.fmap.dal.exception.DataException;
import org.gvsig.fmap.dal.feature.Feature;
import org.gvsig.fmap.dal.feature.FeatureSet;
import org.gvsig.fmap.dal.feature.FeatureStore;
import org.gvsig.fmap.dal.feature.impl.dynobjectutils.DynObjectSetFeatureSetFacade;
import org.gvsig.tools.dispose.DisposableIterator;
import org.gvsig.tools.dispose.DisposeUtils;
import org.gvsig.tools.dynobject.DynObjectSet;
import org.gvsig.tools.exception.BaseException;
import org.gvsig.tools.visitor.VisitCanceledException;
import org.gvsig.tools.visitor.Visitor;
import org.gvsig.tools.visitor.impl.AbstractIndexedVisitable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;


public abstract class AbstractFeatureSet 
    extends AbstractIndexedVisitable 
    implements FeatureSet {

    protected static final Logger LOG = LoggerFactory.getLogger(AbstractFeatureSet.class);

    public abstract FeatureStore getFeatureStore();
    
    @Override
    public final void accept(Visitor visitor, long firstValueIndex, long elements) throws BaseException {
        try {
            doAccept(visitor, firstValueIndex, elements);
        } catch (VisitCanceledException ex) {
            // The visit has been cancelled by the visitor, so we finish here.
            LOG.debug(
                    "The visit, beggining on position {}, has been cancelled "
                    + "by the visitor: {}", new Long(firstValueIndex),
                    visitor);
        }
    }
    
    @Override
    protected void doAccept(Visitor visitor, long firstValueIndex)
        throws VisitCanceledException, BaseException {
        doAccept(visitor, firstValueIndex, 0);
    }

    protected void doAccept(Visitor visitor, long firstValueIndex, long elements)
        throws VisitCanceledException, BaseException {
        DisposableIterator iterator = fastIterator(firstValueIndex, elements);

        try {
            while (iterator.hasNext()) {
                Feature feature = (Feature) iterator.next();
                visitor.visit(feature);
            }
        } finally {
            iterator.dispose();
        }
    }
      
    @Override
    public Feature first() {
        DisposableIterator it = null;
        try {
            it = this.iterator();
            if( it == null ) {
                return null;
            }
            if( it.hasNext() ) {
                Feature f = (Feature) it.next();
                return f;
            }
            return null;
        } finally {
            DisposeUtils.disposeQuietly(it);
        }
    }

    @Override
    public boolean isEmpty() throws DataException {
        return this.getSize() == 0;
    }

    @Override
    public DisposableIterator fastIterator() throws DataException {
        return this.fastIterator(0);
    }   

    @Override
    public DisposableIterator iterator() {
        try {
            return this.fastIterator(0);
        } catch (DataException ex) {
            throw new RuntimeException("Can't obtain itertor.",ex);
        }
    }
    
    @Override
    public DynObjectSet getDynObjectSet() {
        return this.getDynObjectSet(true);
    }

    @Override
    public DynObjectSet getDynObjectSet(boolean fast) {
        return new DynObjectSetFeatureSetFacade(this, this.getFeatureStore(), fast);
    }

    @Override
    public boolean isFromStore(DataStore store) {
        return this.getFeatureStore().equals(store);
    }    

    @Override
    public long size64() {
        try {
            return this.getSize();
        } catch (DataException ex) {
            throw new RuntimeException("Can't get size",ex);
        }
    }

    @Override
    public int size() {
        try {
            long sz = this.getSize();
            if( sz > Integer.MAX_VALUE ) {
                return Integer.MAX_VALUE;
            }
            return (int) sz;
        } catch (DataException ex) {
            throw new RuntimeException("Can't get size",ex);
        }
    }

}
