package org.gvsig.fmap.dal.impl.expressionevaluator;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import org.gvsig.expressionevaluator.Code;
import org.gvsig.expressionevaluator.SymbolTable;
import org.gvsig.expressionevaluator.Expression;
import org.gvsig.expressionevaluator.ExpressionEvaluatorLocator;
import org.gvsig.expressionevaluator.ExpressionEvaluatorManager;
import org.gvsig.expressionevaluator.Function;
import org.gvsig.fmap.dal.expressionevaluator.ExpressionEvaluator;
import org.gvsig.fmap.dal.expressionevaluator.FeatureSymbolTable;
import org.gvsig.fmap.dal.feature.Feature;
import org.gvsig.tools.evaluator.AbstractEvaluator;
import org.gvsig.tools.evaluator.Evaluator;
import org.gvsig.tools.evaluator.EvaluatorData;
import org.gvsig.tools.evaluator.EvaluatorException;
import org.gvsig.tools.evaluator.EvaluatorFieldsInfo;
import org.gvsig.tools.exception.BaseException;
import org.gvsig.tools.visitor.VisitCanceledException;
import org.gvsig.tools.visitor.Visitor;
import org.gvsig.tools.lang.Cloneable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

@SuppressWarnings("UseSpecificCatch")
public class DefaultExpressionEvaluator 
        extends AbstractEvaluator 
        implements ExpressionEvaluator, Cloneable 
    {

    private static final Logger LOGGER = LoggerFactory.getLogger(DefaultExpressionEvaluator.class);
    
    private static class DescriptionAdapter implements Description {

        Function function;

        public DescriptionAdapter(Function function) {
            this.function = function;
        }

        @Override
        public String getName() {
            return this.function.name();
        }

        @Override
        public String getDescription() {
            return this.function.description();
        }

        @Override
        public String getTemplate() {
            return this.function.template();
        }

        @Override
        public int getDataTypeCategories() {
            switch (this.function.group()) {
                case Function.GROUP_STRING:
                    return Description.DATATYPE_CATEGORY_STRING;
                case Function.GROUP_BOOLEAN:
                    return Description.DATATYPE_CATEGORY_BOOLEAN;
                case Function.GROUP_DATETIME:
                    return Description.DATATYPE_CATEGORY_DATETIME;
                case Function.GROUP_NUMERIC:
                    return Description.DATATYPE_CATEGORY_NUMBER;
                case Function.GROUP_OGC:
                    return Description.DATATYPE_CATEGORY_ALL;
                default:
                    return Description.DATATYPE_CATEGORY_ALL;
            }
        }

    }

    private FeatureSymbolTable featureSymbolTable;
    private SymbolTable symbolTable;
    private Expression expression;
    private Description[] availableOperators;
    private Description[] availableFunctions;

    public DefaultExpressionEvaluator(String source) {
        ExpressionEvaluatorManager manager = ExpressionEvaluatorLocator.getManager();
        this.symbolTable = manager.createSymbolTable();
        this.featureSymbolTable = new DefaultFeatureSymbolTable();
        this.symbolTable.addSymbolTable(this.featureSymbolTable);
        
        this.expression = manager.createExpression();
        this.expression.setPhrase(source);
    }

    public DefaultExpressionEvaluator(Expression expression) {
        ExpressionEvaluatorManager manager = ExpressionEvaluatorLocator.getManager();
        this.expression = expression;
        this.symbolTable = manager.createSymbolTable();
        this.featureSymbolTable = new DefaultFeatureSymbolTable(
//                expression.getUserScript(), 
//                expression.getScripts()
        );
        this.symbolTable.addSymbolTable(this.featureSymbolTable);

    }

    @Override
    public String getName() {
        return "Genereric expression";
    }

    @Override
    public String getDescription() {
        return this.expression.getPhrase();
    }

    @Override
    public String getSQL() {
        return this.expression.getPhrase();
    }

    @Override
    public Expression getExpression() {
        return expression;
    }

    @Override
    public SymbolTable getSymbolTable() {
        return this.symbolTable;
    }
    
    @Override
    public Object evaluate(EvaluatorData data) throws EvaluatorException {
        this.featureSymbolTable.setFeature((Feature) data);
        Object v = this.expression.execute(this.symbolTable);
        return v;
    }

    @Override
    public EvaluatorFieldsInfo getFieldsInfo() {
        final Set<String> names = new HashSet<>();
        try {
            this.expression.getCode().accept(new Visitor() {
                @Override
                public void visit(Object code) throws VisitCanceledException, BaseException {
                    if (code instanceof Code.Identifier) {
                        Code.Identifier identifier = (Code.Identifier) code;
                        names.add(identifier.name());
                    }
                }
            });
            EvaluatorFieldsInfo info = new EvaluatorFieldsInfo();
            for (String name : names) {
                info.addFieldValue(name);
            }
            return info;
        } catch (BaseException ex) {
            throw new RuntimeException("Can't calculate fields information.", ex);
        }
    }

    @Override
    public Description[] getAvailableOperators() {
        if (availableOperators == null) {
            List<Description> l = new ArrayList<>();
            for (Function function : getSymbolTable()) {
                if (function.isOperator()) {
                    l.add(new DescriptionAdapter(function));
                }
            }
            this.availableOperators = l.toArray(new Description[l.size()]);
        }
        return availableOperators;
    }

    @Override
    public Description[] getAvailableFunctions() {
        if (availableFunctions == null) {
            List<Description> l = new ArrayList<>();
            for (Function function : getSymbolTable()) {
                if (!function.isOperator()) {
                    l.add(new DescriptionAdapter(function));
                }
            }
            this.availableFunctions = l.toArray(new Description[l.size()]);
        }
        return availableFunctions;
    }

    @Override
    public Evaluator clone() throws CloneNotSupportedException {
        DefaultExpressionEvaluator other = (DefaultExpressionEvaluator) super.clone();
        other.expression = this.expression.clone();
        other.availableFunctions = null;
        other.availableOperators = null;

        if (this.featureSymbolTable != null) {
            other.featureSymbolTable = this.featureSymbolTable.clone();
        }
        if (this.symbolTable != null) {
            other.symbolTable = this.symbolTable.clone();
        }
        return other;
    }

}
