package org.gvsig.fmap.dal.store;

import org.apache.commons.lang3.ArrayUtils;
import org.cresques.cts.IProjection;
import org.gvsig.expressionevaluator.ExpressionBuilder;
import org.gvsig.fmap.crs.CRSFactory;
import org.gvsig.fmap.dal.SQLBuilder;
import org.gvsig.fmap.dal.SQLBuilder.Privilege;
import org.gvsig.fmap.dal.feature.spi.SQLBuilderBase;
import org.gvsig.fmap.geom.DataTypes;
import org.gvsig.fmap.geom.Geometry;
import org.gvsig.fmap.geom.GeometryLocator;
import org.gvsig.fmap.geom.GeometryManager;
import org.gvsig.fmap.geom.primitive.Polygon;
import org.gvsig.tools.junit.AbstractLibraryAutoInitTestCase;

public class SQLBuilderTest extends AbstractLibraryAutoInitTestCase {
    
    @Override
    protected void doSetUp() throws Exception {

    }
        
    private SQLBuilder createSQLBuilder() {
        return new SQLBuilderBase();
    }
    
    public void testCalulateEnvelope() throws Exception {
        GeometryManager geometryManager = GeometryLocator.getGeometryManager();
        IProjection proj = CRSFactory.getCRS("EPSG:4326");
        
        Polygon limit = geometryManager.createPolygon(Geometry.SUBTYPES.GEOM2D);
        limit.addVertex(0, 0);
        limit.addVertex(0, 100);
        limit.addVertex(100, 100);
        limit.addVertex(100, 0);
        limit.addVertex(0, 0);
        
        SQLBuilder sqlbuilder = createSQLBuilder();
        ExpressionBuilder expbuilder = sqlbuilder.expression();
        
        sqlbuilder.select().column().value(
            expbuilder.as_geometry(
              expbuilder.ST_ExtentAggregate(
                expbuilder.column("the_geom")
              )
            )
        ).as("envelope");
        sqlbuilder.select().from().table().database("master").schema("dbo").name("test1");
        sqlbuilder.select().where().set(
            expbuilder.ST_Intersects(
                expbuilder.ST_Envelope(
                    expbuilder.column("the_geom")
                ),
                expbuilder.geometry(limit, proj)
            )
        );
        sqlbuilder.select().where().and(
                expbuilder.custom("x = 27")
        );
        
        // SELECT ST_AsBinary(ST_Extent("the_geom")) AS "envelope" FROM "master"."dbo"."test1" WHERE ST_Intersects((ST_Envelope("the_geom")), (ST_GeomFromText('POLYGON ((0 0, 0 100, 100 100, 100 0, 0 0))', (4326)))) AND x = 27
        
        System.out.println("# Test:: testCalulateEnvelope");
        System.out.println("# SQL:: " + sqlbuilder.toString());
        System.out.println("# Variables:: " + ArrayUtils.toString(expbuilder.variables_names()));
        System.out.println("# Parametros:: " + ArrayUtils.toString(expbuilder.parameters_names()));
        assertEquals(
                "SELECT ST_AsBinary(ST_Extent(\"the_geom\")) AS \"envelope\" FROM \"master\".\"dbo\".\"test1\" WHERE ST_Intersects((ST_Envelope(\"the_geom\")), (ST_GeomFromText('POLYGON ((0 0, 0 100, 100 100, 100 0, 0 0))', (4326)))) AND x = 27",
                sqlbuilder.toString()
        );
        assertEquals(
                "[the_geom]",
                ArrayUtils.toString(expbuilder.variables_names())
        );
        assertEquals(
                "[]",
                ArrayUtils.toString(expbuilder.parameters_names())
        );
    }

    public void testCount() throws Exception {
        SQLBuilder sqlbuilder = createSQLBuilder();
        ExpressionBuilder expbuilder = sqlbuilder.expression();
        
        sqlbuilder.select().column().value(sqlbuilder.count().all());
        sqlbuilder.select().from().table().database("master").schema("dbo").name("test1");
        sqlbuilder.select().from().subquery(null);
        sqlbuilder.select().where().set( expbuilder.custom("pp = 200"));

        System.out.println("# Test:: testCount");
        System.out.println("# SQL:: " + sqlbuilder.toString());
        System.out.println("# Variables:: " + ArrayUtils.toString(expbuilder.variables_names()));
        System.out.println("# Parametros:: " + ArrayUtils.toString(expbuilder.parameters_names()));

        //# Test:: testCount
        //# SQL:: SELECT COUNT(*) FROM "master"."dbo"."test1" WHERE pp = 200
        //# Variables:: []
        //# Parametros:: []

        assertEquals(
                "SELECT COUNT(*) FROM \"master\".\"dbo\".\"test1\" WHERE pp = 200",
                sqlbuilder.toString()
        );
        assertEquals(
                "[]",
                ArrayUtils.toString(expbuilder.variables_names())
        );
        assertEquals(
                "[]",
                ArrayUtils.toString(expbuilder.parameters_names())
        );
    }
    
    public void testCreateTable() throws Exception {
        SQLBuilder sqlbuilder = createSQLBuilder();
        ExpressionBuilder expbuilder = sqlbuilder.expression();

        sqlbuilder.create_table().table().database("master").schema("dbo").name("test1");
        sqlbuilder.create_table().add_column(
                "name",
                DataTypes.STRING,
                45,
                0,
                false,
                false,
                true,
                false,
                null
        );
        sqlbuilder.create_table().add_column(
                "id",
                DataTypes.INT,
                0,
                0,
                true,
                false,
                false,
                true,
                0
        );
        sqlbuilder.create_table().add_column(
                "geom",
                DataTypes.GEOMETRY,
                0,
                0,
                false,
                false,
                true,
                false,
                null
        );

        
        // CREATE TABLE "master"."dbo"."test1" ("name" VARCHAR(45) DEFAULT NULL NULL, "id" SERIAL DEFAULT '0' NOT NULL PRIMARY KEY, "geom" TEXT DEFAULT NULL NULL )
        System.out.println("# Test:: testCreateTable");
        System.out.println("# SQL:: " + sqlbuilder.toString());
        System.out.println("# Variables:: " + ArrayUtils.toString(expbuilder.variables_names()));
        System.out.println("# Parametros:: " + ArrayUtils.toString(expbuilder.parameters_names()));
        assertEquals(
                "CREATE TABLE \"master\".\"dbo\".\"test1\" (\"name\" VARCHAR(45) DEFAULT NULL NULL, \"id\" SERIAL DEFAULT '0' NOT NULL PRIMARY KEY, \"geom\" TEXT DEFAULT NULL NULL )",
                sqlbuilder.toString()
        );
        assertEquals(
                "[]",
                ArrayUtils.toString(expbuilder.variables_names())
        );
        assertEquals(
                "[]",
                ArrayUtils.toString(expbuilder.parameters_names())
        );
    }

    public void testDropTable() throws Exception {
        SQLBuilder sqlbuilder = createSQLBuilder();
        ExpressionBuilder expbuilder = sqlbuilder.expression();
        
        sqlbuilder.drop_table().table().database("master").schema("dbo").name("test1");

        // DROP TABLE "master"."dbo"."test1"; DELETE FROM GEOMETRY_COLUMNS WHERE f_table_schema = 'dbo' AND f_table_name = 'test1'
        
        System.out.println("# Test:: testDropTable");
        System.out.println("# SQL:: " + sqlbuilder.toString());
        System.out.println("# Variables:: " + ArrayUtils.toString(expbuilder.variables_names()));
        System.out.println("# Parametros:: " + ArrayUtils.toString(expbuilder.parameters_names()));
        assertEquals(
                "DROP TABLE \"master\".\"dbo\".\"test1\"; DELETE FROM GEOMETRY_COLUMNS WHERE f_table_schema = 'dbo' AND f_table_name = 'test1'",
                sqlbuilder.toString()
        );
        assertEquals(
                "[]",
                ArrayUtils.toString(expbuilder.variables_names())
        );
        assertEquals(
                "[]",
                ArrayUtils.toString(expbuilder.parameters_names())
        );
    }
    
    public void testFetchFeatureProviderByReference() throws Exception {
        SQLBuilder sqlbuilder = createSQLBuilder();
        ExpressionBuilder expbuilder = sqlbuilder.expression();
        
        String value = "yoyo";
        sqlbuilder.select().column().name("name");
        sqlbuilder.select().column().name("id");
        sqlbuilder.select().column().name("geom").as_geometry();
        sqlbuilder.select().from().table().database("master").schema("dbo").name("test1");
        sqlbuilder.select().where().set(
            expbuilder.eq(
                expbuilder.column("name"),
                expbuilder.parameter(value).as_constant()
            )
        );
        sqlbuilder.select().limit(1);

        // SELECT "name", "id", ST_AsBinary("geom") FROM "master"."dbo"."test1" WHERE ( ("name") = (?) ) LIMIT 1

        System.out.println("# Test:: testFetchFeatureProviderByReference");
        System.out.println("# SQL:: " + sqlbuilder.toString());
        System.out.println("# Variables:: " + ArrayUtils.toString(expbuilder.variables_names()));
        System.out.println("# Parametros:: " + ArrayUtils.toString(expbuilder.parameters_names()));
        assertEquals(
                "SELECT \"name\", \"id\", ST_AsBinary(\"geom\") FROM \"master\".\"dbo\".\"test1\" WHERE ( (\"name\") = (?) ) LIMIT 1",
                sqlbuilder.toString()
        );
        assertEquals(
                "[geom, id, name]",
                ArrayUtils.toString(expbuilder.variables_names())
        );
        assertEquals(
                "['yoyo']",
                ArrayUtils.toString(expbuilder.parameters_names())
        );
    }
    
    public void testFetchFeatureType() throws Exception {
        SQLBuilder sqlbuilder = createSQLBuilder();
        ExpressionBuilder expbuilder = sqlbuilder.expression();

        sqlbuilder.select().column().all();
        sqlbuilder.select().from().table().database("master").schema("dbo").name("test1");
        sqlbuilder.select().limit(1);

        System.out.println("# Test:: testFetchFeatureType");
        System.out.println("# SQL:: " + sqlbuilder.toString());
        System.out.println("# Variables:: " + ArrayUtils.toString(expbuilder.variables_names()));
        System.out.println("# Parametros:: " + ArrayUtils.toString(expbuilder.parameters_names()));
        
        //# Test:: testFetchFeatureType
        //# SQL:: SELECT * FROM [master].[dbo].[test1] LIMMIT 1
        //# Variables:: []
        //# Parametros:: []        
        
        assertEquals(
                "SELECT * FROM \"master\".\"dbo\".\"test1\" LIMIT 1",
                sqlbuilder.toString()
        );
        assertEquals(
                "[]",
                ArrayUtils.toString(expbuilder.variables_names())
        );
        assertEquals(
                "[]",
                ArrayUtils.toString(expbuilder.parameters_names())
        );
    }
        
    public void testPerformDeletes() throws Exception {
        SQLBuilder sqlbuilder = createSQLBuilder();
        ExpressionBuilder expbuilder = sqlbuilder.expression();

        sqlbuilder.delete().table().database("master").schema("dbo").name("test1");
        sqlbuilder.delete().where().and(
            expbuilder.eq( 
                expbuilder.column("id1"),
                expbuilder.parameter("id1").as_variable()
            )
        );
        sqlbuilder.delete().where().and(
            expbuilder.eq( 
                expbuilder.column("id2"),
                expbuilder.parameter("id2").as_variable()
            )
        );

        // DELETE FROM "master"."dbo"."test1" WHERE ( ("id1") = (?) ) AND ( ("id2") = (?) )

        System.out.println("# Test:: testPerformDeletes");
        System.out.println("# SQL:: " + sqlbuilder.toString());
        System.out.println("# Variables:: " + ArrayUtils.toString(expbuilder.variables_names()));
        System.out.println("# Parametros:: " + ArrayUtils.toString(expbuilder.parameters_names()));
        assertEquals(
                "DELETE FROM \"master\".\"dbo\".\"test1\" WHERE ( (\"id1\") = (?) ) AND ( (\"id2\") = (?) )",
                sqlbuilder.toString()
        );
        assertEquals(
                "[id1, id2]",
                ArrayUtils.toString(expbuilder.variables_names())
        );
        assertEquals(
                "[\"id1\", \"id2\"]",
                ArrayUtils.toString(expbuilder.parameters_names())
        );
    }

    public void testPerformInserts1() throws Exception {
        IProjection proj = CRSFactory.getCRS("EPSG:4326");

        SQLBuilder sqlbuilder = createSQLBuilder();
        ExpressionBuilder expbuilder = sqlbuilder.expression();

        sqlbuilder.insert().table().database("master").schema("dbo").name("test1");
        sqlbuilder.insert().column().name("id").with_value(expbuilder.parameter("id"));
        sqlbuilder.insert().column().name("name").with_value(expbuilder.parameter("name"));
        sqlbuilder.insert().column().name("geom").with_value(expbuilder.parameter("geom").as_geometry_variable().srs(proj));
        
        // INSERT INTO "master"."dbo"."test1" ( "id", "name", "geom" ) VALUES ( ?, ?, ST_GeomFromText(?, (4326)) )
        
        System.out.println("# Test:: testPerformInserts1");
        System.out.println("# SQL:: " + sqlbuilder.toString());
        System.out.println("# Variables:: " + ArrayUtils.toString(expbuilder.variables_names()));
        System.out.println("# Parametros:: " + ArrayUtils.toString(expbuilder.parameters_names()));
        assertEquals(
                "INSERT INTO \"master\".\"dbo\".\"test1\" ( \"id\", \"name\", \"geom\" ) VALUES ( ?, ?, ST_GeomFromText(?, (4326)) )",
                sqlbuilder.toString()
        );
        assertEquals(
                "[geom, id, name]",
                ArrayUtils.toString(expbuilder.variables_names())
        );
        assertEquals(
                "[\"id\", \"name\", \"geom\"]",
                ArrayUtils.toString(expbuilder.parameters_names())
        );
    }

    public void testPerformInserts2() throws Exception {
        IProjection proj = CRSFactory.getCRS("EPSG:4326");

        SQLBuilder sqlbuilder = createSQLBuilder();
        ExpressionBuilder expbuilder = sqlbuilder.expression();

        sqlbuilder.insert().table().database("master").schema("dbo").name("test1");
        sqlbuilder.insert().column().name("id").with_value(expbuilder.parameter("id"));
        sqlbuilder.insert().column().name("name").with_value(expbuilder.parameter("name"));
        sqlbuilder.insert().column().name("geom").with_value(expbuilder.parameter("geom").as_geometry_variable().srs(expbuilder.parameter().value(proj)));
        
        System.out.println("# Test:: testPerformInserts2");
        System.out.println("# SQL:: " + sqlbuilder.toString());
        System.out.println("# Variables:: " + ArrayUtils.toString(expbuilder.variables_names()));
        System.out.println("# Parametros:: " + ArrayUtils.toString(expbuilder.parameters_names()));
        assertEquals(
                "INSERT INTO \"master\".\"dbo\".\"test1\" ( \"id\", \"name\", \"geom\" ) VALUES ( ?, ?, ST_GeomFromText(?, (?)) )",
                sqlbuilder.toString()
        );
        assertEquals(
                "[geom, id, name]",
                ArrayUtils.toString(expbuilder.variables_names())
        );
        assertEquals(
                "[\"id\", \"name\", \"geom\", 4326]",
                ArrayUtils.toString(expbuilder.parameters_names())
        );
    }

    public void testPerformUpdates() throws Exception {
        IProjection proj = CRSFactory.getCRS("EPSG:4326");

        SQLBuilder sqlbuilder = createSQLBuilder();
        ExpressionBuilder expbuilder = sqlbuilder.expression();

        sqlbuilder.update().table().database("master").schema("dbo").name("test1");
        sqlbuilder.update().where().and(
            expbuilder.eq(
                expbuilder.column("id"), 
                expbuilder.parameter("id").as_variable()
            )
        );
        sqlbuilder.update().column().name("name").with_value(expbuilder.parameter("name"));
        sqlbuilder.update().column().name("geom").with_value(
                expbuilder.parameter("geom").as_geometry_variable()
                        .srs(expbuilder.parameter().value(proj)) 
        );

        // UPDATE "master"."dbo"."test1" SET "name" = ?, "geom" = ST_GeomFromText(?, (?)) WHERE ( ("id") = (?) )

        System.out.println("# Test:: testPerformUpdates");
        System.out.println("# SQL:: " + sqlbuilder.toString());
        System.out.println("# Variables:: " + ArrayUtils.toString(expbuilder.variables_names()));
        System.out.println("# Parametros:: " + ArrayUtils.toString(expbuilder.parameters_names()));
        assertEquals(
                "UPDATE \"master\".\"dbo\".\"test1\" SET \"name\" = ?, \"geom\" = ST_GeomFromText(?, (?)) WHERE ( (\"id\") = (?) )",
                sqlbuilder.toString()
        );
        assertEquals(
                "[geom, id, name]",
                ArrayUtils.toString(expbuilder.variables_names())
        );
        assertEquals(
                "[\"name\", \"geom\", 4326, \"id\"]",
                ArrayUtils.toString(expbuilder.parameters_names())
        );
    }

    public void testGrant1() throws Exception {

        SQLBuilder sqlbuilder = createSQLBuilder();
        ExpressionBuilder expbuilder = sqlbuilder.expression();

        sqlbuilder.grant().table().database("master").schema("dbo").name("test1");
        sqlbuilder.grant().role("prueba").select().insert().update();
        sqlbuilder.grant().role("gis").all();
                
        
        System.out.println("# Test:: testGrant1");
        System.out.println("# SQL:: " + sqlbuilder.toString());
        System.out.println("# Variables:: " + ArrayUtils.toString(expbuilder.variables_names()));
        System.out.println("# Parametros:: " + ArrayUtils.toString(expbuilder.parameters_names()));
        assertEquals(
                "",
                sqlbuilder.toString()
        );
        assertEquals(
                "[]",
                ArrayUtils.toString(expbuilder.variables_names())
        );
        assertEquals(
                "[]",
                ArrayUtils.toString(expbuilder.parameters_names())
        );
    }

    public void testGrant2() throws Exception {

        SQLBuilder sqlbuilder = createSQLBuilder();
        ExpressionBuilder expbuilder = sqlbuilder.expression();

        sqlbuilder.grant().table().database("master").schema("dbo").name("test1");
        sqlbuilder.grant().role("prueba").privilege(Privilege.SELECT)
                .privilege(Privilege.INSERT)
                .privilege(Privilege.UPDATE);
        sqlbuilder.grant().role("gis").privilege(Privilege.ALL);
                
        
        System.out.println("# Test:: testGrant2");
        System.out.println("# SQL:: " + sqlbuilder.toString());
        System.out.println("# Variables:: " + ArrayUtils.toString(expbuilder.variables_names()));
        System.out.println("# Parametros:: " + ArrayUtils.toString(expbuilder.parameters_names()));
        assertEquals(
                "",
                sqlbuilder.toString()
        );
        assertEquals(
                "[]",
                ArrayUtils.toString(expbuilder.variables_names())
        );
        assertEquals(
                "[]",
                ArrayUtils.toString(expbuilder.parameters_names())
        );
    }


}
