package org.gvsig.fmap.dal.feature.spi;

import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.ImmutablePair;
import org.apache.commons.lang3.tuple.Pair;
import org.cresques.cts.IProjection;
import org.gvsig.expressionevaluator.ExpressionBuilder;
import org.gvsig.expressionevaluator.ExpressionBuilder.AbstractValue;
import org.gvsig.expressionevaluator.ExpressionBuilder.ClassVisitorFilter;
import static org.gvsig.expressionevaluator.ExpressionBuilder.EMPTY_FORMATTER;
import org.gvsig.expressionevaluator.ExpressionBuilder.GeometrySupportType;
import org.gvsig.expressionevaluator.ExpressionBuilder.Parameter;
import org.gvsig.expressionevaluator.ExpressionBuilder.Value;
import org.gvsig.expressionevaluator.ExpressionBuilder.Variable;
import org.gvsig.expressionevaluator.ExpressionBuilder.Visitable;
import org.gvsig.expressionevaluator.ExpressionBuilder.Visitor;
import org.gvsig.expressionevaluator.ExpressionBuilder.VisitorFilter;
import org.gvsig.expressionevaluator.ExpressionEvaluatorLocator;
import org.gvsig.expressionevaluator.Formatter;
import org.gvsig.fmap.dal.DataStoreParameters;
import org.gvsig.fmap.dal.DataTypes;
import org.gvsig.fmap.dal.SQLBuilder;
import org.gvsig.fmap.dal.SQLBuilder.AlterTableBuilder;
import org.gvsig.fmap.dal.SQLBuilder.CreateTableBuilder;
import org.gvsig.fmap.dal.SQLBuilder.DeleteBuilder;
import org.gvsig.fmap.dal.SQLBuilder.DropTableBuilder;
import org.gvsig.fmap.dal.SQLBuilder.FromBuilder;
import org.gvsig.fmap.dal.SQLBuilder.GrantBuilder;
import org.gvsig.fmap.dal.SQLBuilder.InsertBuilder;
import org.gvsig.fmap.dal.SQLBuilder.InsertColumnBuilder;
import org.gvsig.fmap.dal.SQLBuilder.OrderByBuilder;
import org.gvsig.fmap.dal.SQLBuilder.Privilege;
import org.gvsig.fmap.dal.SQLBuilder.SelectBuilder;
import org.gvsig.fmap.dal.SQLBuilder.SelectColumnBuilder;
import org.gvsig.fmap.dal.SQLBuilder.TableNameBuilder;
import org.gvsig.fmap.dal.SQLBuilder.UpdateBuilder;
import org.gvsig.fmap.dal.SQLBuilder.UpdateColumnBuilder;
import org.gvsig.fmap.dal.SQLBuilder.UpdateTableStatisticsBuilder;
import org.gvsig.fmap.dal.feature.FeatureAttributeDescriptor;
import org.gvsig.fmap.geom.Geometry;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

@SuppressWarnings("UseSpecificCatch")
public class SQLBuilderBase implements SQLBuilder {

    protected static final Logger LOGGER = LoggerFactory.getLogger(SQLBuilderBase.class);

    protected SelectBuilder select;
    protected UpdateBuilder update;
    protected InsertBuilder insert;
    protected DeleteBuilder delete;
    protected AlterTableBuilder alter_table;
    protected CreateTableBuilder create_table;
    protected GrantBuilder grant;
    protected DropTableBuilder drop_table;
    protected UpdateTableStatisticsBuilder update_table_statistics;
    protected CreateIndexBuilder create_index;
    protected TableNameBuilder table_name;

    protected abstract class AbstractStatementPart extends AbstractValue {

    }

    protected abstract class AbstractStatement extends AbstractStatementPart {

    }

    protected class ColumnDescriptorBase implements ColumnDescriptor {

        private String name;
        private int type;
        private int type_p;
        private int type_s;
        private boolean isPk;
        private boolean _allowNulls;
        private boolean _isAutomatic;
        private Object defaultValue;
        private int geom_type;
        private int geom_subtype;
        private Object geom_srsdbcode;
        private boolean _isIndexed;
        private DataStoreParameters parameters = null;

        public ColumnDescriptorBase(String name, int type, Object defaultValue) {
            this.name = name;
            this.type = type;
            this.type_p = -1;
            this.type_s = -1;
            this.isPk = false;
            this._allowNulls = true;
            this._isAutomatic = false;
            this.defaultValue = defaultValue;
            this.geom_type = Geometry.TYPES.GEOMETRY;
            this.geom_subtype = Geometry.SUBTYPES.GEOM2D;
            this.geom_srsdbcode = null;
            this._isIndexed = false;
        }

        public ColumnDescriptorBase(String name, int type, int type_p, int type_s, boolean isPk, boolean isIndexed, boolean allowNulls, boolean isAutomatic, Object defaultValue) {
            this.name = name;
            this.type = type;
            this.type_p = type_p;
            this.type_s = type_s;
            this.isPk = isPk;
            this._allowNulls = allowNulls;
            this._isAutomatic = isAutomatic;
            this.defaultValue = defaultValue;
            this.geom_type = Geometry.TYPES.GEOMETRY;
            this.geom_subtype = Geometry.SUBTYPES.GEOM2D;
            this.geom_srsdbcode = null;
            this._isIndexed = isIndexed;
        }

        public ColumnDescriptorBase(String name, int geom_type, int geom_subtype, IProjection proj, boolean isIndexed, boolean allowNulls) {
            this.name = name;
            this.type = DataTypes.GEOMETRY;
            this.type_p = 0;
            this.type_s = 0;
            this.isPk = false;
            this._allowNulls = allowNulls;
            this._isAutomatic = false;
            this.defaultValue = null;
            this.geom_type = geom_type;
            this.geom_subtype = geom_subtype;
            this.geom_srsdbcode = srs_id(proj);
            this._isIndexed = isIndexed;
        }

        public ColumnDescriptorBase(String name, int geom_type, int geom_subtype, Object srsdbcode, boolean isIndexed, boolean allowNulls) {
            this.name = name;
            this.type = DataTypes.GEOMETRY;
            this.type_p = 0;
            this.type_s = 0;
            this.isPk = false;
            this._allowNulls = allowNulls;
            this._isAutomatic = false;
            this.defaultValue = null;
            this.geom_type = geom_type;
            this.geom_subtype = geom_subtype;
            this.geom_srsdbcode = srsdbcode;
            this._isIndexed = isIndexed;
        }

        private ColumnDescriptorBase(FeatureAttributeDescriptor fad) {
            this(fad.getName(), fad.getType(), fad.getDefaultValue());
            this.type_p = fad.getPrecision();
            this.type_s = fad.getSize();
            this.isPk = fad.isPrimaryKey();
            this._allowNulls = fad.allowNull();
            this._isAutomatic = fad.isAutomatic();
            this._isIndexed = fad.isIndexed();

            if (fad.getType() == org.gvsig.fmap.geom.DataTypes.GEOMETRY) {
                this.geom_type = fad.getGeomType().getType();
                this.geom_subtype = fad.getGeomType().getSubType();
                this.geom_srsdbcode = fad.getSRS();
            }
        }
    

        @Override
        public String getName() {
            return this.name;
        }

        @Override
        public void setName(String name) {
            this.name = name;
        }

        @Override
        public int getType() {
            return this.type;
        }

        @Override
        public void setType(int type) {
            this.type = type;
        }

        @Override
        public int getPrecision() {
            return type_p;
        }

        @Override
        public void setPrecision(int precision) {
            this.type_p = precision;
        }

        @Override
        public int getSize() {
            return type_s;
        }

        @Override
        public void setSize(int size) {
            this.type_s = size;
        }

        @Override
        public boolean isPrimaryKey() {
            return isPk;
        }

        @Override
        public void setIsPrimaryKey(boolean isPk) {
            this.isPk = isPk;
        }

        @Override
        public boolean allowNulls() {
            return _allowNulls;
        }

        @Override
        public void setAllowNulls(boolean allowNulls) {
            this._allowNulls = allowNulls;
        }

        @Override
        public boolean isAutomatic() {
            return _isAutomatic;
        }

        @Override
        public boolean isIndexed() {
            return _isIndexed;
        }

        @Override
        public void setIsAutomatic(boolean isAutomatic) {
            this._isAutomatic = isAutomatic;
        }

        @Override
        public Object getDefaultValue() {
            return defaultValue;
        }

        @Override
        public void setDefaultValue(Object defaultValue) {
            this.defaultValue = defaultValue;
        }

        @Override
        public int getGeometryType() {
            return geom_type;
        }

        @Override
        public void setGeometryType(int geom_type) {
            this.geom_type = geom_type;
        }

        @Override
        public int getGeometrySubtype() {
            return geom_subtype;
        }

        @Override
        public void setGeometrySubtype(int geom_subtype) {
            this.geom_subtype = geom_subtype;
        }

        @Override
        public Object getGeometrySRSId() {
            return geom_srsdbcode;
        }

        @Override
        public void setGeometrySRSId(Object geom_srsid) {
            this.geom_srsdbcode = geom_srsid;
        }

        @Override
        public boolean isGeometry() {
            return this.type == DataTypes.GEOMETRY;
        }

        private void setStoreParameters(DataStoreParameters parameters) {
            this.parameters = parameters;
        }

        @Override
        public DataStoreParameters getStoreParameters() {
            return this.parameters;
        }
    }

    public class ColumnBase extends AbstractValue implements Column {

        private String name;
        private TableNameBuilder table;

        public ColumnBase(TableNameBuilder table, String name) {
            this.name = name;
            this.table = table;
        }

        @Override
        public String name() {
            return this.name;
        }

        @Override
        public TableNameBuilder table() {
            return this.table;
        }

        @Override
        public TableNameBuilder table(TableNameBuilder table) {
            this.table = table;
            return this.table;
        }

        @Override
        public String toString() {
            return this.toString(EMPTY_FORMATTER);
        }
        
        @Override
        public String toString(Formatter<Value> formatter) {
            if( formatter!=null && formatter.canApply(this) ) {
                return formatter.format(this);
            }
            if( this.table==null ) {
                return as_identifier(this.name);
            }
            return this.table.toString(formatter) + "." + as_identifier(this.name);
        }

        @Override
        public int compareTo(Variable o) {
            return this.name.compareTo(o.name());
        }

        @Override
        public boolean equals(Object obj) {
            if (!(obj instanceof Variable)) {
                return false;
            }
            return StringUtils.equals(this.toString(), ((Variable) obj).toString());
        }

        @Override
        public int hashCode() {
            int hash = 7;
            hash = 37 * hash + Objects.hashCode(this.toString());
            return hash;
        }
    }

    public class TableNameBuilderBase
            extends AbstractStatementPart
            implements TableNameBuilder {

        public String tableName;
        public String schemaName;
        private String databaseName;

        public TableNameBuilderBase() {
        }

        @Override
        public void accept(Visitor visitor, VisitorFilter filter) {
            if (filter.accept(this)) {
                visitor.visit(this);
            }
        }

        @Override
        public TableNameBuilder database(String name) {
            this.databaseName = name;
            return this;
        }

        @Override
        public TableNameBuilder schema(String name) {
            if (support_schemas()) {
                this.schemaName = name;
            }
            return this;
        }

        @Override
        public TableNameBuilder name(String name) {
            this.tableName = name;
            return this;
        }

        @Override
        public String getDatabase() {
            return this.databaseName;
        }

        @Override
        public String getSchema() {
            return this.schemaName;
        }

        @Override
        public String getName() {
            return this.tableName;
        }

        @Override
        public boolean has_schema() {
            if (!support_schemas()) {
                return false;
            }
            return !StringUtils.isEmpty(this.schemaName);
        }

        @Override
        public boolean has_database() {
            return !StringUtils.isEmpty(this.databaseName);
        }

        @Override
        public String toString() {
            return this.toString(formatter());
        }

        @Override
        public String toString(Formatter<Value> formatter) {
            if (formatter!=null && formatter.canApply(this)) {
                return formatter.format(this);
            }
            if (this.has_database()) {
                if (this.has_schema()) {
                    return as_identifier(this.databaseName) + "."
                            + as_identifier(this.schemaName) + "."
                            + as_identifier(this.tableName);
                }
            } else {
                if (this.has_schema()) {
                    return as_identifier(this.schemaName) + "."
                            + as_identifier(this.tableName);
                }
            }
            return as_identifier(this.tableName);
        }

        @Override
        public boolean equals(Object obj) {
            if( obj==null || !(obj instanceof TableNameBuilder) ) {
                return false;
            }
            TableNameBuilder other = (TableNameBuilder) obj;
            if (this.has_database()) {
                if (this.has_schema()) {
                    return this.databaseName.equals(other.getDatabase()) &&
                           this.schemaName.equals(other.getSchema()) &&
                           this.tableName.equals(other.getName());
                }
            } else {
                if (this.has_schema()) {
                    return this.schemaName.equals(other.getSchema()) &&
                           this.tableName.equals(other.getName());
                }
            }
            return this.tableName.equals(other.getName());
        }

        @Override
        public int hashCode() {
            int hash = 7;
            hash = 37 * hash + Objects.hashCode(this.toString());
            return hash;
        }

    }

    public class CountBuilderBase
            extends AbstractStatementPart
            implements CountBuilder {

        protected Value value;
        protected boolean distinct;
        protected boolean all;

        public CountBuilderBase() {
            this.value = null;
            this.distinct = false;
            this.all = false;
        }

        @Override
        public CountBuilder all() {
            this.all = true;
            return this;
        }

        @Override
        public CountBuilder column(Value value) {
            this.value = value;
            return this;
        }

        @Override
        public CountBuilder distinct() {
            this.distinct = true;
            return this;
        }

        @Override
        public String toString() {
            return this.toString(formatter());
        }

        @Override
        public String toString(Formatter formatter) {
            if (formatter!=null && formatter.canApply(this)) {
                return formatter.format(this);
            }
            if (this.all) {
                return "COUNT(*)";
            }
            if (this.distinct) {
                return MessageFormat.format(
                        "COUNT(DISTINCT {0})",
                        value.toString(formatter)
                );
            }
            return MessageFormat.format(
                    "COUNT({0})",
                    value.toString(formatter)
            );
        }

    }

    protected class JoinBase 
            extends AbstractStatementPart
            implements StatementPart 
        {
        protected String type;
        protected TableNameBuilder table;
        protected Value expression;
        
        public JoinBase(String type, TableNameBuilder table, Value expression) {
            this.type = type;
            this.table = table;
            this.expression = expression;
        }

        @Override
        public String toString() {
            return this.toString(formatter());
        }

        @Override
        public String toString(Formatter<Value> formatter) {
            if (formatter!=null && formatter.canApply(this)) {
                return formatter.format(this);
            }
            StringBuilder builder = new StringBuilder();
            // INNER JOIN Customers ON Orders.CustomerID = Customers.CustomerID
            builder.append(this.type.toUpperCase());
            builder.append(" JOIN ");
            builder.append(this.table.toString(formatter));
            builder.append(" ON ");
            builder.append(this.expression.toString(formatter));
            return builder.toString();
        }
        
        
    }
    
    public class FromBuilderBase
            extends AbstractStatementPart
            implements FromBuilder {

        protected TableNameBuilder tableName;
        private String subquery;
        private String passthrough;
        private List<JoinBase> joins;

        public FromBuilderBase() {
            this.tableName = null;
            this.subquery = null;
            this.passthrough = null;
            this.joins = null;
        }

        @Override
        public FromBuilder left_join(TableNameBuilder table, Value expression) {
            JoinBase join = createJoin("LEFT", table, expression);
            if( this.joins==null ) {
                this.joins = new ArrayList<>();
            }
            this.joins.add(join);
            return this;
        }
        
        @Override
        public TableNameBuilder table() {
            if (tableName == null) {
                this.tableName = createTableNameBuilder();
            }
            return this.tableName;
        }

        @Override
        public void accept(Visitor visitor, VisitorFilter filter) {
            if (filter.accept(this)) {
                visitor.visit(this);
            }
            if (this.tableName != null) {
                this.tableName.accept(visitor, filter);
            }
        }

        @Override
        public FromBuilder custom(String passthrough) {
            this.passthrough = passthrough;
            return this;
        }

        @Override
        public FromBuilder subquery(String subquery) {
            this.subquery = subquery;
            return this;
        }

        @Override
        public String toString() {
            return this.toString(formatter());
        }

        @Override
        public String toString(Formatter<Value> formatter) {
            if (formatter!=null && formatter.canApply(this)) {
                return formatter.format(this);
            }
            if (!StringUtils.isEmpty(passthrough)) {
                return passthrough;
            }
            if (!StringUtils.isEmpty(subquery)) {
                return "( " + this.subquery + ") as _subquery_alias_ ";
            }
            if( this.joins==null || this.joins.isEmpty() ) {
                return this.tableName.toString(formatter);
            }
            StringBuilder builder = new StringBuilder();
            builder.append(this.tableName.toString(formatter));
            for (JoinBase join : this.joins) {
                builder.append(" ");
                builder.append(join.toString(formatter));
            }
            return builder.toString();
        }

    }

    public class SelectColumnBuilderBase
            extends AbstractStatementPart
            implements SelectColumnBuilder {

        private Variable name = null;
        private String alias = null;
        private Value value = null;
        private boolean asGeometry = false;
        private TableNameBuilder table;

        @Override
        public void accept(Visitor visitor, VisitorFilter filter) {
            if (filter.accept(this)) {
                visitor.visit(this);
            }
            if (this.name != null) {
                this.name.accept(visitor, filter);
            }
            if (this.value != null) {
                this.value.accept(visitor, filter);
            }
        }

        @Override
        public void replace(Value target, Value replacement) {
            if (this.name!=null ) {
                if( this.name == target) {
                    this.name = (Variable) replacement;
                }
            }
            if( this.value!=null ) {
                if (this.value == target) {
                    this.value = replacement;
                } else {
                    this.value.replace(target, replacement);
                }
            }
        }
        
        @Override
        public SelectColumnBuilder name(String name) {
            return this.name(null, name);
        }

        @Override
        public SelectColumnBuilder name(TableNameBuilder table, String name) {
            String quote = quote_for_identifiers();
            if (name.startsWith(quote)) {
                // Remove quotes
                name = name.substring(1, name.length() - 1);
            }
            this.name = expression().variable(name);
            this.table = table;
            this.value = null;
            this.asGeometry = false;
            return this;
        }

        @Override
        public SelectColumnBuilder all() {
            this.name = null;
            this.value = expression().custom("*");
            this.asGeometry = false;
            return this;
        }

        @Override
        public SelectColumnBuilder as_geometry() {
            this.asGeometry = true;
            return this;
        }

        @Override
        public SelectColumnBuilder value(Value value) {
            this.value = value;
            this.name = null;
            return this;
        }

        @Override
        public SelectColumnBuilder as(String alias) {
            this.alias = alias;
            return this;
        }

        @Override
        public String getName() {
            return this.name.name();
        }

        @Override
        public String getAlias() {
            return this.alias;
        }

        @Override
        public String getValue() {
            return this.alias;
        }

        @Override
        public String toString() {
            return this.toString(formatter());
        }

        @Override
        public String toString(Formatter<Value> formatter) {
            if (formatter!=null && formatter.canApply(this)) {
                return formatter.format(this);
            }
            StringBuilder builder = new StringBuilder();
            if (this.asGeometry) {
                builder.append(expression().ST_AsBinary(this.name).toString(formatter));
            } else {
                if (this.name != null) {
                    if( this.table==null ) {
                        builder.append(this.name.toString(formatter));
                    } else {
                        builder.append(this.table.toString(formatter));
                        builder.append(".");
                        builder.append(this.name.toString(formatter));
                    }
                } else {
                    builder.append(this.value.toString(formatter));
                }
            }
            if (this.alias != null) {
                builder.append(" AS ");
                builder.append(as_identifier(this.alias));
            }
            return builder.toString();
        }
    }

    public class OrderByBuilderBase
            extends AbstractStatementPart
            implements OrderByBuilder {

        protected String value;
        protected String custom;
        protected boolean ascending;

        public OrderByBuilderBase() {
            this.ascending = true;
        }

        @Override
        public void accept(Visitor visitor, VisitorFilter filter) {
            if (filter.accept(this)) {
                visitor.visit(this);
            }
        }

        @Override
        public OrderByBuilder column(String name) {
            this.value = name;
            return this;
        }

        @Override
        public OrderByBuilder custom(String order) {
            this.custom = order;
            return this;
        }

        @Override
        public OrderByBuilder ascending() {
            this.ascending = true;
            return this;
        }

        @Override
        public OrderByBuilder ascending(boolean asc) {
            this.ascending = asc;
            return this;
        }

        @Override
        public OrderByBuilder descending() {
            this.ascending = false;
            return this;
        }

        @Override
        public String toString() {
            return this.toString(formatter());
        }

        @Override
        public String toString(Formatter<Value> formatter) {
            if (formatter!=null && formatter.canApply(this)) {
                return formatter.format(this);
            }
            if (!StringUtils.isEmpty(this.custom)) {
                return this.custom;
            }
            if (this.ascending) {
                return as_identifier(this.value) + " ASC";
            }
            return as_identifier(this.value) + " DESC";
        }
    }

    public class SelectBuilderBase
            extends AbstractStatement
            implements SelectBuilder {

        protected FromBuilder from;
        protected ExpressionBuilder where;
        protected long limit = -1;
        protected long offset = -1;
        protected List<SelectColumnBuilder> columns;
        protected List<OrderByBuilder> order_by;
        protected boolean distinct;
        protected List<Variable> groupColumn;

        public SelectBuilderBase() {
            this.columns = new ArrayList<>();
            this.distinct = false;
        }

        @Override
        public SelectBuilder group_by(Variable... columns) {
            if( this.groupColumn==null ) {
                this.groupColumn = new ArrayList<>();
            }
            for (Variable column : columns) {
                this.groupColumn.add(column);
            }
            return this;
        }

        @Override
        public void accept(Visitor visitor, VisitorFilter filter) {
            if (filter.accept(this)) {
                visitor.visit(this);
            }
            for (SelectColumnBuilder column : columns) {
                column.accept(visitor, filter);
            }
            if (this.has_from()) {
                this.from.accept(visitor, filter);
            }
            if (this.has_where()) {
                this.where.accept(visitor, filter);
            }
            if (this.has_order_by()) {
                for (OrderByBuilder order : order_by) {
                    order.accept(visitor, filter);
                }
            }
        }

        @Override
        public void replace(Value target, Value replacement) {
            if( this.columns!=null ) {
                for (int i = 0; i < columns.size(); i++) {
                    SelectColumnBuilder column = columns.get(i);
                    if( column == target ) {
                        columns.set(i, (SelectColumnBuilder) replacement);
                    } else {
                        column.replace(target, replacement);
                    }
                }
            }
            if (this.has_from()) {
                if( this.from == target ) {
                    this.from = (FromBuilder) replacement;
                } else {
                    this.from.replace(target, replacement);
                }
            }
            if (this.has_where()) {
                if( this.where == target ) {
                    this.where = (ExpressionBuilder) replacement;
                } else if( this.where.value() == target ) {
                    this.where.value(replacement);
                } else {
                    this.where.value().replace(target, replacement);
                }
            }
            if (this.has_order_by()) {
                for (int i = 0; i < order_by.size(); i++) {
                    OrderByBuilder order = order_by.get(i);
                    if( order == target ) {
                        order_by.set(i, (OrderByBuilder) replacement);
                    } else {
                        order.replace(target, replacement);
                    }
                }
            }
        }

        @Override
        public SelectBuilder distinct() {
            this.distinct = true;
            return this;
        }

        @Override
        public SelectColumnBuilder column() {
            SelectColumnBuilder builder = createSelectColumnBuilder();
            this.columns.add(builder);
            return builder;
        }

        @Override
        public SelectBuilder remove_all_columns() {
            this.columns = new ArrayList<>();
            return this;
        }
        
        @Override
        public boolean has_column(String name) {
            for (SelectColumnBuilder column : columns) {
                if (name.equals(column.getName())) {
                    return true;
                }
            }
            return false;
        }

        @Override
        public FromBuilder from() {
            if (this.from == null) {
                this.from = createFromBuilder();
            }
            return this.from;
        }

        @Override
        public boolean has_from() {
            return this.from != null;
        }

        @Override
        public ExpressionBuilder where() {
            if (this.where == null) {
                this.where = createExpressionBuilder();
            }
            return this.where;
        }

        @Override
        public boolean has_where() {
            if (this.where == null) {
                return false;
            }
            return this.where.value() != null;
        }

        @Override
        public SelectBuilder limit(long limit) {
            this.limit = limit;
            return this;
        }

        @Override
        public SelectBuilder limit(Long limit) {
            if (limit == null) {
                this.limit = 0;
            } else {
                this.limit = limit;
            }
            return this;
        }

        @Override
        public boolean has_limit() {
            return this.limit > 0;
        }

        @Override
        public SelectBuilder offset(long offset) {
            this.offset = offset;
            return this;
        }

        @Override
        public boolean has_offset() {
            return this.offset > 0;
        }

        @Override
        public OrderByBuilder order_by() {
            if (this.order_by == null) {
                this.order_by = new ArrayList<>();
            }
            OrderByBuilder order = createOrderByBuilder();
            this.order_by.add(order);
            return order;
        }

        @Override
        public boolean has_order_by() {
            if (this.order_by == null) {
                return false;
            }
            return !this.order_by.isEmpty();
        }
        
        @Override
        public boolean has_group_by() {
            if (this.groupColumn == null) {
                return false;
            }
            return !this.groupColumn.isEmpty();
        }
        
        protected boolean isValid(StringBuilder message) {
            if (message == null) {
                message = new StringBuilder();
            }
            if (this.has_offset() && !this.has_order_by()) {
                // Algunos gestores de BBDD requieren que se especifique un
                // orden para poder usar OFFSET. Como eso parece buena idea para
                // asegurar que siempre tengamos los mismo resultados, lo exijimos
                // siempre.
                message.append("Can't use OFFSET without an ORDER BY.");
                return false;
            }
            return true;
        }

        @Override
        public String toString() {
            return this.toString(formatter());
        }

        @Override
        public String toString(Formatter<Value> formatter) {
            if (formatter!=null && formatter.canApply(this)) {
                return formatter.format(this);
            }
            StringBuilder builder = new StringBuilder();
            if (!this.isValid(builder)) {
                throw new IllegalStateException(builder.toString());
            }
            builder.append("SELECT ");
            if (this.distinct) {
                builder.append("DISTINCT ");
            }
            boolean first = true;
            for (SelectColumnBuilder column : columns) {
                if (first) {
                    first = false;
                } else {
                    builder.append(", ");
                }
                builder.append(column.toString(formatter));
            }

            if (this.has_from()) {
                builder.append(" FROM ");
                builder.append(this.from.toString(formatter));
            }
            if( this.has_group_by() ) {
                builder.append(" GROUP BY ");
                builder.append(this.groupColumn.get(0).toString(formatter));
                for (int i = 1; i < groupColumn.size(); i++) {
                    builder.append(", ");
                    builder.append(this.groupColumn.get(i).toString(formatter));
                }
            }
            if (this.has_where()) {
                builder.append(" WHERE ");
                builder.append(this.where.toString(formatter));
            }

            if (this.has_order_by()) {
                builder.append(" ORDER BY ");
                first = true;
                for (OrderByBuilder item : this.order_by) {
                    if (first) {
                        first = false;
                    } else {
                        builder.append(", ");
                    }
                    builder.append(item.toString(formatter));
                }
            }

            if (this.has_limit()) {
                builder.append(" LIMIT ");
                builder.append(this.limit);
            }
            if (this.has_offset()) {
                builder.append(" OFFSET ");
                builder.append(this.offset);
            }
            return builder.toString();

        }
    }

    public class DropTableBuilderBase
            extends AbstractStatement
            implements DropTableBuilder {

        protected TableNameBuilder table;

        @Override
        public TableNameBuilder table() {
            if (table == null) {
                table = createTableNameBuilder();
            }
            return table;
        }

        @Override
        public void accept(Visitor visitor, VisitorFilter filter) {
            if (filter.accept(this)) {
                visitor.visit(this);
            }
            this.table.accept(visitor, filter);
        }

        @Override
        public String toString() {
            return this.toString(formatter());
        }

        @Override
        public String toString(Formatter<Value> formatter) {
            if (formatter!=null && formatter.canApply(this)) {
                return formatter.format(this);
            }
            StringBuilder builder = new StringBuilder();
            boolean first = true;
            for (String sql : toStrings(formatter)) {
                if (StringUtils.isEmpty(sql)) {
                    continue;
                }
                if (first) {
                    first = false;
                } else {
                    builder.append("; ");
                }
                builder.append(sql);
            }
            return builder.toString();
        }

        @Override
        public List<String> toStrings() {
            return this.toStrings(formatter());
        }

        @Override
        public List<String> toStrings(Formatter formatter) {
            List<String> sqls = new ArrayList<>();

            sqls.add(
                    MessageFormat.format(
                            STMT_DROP_TABLE_table,
                            this.table.toString(formatter)
                    )
            );
            String sql;
            if (!StringUtils.isBlank(STMT_DELETE_GEOMETRY_COLUMN_FROM_TABLE_schema_table)) {
                if (this.table.has_schema()) {
                    sql = MessageFormat.format(
                            STMT_DELETE_GEOMETRY_COLUMN_FROM_TABLE_schema_table,
                            as_string(this.table.getSchema()),
                            as_string(this.table.getName())
                    );
                } else {
                    sql = MessageFormat.format(
                            STMT_DELETE_GEOMETRY_COLUMN_FROM_TABLE_table,
                            as_identifier(this.table.getName())
                    );
                }
                if (!StringUtils.isEmpty(sql)) {
                    sqls.add(sql);
                }
            }
            return sqls;
        }
    }

    public class GrantRoleBuilderBase
            extends AbstractStatementPart
            implements GrantRoleBuilder {

        protected TableNameBuilder table;
        protected String role;
        protected Set<Privilege> privileges;

        public GrantRoleBuilderBase(TableNameBuilder table, String role) {
            this.table = table;
            this.role = role;
            this.privileges = new HashSet<>();
        }

        @Override
        public GrantRoleBuilder privilege(Privilege privilege) {
            privileges.add(privilege);
            return this;
        }

        @Override
        public GrantRoleBuilder select() {
            privileges.add(Privilege.SELECT);
            return this;
        }

        @Override
        public GrantRoleBuilder update() {
            privileges.add(Privilege.UPDATE);
            return this;
        }

        @Override
        public GrantRoleBuilder insert() {
            privileges.add(Privilege.INSERT);
            return this;
        }

        @Override
        public GrantRoleBuilder delete() {
            privileges.add(Privilege.DELETE);
            return this;
        }

        @Override
        public GrantRoleBuilder truncate() {
            privileges.add(Privilege.TRUNCATE);
            return this;
        }

        @Override
        public GrantRoleBuilder reference() {
            privileges.add(Privilege.REFERENCE);
            return this;
        }

        @Override
        public GrantRoleBuilder trigger() {
            privileges.add(Privilege.TRIGGER);
            return this;
        }

        @Override
        public GrantRoleBuilder all() {
            privileges.add(Privilege.ALL);
            return this;
        }

        protected String getPrivilegeName(Privilege privilege) {
            switch (privilege) {
                case DELETE:
                    return "DELETE";
                case INSERT:
                    return "INSERT";
                case REFERENCE:
                    return "REFERENCE";
                case SELECT:
                    return "SELECT";
                case TRIGGER:
                    return "TRIGGER";
                case TRUNCATE:
                    return "TRUNCATE";
                case UPDATE:
                    return "UPDATE";
                case ALL:
                default:
                    return "ALL";
            }
        }

        @Override
        public String toString() {
            return this.toString(formatter());
        }

        @Override
        public String toString(Formatter<Value> formatter) {
            if (formatter!=null && formatter.canApply(this)) {
                return formatter.format(this);
            }
            StringBuilder builder = new StringBuilder();
            boolean first = true;
            for (Privilege privilege : privileges) {
                if (first) {
                    first = false;
                } else {
                    builder.append(", ");
                }
                builder.append(this.getPrivilegeName(privilege));
            }
            String sql = MessageFormat.format(
                    STMT_GRANT_privileges_ON_table_TO_role,
                    builder.toString(),
                    table.toString(formatter),
                    role
            );
            return sql;
        }
    }

    public class GrantBuilderBase
            extends AbstractStatement
            implements GrantBuilder {

        protected TableNameBuilder table;
        protected Map<String, GrantRoleBuilder> roles;

        public GrantBuilderBase() {
            this.roles = new HashMap<>();
        }

        @Override
        public TableNameBuilder table() {
            if (table == null) {
                table = createTableNameBuilder();
            }
            return table;
        }

        @Override
        public void accept(Visitor visitor, VisitorFilter filter) {
            if (filter.accept(this)) {
                visitor.visit(this);
            }
            if (this.table != null) {
                this.table.accept(visitor, filter);
            }
        }

        @Override
        public GrantRoleBuilder role(String role) {
            GrantRoleBuilder roleBuilder = this.roles.get(role);
            if (roleBuilder == null) {
                roleBuilder = createGrantRoleBuilder(this.table(), role);
                this.roles.put(role, roleBuilder);
            }
            return roleBuilder;
        }

        @Override
        public String toString() {
            return this.toString(formatter());
        }

        @Override
        public String toString(Formatter<Value> formatter) {
            if (formatter!=null && formatter.canApply(this)) {
                return formatter.format(this);
            }
            StringBuilder builder = new StringBuilder();
            boolean first = true;
            for (String sql : toStrings(formatter)) {
                if (StringUtils.isEmpty(sql)) {
                    continue;
                }
                if (first) {
                    first = false;
                } else {
                    builder.append("; ");
                }
                builder.append(sql);
            }
            return builder.toString();
        }

        @Override
        public List<String> toStrings() {
            return this.toStrings(formatter());
        }

        @Override
        public List<String> toStrings(Formatter formatter) {
            List<String> sqls = new ArrayList<>();
            for (GrantRoleBuilder role : roles.values()) {
                sqls.add(role.toString(formatter));
            }
            return sqls;
        }
    }

    public class UpdateColumnBuilderBase
            extends InsertColumnBuilderBase
            implements UpdateColumnBuilder {

        public UpdateColumnBuilderBase() {
            super();
        }

        @Override
        public UpdateColumnBuilder name(String name) {
            return (UpdateColumnBuilder) super.name(name);
        }

        @Override
        public UpdateColumnBuilder with_value(Value value) {
            return (UpdateColumnBuilder) super.with_value(value);
        }

    }

    public class UpdateBuilderBase
            extends AbstractStatement
            implements UpdateBuilder {

        protected ExpressionBuilder where;
        protected List<UpdateColumnBuilder> columns;
        protected TableNameBuilder table;

        public UpdateBuilderBase() {
            this.columns = new ArrayList<>();
        }

        @Override
        public void accept(Visitor visitor, VisitorFilter filter) {
            if (filter.accept(this)) {
                visitor.visit(this);
            }
            if (this.table != null) {
                this.table.accept(visitor, filter);
            }
            for (UpdateColumnBuilder column : columns) {
                column.accept(visitor, filter);
            }
            if (this.has_where()) {
                this.where.accept(visitor, filter);
            }
        }

        @Override
        public ExpressionBuilder where() {
            if (this.where == null) {
                this.where = createExpressionBuilder();
            }
            return this.where;
        }

        @Override
        public TableNameBuilder table() {
            if (table == null) {
                table = createTableNameBuilder();
            }
            return table;
        }

        @Override
        public UpdateColumnBuilder column() {
            UpdateColumnBuilder column = createUpdateColumnBuilder();
            this.columns.add(column);
            return column;
        }

        @Override
        public boolean has_where() {
            return this.where != null;
        }

        @Override
        public String toString() {
            return this.toString(formatter());
        }

        @Override
        public String toString(Formatter<Value> formatter) {
            if (formatter!=null && formatter.canApply(this)) {
                return formatter.format(this);
            }
            /*
             * UPDATE [ ONLY ] table [ [ AS ] alias ] SET { column = { expression |
             * DEFAULT } | ( column [, ...] ) = ( { expression | DEFAULT } [, ...] )
             * } [, ...] [ FROM fromlist ] [ WHERE condition ] [ RETURNING * |
             * output_expression [ AS output_name ] [, ...] ]
             */
            StringBuilder columnsAndValues = new StringBuilder();

            boolean first = true;
            for (UpdateColumnBuilder column : columns) {
                if (first) {
                    first = false;
                } else {
                    columnsAndValues.append(", ");
                }
                columnsAndValues.append(as_identifier(column.getName()));
                columnsAndValues.append(" = ");
                columnsAndValues.append(column.getValue().toString(formatter));
            }

            String sql;
            if (this.has_where()) {
                sql = MessageFormat.format(
                        STMT_UPDATE_table_SET_columnsAndValues_WHERE_expresion,
                        this.table.toString(formatter),
                        columnsAndValues.toString(),
                        this.where.toString(formatter)
                );
            } else {
                sql = MessageFormat.format(
                        STMT_UPDATE_table_SET_columnsAndValues,
                        this.table.toString(formatter),
                        columnsAndValues.toString()
                );
            }
            return sql;
        }
    }

    public class DeleteBuilderBase
            extends AbstractStatement
            implements DeleteBuilder {

        protected ExpressionBuilder where;
        protected TableNameBuilder table;

        public DeleteBuilderBase() {
        }

        @Override
        public void accept(Visitor visitor, VisitorFilter filter) {
            if (filter.accept(this)) {
                visitor.visit(this);
            }
            if (this.table != null) {
                this.table.accept(visitor, filter);
            }
            if (this.has_where()) {
                this.where.accept(visitor, filter);
            }
        }

        @Override
        public ExpressionBuilder where() {
            if (this.where == null) {
                this.where = createExpressionBuilder();
            }
            return this.where;
        }

        @Override
        public TableNameBuilder table() {
            if (table == null) {
                table = createTableNameBuilder();
            }
            return table;
        }

        @Override
        public boolean has_where() {
            return this.where != null;
        }

        @Override
        public String toString() {
            return this.toString(formatter());
        }

        @Override
        public String toString(Formatter<Value> formatter) {
            if (formatter!=null && formatter.canApply(this)) {
                return formatter.format(this);
            }
            /*
             * DELETE FROM table_name
             * WHERE some_column=some_value; 
             */
            String sql;
            if (this.has_where()) {
                sql = MessageFormat.format(
                        STMT_DELETE_FROM_table_WHERE_expresion,
                        this.table.toString(formatter),
                        this.where.toString(formatter)
                );
            } else {
                sql = MessageFormat.format(
                        STMT_DELETE_FROM_table,
                        this.table.toString(formatter)
                );
            }
            return sql;
        }
    }

    public class CreateIndexBuilderBase
            extends AbstractStatement
            implements CreateIndexBuilder {

        protected boolean ifNotExist = false;
        protected boolean isUnique = false;
        protected String indexName;
        protected boolean isSpatial = false;
        protected TableNameBuilder table;
        protected final List<String> columns;

        public CreateIndexBuilderBase() {
            this.columns = new ArrayList<>();
        }

        @Override
        public CreateIndexBuilder unique() {
            this.isUnique = true;
            return this;
        }

        @Override
        public CreateIndexBuilder if_not_exist() {
            this.ifNotExist = true;
            return this;
        }

        @Override
        public CreateIndexBuilder name(String name) {
            this.indexName = name;
            return this;
        }

        @Override
        public CreateIndexBuilder spatial() {
            this.isSpatial = true;
            return this;
        }

        @Override
        public CreateIndexBuilder column(String name) {
            this.columns.add(name);
            return this;
        }

        @Override
        public TableNameBuilder table() {
            if (table == null) {
                table = createTableNameBuilder();
            }
            return table;
        }

        @Override
        public void accept(Visitor visitor, VisitorFilter filter) {
            if (filter.accept(this)) {
                visitor.visit(this);
            }
            if (this.table != null) {
                this.table.accept(visitor, filter);
            }
        }

        @Override
        public String toString() {
            return this.toString(formatter());
        }

        @Override
        public String toString(Formatter<Value> formatter) {
            if (formatter!=null && formatter.canApply(this)) {
                return formatter.format(this);
            }
            StringBuilder builder = new StringBuilder();
            boolean first = true;
            for (String sql : toStrings(formatter)) {
                if (StringUtils.isEmpty(sql)) {
                    continue;
                }
                if (first) {
                    first = false;
                } else {
                    builder.append("; ");
                }
                builder.append(sql);
            }
            return builder.toString();
        }

        @Override
        public List<String> toStrings() {
            return this.toStrings(formatter());
        }

        @Override
        public List<String> toStrings(Formatter formatter) {
            StringBuilder builder = new StringBuilder();
            builder.append("CREATE ");
            if (this.isUnique) {
                builder.append("UNIQUE ");
            }
            builder.append("INDEX ");
            if (this.ifNotExist) {
                builder.append("IF NOT EXISTS ");
            }
            builder.append(as_identifier(this.indexName));
            builder.append(" ON ");
            builder.append(this.table.toString(formatter));
            if (this.isSpatial) {
                builder.append(" USING GIST ");
            }
            builder.append(" ( ");
            boolean is_first_column = true;
            for (String column : this.columns) {
                if (is_first_column) {
                    is_first_column = false;
                } else {
                    builder.append(", ");
                }
                builder.append(column);
            }
            builder.append(" )");

            List<String> sqls = new ArrayList<>();
            sqls.add(builder.toString());
            return sqls;
        }

    }

    public class AlterTableBuilderBase
            extends AbstractStatement
            implements AlterTableBuilder {

        protected TableNameBuilder table;
        protected List<String> drops;
        protected List<ColumnDescriptor> adds;
        protected List<ColumnDescriptor> alters;
        protected List<Pair<String, String>> renames;

        public AlterTableBuilderBase() {
            this.drops = new ArrayList<>();
            this.adds = new ArrayList<>();
            this.alters = new ArrayList<>();
            this.renames = new ArrayList<>();
        }

        @Override
        public boolean isEmpty() {
            return this.drops.isEmpty()
                    && this.adds.isEmpty()
                    && this.alters.isEmpty()
                    && this.renames.isEmpty();
        }

        @Override
        public void accept(Visitor visitor, VisitorFilter filter) {
            if (filter.accept(this)) {
                visitor.visit(this);
            }
            if (this.table != null) {
                this.table.accept(visitor, filter);
            }
        }

        @Override
        public TableNameBuilder table() {
            if (table == null) {
                table = createTableNameBuilder();
            }
            return table;
        }

        @Override
        public AlterTableBuilder drop_column(String columnName) {
            this.drops.add(columnName);
            return this;
        }

        @Override
        public AlterTableBuilder add_column(FeatureAttributeDescriptor fad) {
            this.adds.add(new ColumnDescriptorBase(fad));
            return this;
        }

        @Override
        public AlterTableBuilder add_column(String columnName, int type, int type_p, int type_s, boolean isPk, boolean isIndexed, boolean allowNulls, boolean isAutomatic, Object defaultValue) {
            if (isPk || isAutomatic) {
                allowNulls = false;
            }
            this.adds.add(new ColumnDescriptorBase(columnName, type, type_p, type_s, isPk, isIndexed, allowNulls, isAutomatic, defaultValue));
            return this;
        }

        @Override
        public AlterTableBuilder add_geometry_column(String columnName, int type, int subtype, IProjection proj, boolean isIndexed, boolean allowNulls) {
            if (StringUtils.isEmpty(columnName)) {
                throw new IllegalArgumentException("Argument 'columnName' can't be empty.");
            }
            this.adds.add(new ColumnDescriptorBase(columnName, type, subtype, proj, isIndexed, allowNulls));
            return this;
        }

        @Override
        public AlterTableBuilder add_geometry_column(String columnName, int type, int subtype, Object srsdbcode, boolean isIndexed, boolean allowNulls) {
            if (StringUtils.isEmpty(columnName)) {
                throw new IllegalArgumentException("Argument 'columnName' can't be empty.");
            }
            this.adds.add(new ColumnDescriptorBase(columnName, type, subtype, srsdbcode, isIndexed, allowNulls));
            return this;
        }

        @Override
        public AlterTableBuilder alter_column(FeatureAttributeDescriptor fad) {
            this.alters.add(new ColumnDescriptorBase(fad));
            return this;
        }

        @Override
        public AlterTableBuilder alter_column(String columnName, int type, int type_p, int type_s, boolean isPk, boolean isIndexed, boolean allowNulls, boolean isAutomatic, Object defaultValue) {
            if (isPk || isAutomatic) {
                allowNulls = false;
            }
            this.alters.add(new ColumnDescriptorBase(columnName, type, type_p, type_s, isPk, isIndexed, allowNulls, isAutomatic, defaultValue));
            return this;
        }

        @Override
        public AlterTableBuilder alter_geometry_column(String columnName, int type, int subtype, IProjection proj, boolean isIndexed, boolean allowNulls) {
            if (StringUtils.isEmpty(columnName)) {
                throw new IllegalArgumentException("Argument 'columnName' can't be empty.");
            }
            this.alters.add(new ColumnDescriptorBase(columnName, type, subtype, proj, isIndexed, allowNulls));
            return this;
        }

        @Override
        public AlterTableBuilder alter_geometry_column(String columnName, int type, int subtype, Object srsdbcode, boolean isIndexed, boolean allowNulls) {
            if (StringUtils.isEmpty(columnName)) {
                throw new IllegalArgumentException("Argument 'columnName' can't be empty.");
            }
            this.alters.add(new ColumnDescriptorBase(columnName, type, subtype, srsdbcode, isIndexed, allowNulls));
            return this;
        }

        @Override
        public AlterTableBuilder rename_column(String source, String target) {
            this.renames.add(new ImmutablePair(source, target));
            return this;
        }

        @Override
        public String toString() {
            return this.toString(formatter());
        }

        @Override
        public String toString(Formatter<Value> formatter) {
            if (formatter!=null && formatter.canApply(this)) {
                return formatter.format(this);
            }
            StringBuilder builder = new StringBuilder();
            boolean first = true;
            for (String sql : toStrings(formatter)) {
                if (StringUtils.isEmpty(sql)) {
                    continue;
                }
                if (first) {
                    first = false;
                } else {
                    builder.append("; ");
                }
                builder.append(sql);
            }
            return builder.toString();
        }

        @Override
        public List<String> toStrings() {
            return this.toStrings(formatter());
        }

        @Override
        public List<String> toStrings(Formatter formatter) {
            List<String> sqls = new ArrayList<>();
            if (this.isEmpty()) {
                return sqls;
            }
            for (String column : drops) {
                StringBuilder builder = new StringBuilder();
                builder.append("ALTER TABLE ");
                builder.append(this.table.toString(formatter));
                builder.append(" DROP COLUMN IF EXISTS ");
                builder.append(as_identifier(column));
                sqls.add(builder.toString());
            }
            for (ColumnDescriptor column : adds) {
                StringBuilder builder = new StringBuilder();
                builder.append("ALTER TABLE ");
                builder.append(this.table.toString(formatter));
                builder.append(" ADD COLUMN ");
                builder.append(as_identifier(column.getName()));
                builder.append(" ");
                if (column.getType() == DataTypes.INT && column.isAutomatic()) {
                    builder.append(" SERIAL");
                } else {
                    builder.append(
                            sqltype(
                                    column.getType(),
                                    column.getPrecision(),
                                    column.getSize(),
                                    column.getGeometryType(),
                                    column.getGeometrySubtype()
                            )
                    );
                }
                if (column.getDefaultValue() == null) {
                    if (column.allowNulls()) {
                        builder.append(" DEFAULT NULL");
                    }
                } else {
                    builder.append(" DEFAULT '");
                    builder.append(Objects.toString(column.getDefaultValue(),""));
                    builder.append("'");
                }
                if (column.allowNulls()) {
                    builder.append(" NULL");
                } else {
                    builder.append(" NOT NULL");
                }
                if (column.isPrimaryKey()) {
                    builder.append(" PRIMARY KEY");
                }
                sqls.add(builder.toString());
            }
            for (ColumnDescriptor column : alters) {
                StringBuilder builder = new StringBuilder();
                builder.append("ALTER TABLE ");
                builder.append(this.table.toString(formatter));
                builder.append(" ALTER COLUMN ");
                builder.append(as_identifier(column.getName()));
                builder.append(" SET DATA TYPE ");
                if (column.getType() == DataTypes.INT && column.isAutomatic()) {
                    builder.append(" SERIAL");
                } else {
                    builder.append(
                            sqltype(
                                    column.getType(),
                                    column.getPrecision(),
                                    column.getSize(),
                                    column.getGeometryType(),
                                    column.getGeometrySubtype()
                            )
                    );
                }
                if (column.getDefaultValue() == null) {
                    if (column.allowNulls()) {
                        builder.append(" DEFAULT NULL");
                    } else {
                        builder.append(" DROP DEFAULT");
                    }
                } else {
                    builder.append(" DEFAULT '");
                    builder.append(column.getDefaultValue().toString());
                    builder.append("'");
                }
                sqls.add(builder.toString());
            }
            for (Pair<String, String> pair : renames) {
                StringBuilder builder = new StringBuilder();
                builder.append("ALTER TABLE ");
                builder.append(this.table.toString(formatter));
                builder.append(" RENAME COLUMN ");
                builder.append(as_identifier(pair.getLeft()));
                builder.append(" TO ");
                builder.append(as_identifier(pair.getRight()));
                sqls.add(builder.toString());
            }
            return sqls;
        }

    }

    public class CreateTableBuilderBase
            extends AbstractStatement
            implements CreateTableBuilder {

        protected TableNameBuilder table;
        protected List<ColumnDescriptor> columns;

        public CreateTableBuilderBase() {
            this.columns = new ArrayList<>();
        }

        @Override
        public void accept(Visitor visitor, VisitorFilter filter) {
            if (filter.accept(this)) {
                visitor.visit(this);
            }
            if (this.table != null) {
                this.table.accept(visitor, filter);
            }
        }

        @Override
        public TableNameBuilder table() {
            if (table == null) {
                table = createTableNameBuilder();
            }
            return table;
        }

        @Override
        public CreateTableBuilderBase add_column(FeatureAttributeDescriptor fad) {
            this.columns.add(new ColumnDescriptorBase(fad));
            return this;
        }

        @Override
        public CreateTableBuilderBase add_column(String columnName, int type, int type_p, int type_s, boolean isPk, boolean isIndexed, boolean allowNulls, boolean isAutomatic, Object defaultValue) {
            if (StringUtils.isEmpty(columnName)) {
                throw new IllegalArgumentException("Argument 'columnName' can't be empty.");
            }
            if (isPk || isAutomatic) {
                allowNulls = false;
            }
            this.columns.add(new ColumnDescriptorBase(columnName, type, type_p, type_s, isPk, isIndexed, allowNulls, isAutomatic, defaultValue));
            return this;
        }

        @Override
        public CreateTableBuilder add_geometry_column(String columnName, int type, int subtype, IProjection proj, boolean isIndexed, boolean allowNulls) {
            if (StringUtils.isEmpty(columnName)) {
                throw new IllegalArgumentException("Argument 'columnName' can't be empty.");
            }
            this.columns.add(new ColumnDescriptorBase(columnName, type, subtype, proj, isIndexed, allowNulls));
            return this;
        }

        @Override
        public CreateTableBuilder add_geometry_column(String columnName, int type, int subtype, Object srsdbcode, boolean isIndexed, boolean allowNulls) {
            if (StringUtils.isEmpty(columnName)) {
                throw new IllegalArgumentException("Argument 'columnName' can't be empty.");
            }
            this.columns.add(new ColumnDescriptorBase(columnName, type, subtype, srsdbcode, isIndexed, allowNulls));
            return this;
        }

        @Override
        public ColumnDescriptor getColumnDescriptor(String columnName) {
            if (StringUtils.isEmpty(columnName)) {
                return null;
            }
            for (ColumnDescriptor column : columns) {
                if (columnName.equals(column.getName())) {
                    return column;
                }
            }
            return null;
        }

        @Override
        public String toString() {
            return this.toString(formatter());
        }

        @Override
        public String toString(Formatter<Value> formatter) {
            if (formatter!=null && formatter.canApply(this)) {
                return formatter.format(this);
            }
            StringBuilder builder = new StringBuilder();
            boolean first = true;
            for (String sql : toStrings(formatter)) {
                if (StringUtils.isEmpty(sql)) {
                    continue;
                }
                if (first) {
                    first = false;
                } else {
                    builder.append("; ");
                }
                builder.append(sql);
            }
            return builder.toString();
        }

        @Override
        public List<String> toStrings() {
            return this.toStrings(formatter());
        }

        @Override
        public List<String> toStrings(Formatter formatter) {
            List<String> sqls = new ArrayList<>();
            /**
             * CREATE [ [ GLOBAL | LOCAL ] { TEMPORARY | TEMP } ] TABLE
             * table_name ( { column_name data_type [ DEFAULT default_expr ] [
             * column_constraint [ ... ] ] | table_constraint | LIKE
             * parent_table [ { INCLUDING | EXCLUDING } DEFAULTS ] } [, ... ] )
             * [ INHERITS ( parent_table [, ... ] ) ] [ WITH OIDS | WITHOUT OIDS
             * ] [ ON COMMIT { PRESERVE ROWS | DELETE ROWS | DROP } ]
             *
             * where column_constraint is:
             *
             * [ CONSTRAINT constraint_name ] { NOT NULL | NULL | UNIQUE |
             * PRIMARY KEY | CHECK (expression) | REFERENCES reftable [ (
             * refcolumn ) ] [ MATCH FULL | MATCH PARTIAL | MATCH SIMPLE ] [ ON
             * DELETE action ] [ ON UPDATE action ] } [ DEFERRABLE | NOT
             * DEFERRABLE ] [ INITIALLY DEFERRED | INITIALLY IMMEDIATE ]
             *
             * and table_constraint is:
             *
             * [ CONSTRAINT constraint_name ] { UNIQUE ( column_name [, ... ] )
             * | PRIMARY KEY ( column_name [, ... ] ) | CHECK ( expression ) |
             * FOREIGN KEY ( column_name [, ... ] ) REFERENCES reftable [ (
             * refcolumn [, ... ] ) ] [ MATCH FULL | MATCH PARTIAL | MATCH
             * SIMPLE ] [ ON DELETE action ] [ ON UPDATE action ] } [ DEFERRABLE
             * | NOT DEFERRABLE ] [ INITIALLY DEFERRED | INITIALLY IMMEDIATE ]
             */
            StringBuilder builder = new StringBuilder();

            builder.append("CREATE TABLE ");
            builder.append(this.table.toString(formatter));
            builder.append(" (");
            boolean first = true;
            for (ColumnDescriptor column : columns) {
                if (first) {
                    first = false;
                } else {
                    builder.append(", ");
                }
                builder.append(as_identifier(column.getName()));
                builder.append(" ");
                if (column.isAutomatic() && column.getType() == DataTypes.INT) {
                    builder.append("SERIAL");
                } else if (column.isAutomatic() && column.getType() == DataTypes.LONG) {
                    builder.append("BIGSERIAL");
                } else {
                    builder.append(sqltype(
                            column.getType(),
                            column.getPrecision(),
                            column.getSize(),
                            column.getGeometryType(),
                            column.getGeometrySubtype()
                    )
                    );
                }
                if (column.getDefaultValue() == null) {
                    if (column.allowNulls()) {
                        builder.append(" DEFAULT NULL");
                    }
                } else {
                    builder.append(" DEFAULT '");
                    builder.append(Objects.toString(column.getDefaultValue(),""));
                    builder.append("'");
                }
                if (column.allowNulls()) {
                    builder.append(" NULL");
                } else {
                    builder.append(" NOT NULL");
                }
                if (column.isPrimaryKey()) {
                    builder.append(" PRIMARY KEY");
                }
            }
            builder.append(" )");
            sqls.add(builder.toString());
            return sqls;
        }
    }

    public class InsertColumnBuilderBase
            extends AbstractStatement
            implements InsertColumnBuilder {

        protected Variable name;
        protected Value value;

        public InsertColumnBuilderBase() {
        }

        @Override
        public void accept(Visitor visitor, VisitorFilter filter) {
            if (filter.accept(this)) {
                visitor.visit(this);
            }
            if (this.name != null) {
                this.name.accept(visitor, filter);
            }
            if (this.value != null) {
                this.value.accept(visitor, filter);
            }
        }

        @Override
        public InsertColumnBuilder name(String name) {
            this.name = expression().variable(name);
            return this;
        }

        @Override
        public InsertColumnBuilder with_value(Value value) {
            this.value = value;
            return this;
        }

        @Override
        public String getName() {
            return this.name.name();
        }

        @Override
        public Value getValue() {
            return this.value;
        }

        @Override
        public String toString() {
            return this.toString(formatter());
        }

        @Override
        public String toString(Formatter<Value> formatter) {
            if (formatter!=null && formatter.canApply(this)) {
                return formatter.format(this);
            }
            return this.value.toString(formatter);
        }
    }

    public class InsertBuilderBase
            extends AbstractStatement
            implements InsertBuilder {

        protected List<InsertColumnBuilder> columns;
        protected TableNameBuilder table;

        public InsertBuilderBase() {
            this.columns = new ArrayList<>();
        }

        @Override
        public void accept(Visitor visitor, VisitorFilter filter) {
            if (filter.accept(this)) {
                visitor.visit(this);
            }
            if (this.table != null) {
                this.table.accept(visitor, filter);
            }
            for (InsertColumnBuilder column : columns) {
                column.accept(visitor, filter);
            }
        }

        @Override
        public TableNameBuilder table() {
            if (table == null) {
                table = createTableNameBuilder();
            }
            return table;
        }

        @Override
        public InsertColumnBuilder column() {
            InsertColumnBuilder column = createInsertColumnBuilder();
            this.columns.add(column);
            return column;
        }

        @Override
        public String toString() {
            return this.toString(formatter());
        }

        @Override
        public String toString(Formatter<Value> formatter) {
            if (formatter!=null && formatter.canApply(this)) {
                return formatter.format(this);
            }
            /*
             * INSERT INTO table [ ( column [, ...] ) ] { DEFAULT VALUES | VALUES (
             * { expression | DEFAULT } [, ...] ) [, ...] | query } [ RETURNING * |
             * output_expression [ AS output_name ] [, ...] ]
             */
            StringBuilder builderColumns = new StringBuilder();
            StringBuilder builderValues = new StringBuilder();

            boolean first = true;
            for (InsertColumnBuilder column : columns) {
                if (first) {
                    first = false;
                } else {
                    builderColumns.append(", ");
                }
                builderColumns.append(as_identifier(column.getName()));
            }
            first = true;
            for (InsertColumnBuilder column : columns) {
                if (first) {
                    first = false;
                } else {
                    builderValues.append(", ");
                }
                builderValues.append(column.toString(formatter));
            }

            String sql = MessageFormat.format(
                    STMT_INSERT_INTO_table_columns_VALUES_values,
                    this.table.toString(formatter),
                    builderColumns.toString(),
                    builderValues.toString()
            );
            return sql;

        }
    }

    public class UpdateTableStatisticsBuilderBase
            extends AbstractStatement
            implements UpdateTableStatisticsBuilder {

        protected TableNameBuilder table;

        @Override
        public void accept(Visitor visitor, VisitorFilter filter) {
            if (filter.accept(this)) {
                visitor.visit(this);
            }
            if (this.table != null) {
                this.table.accept(visitor, filter);
            }
        }

        @Override
        public TableNameBuilder table() {
            if (table == null) {
                table = createTableNameBuilder();
            }
            return table;
        }

        @Override
        public String toString() {
            return this.toString(formatter());
        }

        @Override
        public String toString(Formatter<Value> formatter) {
            if (formatter!=null && formatter.canApply(this)) {
                return formatter.format(this);
            }
            StringBuilder builder = new StringBuilder();
            boolean first = true;
            for (String sql : toStrings(formatter)) {
                if (StringUtils.isEmpty(sql)) {
                    continue;
                }
                if (first) {
                    first = false;
                } else {
                    builder.append("; ");
                }
                builder.append(sql);
            }
            return builder.toString();
        }

        @Override
        public List<String> toStrings() {
            return this.toStrings(formatter());
        }

        @Override
        public List<String> toStrings(Formatter formatter) {
            List<String> sqls = new ArrayList<>();

            if (!StringUtils.isBlank(STMT_UPDATE_TABLE_STATISTICS_table)) {
                String sql = MessageFormat.format(
                        STMT_UPDATE_TABLE_STATISTICS_table,
                        table.toString(formatter)
                );
                if (!StringUtils.isEmpty(sql)) {
                    sqls.add(sql);
                }
            }
            return sqls;
        }
    }

    protected ExpressionBuilder expressionBuilder;

    protected String defaultSchema;
    protected boolean supportSchemas;
    protected boolean hasSpatialFunctions;
    protected GeometrySupportType geometrySupportType;
    protected boolean allowAutomaticValues;

    private static Map<Pair<Integer, Integer>, String> sqlgeometrytypes = null;

    protected String constant_true = "(1=1)";
    protected String constant_false = "(1<>1)";

    protected String type_boolean = "BOOLEAN";
    protected String type_byte = "TINYINT";
    protected String type_bytearray = "BYTEA";
    protected String type_geometry = "TEXT";
    protected String type_char = "CHARACTER(1)";
    protected String type_date = "DATE";
    protected String type_double = "DOUBLE PRECISION"; //float con 53 bits de mantisa, float(54)
    protected String type_numeric_p = "NUMERIC({0,Number,#######})";
    protected String type_numeric_ps = "NUMERIC({0,Number,#######},{1,Number,#######})";
    protected String type_bigdecimal = "NUMERIC({0,Number,#######},{1,Number,#######})";
    protected String type_float = "REAL"; //float con 24 bits de mantisa, float(24)
    protected String type_int = "INT";
    protected String type_long = "BIGINT";
    protected String type_string = "TEXT";
    protected String type_string_p = "VARCHAR({0,Number,#######})";
    protected String type_time = "TIME";
    protected String type_timestamp = "TIMESTAMP";
    protected String type_version = "VARCHAR(30)";
    protected String type_URI = "TEXT";
    protected String type_URL = "TEXT";
    protected String type_FILE = "TEXT";
    protected String type_FOLDER = "TEXT";

    protected String STMT_DELETE_FROM_table_WHERE_expresion = "DELETE FROM {0} WHERE {1}";
    protected String STMT_DELETE_FROM_table = "DELETE FROM {0}";
    protected String STMT_INSERT_INTO_table_columns_VALUES_values = "INSERT INTO {0} ( {1} ) VALUES ( {2} )";
    protected String STMT_UPDATE_TABLE_STATISTICS_table = "VACUUM ANALYZE {0}";
    protected String STMT_DROP_TABLE_table = "DROP TABLE {0}";
    protected String STMT_DELETE_GEOMETRY_COLUMN_FROM_TABLE_schema_table = "DELETE FROM GEOMETRY_COLUMNS WHERE f_table_schema = {0} AND f_table_name = {1}";
    protected String STMT_DELETE_GEOMETRY_COLUMN_FROM_TABLE_table = "DELETE FROM GEOMETRY_COLUMNS WHERE f_table_name = {0}";
    protected String STMT_UPDATE_table_SET_columnsAndValues_WHERE_expresion = "UPDATE {0} SET {1} WHERE {2}";
    protected String STMT_UPDATE_table_SET_columnsAndValues = "UPDATE {0} SET {1}";
    protected String STMT_GRANT_privileges_ON_table_TO_role = "GRANT {0} ON {1} TO {2}";

    public SQLBuilderBase() {
        this.expressionBuilder = ExpressionEvaluatorLocator.getManager().createExpressionBuilder();

        this.hasSpatialFunctions = false;
        this.supportSchemas = true;
        this.geometrySupportType = GeometrySupportType.WKT;

        this.defaultSchema = "public";
        this.allowAutomaticValues = true;

    }
    
    @Override
    public void setProperties(Class filter, final Object... values) {
        this.accept(new Visitor() {
            @Override
            public void visit(Visitable v) {
                for (int i = 0; i < values.length; i+=2) {
                    ((Value)v).setProperty((String) values[i], values[i+1]);
                }
            }
        }, new ClassVisitorFilter(filter) );
    }

    public String quote_for_identifiers() {
        return "\"";
    }

    public String quote_for_strings() {
        return "'";
    }

    @Override
    public String as_identifier(String id) {
        String quote = this.quote_for_identifiers();
//        No se porque no esta disponible wrapIfMissing
//        return StringUtils.wrapIfMissing(id,quote);
        if (id.startsWith(quote)) {
            return id;
        }
        return quote + id + quote;

    }

    @Override
    public String as_string(String s) {
        String quote = this.quote_for_strings();
//        No se porque no esta disponible wrapIfMissing
//        return StringUtils.wrapIfMissing(id,quote);
        if (s.startsWith(quote)) {
            return s;
        }
        return quote + s + quote;

    }

    @Override
    public String as_string(byte[] data) {
        return this.expressionBuilder.bytearray_0x(data);
//        return this.expressionBuilder.bytearray_hex(data);
//        return this.expressionBuilder.bytearray_x(data);
    }
    
    @Override
    public String as_string(boolean value) {
        return value? "TRUE" : "FALSE";
    }

    @Override
    public String as_string(Number value) {
        return Objects.toString(value);
    }
    
    @Override
    public String as_string(Object value) {
        if( value == null ) {
            return "NULL";
        }
        if( value instanceof CharSequence ) {
            return as_string(value.toString());
        }
        if( value instanceof Number ) {
            return as_string((Number)value);
        }
        if( value instanceof Boolean ) {
            return as_string((boolean)value);
        }
        if( value instanceof byte[] ) {
            return as_string((byte[])value);
        }
        throw new IllegalArgumentException("Can't support type of value '"+value.getClass().getName()+"'.");
    }
    
    @Override
    public ExpressionBuilder expression() {
        return this.expressionBuilder;
    }

    @Override
    public boolean has_spatial_functions() {
        return this.hasSpatialFunctions;
    }

    @Override
    public GeometrySupportType geometry_support_type() {
        return this.geometrySupportType;
    }

    protected ExpressionBuilder createExpressionBuilder() {
        return ExpressionEvaluatorLocator.getManager().createExpressionBuilder();
    }

    @Override
    public Object srs_id(IProjection projection) {
        String abrev = projection.getAbrev();
        return abrev.split(":")[1].trim();
    }

    @Override
    public String default_schema() {
        return this.defaultSchema;
    }

    @Override
    public boolean support_schemas() {
        return this.supportSchemas;
    }

    @Override
    public String sqltype(int type, int p, int s, int geomType, int geomSubtype) {
        switch (type) {
            case DataTypes.BOOLEAN:
                return type_boolean;
            case DataTypes.BYTE:
                return type_byte;
            case DataTypes.BYTEARRAY:
                return type_bytearray;
            case DataTypes.GEOMETRY:
                return type_geometry;
            case DataTypes.CHAR:
                return type_char;
            case DataTypes.DATE:
                return type_date;
            case DataTypes.DOUBLE:
                // FIXME: Si cargamos la capa "country" al exportarla a
                // SQLServer falla por:
                //  Error de desbordamiento aritmtico al convertir float al tipo de datos numeric.
                // Al parecer la capa declara la columna sqkm_ctry como Numeric(12,6) y para 
                // Algeria intenta asignarle un valor de 2320972.0 y falla.
                // Habria que repasar el proveedor de shape.

//                if (p > 1) {
//                    if (s < 0) {
//                        return MessageFormat.format(config.getString(SQLConfig.type_numeric_p), p);
//                    }
//                    return MessageFormat.format(config.getString(SQLConfig.type_numeric_ps), p,s);
//                }
                return MessageFormat.format(type_double, p, s);
            case DataTypes.BIGDECIMAL:
                if (p < 1) {
                    p = 20;
                }
                if (s < 0) {
                    s = 10;
                }
                return MessageFormat.format(type_bigdecimal, p, s);
            case DataTypes.FLOAT:
                return MessageFormat.format(type_float, p, s);
            case DataTypes.INT:
                return MessageFormat.format(type_int, p, s);
            case DataTypes.LONG:
                return MessageFormat.format(type_long, p, s);
            case DataTypes.STRING:
                if (p < 0) {
                    return type_string;
                } else if (p < 4096) {
                    return MessageFormat.format(type_string_p, p);
                }
                return type_string;
            case DataTypes.TIME:
                return type_time;
            case DataTypes.TIMESTAMP:
                return type_timestamp;
            case DataTypes.VERSION:
                return type_version;
            case DataTypes.URI:
                return type_URI;
            case DataTypes.URL:
                return type_URL;
            case DataTypes.FILE:
                return type_FILE;
            case DataTypes.FOLDER:
                return type_FOLDER;
            default:
                return null;
        }
    }

    @Override
    public Object sqlgeometrytype(int type, int subtype) {
        // Devuelve un Object por que algunos gestores de BBDD utilizan
        // identificadores numericos para el tipo y otros strings.
        // Por defecto vamos a devolver strings.
        if (sqlgeometrytypes == null) {
            sqlgeometrytypes = new HashMap<>();
            sqlgeometrytypes.put(new ImmutablePair<>(Geometry.TYPES.POINT, Geometry.SUBTYPES.GEOM2D), "POINT");
            sqlgeometrytypes.put(new ImmutablePair<>(Geometry.TYPES.POINT, Geometry.SUBTYPES.GEOM3D), "POINTZ");
            sqlgeometrytypes.put(new ImmutablePair<>(Geometry.TYPES.POINT, Geometry.SUBTYPES.GEOM2DM), "POINTM");
            sqlgeometrytypes.put(new ImmutablePair<>(Geometry.TYPES.POINT, Geometry.SUBTYPES.GEOM3DM), "POINTZM");

            sqlgeometrytypes.put(new ImmutablePair<>(Geometry.TYPES.LINE, Geometry.SUBTYPES.GEOM2D), "LINESTRING");
            sqlgeometrytypes.put(new ImmutablePair<>(Geometry.TYPES.LINE, Geometry.SUBTYPES.GEOM3D), "LINESTRINGZ");
            sqlgeometrytypes.put(new ImmutablePair<>(Geometry.TYPES.LINE, Geometry.SUBTYPES.GEOM2DM), "LINESTRINGM");
            sqlgeometrytypes.put(new ImmutablePair<>(Geometry.TYPES.LINE, Geometry.SUBTYPES.GEOM3DM), "LINESTRINGZM");

            sqlgeometrytypes.put(new ImmutablePair<>(Geometry.TYPES.POLYGON, Geometry.SUBTYPES.GEOM2D), "POLYGON");
            sqlgeometrytypes.put(new ImmutablePair<>(Geometry.TYPES.POLYGON, Geometry.SUBTYPES.GEOM3D), "POLYGONZ");
            sqlgeometrytypes.put(new ImmutablePair<>(Geometry.TYPES.POLYGON, Geometry.SUBTYPES.GEOM2DM), "POLYGONM");
            sqlgeometrytypes.put(new ImmutablePair<>(Geometry.TYPES.POLYGON, Geometry.SUBTYPES.GEOM3DM), "POLYGONZM");

            sqlgeometrytypes.put(new ImmutablePair<>(Geometry.TYPES.MULTIPOINT, Geometry.SUBTYPES.GEOM2D), "MULTIPOINT");
            sqlgeometrytypes.put(new ImmutablePair<>(Geometry.TYPES.MULTIPOINT, Geometry.SUBTYPES.GEOM3D), "MULTIPOINTZ");
            sqlgeometrytypes.put(new ImmutablePair<>(Geometry.TYPES.MULTIPOINT, Geometry.SUBTYPES.GEOM2DM), "MULTIPOINTM");
            sqlgeometrytypes.put(new ImmutablePair<>(Geometry.TYPES.MULTIPOINT, Geometry.SUBTYPES.GEOM3DM), "MULTIPOINTZM");

            sqlgeometrytypes.put(new ImmutablePair<>(Geometry.TYPES.MULTILINE, Geometry.SUBTYPES.GEOM2D), "MULTILINESTRING");
            sqlgeometrytypes.put(new ImmutablePair<>(Geometry.TYPES.MULTILINE, Geometry.SUBTYPES.GEOM3D), "MULTILINESTRINGZ");
            sqlgeometrytypes.put(new ImmutablePair<>(Geometry.TYPES.MULTILINE, Geometry.SUBTYPES.GEOM2DM), "MULTILINESTRINGM");
            sqlgeometrytypes.put(new ImmutablePair<>(Geometry.TYPES.MULTILINE, Geometry.SUBTYPES.GEOM3DM), "MULTILINESTRINGZM");

            sqlgeometrytypes.put(new ImmutablePair<>(Geometry.TYPES.MULTICURVE, Geometry.SUBTYPES.GEOM2D), "MULTILINESTRING");
            sqlgeometrytypes.put(new ImmutablePair<>(Geometry.TYPES.MULTICURVE, Geometry.SUBTYPES.GEOM3D), "MULTILINESTRINGZ");
            sqlgeometrytypes.put(new ImmutablePair<>(Geometry.TYPES.MULTICURVE, Geometry.SUBTYPES.GEOM2DM), "MULTILINESTRINGM");
            sqlgeometrytypes.put(new ImmutablePair<>(Geometry.TYPES.MULTICURVE, Geometry.SUBTYPES.GEOM3DM), "MULTILINESTRINGZM");

            sqlgeometrytypes.put(new ImmutablePair<>(Geometry.TYPES.MULTIPOLYGON, Geometry.SUBTYPES.GEOM2D), "MULTIPOLYGON");
            sqlgeometrytypes.put(new ImmutablePair<>(Geometry.TYPES.MULTIPOLYGON, Geometry.SUBTYPES.GEOM3D), "MULTIPOLYGONZ");
            sqlgeometrytypes.put(new ImmutablePair<>(Geometry.TYPES.MULTIPOLYGON, Geometry.SUBTYPES.GEOM2DM), "MULTIPOLYGONM");
            sqlgeometrytypes.put(new ImmutablePair<>(Geometry.TYPES.MULTIPOLYGON, Geometry.SUBTYPES.GEOM3DM), "MULTIPOLYGONZM");

            sqlgeometrytypes.put(new ImmutablePair<>(Geometry.TYPES.MULTISURFACE, Geometry.SUBTYPES.GEOM2D), "MULTIPOLYGON");
            sqlgeometrytypes.put(new ImmutablePair<>(Geometry.TYPES.MULTISURFACE, Geometry.SUBTYPES.GEOM3D), "MULTIPOLYGONZ");
            sqlgeometrytypes.put(new ImmutablePair<>(Geometry.TYPES.MULTISURFACE, Geometry.SUBTYPES.GEOM2DM), "MULTIPOLYGONM");
            sqlgeometrytypes.put(new ImmutablePair<>(Geometry.TYPES.MULTISURFACE, Geometry.SUBTYPES.GEOM3DM), "MULTIPOLYGONZM");

            sqlgeometrytypes.put(new ImmutablePair<>(Geometry.TYPES.GEOMETRY, Geometry.SUBTYPES.GEOM2D), "GEOMETRY");
            sqlgeometrytypes.put(new ImmutablePair<>(Geometry.TYPES.GEOMETRY, Geometry.SUBTYPES.GEOM3D), "GEOMETRYZ");
            sqlgeometrytypes.put(new ImmutablePair<>(Geometry.TYPES.GEOMETRY, Geometry.SUBTYPES.GEOM2DM), "GEOMETRYM");
            sqlgeometrytypes.put(new ImmutablePair<>(Geometry.TYPES.GEOMETRY, Geometry.SUBTYPES.GEOM3DM), "GEOMETRYZM");
        }
        return sqlgeometrytypes.get(new ImmutablePair<>(type, subtype));
    }

    @Override
    public Object sqlgeometrydimension(int type, int subtype) {
        // Devuelve un Object por que algunos gestortes de BBDD utilizan
        // identificadores numericos para las dimensiones y otros strings.
        // Por defecto vamos a devolver enteros.
        switch (subtype) {
            case Geometry.SUBTYPES.GEOM3D:
                return 3;
            case Geometry.SUBTYPES.GEOM2DM:
                return 3;
            case Geometry.SUBTYPES.GEOM3DM:
                return 4;
            case Geometry.SUBTYPES.GEOM2D:
            default:
                return 2;
        }
    }

    @Override
    public TableNameBuilder createTableNameBuilder() {
        return new TableNameBuilderBase();
    }

    protected SelectColumnBuilder createSelectColumnBuilder() {
        return new SelectColumnBuilderBase();
    }

    protected UpdateColumnBuilder createUpdateColumnBuilder() {
        return new UpdateColumnBuilderBase();
    }

    protected InsertColumnBuilder createInsertColumnBuilder() {
        return new InsertColumnBuilderBase();
    }

    protected OrderByBuilder createOrderByBuilder() {
        return new OrderByBuilderBase();
    }

    protected FromBuilder createFromBuilder() {
        return new FromBuilderBase();
    }

    protected SelectBuilder createSelectBuilder() {
        return new SelectBuilderBase();
    }

    protected UpdateBuilder createUpdateBuilder() {
        return new UpdateBuilderBase();
    }

    protected DeleteBuilder createDeleteBuilder() {
        return new DeleteBuilderBase();
    }

    protected GrantBuilder createGrantBuilder() {
        return new GrantBuilderBase();
    }

    protected GrantRoleBuilder createGrantRoleBuilder(TableNameBuilder table, String role) {
        return new GrantRoleBuilderBase(table, role);
    }

    protected DropTableBuilder createDropTableBuilder() {
        return new DropTableBuilderBase();
    }

    protected CreateTableBuilder createCreateTableBuilder() {
        return new CreateTableBuilderBase();
    }

    protected AlterTableBuilder createAlterTableBuilder() {
        return new AlterTableBuilderBase();
    }

    protected InsertBuilder createInsertBuilder() {
        return new InsertBuilderBase();
    }

    protected UpdateTableStatisticsBuilder createUpdateTableStatisticsBuilder() {
        return new UpdateTableStatisticsBuilderBase();
    }

    protected CreateIndexBuilder createCreateIndexBuilder() {
        return new CreateIndexBuilderBase();
    }

    @Override
    public SelectBuilder select() {
        if (this.select == null) {
            this.select = this.createSelectBuilder();
        }
        return this.select;
    }

    @Override
    public UpdateBuilder update() {
        if (this.update == null) {
            this.update = this.createUpdateBuilder();
        }
        return this.update;
    }

    @Override
    public UpdateTableStatisticsBuilder update_table_statistics() {
        if (this.update_table_statistics == null) {
            this.update_table_statistics = this.createUpdateTableStatisticsBuilder();
        }
        return this.update_table_statistics;
    }

    @Override
    public DropTableBuilder drop_table() {
        if (this.drop_table == null) {
            this.drop_table = this.createDropTableBuilder();
        }
        return this.drop_table;
    }

    @Override
    public CreateIndexBuilder create_index() {
        if (this.create_index == null) {
            this.create_index = this.createCreateIndexBuilder();
        }
        return this.create_index;
    }

    @Override
    public DeleteBuilder delete() {
        if (this.delete == null) {
            this.delete = this.createDeleteBuilder();
        }
        return this.delete;
    }

    @Override
    public InsertBuilder insert() {
        if (this.insert == null) {
            this.insert = this.createInsertBuilder();
        }
        return this.insert;
    }

    @Override
    public TableNameBuilder table_name() {
        if (this.table_name == null) {
            this.table_name = this.createTableNameBuilder();
        }
        return this.table_name;
    }

    
    @Override
    public AlterTableBuilder alter_table() {
        if (this.alter_table == null) {
            this.alter_table = this.createAlterTableBuilder();
        }
        return this.alter_table;
    }

    @Override
    public CreateTableBuilder create_table() {
        if (this.create_table == null) {
            this.create_table = this.createCreateTableBuilder();
        }
        return this.create_table;
    }

    @Override
    public GrantBuilder grant() {
        if (this.grant == null) {
            this.grant = this.createGrantBuilder();
        }
        return this.grant;
    }
    
    @Override
    public Column column(String name) {
        ColumnBase col = new ColumnBase(null, name);
        return col;
    }

    @Override
    public Column column(TableNameBuilder table, String name) {
        ColumnBase col = new ColumnBase(table, name);
        return col;
    }
    
    protected JoinBase createJoin(String type, TableNameBuilder table, Value expression) {
        return new JoinBase(type, table, expression);
    }

    public void accept(Visitor visitor, VisitorFilter filter) {
        if (this.select != null) {
            this.select.accept(visitor, filter);
        }
        if (this.update != null) {
            this.update.accept(visitor, filter);
        }
        if (this.insert != null) {
            this.insert.accept(visitor, filter);
        }
        if (this.delete != null) {
            this.delete.accept(visitor, filter);
        }
        if (this.alter_table != null) {
            this.alter_table.accept(visitor, filter);
        }
        if (this.create_table != null) {
            this.create_table.accept(visitor, filter);
        }
        if (this.drop_table != null) {
            this.drop_table.accept(visitor, filter);
        }
        if (this.table_name != null) {
            this.table_name.accept(visitor, filter);
        }
    }

    protected Formatter formatter() {
        return ExpressionBuilder.EMPTY_FORMATTER;
    }

    @Override
    public String toString() {
        return this.toString(formatter());
    }

    @Override
    public String toString(Formatter formatter) {
        if (this.select != null) {
            return this.select.toString(formatter);
        }
        if (this.update != null) {
            return this.update.toString(formatter);
        }
        if (this.insert != null) {
            return this.insert.toString(formatter);
        }
        if (this.delete != null) {
            return this.delete.toString(formatter);
        }
        if (this.alter_table != null) {
            return this.alter_table.toString(formatter);
        }
        if (this.create_table != null) {
            return this.create_table.toString(formatter);
        }
        if (this.drop_table != null) {
            return this.drop_table.toString(formatter);
        }
        if (this.update_table_statistics != null) {
            return this.update_table_statistics.toString(formatter);
        }
        if (this.table_name != null) {
            return this.table_name.toString(formatter);
        }
        return "";
    }

    @Override
    public CountBuilder count() {
        return new CountBuilderBase();
    }

    @Override
    public List<Parameter> parameters() {
        final List<Parameter> params = new ArrayList<>();
        this.accept(new Visitor() {
            @Override
            public void visit(Visitable value) {
                params.add((Parameter) value);
            }
        }, new ClassVisitorFilter(Parameter.class));
        return params;
    }

    @Override
    public List<Variable> variables() {
        final List<Variable> vars = new ArrayList<>();
        this.accept(new Visitor() {
            @Override
            public void visit(Visitable value) {
                if (!vars.contains((Variable) value)) {
                    vars.add((Variable) value);
                }
            }
        }, new ClassVisitorFilter(Variable.class));
        return vars;
    }

    @Override
    public List<String> parameters_names() {
        List<String> params = new ArrayList<>();
        for (Parameter param : parameters()) {
            String s;
            switch (param.type()) {
                case Constant:
                    Object theValue = param.value();
                    if (theValue == null) {
                        s = "null";
                    } else if (theValue instanceof String) {
                        s = "'" + (String) theValue + "'";
                    } else {
                        s = theValue.toString();
                    }
                    break;
                case Geometry:
                case Variable:
                default:
                    s = "\"" + param.name() + "\"";
            }
            params.add(s);
        }
        return params;
    }

    @Override
    public List<String> variables_names() {
        List<String> vars = new ArrayList<>();
        for (Variable var : this.variables()) {
            vars.add(var.name());
        }
        Collections.sort(vars);
        return vars;
    }
}
