package org.gvsig.fmap.dal.swing.impl.featuretable;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import javax.swing.table.AbstractTableModel;
import org.gvsig.fmap.dal.complements.Search;
import org.gvsig.fmap.dal.exception.DataException;
import org.gvsig.fmap.dal.feature.Feature;
import org.gvsig.fmap.dal.feature.FeatureAttributeDescriptor;
import org.gvsig.fmap.dal.feature.FeatureStore;
import org.gvsig.fmap.dal.feature.FeatureType;
import org.gvsig.tools.ToolsLocator;
import org.gvsig.tools.logger.FilteredLogger;
import org.gvsig.tools.util.UnmodifiableBasicList;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 *
 * @author jjdelcerro
 */
public class SimpleFeaturesTableModel 
        extends AbstractTableModel
        implements UnmodifiableBasicList<Feature> {

        private static final Logger LOGGER = LoggerFactory.getLogger(SimpleFeaturesTableModel.class);
    
        private final List<Feature> features;
        private final List<String> columnNames;
        private final FeatureType featureType;
        private FilteredLogger logger;

        public SimpleFeaturesTableModel(FeatureStore store) throws DataException {
            this(store.getDefaultFeatureType(), null, store.getFeatures());
        }
        
        public SimpleFeaturesTableModel(FeatureType featureType) {
            this(featureType, null, null);
            this.logger = new FilteredLogger(LOGGER, "SimpleFeaturesTableModel", 10);
        }
        
        public SimpleFeaturesTableModel(FeatureType featureType, List<String> columnNames, List<Feature> features) {
            this.logger = new FilteredLogger(LOGGER, "SimpleFeaturesTableModel", 10);
            this.features = features;
            this.featureType = featureType;
            if (columnNames == null || columnNames.isEmpty()) {
                this.columnNames = new ArrayList<>();
                Search search = (Search) ToolsLocator.getComplementsManager().get(
                        Search.COMPLEMENT_MANE, featureType
                );
                List<Search.OrderedAttribute> attributos = search.getOrderedAttributes(
                        Search.BASIC_TYPES_FILTER,
                        Search.STR_INT_LONG_LABEL_ORDER,
                        12
                );
                for (Search.OrderedAttribute attrdesc : attributos) {
                    this.columnNames.add(attrdesc.getDescriptor().getName());
                }
            } else {
                this.columnNames = columnNames;
            }
        }

        public List<Feature> getFeatures() {
            return this.features;
        }

        @Override
        public int getRowCount() {
            if (this.features == null) {
                return 0;
            }
            return this.features.size();
        }

        @Override
        public int getColumnCount() {
            return this.columnNames.size();
        }

        @Override
        public String getColumnName(int columnIndex) {
            String attrName = this.columnNames.get(columnIndex);
            if (this.featureType == null) {
                return attrName;
            }
            FeatureAttributeDescriptor attrdesc = this.featureType.getAttributeDescriptor(attrName);
            if (attrdesc == null) {
                return "C" + columnIndex;
            }
            return attrdesc.getLocalizedShortLabel();
        }

        @Override
        public Class<?> getColumnClass(int columnIndex) {
            if (this.featureType == null) {
                return String.class;
            }
            String attrName = this.columnNames.get(columnIndex);
            FeatureAttributeDescriptor attrdesc = this.featureType.getAttributeDescriptor(attrName);
            if (attrdesc == null) {
                return String.class;
            }
            if( attrdesc.isForeingKey() && attrdesc.getForeingKey().isClosedList() ) {
                return String.class;
            }
            Class theClass = attrdesc.getDataType().getDefaultClass();
            if( theClass==null ) {
                return String.class;
            }
            return theClass;
        }

        @Override
        public boolean isCellEditable(int rowIndex, int columnIndex) {
            return false;
        }

        @Override
        public Feature get(int position) {
            if (this.features == null) {
                return null;
            }
            Feature feature = this.features.get(position);
            return feature;
        }

        @Override
        public Object getValueAt(int rowIndex, int columnIndex) {
            if (this.features == null) {
                return null;
            }
            try {
                Feature feature = this.features.get(rowIndex);
                String attrName = this.columnNames.get(columnIndex);
                Object value = feature.get(attrName);
                FeatureAttributeDescriptor attrdesc = this.featureType.getAttributeDescriptor(attrName);
                if (attrdesc != null) {
                    if( attrdesc.isForeingKey() && attrdesc.getForeingKey().isClosedList() ) {
                        value = attrdesc.getForeingKey().getLabelForValue(value);
                    }
                }
                return value;
            } catch (Throwable th) {
                logger.warn("Can't get cell value at "+rowIndex+", "+columnIndex+".", th);
                return null;
            }
        }

        @Override
        public void setValueAt(Object aValue, int rowIndex, int columnIndex) {

        }

    @Override
    public List<Feature> toList() {
        return this.features;
    }

    @Override
    public boolean isEmpty() {
        return this.features.isEmpty();
    }

    @Override
    public int size() {
        return this.features.size();
    }

    @Override
    public Iterator<Feature> iterator() {
        return this.features.iterator();
    }
    
}
