/**
 * gvSIG. Desktop Geographic Information System.
 *
 * Copyright (C) 2007-2013 gvSIG Association.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 *
 * For any additional information, do not hesitate to contact us
 * at info AT gvsig.com, or visit our website www.gvsig.com.
 */
package org.gvsig.fmap.geom;

import java.util.logging.Level;
import java.util.logging.Logger;
import org.apache.commons.lang3.StringUtils;
import org.cresques.cts.IProjection;
import org.gvsig.fmap.geom.aggregate.MultiLine;
import org.gvsig.fmap.geom.aggregate.MultiPolygon;
import org.gvsig.fmap.geom.exception.CreateEnvelopeException;
import org.gvsig.fmap.geom.exception.CreateGeometryException;
import org.gvsig.fmap.geom.primitive.Envelope;
import org.gvsig.fmap.geom.primitive.Line;
import org.gvsig.fmap.geom.primitive.Point;
import org.gvsig.fmap.geom.primitive.Polygon;
import org.gvsig.fmap.geom.type.GeometryType;
import org.gvsig.fmap.geom.type.GeometryTypeNotSupportedException;
import org.gvsig.fmap.geom.type.GeometryTypeNotValidException;

/**
 *
 * @author jjdelcerro
 */
@SuppressWarnings("UseSpecificCatch")
public class GeometryUtils {

    private GeometryUtils() {
        
    }
    
    public static GeometryType getGeometryType(int geometryType, int geometrySubType) {
        GeometryManager geomManager = GeometryLocator.getGeometryManager();
        try {
            return geomManager.getGeometryType(geometryType, geometrySubType);
        } catch (Exception ex) {
            return null;
        }
    }
    
    public static boolean isSubtype(int geomTypeParent, int geomTypeChild) {
        GeometryManager geomManager = GeometryLocator.getGeometryManager();
        return geomManager.isSubtype(geomTypeParent, geomTypeChild);
    }
    
    public static Envelope createEnvelope(int subType) {
        GeometryManager geomManager = GeometryLocator.getGeometryManager();
        try {
            return geomManager.createEnvelope(subType);
        } catch (CreateEnvelopeException ex) {
            return null;
        }
    }
    
    public static Line createLine(int subType) {
        GeometryManager geomManager = GeometryLocator.getGeometryManager();
        try {
            return geomManager.createLine(subType);
        } catch (CreateGeometryException ex) {
            return null;
        }
    }
    
    public static MultiLine createMultiLine(int subType) {
        GeometryManager geomManager = GeometryLocator.getGeometryManager();
        try {
            return geomManager.createMultiLine(subType);
        } catch (CreateGeometryException ex) {
            return null;
        }
    }
    
    public static Polygon createPolygon(int subType) {
        GeometryManager geomManager = GeometryLocator.getGeometryManager();
        try {
            return geomManager.createPolygon(subType);
        } catch (CreateGeometryException ex) {
            return null;
        }
    }
        
    public static MultiPolygon createMultiPolygon(int subType) {
        GeometryManager geomManager = GeometryLocator.getGeometryManager();
        try {
            return geomManager.createMultiPolygon(subType);
        } catch (CreateGeometryException ex) {
            return null;
        }
    }
        
    public static Point createPoint(double x, double y) {
        GeometryManager geomManager = GeometryLocator.getGeometryManager();
        try {
            return geomManager.createPoint(x, y, Geometry.SUBTYPES.GEOM2D);
        } catch (CreateGeometryException ex) {
            return null;
        }
    }
    
    public static Point createPoint(double x, double y, double z) {
        GeometryManager geomManager = GeometryLocator.getGeometryManager();
        try {
            Point p = geomManager.createPoint(x, y, Geometry.SUBTYPES.GEOM3D);
            p.setCoordinateAt(Geometry.DIMENSIONS.Z, z);
            return p;
        } catch (CreateGeometryException ex) {
            return null;
        }
    }
    
    public static Point createPoint(double x, double y, double z, double m) {
        GeometryManager geomManager = GeometryLocator.getGeometryManager();
        try {
            Point p = geomManager.createPoint(x, y, Geometry.SUBTYPES.GEOM3DM);
            p.setCoordinateAt(Geometry.DIMENSIONS.Z, z);
            p.setCoordinateAt(p.getDimension()-1, m);
            return p;
        } catch (CreateGeometryException ex) {
            return null;
        }
    }
    
    public static Geometry createFrom(Object data) {
        GeometryManager geomManager = GeometryLocator.getGeometryManager();
        try {
            return geomManager.createFrom(data);
        } catch (GeometryException ex) {
            return null;
        }
    }
    
    public static Geometry createFrom(String wkt, String srs) {
        GeometryManager geomManager = GeometryLocator.getGeometryManager();
        try {
            return geomManager.createFrom(wkt, srs);
        } catch (GeometryException ex) {
            return null;
        }
    }

    public static Geometry createFrom(String wkt, IProjection srs) {
        GeometryManager geomManager = GeometryLocator.getGeometryManager();
        try {
            return geomManager.createFrom(wkt, srs);
        } catch (GeometryException ex) {
            return null;
        }
    }
    
    public static Geometry createFrom(String wkt) {
        GeometryManager geomManager = GeometryLocator.getGeometryManager();
        try {
            return geomManager.createFrom(wkt);
        } catch (GeometryException ex) {
            return null;
        }
    }
    
    public static Object convertTo(Geometry geom, String format) {
        try {
            return geom.convertTo(format);
        } catch (Exception ex) {
            return null;
        }
    }
        
    public static String toWKT(Geometry geom) {
        try {
            return geom.convertToWKT();
        } catch (Exception ex) {
            return null;
        }
    }

    public static byte[] toWKB(Geometry geom) {
        try {
            return geom.convertToWKB();
        } catch (Exception ex) {
            return null;
        }
    }
    
    public static byte[] toEWKB(Geometry geom) {
        try {
            return geom.convertToEWKB();
        } catch (Exception ex) {
            return null;
        }
    }
    
    public static boolean intersects(Geometry geom1, Geometry geom2) {
        try {
            return geom1.intersects(geom2);
        } catch (Exception ex) {
            return false;
        }
    }

    public static int getGeometryType(String typeName) {
        if( StringUtils.isBlank(typeName) ) {
            return Geometry.TYPES.UNKNOWN;
        }
        switch(typeName.toLowerCase()) {
            case "geometry":
                return Geometry.TYPES.GEOMETRY;
            case "point":
                    return Geometry.TYPES.POINT;
            case "curve":
                    return Geometry.TYPES.CURVE;
            case "surface":
                    return Geometry.TYPES.SURFACE;
            case "solid":
                    return Geometry.TYPES.SOLID;
            case "aggregate":
                    return Geometry.TYPES.AGGREGATE;
            case "multipoint":
                    return Geometry.TYPES.MULTIPOINT;
            case "multicurve":
                    return Geometry.TYPES.MULTICURVE;
            case "multisurface":
                    return Geometry.TYPES.MULTISURFACE;
            case "multisolid":
                    return Geometry.TYPES.MULTISOLID;
            case "circle":
                    return Geometry.TYPES.CIRCLE;
            case "arc":
                    return Geometry.TYPES.ARC;
            case "ellipse":
                    return Geometry.TYPES.ELLIPSE;
            case "spline":
                    return Geometry.TYPES.SPLINE;
            case "ellipticarc":
                    return Geometry.TYPES.ELLIPTICARC;
            case "complex":
                    return Geometry.TYPES.COMPLEX;
            case "line":
                    return Geometry.TYPES.LINE;
            case "polygon":
                    return Geometry.TYPES.POLYGON;
            case "ring":
                    return Geometry.TYPES.RING;
            case "multiline":
                    return Geometry.TYPES.MULTILINE;
            case "multipolygon":
                    return Geometry.TYPES.MULTIPOLYGON;
            case "circumference":
                    return Geometry.TYPES.CIRCUMFERENCE;
            case "periellipse":
                    return Geometry.TYPES.PERIELLIPSE;
            case "filledspline":
                    return Geometry.TYPES.FILLEDSPLINE;
            default:
                return Geometry.TYPES.UNKNOWN;
        }
    }

    public static int getGeometrySubtype(String subtype) {
        if( StringUtils.isBlank(subtype) ) {
            return Geometry.SUBTYPES.UNKNOWN;
        }
        switch(subtype.toUpperCase()) {
            case "GEOM2D":
            case "2D":
                    return Geometry.SUBTYPES.GEOM2D;
            case "GEOM3D":
            case "3D":
                    return Geometry.SUBTYPES.GEOM3D;
            case "GEOM2DM":
            case "2DM":
                    return Geometry.SUBTYPES.GEOM2DM;
            case "GEOM3DM":
            case "3DM":
                    return Geometry.SUBTYPES.GEOM3DM;
            default:
                return Geometry.SUBTYPES.UNKNOWN;
        }
    }    
}
