package org.gvsig.fmap.mapcontext.layers;

import java.util.ArrayList;
import java.util.List;

import org.apache.commons.lang3.StringEscapeUtils;
import org.apache.commons.lang3.StringUtils;
import org.cresques.cts.ICRSFactory;
import org.cresques.cts.IProjection;

import org.gvsig.fmap.geom.Geometry;
import org.gvsig.fmap.geom.Geometry.DIMENSIONS;
import org.gvsig.fmap.geom.primitive.Envelope;
import org.gvsig.tools.ToolsLocator;

public class DefaultLayerInformationBuilder implements LayerInformationBuilder {

    private String backgroundColor = "\"#FFFFFF\"";
    private String backgroundColorTitle = "\"#FBFFE1\""; // light yellow
    private String backgroundColorPropertyLabel = "\"#D6D6D6\""; // Gris
    private String backgroundColorPropertyValue1 = "\"#FEEDD6\""; // light salmon
    private String backgroundColorPropertyValue2 = "\"#EAEAEA\""; // light grey

    private int propertyCount = 0;
    private final List<Element> elements;

    private class DefaultTitleElement implements TitleElement {

        private String label;

        @Override
        public DefaultTitleElement label(String label) {
            this.label = StringEscapeUtils.escapeHtml3(label);
            return this;
        }

        @Override
        public TitleElement labelkey(String labelkey) {
            return label(ToolsLocator.getI18nManager().getTranslation(labelkey));
        }

        @Override
        public String build() {
            propertyCount = 0;
            return "<tr valign=\"middle\" >"
                    + "<td bgcolor=" + backgroundColorTitle + " align=\"center\" colspan=\"2\"><font face=\"Arial\" size=\"3\"><b> " + this.label + "</b></font></td>"
                    + "</tr>";
        }
    }

    private class DefaultRawElement implements RawElement {

        protected String value;

        @Override
        public RawElement value(String format, Object... args) {
            if (args == null) {
                this.value = format;
            } else {
                this.value = String.format(format, args);
            }
            return this;
        }

        @Override
        public RawElement value(Envelope env) {
            return asWKT(env);
        }

        @Override
        public RawElement value(Geometry geom) {
            return asWKT(geom);
        }

        @Override
        public RawElement value(IProjection proj) {
            if( proj == null ) {
                return value("");
            }
            return value(proj.getAbrev());
        }

        @Override
        public RawElement asWKT(IProjection proj) {
            if( proj == null ) {
                return value("");
            }
            String s = proj.export(ICRSFactory.FORMAT_WKT);
            if (s == null) {
                s = proj.export(ICRSFactory.FORMAT_WKT_ESRI);
            }
            if (s == null) {
                s = proj.export(ICRSFactory.FORMAT_PROJ4);
            }
            if (s == null) {
                s = proj.getAbrev();
            }
            if (s != null) {
                s = s.replaceAll("\\[", "[\n  ");
                s = s.replaceAll("]", "\n]");
            }
            return value(s);
        }

        @Override
        public RawElement asWKT(Envelope env) {
            if( env == null ) {
                return value("");
            }
            return asWKT(env.getGeometry());
        }

        @Override
        public RawElement asWKT(Geometry geom) {
            String s = "";
            try {
                s = geom.convertToWKT();
            } catch (Exception ex) {
            }
            return value(s);
        }

        @Override
        public String build() {
            String color = ((propertyCount++ % 2) == 0) ? backgroundColorPropertyValue1 : backgroundColorPropertyValue2;
            String content = "<tr valign=\"top\">\n";
            content += "<td bgcolor=" + color + "align=\"left\" colspan=\"2\"><font face=\"Arial\" size=\"3\"><div>" + value.replace("\\n",  "<br>\n") + "</div></font></td>\n";
            content += "</tr>\n";
            return content;
        }
    }

    private class DefaultTextElement extends DefaultRawElement implements TextElement {

        @Override
        public TextElement value(String format, Object... args) {
            if (args == null) {
                this.value = StringEscapeUtils.escapeHtml3(format);
            } else {
                this.value = StringEscapeUtils.escapeHtml3(String.format(format, args));
            }
            return this;
        }

        @Override
        public TextElement value(IProjection proj) {
            return (TextElement) super.value(proj);
        }

        @Override
        public TextElement value(Envelope env) {
            return (TextElement) super.value(env);
        }

        @Override
        public TextElement value(Geometry geom) {
            return (TextElement) super.value(geom);
        }

        @Override
        public TextElement asWKT(IProjection proj) {
            return (TextElement) super.asWKT(proj);
        }

        @Override
        public TextElement asWKT(Envelope env) {
            return (TextElement) super.asWKT(env);
        }

        @Override
        public TextElement asWKT(Geometry geom) {
            return (TextElement) super.asWKT(geom);
        }

        @Override
        public String build() {
            String color = ((propertyCount++ % 2) == 0) ? backgroundColorPropertyValue1 : backgroundColorPropertyValue2;
            String content = "<tr valign=\"top\">\n";
            content += "<td bgcolor=" + color + "align=\"left\" colspan=\"2\"><font face=\"Arial\" size=\"3\">" + value.replace("\n",  "<br>\n") + "</font></td>\n";
            content += "</tr>\n";
            return content;
        }
    }

    private class DefaultPropertyElement extends DefaultTextElement implements PropertyElement {

        private String label;
        private boolean monospace = false;

        @Override
        public DefaultPropertyElement label(String label) {
            this.label = StringEscapeUtils.escapeHtml3(label);
            return this;
        }

        @Override
        public PropertyElement labelkey(String labelkey) {
            return label(ToolsLocator.getI18nManager().getTranslation(labelkey));
        }

        @Override
        public PropertyElement value(IProjection proj) {
            if (this.label == null) {
                this.labelkey("_CRS");
            }
            if( proj == null ) {
                return (PropertyElement) value("");
            }
            return (PropertyElement) value(proj.getAbrev());
        }

        @Override
        public String build() {
            String color = ((propertyCount++ % 2) == 0) ? backgroundColorPropertyValue1 : backgroundColorPropertyValue2;
            String content = "<tr valign=\"top\">\n";
            if (!StringUtils.isEmpty(label)) {
                content += "<td nowrap bgcolor=" + backgroundColorPropertyLabel + "align=\"right\" width=\"140\"><font face=\"Arial\" size=\"3\">" + label + ":&nbsp;</font></td>\n";
            }
            if( this.monospace ) {
                content += "<td bgcolor=" + color + "align=\"left\"><font face=\"Monospaced\" size=\"3\">" + value.replace("\\n",  "<br>\n") + "</font></td>";
            } else {
                content += "<td bgcolor=" + color + "align=\"left\"><font face=\"Arial\" size=\"3\">" + value.replace("\\n",  "<br>\n") + "</font></td>";
            }
            content += "</tr>\n";
            return content;
        }

        @Override
        public PropertyElement value(String format, Object... args) {
            return (PropertyElement) super.value(format, args);
        }

        @Override
        public PropertyElement value(Envelope env) {
            return (PropertyElement) super.value(env);
        }

        @Override
        public PropertyElement value(Geometry geom) {
            return (PropertyElement) super.value(geom);
        }

        @Override
        public PropertyElement asWKT(IProjection proj) {
            return (PropertyElement) super.asWKT(proj);
        }

        @Override
        public PropertyElement asWKT(Envelope env) {
            return (PropertyElement) super.asWKT(env);
        }

        @Override
        public PropertyElement asWKT(Geometry geom) {
            return (PropertyElement) super.asWKT(geom);
        }

        @Override
        public PropertyElement monospace() {
            this.monospace = true;
            return this;
        }
    }

    private class DefaultEnvelopeElement implements EnvelopeElement {

        private Envelope value;

        @Override
        public DefaultEnvelopeElement value(Envelope value) {
            this.value = value;
            return this;
        }

        @Override
        public String build() {
            if( value == null ) {
                return new DefaultTextElement().value("None").build();
            } else {
                PropertyElement p1 = new DefaultPropertyElement();
                p1.labelkey("_upper_left_corner");
                p1.value("%+f, %+f", value.getMinimum(DIMENSIONS.X), value.getMaximum(DIMENSIONS.Y));
                p1.monospace();

                PropertyElement p2 = new DefaultPropertyElement();
                p2.labelkey("_upper_right_corner");
                p2.value("%+f, %+f", value.getMaximum(DIMENSIONS.X), value.getMaximum(DIMENSIONS.Y));
                p2.monospace();

                PropertyElement p3 = new DefaultPropertyElement();
                p3.labelkey("_lower_right_corner");
                p3.value("%+f, %+f", value.getMaximum(DIMENSIONS.X), value.getMinimum(DIMENSIONS.Y));
                p3.monospace();

                PropertyElement p4 = new DefaultPropertyElement();
                p4.labelkey("_lower_left_corner");
                p4.value("%+f, %+f", value.getMinimum(DIMENSIONS.X), value.getMinimum(DIMENSIONS.Y));
                p4.monospace();

                return p1.build() + p2.build() + p3.build() + p4.build();
            }
        }

    }

    public DefaultLayerInformationBuilder() {
        this.elements = new ArrayList<>();
    }

    @Override
    public DefaultLayerInformationBuilder backgroundColor(String color) {
        this.backgroundColor = color;
        return this;
    }

    @Override
    public DefaultLayerInformationBuilder backgroundColorTitle(String color) {
        this.backgroundColorTitle = color;
        return this;
    }

    @Override
    public DefaultLayerInformationBuilder backgroundColorPropertyLabel(String color) {
        this.backgroundColorPropertyLabel = color;
        return this;
    }

    @Override
    public DefaultLayerInformationBuilder backgroundColorPropertyValue1(String color) {
        this.backgroundColorPropertyValue1 = color;
        return this;
    }

    @Override
    public DefaultLayerInformationBuilder backgroundColorPropertyValue2(String color) {
        this.backgroundColorPropertyValue2 = color;
        return this;
    }

    @Override
    public TitleElement title() {
        TitleElement element = new DefaultTitleElement();
        this.elements.add(element);
        return element;
    }

    @Override
    public PropertyElement property() {
        PropertyElement element = new DefaultPropertyElement();
        this.elements.add(element);
        return element;
    }

    @Override
    public TextElement text() {
        TextElement element = new DefaultTextElement();
        this.elements.add(element);
        return element;
    }

    @Override
    public EnvelopeElement envelope() {
        EnvelopeElement element = new DefaultEnvelopeElement();
        this.elements.add(element);
        return element;
    }

    @Override
    public RawElement raw() {
        RawElement element = new DefaultRawElement();
        this.elements.add(element);
        return element;
    }

    @Override
    public String build() {
        StringBuilder html = new StringBuilder();
        html.append("<html>\n");
        html.append("<body bgcolor=").append(backgroundColor).append(" topmargin=\"0\" marginheight=\"0\">\n");
        html.append("<table cellpadding=\"0\" cellspacing=\"0\" align=\"center\" width=\"100%\">");
        boolean first = true;
        for (Element element : elements) {
            if( element instanceof TitleElement ) {
                if( first ) {
                    first = false;
                } else {
                    html.append("<tr>\n<td><br></td>\n<td></td></tr>\n");
                }
            }
            html.append(element.build());
        }
        html.append("</table>");
        html.append("</body>");
        html.append("</html>");
        return html.toString();
    }

    @Override
    public String toString() {
        return this.build();
    }
}
