/**
 * gvSIG. Desktop Geographic Information System.
 *
 * Copyright (C) 2007-2013 gvSIG Association.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * For any additional information, do not hesitate to contact us at info AT
 * gvsig.com, or visit our website www.gvsig.com.
 */
package org.gvsig.export.impl;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import org.apache.commons.io.IOUtils;
import org.cresques.cts.IProjection;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.gvsig.export.ExportManager;
import org.gvsig.export.ExportParameters;
import org.gvsig.export.ExportProcess;
import org.gvsig.fmap.dal.feature.FeatureStore;
import org.gvsig.tools.bookmarksandhistory.Bookmarks;
import org.gvsig.tools.bookmarksandhistory.History;
import org.gvsig.tools.bookmarksandhistory.impl.BaseBookmarks;
import org.gvsig.tools.bookmarksandhistory.impl.BaseHistory;

/**
 * Default {@link ExportManager} implementation.
 *
 * @author gvSIG Team
 * @version $Id$
 */
public class DefaultExportManager implements ExportManager {

    private static final Logger LOG = LoggerFactory.getLogger(DefaultExportManager.class);
    
    private static class DefaultExportFilter implements ExportFilter {

        private final String name;
        private final String expression;
    
        public DefaultExportFilter(String name, String expression) {
            this.name = name;
            this.expression = expression;
        }

        @Override
        public String getName() {
            return name;
        }

        @Override
        public String getExpression() {
            return expression;
        }
    }
    
    private Map<String, ExportFilter> filters = null;
    private File homeFolder = null;
    private final History<ExportParameters> history;
    private final Bookmarks<ExportParameters> bookmarks;


    public DefaultExportManager() {
        super();
        this.history = new BaseHistory<>(10);
        this.bookmarks = new BaseBookmarks<>();
    }

    private void loadFilters() {
        this.filters = new HashMap<>();
        if (this.homeFolder != null) {
            File filtersFile = new File(this.homeFolder, "filters");
            if (filtersFile.exists()) {
                FileInputStream fis = null;
                try {
                    fis = new FileInputStream(filtersFile);
                    Properties properties = new Properties();
                    properties.load(fis);
                    for (Object name : properties.keySet()) {
                        String value = (String) properties.get(name);
                        ExportFilter filter = new DefaultExportFilter((String) name, value);
                        this.filters.put(filter.getName(), filter);
                    }
                } catch (IOException ex) {
                    LOG.warn("Can't load filters file '" + filtersFile.getAbsolutePath() + "'.", ex);
                } finally {
                    IOUtils.closeQuietly(fis);
                }
            }
        }
    }

    public void saveFilters() {
        if (this.homeFolder != null) {
            Properties properties = new Properties();
            for (ExportFilter filter : this.filters.values()) {
                properties.setProperty(filter.getName(), filter.getExpression());
            }
            File filtersFile = new File(this.homeFolder, "filters");
            FileOutputStream fos = null;
            try {
                fos = new FileOutputStream(filtersFile);
                properties.store(fos, "");
            } catch (IOException ex) {
                LOG.warn("Can't load filters file '" + filtersFile.getAbsolutePath() + "'.", ex);
            } finally {
                IOUtils.closeQuietly(fos);
            }
        }
        
    }
    
    public Map<String,ExportFilter> getFiltersAsMap() {
        if (this.filters == null) {
            loadFilters();
        }
        return this.filters;
    }
    
    @Override
    public List<ExportFilter> getFilters() {
        List<ExportFilter> l = new ArrayList<>();
        l.addAll(this.getFiltersAsMap().values());
        return l;
    }

    @Override
    public ExportFilter getFilter(String filterName) {
        return this.getFiltersAsMap().get(filterName);
    }

    @Override
    public ExportFilter addFilter(String name, String expression) {
        ExportFilter filter = new DefaultExportFilter(name, expression);
        this.filters.put(filter.getName(), filter);
        return filter;
    }

    @Override
    public void setHomeFolder(File homeFolder) {
        this.homeFolder = homeFolder;
    }

    @Override
    public ExportProcess createProcess() {
        ExportProcess process = new DefaultExportProcess();
        return process;
    }

    @Override
    public ExportProcess createProcess(FeatureStore store) {
        ExportProcess process = new DefaultExportProcess();
        process.setSourceFeatureStore(store);
        return process;
    }

    @Override
    public ExportProcess createProcess(FeatureStore store, IProjection contextProjection) {
        ExportProcess process = new DefaultExportProcess();
        process.setSourceFeatureStore(store);
        process.setContextProjection(contextProjection);
        return process;
    }


    @Override
    public History<ExportParameters> getHistory() {
        return this.history;
    }

    @Override
    public Bookmarks<ExportParameters> getBookmarks() {
        return this.bookmarks;
    }
}
