/**
 * gvSIG. Desktop Geographic Information System.
 *
 * Copyright (C) 2007-2013 gvSIG Association.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 *
 * For any additional information, do not hesitate to contact us
 * at info AT gvsig.com, or visit our website www.gvsig.com.
 */
package org.gvsig.export.swing.impl;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Date;
import java.util.HashSet;
import java.util.Set;
import javax.swing.JButton;

import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.SwingUtilities;

import jwizardcomponent.DefaultJWizardComponents;
import org.apache.commons.lang3.StringUtils;

import org.gvsig.export.ExportException;
import org.gvsig.export.ExportLocator;
import org.gvsig.export.ExportManager;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.gvsig.fmap.dal.feature.Feature;
import org.gvsig.gui.beans.wizard.WizardPanel;
import org.gvsig.gui.beans.wizard.WizardPanelActionListener;
import org.gvsig.tools.ToolsLocator;
import org.gvsig.tools.i18n.I18nManager;
import org.gvsig.export.ExportParameters;
import org.gvsig.export.ExportProcess;
import org.gvsig.export.Filter;
import org.gvsig.export.spi.ExportServiceFactory;
import org.gvsig.export.swing.ExportSwingLocator;
import org.gvsig.export.swing.JExportProcessPanel;
import org.gvsig.export.swing.impl.panels.ExportProgressPanel;
import org.gvsig.export.swing.impl.panels.SelectFormatPanel;
import org.gvsig.export.swing.spi.ExportPanel;
import org.gvsig.export.swing.spi.ExportPanels;
import org.gvsig.export.swing.spi.ExportPanelsFactory;
import org.gvsig.export.swing.spi.ExportPanelsManager;
import org.gvsig.gui.beans.wizard.WizardPanelWithLogo;
import org.gvsig.gui.beans.wizard.panel.OptionPanel;
import org.gvsig.tools.bookmarksandhistory.History;

/**
 * Default implementation for the {@link JExportProcessPanel}.
 *
 * @author gvSIG Team
 * @version $Id$
 */
@SuppressWarnings("UseSpecificCatch")
public class DefaultJExportProcessPanel
        extends JPanel
        implements JExportProcessPanel, WizardPanel {

    private static final Logger LOG = LoggerFactory.getLogger(DefaultJExportProcessPanel.class);

    private final ExportProcess process;
    private final Filter<ExportServiceFactory> serviceFilter;
    private final Set<ExportFinishListener> finishListeners;
    private WizardPanelWithLogo wizardPanel;
    private WizardPanelActionListener wizardListener;
    private int lastAction;

    public DefaultJExportProcessPanel(
            ExportProcess process,
            Filter<ExportServiceFactory> serviceFilter
    ) {
        this.process = process;
        this.serviceFilter = serviceFilter;
        this.finishListeners = new HashSet<>();
        this.initComponents();
    }

    private void initComponents() {
        I18nManager i18n = ToolsLocator.getI18nManager();

        this.wizardPanel = new WizardPanelWithLogo();
        this.addPanel(
                new SelectFormatPanel(
                        this,
                        this.process.getParameters(),
                        this.serviceFilter,
                        new ActionListener() {
                    @Override
                    public void actionPerformed(ActionEvent e) {
                        doAddPanels(((SelectFormatPanel) (e.getSource())).getSelectedService());
                    }
                }
                )
        );
        this.setLayout(new BorderLayout());
        this.add(this.wizardPanel, BorderLayout.CENTER);
        this.wizardPanel.setFinishButtonText(i18n.getTranslation("export"));
        this.wizardPanel.setWizardListener(this);
        this.setWizardPanelActionListener(this.getWizardPanelActionListener());

        this.wizardPanel.setNextButtonEnabled(false);
        this.wizardPanel.setFinishButtonEnabled(false);
        this.setPreferredSize(new Dimension(600, 350));
    }

    private void addPanel(ExportPanel panel) {
        OptionPanel optionPanel = new WizardOptionPanelAdapter(this, panel);
        this.wizardPanel.addOptionPanel(optionPanel);
    }

    private void doAddPanels(ExportServiceFactory serviceFactory) {

        try {
            removePreviousPanels();
            if (this.process.getParameters() != null) {
                // Check parameters are for the actual serviceFactory, if not, reset.
                if ( !StringUtils.equals(this.process.getParameters().getServiceName(),serviceFactory.getName())) {
                    this.process.setOutputFormat(serviceFactory.getName());
                }
            } else if (this.process.getParameters()==null) {
                this.process.setOutputFormat(serviceFactory.getName());
            }
            

            ExportParameters parameters = this.process.getParameters();

            ExportPanelsManager panelsManager = ExportSwingLocator.getExportPanelsManager();
            ExportPanelsFactory panelsFactory = panelsManager.getPanelsFactory(serviceFactory.getName());
            ExportPanels panels = panelsFactory.createPanels(this, parameters);
            for (ExportPanel panel : panels) {
                this.addPanel(panel);
            }
            if (parameters.needsSelectTargetProjection()) {
                if (!panels.contains(ExportPanelsManager.PANEL_SELECT_TARGET_PROJECTION)) {
                    this.addPanel(panelsManager.createStandardPanel(
                            ExportPanelsManager.PANEL_SELECT_TARGET_PROJECTION,
                            this,
                            parameters
                    )
                    );
                }
            }
            if (!panels.contains(ExportPanelsManager.PANEL_SELECT_FILTER)) {
                this.addPanel(panelsManager.createStandardPanel(
                        ExportPanelsManager.PANEL_SELECT_FILTER,
                        this,
                        parameters
                )
                );
            }
            this.addPanel(new ExportProgressPanel(this, parameters));

        } catch (ExportException ex) {
            LOG.warn("Can't add panels to wizard.", ex);
            showError(ex);
        }
    }

    @Override
    public ExportProcess getProcess() {
        return this.process;
    }

    @Override
    public ExportParameters getParameters() {
        return this.process.getParameters();
    }

    @Override
    public void setWizardPanelActionListener(
            WizardPanelActionListener wizardListener
    ) {
        this.wizardListener = wizardListener;
    }

    @Override
    public WizardPanelActionListener getWizardPanelActionListener() {
        if (wizardListener == null) {
            wizardListener = new WizardPanelActionListener() {
                @Override
                public void finish(WizardPanel wizardPanel) {
                    doExport();
                }

                @Override
                public void cancel(WizardPanel wizardPanel) {
                    doCancel();
                }
            };
        }
        return wizardListener;
    }

    private void removePreviousPanels() {
        DefaultJWizardComponents wizardComponents = wizardPanel.getWizardComponents();
        for (int i = wizardComponents.getWizardPanelList().size() - 1; i >= 1; i--) {
            wizardComponents.removeWizardPanel(i);
        }
    }

    private void lastWizard() {
        this.wizardPanel.getWizardComponents().getNextButton()
                .getActionListeners()[0].actionPerformed(null);
    }

    private void showError(Exception e) {
        Feature f = null;
        if (e instanceof ExportException) {
            f = ((ExportException) e).getFeature();
        }
        I18nManager i18nManager = ToolsLocator.getI18nManager();
        MessagePanel.showMessage(
                i18nManager.getTranslation("_Warning"),
                i18nManager.getTranslation("_There_have_been_problems_exporting_data"),
                e,
                f
        );
    }

    @Override
    public void addFinishListener(ExportFinishListener finishListener) {
        this.finishListeners.add(finishListener);
    }

    private void fireCancelled() {
        if (!SwingUtilities.isEventDispatchThread()) {
            SwingUtilities.invokeLater(new Runnable() {
                @Override
                public void run() {
                    fireCancelled();
                }
            });
        }
        for (ExportFinishListener listener : this.finishListeners) {
            try {
                listener.cancelled(this);
            } catch (Exception ex) {

            }
        }
    }

    private void fireFinished() {
        if (!SwingUtilities.isEventDispatchThread()) {
            SwingUtilities.invokeLater(new Runnable() {
                @Override
                public void run() {
                    fireFinished();
                }
            });
            return;
        }
        for (ExportFinishListener listener : this.finishListeners) {
            try {
                listener.finished(this);
            } catch (Exception ex) {

            }
        }
    }

    private void doCancel() {
        this.setVisible(false);
        this.fireCancelled();
    }

    private void doExport() {
        lastWizard();
        ExportParameters params = this.process.getParameters();
        ExportManager exportManager = ExportLocator.getManager();

        ExportParameters copyParams = null;
        try {
            copyParams = (ExportParameters) params.clone();
            copyParams.setCreationDate(new Date());
            History<ExportParameters> h = exportManager.getHistory();
            h.add(copyParams);
        } catch (Exception ex) {
            LOG.warn("Not able to create a clone object for export parameters", ex);
        }

        Thread th = new Thread(
                new Runnable() {
            @Override
            public void run() {
                try {
                    process.run();
                    doProcessFinished();
                } catch (Exception ex) {
                    LOG.warn("Can't export", ex);
                    showError(ex);
                }
            }
        },"exportProcess");
        th.start();
    }

    private void doProcessFinished() {
        fireFinished();
        this.setButtonText(JExportProcessPanel.BUTTON_CANCEL, ToolsLocator.getI18nManager().getTranslation("_Close"));
    }

    @Override
    public JComponent asJComponent() {
        return this;
    }

    private JButton getButton(int button) {
        switch (button) {
            case BUTTON_BACK:
                return this.wizardPanel.getWizardComponents().getBackButton();
            case BUTTON_NEXT:
                return this.wizardPanel.getWizardComponents().getNextButton();
            case BUTTON_FINISH:
                return this.wizardPanel.getWizardComponents().getFinishButton();
            case BUTTON_CANCEL:
                return this.wizardPanel.getWizardComponents().getCancelButton();
        }
        throw new IllegalArgumentException("Illegal button value (" + button + ").");
    }

    @Override
    public void setButtonEnabled(int button, boolean enabled) {
        this.getButton(button).setEnabled(enabled);
    }

    @Override
    public void setButtonText(int button, String text) {
        this.getButton(button).setText(text);
    }

    @Override
    public boolean isButtonEnabled(int button) {
        return this.getButton(button).isEnabled();
    }

    @Override
    public String getButtonText(int button) {
        return this.getButton(button).getText();
    }

    public void setLastAcction(int action) {
        this.lastAction = action;
    }

    public int getLastAction() {
        return this.lastAction;
    }

    @Override
    public void nextPanel() {
        this.wizardPanel.doAction(WizardPanelWithLogo.ACTION_NEXT);
    }

    @Override
    public void previousPanel() {
        this.wizardPanel.doAction(WizardPanelWithLogo.ACTION_PREVIOUS);
    }

}
