package org.gvsig.export.swing.impl.panels;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.List;
import java.util.function.Predicate;
import javax.swing.DefaultListModel;
import javax.swing.JComponent;
import javax.swing.JScrollPane;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import org.apache.commons.lang3.StringUtils;
import org.gvsig.export.ExportLocator;
import org.gvsig.export.ExportParameters;
import org.gvsig.export.Filter;
import org.gvsig.export.spi.ExportServiceFactory;
import org.gvsig.export.swing.JExportProcessPanel;
import org.gvsig.export.swing.spi.ExportPanel;
import org.gvsig.export.swing.spi.ExportPanelValidationException;
import org.gvsig.tools.ToolsLocator;
import org.gvsig.tools.bookmarksandhistory.Bookmark;
import org.gvsig.tools.bookmarksandhistory.Bookmarks;
import org.gvsig.tools.bookmarksandhistory.History;
import org.gvsig.tools.i18n.I18nManager;
import org.gvsig.tools.swing.api.ListElement;
import org.gvsig.tools.swing.api.ToolsSwingLocator;
import org.gvsig.tools.swing.api.bookmarkshistory.ActionEventWithCurrentValue;
import static org.gvsig.tools.swing.api.bookmarkshistory.ActionEventWithCurrentValue.ID_GETVALUE;
import static org.gvsig.tools.swing.api.bookmarkshistory.ActionEventWithCurrentValue.ID_SETVALUE;
import org.gvsig.tools.swing.api.bookmarkshistory.BookmarksController;
import org.gvsig.tools.swing.api.bookmarkshistory.HistoryController;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 *
 * @author jjdelcerro
 */
@SuppressWarnings("UseSpecificCatch")
public class SelectFormatPanel
        extends SelectFormatPanelView
        implements ExportPanel {

    private static final Logger LOG = LoggerFactory.getLogger(SelectFormatPanel.class);

    private final JExportProcessPanel processPanel;
    private final Filter<ExportServiceFactory> serviceFilter;
    private final ActionListener onExitPanel;

    private ExportParameters parameters;
    private BookmarksController<ExportParameters> bookmarksController;
    private HistoryController<ExportParameters> historyController;

    public SelectFormatPanel(
            JExportProcessPanel processPanel,
            ExportParameters parameters,
            Filter<ExportServiceFactory> serviceFilter,
            ActionListener onExitPanel
    ) {
        this.processPanel = processPanel;
        this.parameters = parameters;
        this.serviceFilter = serviceFilter;
        this.onExitPanel = onExitPanel;
        initComponents();
    }

    private void initComponents() {
        this.txtDescription.setBackground(this.lblEmpty.getBackground());
        this.txtDescription.setForeground(this.lblEmpty.getForeground());
        if (this.txtDescription.getParent().getParent() instanceof JScrollPane) {
            JScrollPane scrollPanel = (JScrollPane) this.txtDescription.getParent().getParent();
            scrollPanel.setBorder(null);
        }

        List<ExportServiceFactory> factories = ExportLocator.getServiceManager()
                .getServiceFactories(serviceFilter);
        DefaultListModel model = new DefaultListModel();
        for (ExportServiceFactory factory : factories) {
            model.addElement(new ListElement(factory.getLabel(), factory));
        }
        this.lstServices.setModel(model);
        this.lstServices.addListSelectionListener(new ListSelectionListener() {
            @Override
            public void valueChanged(ListSelectionEvent e) {
                if (e.getValueIsAdjusting()) {
                    return;
                }
                doUpdateDescription();
            }
        });

        Predicate<ExportParameters> filter = new Predicate<ExportParameters>() {
            @Override
            public boolean test(ExportParameters params) {
                if (params == null) {
                    return false;
                }
                ExportServiceFactory selectedService = getSelectedService();
                if (selectedService == null) {
                    return false;
                }
                String selectedServiceName = selectedService.getName();
                String paramsServiceName = params.getServiceName();
                return StringUtils.equals(paramsServiceName, selectedServiceName);
            }
        };
        ActionListener historyAndBookmarkListener = new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                ActionEventWithCurrentValue<ExportParameters> b = (ActionEventWithCurrentValue<ExportParameters>) e;
                switch (b.getID()) {
                    case ID_GETVALUE:
                        // Esta deshabilitado esta opcion
                        break;
                    case ID_SETVALUE:
                        if (b.getCurrentValue() == null) {
                            return;
                        }
                        ExportParameters exportParams;
                        try {
                            exportParams = b.getCurrentValue().clone();
                        } catch (Exception ex) {
                            LOG.warn("Not been able to clone export parameters", ex);
                            return;
                        }
                        setSelectedService(exportParams.getFactory());

                        processPanel.getProcess().setParameters(exportParams);
                        parameters = exportParams;
                        break;
                }

            }
        };
        Bookmarks<ExportParameters> bookmarks = ExportLocator.getManager().getBookmarks();
        this.bookmarksController = ToolsSwingLocator.getToolsSwingManager().createBookmarksController(bookmarks, this.btnBookmark);
        this.bookmarksController.setAllowAddBookmarks(false);
        this.bookmarksController.setFilter(new Predicate<Bookmark<ExportParameters>>() {
            @Override
            public boolean test(Bookmark<ExportParameters> bookmark) {
                ExportParameters params = bookmark.getValue();
                if (params == null) {
                    return false;
                }
                ExportServiceFactory selectedService = getSelectedService();
                if (selectedService == null) {
                    return true;
                }
                String selectedServiceName = selectedService.getName();
                String paramsServiceName = params.getServiceName();
                return StringUtils.equals(paramsServiceName, selectedServiceName);
            }
        });       
        this.bookmarksController.addActionListener(historyAndBookmarkListener);
        History<ExportParameters> history = ExportLocator.getManager().getHistory();
        this.historyController = ToolsSwingLocator.getToolsSwingManager().createHistoryController(history, this.btnHistory);
        this.historyController.setFilter(filter);
        this.historyController.addActionListener(historyAndBookmarkListener);
    }

    private void doUpdateDescription() {
        ExportServiceFactory factory = this.getSelectedService();
        if (factory == null) {
            return;
        }
        this.txtDescription.setText(factory.getDescription());
        this.processPanel.setButtonEnabled(JExportProcessPanel.BUTTON_NEXT, true);
    }

    @Override
    public String getIdPanel() {
        return "SelectFormatPanel";
    }

    @Override
    public String getTitlePanel() {
        I18nManager i18n = ToolsLocator.getI18nManager();
        return i18n.getTranslation("select_format_to_export");
    }

    @Override
    public boolean validatePanel() throws ExportPanelValidationException {
        I18nManager i18n = ToolsLocator.getI18nManager();
        ExportServiceFactory factory = this.getSelectedService();
        if (factory == null) {
            throw new ExportPanelValidationException(
                    i18n.getTranslation("_You_must_select_an_item_from_the_list_of-formats")
            );
        }
        return true;
    }

    @Override
    public void enterPanel() {

    }

    @Override
    public void previousPanel() {

    }

    @Override
    public void nextPanel() {
        this.onExitPanel.actionPerformed(new ActionEvent(this, 0, "exitPanel"));
//        ExportEntry currentEntry = this.exportationPanelParameters.getCurrentEntry();
//        if (currentEntry != null) {
//            ExportParameters exportParams = currentEntry.getParameters();
//            if (exportParams != null) {
//                this.processPanel.getProcess().setParameters(exportParams);
//            }
//        }
    }

    @Override
    public JComponent asJComponent() {
        return this;
    }

    public ExportServiceFactory getSelectedService() {
        ExportServiceFactory factory = (ExportServiceFactory) ListElement.getSelected(lstServices);
        return factory;
    }

    public void setSelectedService(ExportServiceFactory factory) {
        ListElement.setSelected(lstServices, factory);
    }

}
