package org.gvsig.expressionevaluator.spi;

import java.io.File;
import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.temporal.TemporalAccessor;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Locale;
import java.util.Objects;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.BooleanUtils;
import org.apache.commons.lang3.Range;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.math.util.MathUtils;
import org.gvsig.expressionevaluator.Code;
import org.gvsig.expressionevaluator.Codes;
import org.gvsig.expressionevaluator.Function;
import org.gvsig.expressionevaluator.I18N;
import org.gvsig.expressionevaluator.Interpreter;
import org.gvsig.fmap.geom.Geometry;
import org.gvsig.fmap.geom.primitive.Point;
import org.json.JSONArray;
import org.json.JSONObject;

@SuppressWarnings("UseSpecificCatch")
public abstract class AbstractFunction implements Function {
 
    private final String name;
    private String group;
    private Range argc;
    private String description;
    private String[] descriptionArgs;
    private List<String> alias;
    private String template;
    private String returnType;
    private boolean sqlCompatible;

    protected AbstractFunction(String group, String name, Range argc, String description, String template, String[] descriptionArgs, String returnType, boolean sqlCompatible) {
        this.name = name;
        this.group = group;
        this.argc = argc;
        this.description = description;
        this.template = template;
        this.descriptionArgs = descriptionArgs;
        this.returnType = returnType;
        this.sqlCompatible = sqlCompatible;
        load_from_resource();
    }
    protected AbstractFunction(String group, String name, Range argc, String description, String template, String[] descriptionArgs, String returnType) {
        this(group, name, argc, description, template, descriptionArgs, returnType, false);
    }
    
    protected AbstractFunction(String group, String name, Range argc, String description, String template, String[] descriptionArgs) {
        this(group, name, argc, description, template, null, null);
    }

    protected AbstractFunction(String group, String name, Range argc, String description, String template) {
        this(group, name, argc, description, template, null, null);
    }

    protected AbstractFunction(String group, String name, Range argc) {
        this(group, name, argc, null, null, null, null);
    }

    @Override
    public String name() {
        return this.name;
    }

    @Override
    public String returnType() {
        return this.returnType;
    }

    @Override
    public String group() {
        return this.group;
    }

    @Override
    public Range argc() {
        return argc;
    }

    @Override
    public String description() {
        return description;
    }

    @Override
    public String[] descriptionArgs() {
        return descriptionArgs;
    }

    @Override
    public void addAlias(String name) {
        if( StringUtils.isBlank(name) ) {
            return;
        }
        if( this.alias == null ) {
            this.alias = new ArrayList<>();
        }
        if( this.alias.contains(name) ) {
            return;
        }
        this.alias.add(name);
    }

    @Override
    public List<String> aliases() {
        return this.alias;
    }

    @Override
    public String template() {
        return this.template;
    }

    @Override
    public boolean isOperator() {
        return false;
    }

    @Override
    public boolean useArgumentsInsteadObjects() {
        return false;
    }

    @Override
    public boolean isSQLCompatible() {
        return sqlCompatible;
    }

    @Override
    public boolean allowConstantFolding() {
        return false;
    }
    
    @Override
    public Object call(Interpreter interpreter, Codes args) throws Exception {
        return null;
    }
    
    protected int getInt(Object args[], int n) {
        if( args.length < n  ) {
            throw new IllegalArgumentException(I18N.Required_argument_XargnX_and_only_found_XargcX_in_call_to_XIdentifierX(name(), args.length, n));
        }
        Object value = args[n];
        if( value == null ) {
            throw new IllegalArgumentException(I18N.Illegal_null_value_for_argument_XargnX_of_XIdentifierX_function(name(), n));
        }
        if( !(value instanceof Number) ) {
            String type = value.getClass().getCanonicalName();
            throw new IllegalArgumentException(
                    I18N.The_type_of_the_argument_XargnX_for_the_XIdentifierX_function_is_incorrect(name(), n) + " " +
                    I18N.Expected_XexpectedX_and_found_XfoundX("Number",type)
            );
        }
        return ((Number)value).intValue();
    }

    protected long getLong(Object args[], int n) {
        if( args.length < n  ) {
            throw new IllegalArgumentException(I18N.Required_argument_XargnX_and_only_found_XargcX_in_call_to_XIdentifierX(name(), args.length, n));
        }
        Object value = args[n];
        if( value == null ) {
            throw new IllegalArgumentException(I18N.Illegal_null_value_for_argument_XargnX_of_XIdentifierX_function(name(), n));
        }
        if( !(value instanceof Number) ) {
            String type = value.getClass().getCanonicalName();
            throw new IllegalArgumentException(
                    I18N.The_type_of_the_argument_XargnX_for_the_XIdentifierX_function_is_incorrect(name(), n) + " " +
                    I18N.Expected_XexpectedX_and_found_XfoundX("Number",type)
            );
        }
        return ((Number)value).longValue();
    }

    protected double getDouble(Object args[], int n) {
        if( args.length < n  ) {
            throw new IllegalArgumentException(I18N.Required_argument_XargnX_and_only_found_XargcX_in_call_to_XIdentifierX(name(), args.length, n));
        }
        Object value = args[n];
        if( value == null ) {
            throw new IllegalArgumentException(I18N.Illegal_null_value_for_argument_XargnX_of_XIdentifierX_function(name(), n));
        }
        if( !(value instanceof Number) ) {
            String type = value.getClass().getCanonicalName();
            throw new IllegalArgumentException(
                    I18N.The_type_of_the_argument_XargnX_for_the_XIdentifierX_function_is_incorrect(name(), n) + " " +
                    I18N.Expected_XexpectedX_and_found_XfoundX("Number",type)
            );
        }
        return ((Number)value).doubleValue();
    }
    
    protected String getStr(Object args[], int n) {
        if( args.length < n  ) {
            throw new IllegalArgumentException(I18N.Required_argument_XargnX_and_only_found_XargcX_in_call_to_XIdentifierX(name(), args.length, n));
        }
        return Objects.toString(args[n], "");
    }
    
    protected File getFile(Object args[], int n) {
        if( args.length < n  ) {
            throw new IllegalArgumentException(I18N.Required_argument_XargnX_and_only_found_XargcX_in_call_to_XIdentifierX(name(), args.length, n));
        }
        Object arg = args[n];
        if( arg == null ) {
            return null;
        }
        if( arg instanceof File ) {
            return (File)arg;
        }
        if( arg instanceof URL ) {
            try {
                return new File(((URL)arg).toURI());
            } catch (URISyntaxException ex) {
                return null;
            }
        }
        if( arg instanceof URI ) {
            return new File(((URI)arg));
        }
        String s = Objects.toString(arg, null);
        if( s == null ) {
            return null;
        }
        File f = new File(s);
        return f;
    }
    
    protected Object getObject(Object args[], int n) {
        if( args.length < n  ) {
            throw new IllegalArgumentException(I18N.Required_argument_XargnX_and_only_found_XargcX_in_call_to_XIdentifierX(name(), args.length, n));
        }
        return args[n];
    }
    
    protected Object getObject(Interpreter interpreter, Codes args, int n) {
        if( args.size() < n  ) {
            throw new IllegalArgumentException(I18N.Required_argument_XargnX_and_only_found_XargcX_in_call_to_XIdentifierX(name(), args.size(), n));
        }
        Code arg = args.get(n);
        Object value = interpreter.run(arg);
        return value;
    }
    
    protected Geometry getGeom(Object[] args, int n) {
        return this.getGeom(args, n, false);
    }
    
    protected Geometry getGeom(Object[] args, int n, boolean allowNull) {
        if( args.length < n  ) {
            throw new IllegalArgumentException(I18N.Required_argument_XargnX_and_only_found_XargcX_in_call_to_XIdentifierX(name(), args.length, n));
        }
        Object value = args[n];
        if( value == null ) {
            if( allowNull ) {
                return null;
            }
            throw new IllegalArgumentException(I18N.Illegal_null_value_for_argument_XargnX_of_XIdentifierX_function(name(), n));
        }
        if( !(value instanceof Geometry) ) {
            String type = value.getClass().getCanonicalName();
            throw new IllegalArgumentException(
                    I18N.The_type_of_the_argument_XargnX_for_the_XIdentifierX_function_is_incorrect(name(), n) + " " +
                    I18N.Expected_XexpectedX_and_found_XfoundX("Geometry",type)
            );
        }
        return (Geometry)value;
    }

    protected Point getPoint(Object[] args, int n) {
        if( args.length < n  ) {
            throw new IllegalArgumentException(I18N.Required_argument_XargnX_and_only_found_XargcX_in_call_to_XIdentifierX(name(), args.length, n));
        }
        Object value = args[n];
        if( value == null ) {
            return null;
        }
        if( !(value instanceof Point) ) {
            String type = value.getClass().getCanonicalName();
            throw new IllegalArgumentException(
                    I18N.The_type_of_the_argument_XargnX_for_the_XIdentifierX_function_is_incorrect(name(), n) + " " +
                    I18N.Expected_XexpectedX_and_found_XfoundX("Point",type)
            );
        }
        return (Point)value;
    }
    
    protected Date getDate(Object[] args, int n) {
        if( args.length < n  ) {
            throw new IllegalArgumentException(I18N.Required_argument_XargnX_and_only_found_XargcX_in_call_to_XIdentifierX(name(), args.length, n));
        }
        Object value = args[n];
        if( value == null ) {
            return null;
        }
        if( !(value instanceof Date) ) {
            String type = value.getClass().getCanonicalName();
            throw new IllegalArgumentException(
                    I18N.The_type_of_the_argument_XargnX_for_the_XIdentifierX_function_is_incorrect(name(), n) + " " +
                    I18N.Expected_XexpectedX_and_found_XfoundX("Date",type)
            );
        }
        return (Date)value;
    }
    
    protected LocalDateTime getLocalDateTime(Object[] args, int n) {
        if( args.length < n  ) {
            throw new IllegalArgumentException(I18N.Required_argument_XargnX_and_only_found_XargcX_in_call_to_XIdentifierX(name(), args.length, n));
        }
        Object value = args[n];
        if( value == null ) {
            return null;
        }
        if( value instanceof Date ) {
            Date date = ((Date)value);
            return date.toInstant().atZone(ZoneId.systemDefault()).toLocalDateTime();
        }
        if( value instanceof LocalDateTime ) {
            return (LocalDateTime) value;
        }
        if( value instanceof TemporalAccessor ) {
            return LocalDateTime.from(((TemporalAccessor)value));
        }
        String type = value.getClass().getCanonicalName();
        throw new IllegalArgumentException(
                I18N.The_type_of_the_argument_XargnX_for_the_XIdentifierX_function_is_incorrect(name(), n) + " " +
                I18N.Expected_XexpectedX_and_found_XfoundX("Temporal/Date",type)
        );
    }
    
    protected boolean getBoolean(Object args[], int n, Double accuracy) {
        if( args.length < n  ) {
            throw new IllegalArgumentException(I18N.Required_argument_XargnX_and_only_found_XargcX_in_call_to_XIdentifierX(name(), args.length, n));
        }
        Object value = args[n];
        return toBoolean(value, accuracy);
    }

    protected boolean getBoolean(Interpreter interpreter, Codes args, int n) {
        Object value = getObject(interpreter, args, n);
        return toBoolean(value, interpreter.getAccuracy());
    }

    protected boolean toBoolean(Object value, Double accuracy) {
        if( value == null ) {
            return false;
        }
        if( value instanceof Boolean ) {
            return (Boolean)value;
        }        
        if( value instanceof Number ) {
            return MathUtils.compareTo(
                ((Number) value).doubleValue(), 
                0,
                accuracy==null? MathUtils.EPSILON:accuracy
            ) == 0;
        }
        return BooleanUtils.toBoolean(value.toString());
    } 

    private void load_from_resource() {
        String lang = Locale.getDefault().getLanguage();
        URL url = this.getClass().getResource("/org/gvsig/expressionevaluator/functions/"+lang+"/"+this.name()+".json");
        if( url == null ) {
            url = this.getClass().getResource("/org/gvsig/expressionevaluator/functions/en/"+this.name()+".json");
            if( url == null ) {
                return;
            }
        }
        InputStream is = null;
        JSONObject json;
        try {
            is = url.openStream();
            List<String> lines = IOUtils.readLines(is);
            json = new JSONObject(StringUtils.join(lines,  "\n"));
        } catch (Exception ex) {
            return;
        } finally {
            IOUtils.closeQuietly(is);
        }
        
        if( json.has("group") ) {
            this.group = json.getString("group");
        }
        if( json.has("description") ) {
            Object x = json.get("description");
            if( x instanceof String ) {
                this.description = (String) x;
            } else if( x instanceof JSONArray ) {
                StringBuilder builder = new StringBuilder();
                for (int i = 0; i < ((JSONArray)x).length(); i++) {
                    if( i>0 ) {
                        builder.append(" ");
                    }
                    builder.append(((JSONArray)x).getString(i));
                }
                this.description = builder.toString();
            } else {
                this.description = x.toString();
            }
            this.description = StringUtils.replace(
                    this.description, 
                    "@@@", 
                    url.toString()
            );
        }
        if( json.has("template") ) {
            this.template = json.getString("template");
        }
        if( json.has("returnType") ) {
            this.returnType = json.getString("returnType");
        }
        if( json.has("sqlCompatible") ) {
            this.sqlCompatible = json.getBoolean("sqlCompatible");
        }
        if( json.has("args") ) {
            JSONArray x = json.getJSONArray("args");
            String[] args = new String[x.length()];
            for (int i = 0; i < x.length(); i++) {
                args[i] = x.getString(i);
            }
            this.descriptionArgs = args;
        }
        if( json.has("alias") ) {
            JSONArray x = json.getJSONArray("alias");
            for (int i = 0; i < x.length(); i++) {
                this.addAlias(x.getString(i));
            }
        }
    }
}
