package org.gvsig.expressionevaluator.impl.grammars;

import org.gvsig.expressionevaluator.Code;
import org.gvsig.expressionevaluator.Code.Caller;
import static org.gvsig.expressionevaluator.ExpressionBuilder.FUNCTION_CAST;
import static org.gvsig.expressionevaluator.ExpressionBuilder.FUNCTION_LIST;
import static org.gvsig.expressionevaluator.ExpressionBuilder.OPERATOR_REGEXP;
import org.gvsig.expressionevaluator.ExpressionEvaluatorLocator;
import org.gvsig.expressionevaluator.ExpressionEvaluatorManager;
import org.gvsig.expressionevaluator.Grammar;
import org.gvsig.expressionevaluator.Statement;
import org.gvsig.expressionevaluator.Statement.StatementContext;
import org.gvsig.expressionevaluator.impl.DefaultCodeBuilder.BaseCodes;
import org.gvsig.expressionevaluator.impl.DefaultStatement.StatementBuilderBase;
import org.gvsig.expressionevaluator.impl.DefaultStatement.ArgsBuilderFromNames;
import org.gvsig.expressionevaluator.impl.function.programming.IIFFunction;
import org.gvsig.expressionevaluator.impl.function.programming.LetFunction;
import org.gvsig.expressionevaluator.impl.function.programming.CaseFunction;
import org.gvsig.expressionevaluator.impl.function.programming.CodeBlockFunction;
import org.gvsig.expressionevaluator.impl.function.programming.CodeBlockWithExceptFunction;
import org.gvsig.expressionevaluator.impl.function.programming.CreateFnFunction;
import org.gvsig.expressionevaluator.impl.function.programming.ForEachFunction;
import org.gvsig.expressionevaluator.impl.function.programming.ReturnFunction;
import org.gvsig.expressionevaluator.impl.function.programming.WhileFunction;
import org.gvsig.expressionevaluator.spi.AbstractGrammarFactory;

/**
 *
 * @author jjdelcerro
 */
public class BasicGrammarFactory extends AbstractGrammarFactory {

    private Grammar grammar;

    private class BeginStatementBuilder extends StatementBuilderBase {

        private final String blockID;
        private final String exceptID;

        public BeginStatementBuilder(String blockID, String exceptID) {
            super(CodeBlockFunction.NAME, new ArgsBuilderFromNames(blockID));
            this.blockID = blockID;
            this.exceptID = exceptID;
        }

        @Override
        public Code build(StatementContext context) {
            Caller code = (Caller) super.build(context);
            Code exceptCode = context.getCode(this.exceptID);
            if( exceptCode==null ) {
                return code;
            }
            BaseCodes args = (BaseCodes) code.parameters();
            args.add(exceptCode);
            code = context.getCodeBuilder().function(CodeBlockWithExceptFunction.NAME, args);
            return code;
        }
        
    }
    public BasicGrammarFactory() {
        super("Basic", true);
    }
    
    @Override
    public Grammar create(Object... parameters) {
        ExpressionEvaluatorManager manager = ExpressionEvaluatorLocator.getManager();
        if( this.grammar==null ) {
            Grammar theGrammar = manager.createGrammar(this.getName());
            Statement stmt;
        
            theGrammar.addReservedWord("BEGIN");
            theGrammar.addReservedWord("EXCEPT");
            theGrammar.addReservedWord("END");
            theGrammar.addReservedWord("IF");
            theGrammar.addReservedWord("THEN");
            theGrammar.addReservedWord("ELSE");
            theGrammar.addReservedWord("WHILE");
            theGrammar.addReservedWord("CASE");
            theGrammar.addReservedWord("WHEN");
            theGrammar.addReservedWord("FOR");
            theGrammar.addReservedWord("CREATE");
            theGrammar.addReservedWord("FUNCTION");
            theGrammar.addReservedWord("PROCEDURE");
            theGrammar.addReservedWord("AS");
            theGrammar.addReservedWord("MATCH");
            theGrammar.addReservedWord("CAST");
            theGrammar.addReservedWord("SET");
            theGrammar.addReservedWord("TO");
            theGrammar.addReservedWord("$$"); // ?? Asi van las funciones
            
            stmt = theGrammar.createStatement("BLOCK");
            stmt.addRule(stmt.require_any_token("BEGIN"));
            stmt.addRule(stmt.require_expressions("VALUES", ";"));
            stmt.addRule(stmt.optional_any_token("EXCEPT")
                    .addRuleOnTrue(stmt.require_expressions("EXCEPT", ";"))
            );
            stmt.addRule(stmt.require_any_token("END"));
            stmt.builder(new BeginStatementBuilder("VALUES","EXCEPT"));
            theGrammar.addStatement(stmt);

            stmt = theGrammar.createStatement("IF");
            stmt.addRule(stmt.require_any_token("IF"));
            stmt.addRule(stmt.require_expression("CONDITION"));
            stmt.addRule(stmt.require_any_token("THEN", ":"));
            stmt.addRule(stmt.require_expressions("ONTRUE", ";"));
            stmt.addRule(stmt.optional_any_token("ELSE")
                    .addRuleOnTrue(stmt.optional_any_token(":"))
                    .addRuleOnTrue(stmt.require_expressions("ONFALSE", ";"))
            );
            stmt.addRule(stmt.require_any_token("END"));
            stmt.addRule(stmt.optional_any_token("IF"));
            stmt.code(
                    IIFFunction.NAME,
                    stmt.args_names("CONDITION","ONTRUE","ONFALSE")
            ); 
            theGrammar.addStatement(stmt);

            stmt = theGrammar.createStatement("WHILE");
            stmt.addRule(stmt.require_any_token("WHILE"));
            stmt.addRule(stmt.require_expression("CONDITION"));
            stmt.addRule(stmt.require_any_token("BEGIN", ":"));
            stmt.addRule(stmt.require_expressions("BODY", ";"));
            stmt.addRule(stmt.require_any_token("END"));
            stmt.addRule(stmt.optional_any_token("WHILE"));
            stmt.code(
                    WhileFunction.NAME,
                    stmt.args_names("CONDITION","BODY")
            ); 
            theGrammar.addStatement(stmt);

//            stmt = theGrammar.createStatement("FOR");
//            stmt.addRule(stmt.require_token("FOR"));
//            stmt.addRule(stmt.require_expression("INIT"));
//            stmt.addRule(stmt.require_token(";"));
//            stmt.addRule(stmt.require_expression("CONDITION"));
//            stmt.addRule(stmt.require_token(";"));
//            stmt.addRule(stmt.require_expression("STEP"));
//            stmt.addRule(stmt.optional_token(":"));
//            stmt.addRule(stmt.require_expression("BODY"));
//            stmt.addRule(stmt.optional_token("END"));
//            stmt.code(ForFunction.NAME,"INIT","CONDITION","STEP","BODY"); 
//            theGrammar.addStatement(stmt);

            stmt = theGrammar.createStatement("FOREACH");
            stmt.addRule(stmt.require_any_token("FOR"));
            stmt.addRule(stmt.require_identifier("VAR"));
            stmt.addRule(stmt.require_any_token("IN"));
            stmt.addRule(stmt.require_expression("ITERABLE"));
            stmt.addRule(stmt.require_any_token("BEGIN", ":"));
            stmt.addRule(stmt.require_expressions("BODY", ";"));
            stmt.addRule(stmt.require_any_token("END"));
            stmt.addRule(stmt.optional_any_token("FOR"));
            stmt.code(
                    ForEachFunction.NAME,
                    stmt.args_names("VAR","ITERABLE","BODY")
            ); 
            theGrammar.addStatement(stmt);

            stmt = theGrammar.createStatement("CASE");
            stmt.addRule(stmt.require_any_token("CASE"));
            stmt.addRule(
                stmt.repeat_until_any_tokens("END","ELSE")
                    .addRule(stmt.require_any_token("WHEN"))
                    .addRule(stmt.require_expressions("CONDITION#",";"))
                    .addRule(stmt.require_any_token("THEN",":"))
                    .addRule(stmt.require_expressions("ONTRUE#", ";"))
            );
            stmt.addRule(stmt.optional_any_token("ELSE")
                    .addRuleOnTrue(stmt.optional_any_token(":"))
                    .addRuleOnTrue(stmt.require_expressions("ELSE",";"))
            );
            stmt.addRule(stmt.require_any_token("END"));
            stmt.addRule(stmt.optional_any_token("CASE"));
            stmt.code(
                    CaseFunction.NAME,
                    stmt.args_names("CONDITION#","ONTRUE#","ELSE")
            ); 
            theGrammar.addStatement(stmt);
            
            stmt = theGrammar.createStatement("USERFUNCTION");
            stmt.addRule(stmt.require_any_token("CREATE"));
            stmt.addRule(stmt.optional_any_token("FUNCTION", "PROCEDURE"));
            stmt.addRule(stmt.require_identifier("FUNCTION_NAME"));
            stmt.addRule(stmt.optional_any_token("("));
            stmt.addRule(stmt.optional_identifiers("PARAM_NAMES", ","));
            stmt.addRule(stmt.optional_any_token(")"));
            stmt.addRule(stmt.require_any_token("AS"));
            stmt.addRule(stmt.optional_any_token("$$","BEGIN")
                    .addRuleOnTrue(stmt.require_expressions("BODY", ";"))
                    .addRuleOnTrue(stmt.require_any_token("$$", "END"))
                    .addRuleOnTrue(stmt.optional_any_token("FUNCTION", "PROCEDURE"))
                    .addRuleOnFalse(stmt.optional_literal_string("SCRIPT_PATH")
                        .addRuleOnTrue(stmt.require_any_token(","))
                        .addRuleOnTrue(stmt.require_identifier("SCRIPT_FUNCTION"))
                        .addRuleOnTrue(stmt.optional_any_token("LANGUAGE")
                                .addRuleOnTrue(stmt.require_literal_string("LANGUAGE"))
                        )
                    )
            );
                    
//            stmt.addRule(stmt.require_expressions("BODY", ";"));
//            stmt.addRule(stmt.require_any_token("$$", "END"));
//            stmt.addRule(stmt.optional_any_token("FUNCTION", "PROCEDURE"));
            stmt.code(
                    CreateFnFunction.NAME,
                    stmt.args_names("FUNCTION_NAME","PARAM_NAMES","BODY", "SCRIPT_PATH", "SCRIPT_FUNCTION", "LANGUAGE")
            ); 
            theGrammar.addStatement(stmt);

            stmt = theGrammar.createStatement("RETURN");
            stmt.addRule(stmt.require_any_token("RETURN"));
            stmt.addRule(stmt.optional_any_token(";")
                    .addRuleOnTrue(stmt.set_expression("VALUES", null))
                    .addRuleOnFalse(stmt.require_expression("VALUES"))
                    .addRuleOnFalse(stmt.require_any_token(";"))
            );
            stmt.code(
                    ReturnFunction.NAME,
                    stmt.args_expand("VALUES")
            ); 
            theGrammar.addStatement(stmt);

//            stmt = theGrammar.createStatement("TRY");
//            stmt.addRule(stmt.require_any_token("TRY"));
//            stmt.addRule(stmt.optional_any_token(":"));
//            stmt.addRule(stmt.require_expressions("VALUE", ";"));
//            stmt.addRule(stmt.optional_any_token(";"));
//            stmt.addRule(stmt.require_any_token("EXCEPT","CATH"));
//            stmt.addRule(stmt.optional_any_token(":"));
//            stmt.addRule(stmt.require_expressions("ONERROR", ";"));
//            stmt.addRule(stmt.optional_any_token(";"));
//            stmt.addRule(stmt.require_any_token("END"));
//            stmt.addRule(stmt.optional_any_token("TRY"));
//            stmt.code(
//                    TryFunction.NAME,
//                    stmt.args_names("VALUE","ONERROR")
//            ); 
//            theGrammar.addStatement(stmt);

            stmt = theGrammar.createStatement("LET");
            stmt.addRule(stmt.require_any_token("LET", "SET"));
            stmt.addRule(stmt.require_identifier("NAME"));
            stmt.addRule(stmt.optional_any_token("=","TO"));
            stmt.addRule(stmt.require_expression("VALUE"));
            stmt.code(
                    LetFunction.NAME, 
                    stmt.args_names("NAME", "VALUE")
            ); 
            theGrammar.addStatement(stmt);

            stmt = theGrammar.createStatement("ARRAY");
            stmt.addRule(stmt.require_any_token("ARRAY"));
            stmt.addRule(stmt.require_any_token("["));
            stmt.addRule(stmt.require_expressions("VALUES",","));
            stmt.addRule(stmt.require_any_token("]"));
            stmt.code(
                    FUNCTION_LIST, 
                    stmt.args_expand("VALUES")
            ); 
            theGrammar.addStatement(stmt);

            stmt = theGrammar.createStatement("CAST");
            stmt.addRule(stmt.require_any_token("CAST"));
            stmt.addRule(stmt.optional_any_token("(")
                .addRuleOnTrue(stmt.require_expression("VALUE"))
                .addRuleOnTrue(stmt.require_any_token("AS"))
                .addRuleOnTrue(stmt.require_identifier("TYPE"))
                .addRuleOnTrue(stmt.require_any_token(")"))                    
                .addRuleOnFalse(stmt.require_expression("VALUE"))
                .addRuleOnFalse(stmt.require_any_token("AS"))
                .addRuleOnFalse(stmt.require_identifier("TYPE"))
            );
            stmt.code(
                    FUNCTION_CAST,
                    stmt.args_names("VALUE","TYPE")
            );
            theGrammar.addStatement(stmt);
            
            stmt = theGrammar.createStatement("MATCH");
            stmt.addRule(stmt.require_any_token("MATCH"));
            stmt.addRule(stmt.optional_any_token("(")
                .addRuleOnTrue(stmt.require_expression("VALUE"))
                .addRuleOnTrue(stmt.require_any_token(","))
                .addRuleOnTrue(stmt.require_expression("PATTERN"))
                .addRuleOnTrue(stmt.require_any_token(")"))                    
                .addRuleOnFalse(stmt.require_expression("VALUE"))
                .addRuleOnFalse(stmt.optional_any_token(","))
                .addRuleOnFalse(stmt.require_expression("PATTERN"))
            );
            stmt.code(
                    OPERATOR_REGEXP,
                    stmt.args_names("VALUE","PATTERN")
            );
            theGrammar.addStatement(stmt);
            
//            stmt = theGrammar.createStatement("FETCHFIRST");
//            stmt.addRule(stmt.require_token("FETCH"));
//            stmt.addRule(stmt.require_token("FIRST"));
//            stmt.addRule(stmt.require_expression("VALUE"));
//            stmt.addRule(stmt.optional_token("FROM"));
//            stmt.addRule(stmt.require_expression("FROM"));
//            stmt.addRule(stmt.optional_token("WHERE").addRuleOnTrue(stmt.require_expression("WHERE")));
//            stmt.addRule(stmt.optional_token("ORDER","BY").addRuleOnTrue(stmt.require_expression("ORDER")));
//            stmt.code("FETCH_FIRST","VALUE","FROM","WHERE");
//            theGrammar.addStatement(stmt);

            this.grammar = theGrammar;
        }
        return grammar;
    }
    
}
