package org.gvsig.expresionevaluator.impl;

import java.text.MessageFormat;
import static junit.framework.Assert.assertEquals;
import junit.framework.TestCase;
import org.apache.commons.lang3.StringUtils;
import org.cresques.cts.IProjection;
import org.gvsig.expressionevaluator.Code;
import org.gvsig.expressionevaluator.Code.Caller;
import org.gvsig.expressionevaluator.Code.Constant;
import org.gvsig.expressionevaluator.Codes;
import org.gvsig.expressionevaluator.ExpressionBuilder;
import org.gvsig.expressionevaluator.ExpressionEvaluatorLocator;
import org.gvsig.expressionevaluator.ExpressionEvaluatorManager;
import org.gvsig.expressionevaluator.ExpressionUtils;
import org.gvsig.expressionevaluator.Formatter;
import org.gvsig.expressionevaluator.LexicalAnalyzer;
import org.gvsig.fmap.crs.CRSFactory;
import org.gvsig.fmap.geom.Geometry;
import org.gvsig.fmap.geom.GeometryUtils;
import org.gvsig.fmap.geom.exception.CreateGeometryException;
import org.gvsig.fmap.geom.primitive.Point;
import org.gvsig.tools.library.impl.DefaultLibrariesInitializer;

public class TestCodeFormatter extends TestCase {

    private static class MyFormatter implements Formatter<Code> {
    
        private class Formatter_constant_bytearray implements Formatter<Code> {

            @Override
            public boolean canApply(Code code) {
                if( code instanceof Constant ) {
                    return ((Constant)code).value() instanceof byte[];
                }
                return false;
            }

            @Override
            public String format(Code code) {
                return builder.bytearray_x((byte[]) ((Constant)code).value());
            }
        }

        private class Formatter_constant_geometry implements Formatter<Code> {

            @Override
            public boolean canApply(Code code) {
                if( code instanceof Constant ) {
                    return ((Constant)code).value() instanceof Geometry;
                }
                return false;
            }

            @Override
            public String format(Code code) {
                Geometry geometry = (Geometry) ((Constant)code).value();
                return MessageFormat.format(
                        "ST_GeomFromWKB(({0}), ({1}))",
                        builder.bytearray_x(GeometryUtils.toWKB(geometry)),
                        String.valueOf(builder.srs_id(geometry.getProjection()))
                );
                
            }
        }

        
        private class Formatter_ST_intersects_H2Spatial implements Formatter<Code> {
            // Need for H2Spatial
            @Override
            public boolean canApply(Code code) {
                if( code instanceof Caller ) {
                    return StringUtils.equalsIgnoreCase("ST_intersects",((Caller)code).name());
                }
                return false;
            }

            @Override
            public String format(Code code) {
                Codes parameters = ((Caller)code).parameters();
                String p1 = parameters.get(0).toString(MyFormatter.this);
                String p2 = parameters.get(1).toString(MyFormatter.this);
                String r = MessageFormat.format(
                        "( (({0}) && ({1})) AND ST_Intersects(({0}),({1}) ))", 
                        p1,
                        p2
                );
                return r;
            }
        }
  
        private class Formatter_ST_intersects_SpatiaLite implements Formatter<Code> {
            // Need for SpatiaLite

            private final String table;
            private final String geomcolumn;

            public Formatter_ST_intersects_SpatiaLite(String table, String geomcolumn) {
                this.table = table;
                this.geomcolumn = geomcolumn;
            }

            @Override
            public boolean canApply(Code code) {
                if( code instanceof Caller ) {
                    return StringUtils.equalsIgnoreCase("ST_intersects2",((Caller)code).name());
                }
                return false;
            }

            @Override
            public String format(Code code) {
                Codes parameters = ((Caller)code).parameters();
                String p1 = parameters.get(0).toString(MyFormatter.this);
                String p2 = parameters.get(1).toString(MyFormatter.this);
                String r = MessageFormat.format(
                    "(ST_Intersects({0},{1}) AND ROWID IN ( SELECT ROWID FROM SpatialIndex WHERE f_table_name = ''{2}'' AND f_geometry_column = ''{3}'' AND search_frame = \"{2}\".\"{3}\"))",
                    p1,
                    p2,
                    table,
                    geomcolumn
                );
                return r;
            }
        }
  
        private class Formatter_DECODE implements Formatter<Code> {
            // Need for SpatiaLite 
            @Override
            public boolean canApply(Code code) {
                if( code instanceof Caller ) {
                    if( StringUtils.equalsIgnoreCase("DECODE",((Caller)code).name()) ) {
                        Codes parameters = ((Caller)code).parameters();
                        Code code_p1 = parameters.get(0);
                        Code code_p2 = parameters.get(1);
                        if( code_p1 instanceof Constant && code_p2 instanceof Constant ) {
                            Object p1 = ((Constant)code_p1).value();
                            Object p2 = ((Constant)code_p2).value();
                            if( p1 instanceof String && p1 instanceof String && 
                                StringUtils.equalsIgnoreCase((CharSequence) p2,"hex") ) {
                                return true;
                            }
                        }
                    }
                }
                return false;
            }

            @Override
            public String format(Code code) {
                Codes parameters = ((Caller)code).parameters();
                Object p1 = ((Constant)parameters.get(0)).value();
                return "x'"+ (String)p1 + "'";
            }
        }
  
        private class Formatter_IFNULL implements Formatter<Code> {
            // Need for H2Spatial
            @Override
            public boolean canApply(Code code) {
                if( code instanceof Caller ) {
                    return StringUtils.equalsIgnoreCase("IFNULL",((Caller)code).name());
                }
                return false;
            }

            @Override
            public String format(Code code) {
                Codes parameters = ((Caller)code).parameters();
                String p1 = parameters.get(0).toString(MyFormatter.this);
                String p2 = parameters.get(1).toString(MyFormatter.this);
                String p3 = parameters.get(2).toString(MyFormatter.this);
                String r = MessageFormat.format(
                        "NVL2({0}, {1}, {2})", 
                        p1,
                        p3,
                        p2
                );
                return r;
            }
        }
    
        private class Formatter_ILIKE implements Formatter<Code> {
            // Need for SpatiaLite
            @Override
            public boolean canApply(Code code) {
                if( code instanceof Caller ) {
                    return StringUtils.equalsIgnoreCase("ILIKE",((Caller)code).name());
                }
                return false;
            }

            @Override
            public String format(Code code) {
                Codes parameters = ((Caller)code).parameters();
                String p1 = parameters.get(0).toString(MyFormatter.this);
                String p2 = parameters.get(1).toString(MyFormatter.this);
                String r = MessageFormat.format(
                        "LOWER({0}) LIKE LOWER({1})", 
                        p1,
                        p2
                );
                return r;
            }
        }
    
        private class Formatter_NOT_IS_NULL implements Formatter<Code> {
            // Need for H2Spatial
            @Override
            public boolean canApply(Code code) {
                if( code instanceof Caller ) {
                    return StringUtils.equalsIgnoreCase("NOT_IS_NULL",((Caller)code).name());
                }
                return false;
            }

            @Override
            public String format(Code code) {
                Codes parameters = ((Caller)code).parameters();
                String p1 = parameters.get(0).toString(MyFormatter.this);
                String r = MessageFormat.format(
                        "( ({0}) IS NOT NULL )", 
                        p1
                );
                return r;
            }
        }

        private final Formatter<Code>[] formatters;
        private final ExpressionBuilder builder;

        public MyFormatter() {
            this.builder = ExpressionUtils.createExpressionBuilder();
            this.formatters = new Formatter[] {
//                new Formatter_constant_bytearray(),
//                new Formatter_constant_geometry(),
                new Formatter_IFNULL(),
                new Formatter_NOT_IS_NULL(),
                new Formatter_ST_intersects_H2Spatial(),
                new Formatter_ST_intersects_SpatiaLite("mytable", "the_geom"),
                new Formatter_ILIKE(),
                new Formatter_DECODE(),
            };
        }

        @Override
        public boolean canApply(Code code) {
            for (Formatter<Code> formatter : formatters) {
                if( formatter.canApply(code) ) {
                    return true;
                }
            }
            return false;
        }

        @Override
        public String format(Code code) {
            for (Formatter<Code> formatter : formatters) {
                if( formatter.canApply(code) ) {
                    return formatter.format(code);
                }
            }
            return code.toString(this);
        }
        
    }
    
    public TestCodeFormatter(String testName) {
        super(testName);
    }

    @Override
    protected void setUp() throws Exception {
        super.setUp();
        new DefaultLibrariesInitializer().fullInitialize();
    }

    @Override
    protected void tearDown() throws Exception {
        super.tearDown();
    }

    // TODO add test methods here. The name must begin with 'test'. For example:
    // public void testHello() {}
    
    protected LexicalAnalyzer createLexicalAnalyzer() {
        ExpressionEvaluatorManager manager = ExpressionEvaluatorLocator.getManager();
        LexicalAnalyzer lexer = manager.createLexicalAnalyzer();
        return lexer;
    }

    protected org.gvsig.expressionevaluator.Compiler createCompiler() {
        ExpressionEvaluatorManager manager = ExpressionEvaluatorLocator.getManager();
        org.gvsig.expressionevaluator.Compiler compiler = manager.createCompiler();
        compiler.setLexicalAnalyzer(createLexicalAnalyzer());
        return compiler;
    }
    
    private Formatter<Code> formatter() {
        return new MyFormatter();
    }

    public Code compileExpression(String source) {
        org.gvsig.expressionevaluator.Compiler compiler = createCompiler();
        Code code = compiler.compileExpression(source);
        return code;
    }
    
    public void testIdentifier1() {
        String source = "precio";

        Code code = compileExpression(source);
        assertEquals("\"precio\"", code.toString(formatter()));
    }

    public void testIdentifier2() {
        org.gvsig.expressionevaluator.Compiler compiler = createCompiler();
        compiler.getLexicalAnalyzer().setUseBracketsForIdentifiers(true);

        String source = "[precio]";

        Code code = compiler.compileExpression(source);
        assertEquals("\"precio\"", code.toString(formatter()));
    }

    public void testIdentifier3() {
        String source = "\"precio\"";

        Code code = compileExpression(source);
        assertEquals("\"precio\"", code.toString(formatter()));
    }


    public void testTrue() {
        String source = "true";

        Code code = compileExpression(source);
        assertEquals("TRUE", code.toString(formatter()));
    }

    public void testFalse() {
        String source = "false";

        Code code = compileExpression(source);
        assertEquals("FALSE", code.toString(formatter()));
    }

    public void testNull() {
        String source = "null";

        Code code = compileExpression(source);
        assertEquals("NULL", code.toString(formatter()));
    }

    public void testNotTrue() {
        String source = "not true";

        Code code = compileExpression(source);
        assertEquals("NOT(TRUE)", code.toString(formatter()));
    }

    public void testInteger() {
        String source = "23";

        Code code = compileExpression(source);
        assertEquals("23", code.toString(formatter()));
    }

    public void operator(String operatorName) {
        String source = "precio " + operatorName + " 23";

        Code code = compileExpression(source);
        assertEquals("(\"precio\" "+ operatorName + " 23)", code.toString(formatter()));
    }
    
    public void testOperators() {

        operator("=");
        operator("<>");
        operator(">");
        operator(">=");
        operator("<");
        operator("<=");
        operator("LIKE");
//        operator("ILIKE");
        operator("||");
        operator("+");
        operator("-");
        operator("*");
        operator("OR");
        operator("AND");
        operator("%");
        operator("IS");

        operator("~");
    }

    public void testILike() {
        String source = "precio ILike 23";

        Code code = compileExpression(source);
        assertEquals("LOWER(\"precio\") LIKE LOWER(23)", code.toString(formatter()));
    }
    
    public void testAddMul() {
        String source = "precio + 10 * 2 + 20 + 30";

        Code code = compileExpression(source);
        assertEquals("(((\"precio\" + (10 * 2)) + 20) + 30)", code.toString(formatter()));
    }
    
    public void testAddMulPar() {
        String source = "(precio + 10) * 2 + 20 + 30";

        Code code = compileExpression(source);
        assertEquals("((((\"precio\" + 10) * 2) + 20) + 30)", code.toString(formatter()));
    }
    
    public void testAbs() {
        String source = "precio + abs(10)";

        Code code = compileExpression(source);
        assertEquals("(\"precio\" + abs(10))", code.toString(formatter()));
    }
    
    public void testAbs2() {
        String source = "precio + abs(-10)";

        Code code = compileExpression(source);
        assertEquals("(\"precio\" + abs(-10))", code.toString(formatter()));
    }
    
    public void testPI() {
        String source = "precio + PI()";

        Code code = compileExpression(source);
        assertEquals("(\"precio\" + PI())", code.toString(formatter()));
    }
    
    public void testCeil() {
        String source = "precio + CEIL(PI())";

        Code code = compileExpression(source);
        assertEquals("(\"precio\" + CEIL(PI()))", code.toString(formatter()));
    }
    
    public void testGetitem1() {
        String source = "LIST('uno','dos','tres')[1]" ;

        Code code = compileExpression(source);
        assertEquals("GETITEM(LIST('uno', 'dos', 'tres'), 1)", code.toString(formatter()));
    }
    
    public void testConcat() {
        String source = "CONCAT(precio,' euros')";

        Code code = compileExpression(source);
        assertEquals("CONCAT(\"precio\", ' euros')", code.toString(formatter()));
    }
    
    public void test1() {
        String source = "NOMBRE03 = 'Torre d''En Besora (la)'";

        Code code = compileExpression(source);
        assertEquals("(\"NOMBRE03\" = 'Torre d''En Besora (la)')", code.toString(formatter()));
    }
    
    public void test2() {
        org.gvsig.expressionevaluator.Compiler compiler = createCompiler();
        compiler.getLexicalAnalyzer().setUseBracketsForIdentifiers(true);

        String source = "[1990] = 0.168873933773767";

        Code code = compiler.compileExpression(source);
        assertEquals("(\"1990\" = 0.168873933773767)", code.toString(formatter()));
    }

    public void test2fields() {
        org.gvsig.expressionevaluator.Compiler compiler = createCompiler();
        compiler.getLexicalAnalyzer().setUseBracketsForIdentifiers(true);

        String source = "[1990] = [precio]";

        Code code = compiler.compileExpression(source);
        assertEquals("(\"1990\" = \"precio\")", code.toString(formatter()));
    }
    
    
    public void testInvokeFunction1() {
        String source = "ST_Area(GEOMETRY)";

        Code code = compileExpression(source);
        assertEquals("ST_Area(\"GEOMETRY\")", code.toString(formatter()));
    }

    public void test3() throws CreateGeometryException {
        ExpressionBuilder builder = ExpressionUtils.createExpressionBuilder();

        IProjection proj = CRSFactory.getCRS("EPSG:4326");
        Point point = GeometryUtils.createPoint(10, 20);

        builder.set(        
            builder.ST_Intersects(
              builder.geometry(point,proj),
              builder.variable("the_geom")
            )
        );
        Code code = compileExpression(builder.toString());
        assertEquals(
                "ST_Intersects(ST_GeomFromWKB(DECODE('000000000140240000000000004034000000000000', 'hex'), 4326), \"the_geom\")",
                code.toString()
        );
        assertEquals(
                "( ((ST_GeomFromWKB(x'000000000140240000000000004034000000000000', 4326)) && (\"the_geom\")) AND ST_Intersects((ST_GeomFromWKB(x'000000000140240000000000004034000000000000', 4326)),(\"the_geom\") ))",
                code.toString(formatter())
        );
    }
    
    public void test4() throws CreateGeometryException {
        ExpressionBuilder builder = ExpressionUtils.createExpressionBuilder();

        IProjection proj = CRSFactory.getCRS("EPSG:4326");
        Point point = GeometryUtils.createPoint(10, 20);

        builder.set(        
            builder.function("ST_Intersects2",
                builder.geometry(point,proj),
                builder.variable("the_geom")
            )
        );
        Code code = compileExpression(builder.toString());
        assertEquals(
                "ST_Intersects2(ST_GeomFromWKB(DECODE('000000000140240000000000004034000000000000', 'hex'), 4326), \"the_geom\")",
                code.toString()
        );
        assertEquals(
                "(ST_Intersects(ST_GeomFromWKB(x'000000000140240000000000004034000000000000', 4326),\"the_geom\") AND ROWID IN ( SELECT ROWID FROM SpatialIndex WHERE f_table_name = 'mytable' AND f_geometry_column = 'the_geom' AND search_frame = \"mytable\".\"the_geom\"))",
                code.toString(formatter())
        );
    }
    
}
