package org.gvsig.expresionevaluator.impl;

import junit.framework.TestCase;
import org.gvsig.expressionevaluator.LexicalAnalyzer;
import org.gvsig.expressionevaluator.Compiler;
import org.gvsig.expressionevaluator.Code;
import org.gvsig.expressionevaluator.ExpressionEvaluatorLocator;
import org.gvsig.expressionevaluator.ExpressionEvaluatorManager;
import org.gvsig.expressionevaluator.ExpressionSyntaxException;
import org.gvsig.tools.library.impl.DefaultLibrariesInitializer;


public class TestCompiler extends TestCase {
    
    public TestCompiler(String testName) {
        super(testName);
    }
    
    @Override
    protected void setUp() throws Exception {
        super.setUp();
        new DefaultLibrariesInitializer().fullInitialize();
    }
    
    @Override
    protected void tearDown() throws Exception {
        super.tearDown();
    }

    // TODO add test methods here. The name must begin with 'test'. For example:
    // public void testHello() {}

    
    protected LexicalAnalyzer createLexicalAnalyzer() {
        ExpressionEvaluatorManager manager = ExpressionEvaluatorLocator.getManager();
        LexicalAnalyzer lexer = manager.createLexicalAnalyzer();
        return lexer;
    }

    protected org.gvsig.expressionevaluator.Compiler createCompiler() {
        ExpressionEvaluatorManager manager = ExpressionEvaluatorLocator.getManager();
        Compiler compiler = manager.createCompiler();
        compiler.setLexicalAnalyzer(createLexicalAnalyzer());
        return compiler;
    }
    
    public void testIdentifier1() {
        String source = "precio";

        Compiler compiler = createCompiler();
        Code code = compiler.compileExpression(source);
        assertEquals("\"precio\"", code.toString());
    }

    public void testIdentifier2() {
        Compiler compiler = createCompiler();
        compiler.getLexicalAnalyzer().setUseBracketsForIdentifiers(true);

        String source = "[precio]";

        Code code = compiler.compileExpression(source);
        assertEquals("\"precio\"", code.toString());
    }

    public void testIdentifier3() {
        Compiler compiler = createCompiler();

        String source = "\"precio\"";

        Code code = compiler.compileExpression(source);
        assertEquals("\"precio\"", code.toString());
    }


    public void testTrue() {
        String source = "true";

        Compiler compiler = createCompiler();
        Code code = compiler.compileExpression(source);
        assertEquals("TRUE", code.toString());
    }

    public void testFalse() {
        String source = "false";

        Compiler compiler = createCompiler();
        Code code = compiler.compileExpression(source);
        assertEquals("FALSE", code.toString());
    }

    public void testNull() {
        String source = "null";

        Compiler compiler = createCompiler();
        Code code = compiler.compileExpression(source);
        assertEquals("NULL", code.toString());
    }

    public void testIsNull() {
        String source = "10 is null";

        Compiler compiler = createCompiler();
        Code code = compiler.compileExpression(source);
        assertEquals("(10 IS NULL)", code.toString());
    }

    public void testIsNotNull() {
        String source = "10 is not null";

        Compiler compiler = createCompiler();
        Code code = compiler.compileExpression(source);
        assertEquals("NOT((10 IS NULL))", code.toString());
    }

    public void testNotTrue() {
        String source = "not true";

        Compiler compiler = createCompiler();
        Code code = compiler.compileExpression(source);
        assertEquals("NOT(TRUE)", code.toString());
    }

    public void testInteger() {
        String source = "23";

        Compiler compiler = createCompiler();
        Code code = compiler.compileExpression(source);
        assertEquals("23", code.toString());
    }

    public void operator(String operatorName) {
        String source = "precio " + operatorName + " 23";

        Compiler compiler = createCompiler();
        Code code = compiler.compileExpression(source);
        assertEquals("(\"precio\" "+ operatorName + " 23)", code.toString());
    }
    
    public void testOperators() {

        operator("=");
        operator("<>");
        operator(">");
        operator(">=");
        operator("<");
        operator("<=");
        operator("LIKE");
        operator("ILIKE");
        operator("||");
        operator("+");
        operator("-");
        operator("*");
        operator("OR");
        operator("AND");
        operator("%");
        operator("IS");

        operator("~");
    }

    public void testAddMul() {
        String source = "precio + 10 * 2 + 20 + 30";

        Compiler compiler = createCompiler();
        
        Code code = compiler.compileExpression(source);
        assertEquals("(((\"precio\" + (10 * 2)) + 20) + 30)", code.toString());
    }
    
    public void testAddMulPar() {
        String source = "(precio + 10) * 2 + 20 + 30";

        Compiler compiler = createCompiler();
        
        Code code = compiler.compileExpression(source);
        assertEquals("((((\"precio\" + 10) * 2) + 20) + 30)", code.toString());
    }
    
    public void testAbs() {
        String source = "precio + abs(10)";

        Compiler compiler = createCompiler();
        
        Code code = compiler.compileExpression(source);
        assertEquals("(\"precio\" + abs(10))", code.toString());
    }
    
    public void testAbs2() {
        String source = "precio + abs(-10)";

        Compiler compiler = createCompiler();
        
        Code code = compiler.compileExpression(source);
        assertEquals("(\"precio\" + abs(-10))", code.toString());
    }
    
    public void testPI() {
        String source = "precio + PI()";

        Compiler compiler = createCompiler();
        
        Code code = compiler.compileExpression(source);
        assertEquals("(\"precio\" + PI())", code.toString());
    }
    
    public void testCeil() {
        String source = "precio + CEIL(PI())";

        Compiler compiler = createCompiler();
        
        Code code = compiler.compileExpression(source);
        assertEquals("(\"precio\" + CEIL(PI()))", code.toString());
    }
    
    public void testGetitem1() {
        String source = "LIST('uno','dos','tres')[1]" ;

        Compiler compiler = createCompiler();
        
        Code code = compiler.compileExpression(source);
        assertEquals("GETITEM(LIST('uno', 'dos', 'tres'), 1)", code.toString());
    }
    
    public void testConcat() {
        String source = "CONCAT(precio,' euros')";

        Compiler compiler = createCompiler();
        
        Code code = compiler.compileExpression(source);
        assertEquals("CONCAT(\"precio\", ' euros')", code.toString());
    }
    
    public void test1() {
        String source = "NOMBRE03 = 'Torre d''En Besora (la)'";

        Compiler compiler = createCompiler();
        
        Code code = compiler.compileExpression(source);
        assertEquals("(\"NOMBRE03\" = 'Torre d''En Besora (la)')", code.toString());
    }
    
    public void test2() {
        Compiler compiler = createCompiler();
        compiler.getLexicalAnalyzer().setUseBracketsForIdentifiers(true);

        String source = "[1990] = 0.168873933773767";

        Code code = compiler.compileExpression(source);
        assertEquals("(\"1990\" = 0.168873933773767)", code.toString());
    }

    public void test2fields() {
        Compiler compiler = createCompiler();
        compiler.getLexicalAnalyzer().setUseBracketsForIdentifiers(true);

        String source = "[1990] = [precio]";

        Code code = compiler.compileExpression(source);
        assertEquals("(\"1990\" = \"precio\")", code.toString());
    }
    
    
    public void testInvokeFunction1() {
        String source = "ST_Area(GEOMETRY)";

        Compiler compiler = createCompiler();
        
        Code code = compiler.compileExpression(source);
        assertEquals("ST_Area(\"GEOMETRY\")", code.toString());
    }

    public void testInvokeMethod1() {
        String source = "'hola'.length()";

        Compiler compiler = createCompiler();
        
        Code code = compiler.compileExpression(source);
        assertEquals("'hola'->length()", code.toString());
    }
    
    
    public void testInvokeMethod2() {
        String source = "'hola'.indexOf('l')    ";
        
        Compiler compiler = createCompiler();
        
        Code code = compiler.compileExpression(source);
        assertEquals("'hola'->indexOf('l')", code.toString());
    }

    public void testSyntaxError1() {
        Compiler compiler = createCompiler();
        String source;
        
        source = "3 + (2*)";
        try {
            compiler.compileExpression(source);
            assertTrue("A syntax exception should have been thrown compiling '"+source+"'.", false);
        } catch(ExpressionSyntaxException ex) {
            // Ok
        }
        source = "3 + 2*";
        try {
            compiler.compileExpression(source);
            assertTrue("A syntax exception should have been thrown compiling '"+source+"'.", false);
        } catch(ExpressionSyntaxException ex) {
            // Ok
        }
        source = "3 + (2*2))";
        try {
            compiler.compileExpression(source);
            assertTrue("A syntax exception should have been thrown compiling '"+source+"'.", false);
        } catch(ExpressionSyntaxException ex) {
            // Ok
        }
        source = "3 and or 2";
        try {
            compiler.compileExpression(source);
            assertTrue("A syntax exception should have been thrown compiling '"+source+"'.", false);
        } catch(ExpressionSyntaxException ex) {
            // Ok
        }
        source = "3*/2";
        try {
            compiler.compileExpression(source);
            assertTrue("A syntax exception should have been thrown compiling '"+source+"'.", false);
        } catch(ExpressionSyntaxException ex) {
            // Ok
        }
    }

    public void testInvokeFunction2paramters1() {
        String source = "LEFT('NOMBRE', 2)";

        Compiler compiler = createCompiler();
        
        Code code = compiler.compileExpression(source);
        assertEquals("LEFT('NOMBRE', 2)", code.toString());
    }

    public void testInvokeFunction2paramters2() {
        String source = "LEFT('NOMBRE',2)";

        Compiler compiler = createCompiler();
        
        Code code = compiler.compileExpression(source);
        assertEquals("LEFT('NOMBRE', 2)", code.toString());
    }

    public void testInvokeFunction3paramters1() {
        String source = "MID('NOMBRE', 2, 2)";

        Compiler compiler = createCompiler();
        
        Code code = compiler.compileExpression(source);
        assertEquals("MID('NOMBRE', 2, 2)", code.toString());
    }

    public void testInvokeFunction3paramters2() {
        String source = "MID('NOMBRE', 2,2)";

        Compiler compiler = createCompiler();
        
        Code code = compiler.compileExpression(source);
        assertEquals("MID('NOMBRE', 2, 2)", code.toString());
    }

    public void testDecimal1() {
        String source = "23.1";

        Compiler compiler = createCompiler();
        Code code = compiler.compileExpression(source);
        assertEquals("23.1", code.toString());
    }

    public void testDecimal2() {
        String source = "23,1";

        Compiler compiler = createCompiler();
        try {
            Code code = compiler.compileExpression(source);
            fail("I expected 23,1 to be erroneous.");
        } catch(ExpressionSyntaxException ex) {
            // This is OK.
        }
    }
    public void testDecimal3() {
        String source = "1.23E4";

        Compiler compiler = createCompiler();
        Code code = compiler.compileExpression(source);
        assertEquals("12300", code.toString());
    }

    public void testDecimal4() {
        String source = "1.23E-2";

        Compiler compiler = createCompiler();
        Code code = compiler.compileExpression(source);
        assertEquals("0.0123", code.toString());
    }

    public void testDecimal5() {
        String source = "1.23E-4";

        Compiler compiler = createCompiler();
        Code code = compiler.compileExpression(source);
        assertEquals("1.23E-4", code.toString());
    }

    public void testDecimal6() {
        String source = "-23.1";

        Compiler compiler = createCompiler();
        Code code = compiler.compileExpression(source);
        assertEquals("-23.1", code.toString());
    }

    public void testDecimal7() {
        String source = "-1.23E-2";

        Compiler compiler = createCompiler();
        Code code = compiler.compileExpression(source);
        assertEquals("-0.0123", code.toString());
    }

    public void testDecimal8() {
        String source = "-1.23E-4";

        Compiler compiler = createCompiler();
        Code code = compiler.compileExpression(source);
        assertEquals("-1.23E-4", code.toString());
    }


}
