package org.gvsig.expresionevaluator.impl;

import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.List;
import junit.framework.TestCase;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import org.cresques.cts.IProjection;
import org.gvsig.expressionevaluator.Formatter;
import org.gvsig.expressionevaluator.ExpressionBuilder;
import org.gvsig.expressionevaluator.ExpressionBuilder.Constant;
import org.gvsig.expressionevaluator.ExpressionBuilder.Function;
import org.gvsig.expressionevaluator.ExpressionBuilder.GeometrySupportType;
import org.gvsig.expressionevaluator.ExpressionBuilder.Parameter;
import org.gvsig.expressionevaluator.ExpressionBuilder.Value;
import org.gvsig.expressionevaluator.ExpressionBuilder.Variable;
import org.gvsig.expressionevaluator.impl.DefaultExpressionBuilder;
import org.gvsig.fmap.crs.CRSFactory;
import org.gvsig.fmap.geom.Geometry;
import org.gvsig.fmap.geom.GeometryLocator;
import org.gvsig.fmap.geom.GeometryManager;
import org.gvsig.fmap.geom.GeometryUtils;
import org.gvsig.fmap.geom.exception.CreateGeometryException;
import org.gvsig.fmap.geom.primitive.Point;
import org.gvsig.tools.library.impl.DefaultLibrariesInitializer;

public class TestExpressionBuilderFormatter extends TestCase {

    private static class MyFormatter implements Formatter<Value> {
    
        private class Formatter_constant_bytearray implements Formatter<Value> {

            @Override
            public boolean canApply(Value value) {
                if( value instanceof Constant ) {
                    return ((Constant)value).value() instanceof byte[];
                }
                return false;
            }

            @Override
            public String format(Value constant) {
                return builder.bytearray_x((byte[]) ((Constant)constant).value());
            }
        }

        private class Formatter_constant_geometry implements Formatter<Value> {

            @Override
            public boolean canApply(Value value) {
                if( value instanceof Constant ) {
                    return ((Constant)value).value() instanceof Geometry;
                }
                return false;
            }

            @Override
            public String format(Value constant) {
                Geometry geometry = (Geometry) ((Constant)constant).value();
                switch (builder.geometry_support_type()) {
                    case EWKB:
                        return MessageFormat.format(
                                "ST_GeomFromEWKB(({0}), ({1}))",
                                builder.bytearray_x(GeometryUtils.toEWKB(geometry)),
                                String.valueOf(builder.srs_id(geometry.getProjection()))
                        );
                    case WKB:
                        return MessageFormat.format(
                                "ST_GeomFromWKB(({0}), ({1}))",
                                builder.bytearray_x(GeometryUtils.toWKB(geometry)),
                                String.valueOf(builder.srs_id(geometry.getProjection()))
                        );
                    case WKT:
                    default:
                        return MessageFormat.format(
                                "ST_GeomFromText({0}, ({1}))",
                                builder.string(GeometryUtils.toWKT(geometry)),
                                String.valueOf(builder.srs_id(geometry.getProjection()))
                        );
                }
            }
        }

        private class Formatter_ST_intersects implements Formatter<Value> {

            @Override
            public boolean canApply(Value value) {
                if( value instanceof Function ) {
                    return StringUtils.equalsIgnoreCase("ST_intersects",((Function)value).name());
                }
                return false;
            }

            @Override
            public String format(Value function) {
                List<Value> parameters = ((Function)function).parameters();
                String p1 = parameters.get(0).toString(formatter());
                String p2 = parameters.get(1).toString(formatter());
                String r = MessageFormat.format(
                        "( (({0}) && ({1})) AND ST_Intersects(({0}),({1}) ))", 
                        p1,
                        p2
                );
                return r;
            }
        }

        private class Formatter_ST_GeomFromEWKB implements Formatter<Value> {

            @Override
            public boolean canApply(Value value) {
                if( value instanceof Function ) {
                    return StringUtils.equalsIgnoreCase("ST_GeomFromEWKB",((Function)value).name());
                }
                return false;
            }

            @Override
            public String format(Value function) {
                List<Value> parameters = ((Function)function).parameters();
                String p1 = parameters.get(0).toString(formatter());
                String p2 = parameters.get(1).toString(formatter());
                String r = MessageFormat.format(
                        "GeomFromWKB({0}, {1})", 
                        p1,
                        p2
                );
                return r;
            }
        }
    
        private class Formatter_IFNULL implements Formatter<Value> {

            @Override
            public boolean canApply(Value value) {
                if( value instanceof Function ) {
                    return StringUtils.equalsIgnoreCase("IFNULL",((Function)value).name());
                }
                return false;
            }

            @Override
            public String format(Value function) {
                List<Value> parameters = ((Function)function).parameters();
                String p1 = parameters.get(0).toString(formatter());
                String p2 = parameters.get(1).toString(formatter());
                String p3 = parameters.get(2).toString(formatter());
                String r = MessageFormat.format(
                        "NVL2({0}, {1}, {2})", 
                        p1,
                        p3,
                        p2
                );
                return r;
            }
        }
    
        private class Formatter_ST_AsEWKB implements Formatter<Value> {

            @Override
            public boolean canApply(Value value) {
                if( value instanceof Function ) {
                    return StringUtils.equalsIgnoreCase("ST_AsEWKB",((Function)value).name());
                }
                return false;
            }

            public String format(Value function) {
                List<Value> parameters = ((Function)function).parameters();
                String p1 = parameters.get(0).toString(formatter());
                String r = MessageFormat.format(
                        "AsWKB(({0}))", 
                        p1
                );
                return r;
            }
        }
        
        private class Formatter_NOT_IS_NULL implements Formatter<Value> {

            @Override
            public boolean canApply(Value value) {
                if( value instanceof Function ) {
                    return StringUtils.equalsIgnoreCase("NOT_IS_NULL",((Function)value).name());
                }
                return false;
            }

            @Override
            public String format(Value function) {
                List<Value> parameters = ((Function)function).parameters();
                String p1 = parameters.get(0).toString(MyFormatter.this);
                String r = MessageFormat.format(
                        "( ({0}) IS NOT NULL )", 
                        p1
                );
                return r;
            }
        }

        private final Formatter<Value>[] formatters;
        private final ExpressionBuilder builder;

        public MyFormatter(ExpressionBuilder builder) {
            this.builder = builder;
            this.formatters = new Formatter[] {
                new Formatter_IFNULL(),
                new Formatter_NOT_IS_NULL(),
                new Formatter_ST_AsEWKB(),
                new Formatter_ST_GeomFromEWKB(),
                new Formatter_constant_bytearray(),
                new Formatter_constant_geometry(),
                new Formatter_ST_intersects()
            };
        }

        @Override
        public boolean canApply(Value value) {
            for (Formatter<Value> formatter : formatters) {
                if( formatter.canApply(value) ) {
                    return true;
                }
            }
            return false;
        }

        @Override
        public String format(Value value) {
            for (Formatter<Value> formatter : formatters) {
                if( formatter.canApply(value) ) {
                    return formatter.format(value);
                }
            }
            return value.toString(this);
        }
        
        private Formatter formatter() {
            return this;
        }
    }
    
    public TestExpressionBuilderFormatter(String testName) {
        super(testName);
    }

    @Override
    protected void setUp() throws Exception {
        super.setUp();
        new DefaultLibrariesInitializer().fullInitialize();
    }

    @Override
    protected void tearDown() throws Exception {
        super.tearDown();
    }

    // TODO add test methods here. The name must begin with 'test'. For example:
    // public void testHello() {}
    
    List<String> getVariableNames(ExpressionBuilder builder) {
        List<String> vars = new ArrayList<>();
        for (Variable var : builder.variables()) {
            vars.add(var.name());
        }
        return vars;
    }
    
    List<String> getParameterNames(ExpressionBuilder builder) {
        List<String> params = new ArrayList<>();
        for (Parameter param : builder.parameters()) {
            String s;
            switch(param.type()) {
                case Constant:
                    Object value = param.value();
                    if( value==null ) {
                        s = "null";
                    } else if( value instanceof String ) {
                        s = "'" + (String)value + "'";
                    } else {
                        s = value.toString();
                    }    
                    break;
                case Geometry:
                case Variable:
                default:
                    s = "\"" + param.name() + "\"";
            }
            params.add(s);
        }
        return params;
    }
    
    public void test2() {
        ExpressionBuilder builder = new DefaultExpressionBuilder();
        
        builder.and(
          builder.eq(
            builder.lower(builder.variable("colum_name_c")),
            builder.parameter("colum_name_p")
          )
        );
        builder.and(
            builder.group(
                builder.or(
                    builder.like(
                        builder.lower( builder.variable("uno")),
                        builder.constant("%10")
                    ),
                    builder.lt(
                        builder.variable("dos"),
                        builder.constant(-3.5)
                    )
                )
            )
        );
        builder.and(
                builder.ST_Intersects(
                    builder.variable("geom1"), 
                    builder.ST_Envelope(
                        builder.ST_GeomFromWKB(
                                builder.parameter("geom2"), 
                                builder.parameter().value(4326).as_constant()
                        )
                    )
                )
        );
        builder.and(
            builder.gt(
                builder.variable("tres"),
                builder.constant(123456789)
            )
        );
        assertEquals(
                "(((( (LOWER(\"colum_name_c\")) = (?) ) AND ( ( (LOWER(\"uno\")) LIKE ('%10') ) OR ( (\"dos\") < (-3.5) ) )) AND ST_Intersects((\"geom1\"), (ST_Envelope(ST_GeomFromWKB((?), (?)))))) AND ( (\"tres\") > (123456789) ))",
                builder.toString()
        );
        assertEquals(
                "(((( (LOWER(\"colum_name_c\")) = (?) ) AND ( ( (LOWER(\"uno\")) LIKE ('%10') ) OR ( (\"dos\") < (-3.5) ) )) AND ( ((\"geom1\") && (ST_Envelope(ST_GeomFromWKB((?), (?))))) AND ST_Intersects((\"geom1\"),(ST_Envelope(ST_GeomFromWKB((?), (?)))) ))) AND ( (\"tres\") > (123456789) ))",
                builder.toString(new MyFormatter(builder))
        );
        
        assertEquals(
                "[colum_name_c, dos, geom1, tres, uno]",
                ArrayUtils.toString(getVariableNames(builder))
        );
        assertEquals(
                "[\"colum_name_p\", \"geom2\", 4326]",
                ArrayUtils.toString(getParameterNames(builder))
        );
    }
    
    public void test3() throws CreateGeometryException {
        ExpressionBuilder builder = new DefaultExpressionBuilder();
                
        GeometryManager geometryManager = GeometryLocator.getGeometryManager();
        IProjection proj = CRSFactory.getCRS("EPSG:4326");
        
        Point point = geometryManager.createPoint(10, 20, Geometry.SUBTYPES.GEOM2D);
        builder.set(        
            builder.ST_Intersects(
              builder.geometry(point,proj),
              builder.variable("the_geom")
            )
        );
        builder.geometry_support_type(GeometrySupportType.WKT);
        System.out.println(builder.toString());
        assertEquals(
                "ST_Intersects((ST_GeomFromText(('POINT (10 20)'), (4326))), (\"the_geom\"))",
                builder.toString()
        );
        assertEquals(
                "( ((ST_GeomFromText('POINT (10 20)', (4326))) && (\"the_geom\")) AND ST_Intersects((ST_GeomFromText('POINT (10 20)', (4326))),(\"the_geom\") ))",
                builder.toString(new MyFormatter(builder))
        );

        builder.geometry_support_type(GeometrySupportType.WKB);
        System.out.println(builder.toString());
        assertEquals(
                "ST_Intersects((ST_GeomFromWKB((DECODE('000000000140240000000000004034000000000000','hex')), (4326))), (\"the_geom\"))",
                builder.toString()
        );
        assertEquals(
                "( ((ST_GeomFromWKB((x'000000000140240000000000004034000000000000'), (4326))) && (\"the_geom\")) AND ST_Intersects((ST_GeomFromWKB((x'000000000140240000000000004034000000000000'), (4326))),(\"the_geom\") ))",
                builder.toString(new MyFormatter(builder))
        );
        assertEquals(
                "[the_geom]",
                ArrayUtils.toString(getVariableNames(builder))
        );
        assertEquals(
                "[]",
                ArrayUtils.toString(getParameterNames(builder))
        );
    }
    
    public void test4() throws CreateGeometryException {
        ExpressionBuilder builder = new DefaultExpressionBuilder();
                
        IProjection proj = CRSFactory.getCRS("EPSG:4326");
        
        builder.and(
                builder.ST_Intersects(
                    builder.variable("geom1"), 
                    builder.ST_Envelope(
                        builder.ST_GeomFromWKB(
                                builder.parameter("geom2"), 
                                builder.parameter().value(proj).as_constant()
                        )
                    )
                )
        );
        
        System.out.println(builder.toString());
        assertEquals(
                "ST_Intersects((\"geom1\"), (ST_Envelope(ST_GeomFromWKB((?), (?)))))",
                builder.toString()
        );
        assertEquals(
                "( ((\"geom1\") && (ST_Envelope(ST_GeomFromWKB((?), (?))))) AND ST_Intersects((\"geom1\"),(ST_Envelope(ST_GeomFromWKB((?), (?)))) ))",
                builder.toString(new MyFormatter(builder))
        );
        assertEquals(
                "[geom1]",
                ArrayUtils.toString(getVariableNames(builder))
        );
        assertEquals(
                "[\"geom2\", 4326]",
                ArrayUtils.toString(getParameterNames(builder))
        );
    }
        
    public void test5() throws CreateGeometryException {
        ExpressionBuilder builder = new DefaultExpressionBuilder();
                
        IProjection proj = CRSFactory.getCRS("EPSG:4326");
        
        builder.and(
                builder.eq(
                        builder.ST_SRID( builder.variable("geom") ),
                        builder.srs(proj)
                )
        );
        
        System.out.println(builder.toString());
        assertEquals(
                "( (ST_SRID(\"geom\")) = (4326) )",
                builder.toString()
        );
        assertEquals(
                "( (ST_SRID(\"geom\")) = (4326) )",
                builder.toString(new MyFormatter(builder))
        );
        assertEquals(
                "[geom]",
                ArrayUtils.toString(getVariableNames(builder))
        );
        assertEquals(
                "[]",
                ArrayUtils.toString(getParameterNames(builder))
        );
    }
}
