package org.gvsig.expresionevaluator.impl;

import junit.framework.TestCase;
import org.gvsig.expressionevaluator.Code;
import org.gvsig.expressionevaluator.ExpressionEvaluatorLocator;
import org.gvsig.expressionevaluator.ExpressionEvaluatorManager;
import org.gvsig.expressionevaluator.Interpreter;
import org.gvsig.expressionevaluator.LexicalAnalyzer;
import org.gvsig.expressionevaluator.SymbolTable;
import org.gvsig.expressionevaluator.Compiler;
import org.gvsig.expressionevaluator.MutableSymbolTable;
import org.gvsig.tools.library.impl.DefaultLibrariesInitializer;

/**
 *
 * @author jjdelcerro
 */
public class TestGrammarCompiler extends TestCase {

    public TestGrammarCompiler(String testName) {
        super(testName);
    }

    @Override
    protected void setUp() throws Exception {
        super.setUp();
        new DefaultLibrariesInitializer().fullInitialize();
    }

    @Override
    protected void tearDown() throws Exception {
        super.tearDown();
    }

    // TODO add test methods here. The name must begin with 'test'. For example:
    // public void testHello() {}
    
    protected LexicalAnalyzer createLexicalAnalyzer() {
        ExpressionEvaluatorManager manager = ExpressionEvaluatorLocator.getManager();
        LexicalAnalyzer lexer = manager.createLexicalAnalyzer();
        return lexer;
    }

    protected org.gvsig.expressionevaluator.Compiler createCompiler() {
        ExpressionEvaluatorManager manager = ExpressionEvaluatorLocator.getManager();
        Compiler compiler = manager.createCompiler();
        compiler.setLexicalAnalyzer(createLexicalAnalyzer());
        return compiler;
    }

    protected SymbolTable createSymbolTable() {
        ExpressionEvaluatorManager manager = ExpressionEvaluatorLocator.getManager();
        MutableSymbolTable symbolTable = manager.createSymbolTable();
        symbolTable.setVar("precio", 200);
        symbolTable.setVar("1990", 0.168873933773767);
        return symbolTable;
    }

    protected Interpreter createInterpreter(SymbolTable symbolTable) {
        ExpressionEvaluatorManager manager = ExpressionEvaluatorLocator.getManager();
        Interpreter interpreter = manager.createInterpreter();
        interpreter.setSymbolTable(symbolTable);
        return interpreter;
    }

    public void testLet() {
        String source = "LET V1 = 23";

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("LET('V1', 23)", code.toString());
    }

    public void testBlock() {
        String source = "BEGIN LET V1 = 23; END";

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("BLOCK(LET('V1', 23))", code.toString());
    }

    public void testBlockExcept() {
        String source = "BEGIN LET V1 = 23; EXCEPT NULL; END";

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("BLOCK_EXCEPT(LET('V1', 23), NULL)", code.toString());
    }

    public void testBlock2() {
        String source = "BEGIN LET V1 = 11; LET V2 = 22;  END";

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("BLOCK(LET('V1', 11), LET('V2', 22))", code.toString());
    }

    public void testBlockExcept2() {
        String source = "BEGIN LET V1 = 11; LET V2 = 22; EXCEPT -1; END";

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("BLOCK_EXCEPT(LET('V1', 11), LET('V2', 22), -1)", code.toString());
    }

    public void testBlock3() {
        Compiler compiler = createCompiler();

        String source = "BEGIN LET V1 = 11; LET V2 = 22;  END";
        Code code = compiler.compileExpression(source);
        assertEquals("BLOCK(LET('V1', 11), LET('V2', 22))", code.toString());

        source = "BEGIN LET V1 = 11;; LET V2 = 22;;;  END";
        code = compiler.compileExpression(source);
        assertEquals("BLOCK(LET('V1', 11), LET('V2', 22))", code.toString());
    }

    public void testIfThen() {
        String source = "IF V1 = 11 THEN LET V2 = 22; END IF";

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("IFF((\"V1\" = 11), LET('V2', 22), NULL)", code.toString());
    }

    public void testIfThenElse() {
        String source = "IF V1 = 11 THEN LET V2 = 11; ELSE LET V2 = 22; END IF";

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("IFF((\"V1\" = 11), LET('V2', 11), LET('V2', 22))", code.toString());
    }

    public void testIfThenBlock() {
        String source = "IF V1 = 11 THEN LET V2 = 22; LET V1 = 10; END IF";

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("IFF((\"V1\" = 11), BLOCK(LET('V2', 22), LET('V1', 10)), NULL)", code.toString());
    }

    public void testWhile() {
        String source = "WHILE n < 10 BEGIN LET n = n + 1; END WHILE";

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("WHILE((\"n\" < 10), LET('n', (\"n\" + 1)))", code.toString());
    }

    public void testWhileBlock() {
        String source = "WHILE n < 10 BEGIN LET n = n + 1; print(n); END WHILE";

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("WHILE((\"n\" < 10), BLOCK(LET('n', (\"n\" + 1)), print(\"n\")))", code.toString());
    }

    public void testCast() {
        String source = "CAST(V1 AS INTEGER)";

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("CAST(\"V1\", 'INTEGER')", code.toString());
    }

    public void testCast2() {
        String source = "CAST V1 AS INTEGER";

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("CAST(\"V1\", 'INTEGER')", code.toString());
    }

    public void testMatch() {
        Compiler compiler = createCompiler();
        String source;
        Code code;

        source = "MATCH 'Hola', '[Hh]ola'";
        code = compiler.compileExpression(source);
        assertEquals("~('Hola', '[Hh]ola')", code.toString());

        source = "MATCH 'Hola' '[Hh]ola'";
        code = compiler.compileExpression(source);
        assertEquals("~('Hola', '[Hh]ola')", code.toString());

        source = "MATCH('Hola', '[Hh]ola')";
        code = compiler.compileExpression(source);
        assertEquals("~('Hola', '[Hh]ola')", code.toString());
    }

    public void testForEach() {
        String source = "FOR n in RANGE(10): print(n); END FOR";

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("FOREACH('n', RANGE(10), print(\"n\"))", code.toString());
    }

//    public void testFor() {
//        String source = "FOR LET n = 0 ; n < 10 ; LET n = n + 1 print(n)";
//
//        Compiler compiler = createCompiler();
//
//        Code code = compiler.compileExpression(source);
//        assertEquals("FOR(LET('n', 0), <(\"n\", 10), LET('n', +(\"n\", 1)), print(\"n\"))", code.toString());
//    }

    public void testCase() {
        String source = "CASE WHEN software LIKE '%gvSIG%' THEN 'gvSIG'; ELSE 'Other'; END CASE"; // ;)

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("CASE((\"software\" LIKE '%gvSIG%'), 'gvSIG', 'Other')", code.toString());
    }

    public void testCase1() {
        String source = "CASE WHEN software LIKE '%gvSIG%' THEN 'gvSIG'; ELSE 'Other'; END CASE"; // ;)

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("CASE((\"software\" LIKE '%gvSIG%'), 'gvSIG', 'Other')", code.toString());
    }

    public void testCase2() {
        String source = "CASE WHEN Field_1 >= 75 AND Field_1 <=79 THEN 100; WHEN Field_1 >=80 AND Field_1 <=84 THEN 110; END CASE"; // ;)

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("CASE(((\"Field_1\" >= 75) AND (\"Field_1\" <= 79)), ((\"Field_1\" >= 80) AND (\"Field_1\" <= 84)), 100, 110, NULL)", code.toString());
    }
    
    public void testCase3() {
        String source = "CASE WHEN Field_1 >= 75 AND Field_1 <=79 THEN 100; WHEN Field_1 >=80 AND Field_1 <=84 THEN 110; ELSE 120; END CASE"; // ;)

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("CASE(((\"Field_1\" >= 75) AND (\"Field_1\" <= 79)), ((\"Field_1\" >= 80) AND (\"Field_1\" <= 84)), 100, 110, 120)", code.toString());
    }
    
    public void testDef1() {
        String source = "CREATE PROCEDURE test1 AS BEGIN print('Hola'); END PROCEDURE"; 

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("CREATE_FUNCTION('test1', NULL, print('Hola'), NULL, NULL, NULL)", code.toString());
    }
    
    public void testDef2() {
        String source = "CREATE PROCEDURE test1 nombre AS BEGIN print('Hola '+nombre); END PROCEDURE";

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("CREATE_FUNCTION('test1', LIST('nombre'), print(('Hola ' + \"nombre\")), NULL, NULL, NULL)", code.toString());
    }

    public void testDef3() {
        String source = "CREATE PROCEDURE test1 nombre AS BEGIN RETURN 'Hola '||nombre; END PROCEDURE";

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("CREATE_FUNCTION('test1', LIST('nombre'), RETURN(('Hola ' || \"nombre\")), NULL, NULL, NULL)", code.toString());
    }
    
    public void testCreateFunctionExtern1() {
        String source = "CREATE FUNCTION test1(nombre) AS 'addons/test/test', test1";

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("CREATE_FUNCTION('test1', LIST('nombre'), NULL, 'addons/test/test', 'test1', NULL)", code.toString());
    }
    
    public void testList() {
        String source = "set x TO LIST('hola','adios','fin')";

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("LET('x', LIST('hola', 'adios', 'fin'))", code.toString());
    }
    
    public void testArray() {
        String source = "set x TO ARRAY['hola','adios','fin']";

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("LET('x', LIST('hola', 'adios', 'fin'))", code.toString());
    }
    
    public void testUseCase1() {
        String source = "begin set x = ''; for n in list('hola','adios','fin'): let x = x || ' ' || n; end for; end";

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("BLOCK(LET('x', ''), FOREACH('n', list('hola', 'adios', 'fin'), LET('x', ((\"x\" || ' ') || \"n\"))))", code.toString());
    }

    public void testUseCase2() {
        String source = "begin set s = ''; set x = LIST('hola','adios','fin'); FOR n in RANGE(3): set s = s || ' ' || x[n]; end for; end";

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("BLOCK(LET('s', ''), LET('x', LIST('hola', 'adios', 'fin')), FOREACH('n', RANGE(3), LET('s', ((\"s\" || ' ') || GETITEM(\"x\", \"n\")))))", code.toString());
    }


}
