package org.gvsig.expresionevaluator.impl;

import static junit.framework.Assert.assertEquals;
import org.gvsig.expressionevaluator.impl.SQLLexicalAnalyzer;
import org.gvsig.expressionevaluator.impl.DefaultCompiler;
import junit.framework.TestCase;
import org.cresques.cts.IProjection;
import org.gvsig.expressionevaluator.LexicalAnalyzer;
import org.gvsig.expressionevaluator.Compiler;
import org.gvsig.expressionevaluator.Code;
import org.gvsig.expressionevaluator.ExpressionBuilder;
import org.gvsig.expressionevaluator.Interpreter;
import org.gvsig.expressionevaluator.MutableSymbolTable;
import org.gvsig.expressionevaluator.Optimizer;
import org.gvsig.expressionevaluator.SymbolTable;
import org.gvsig.expressionevaluator.impl.DefaultExpressionBuilder;
import org.gvsig.expressionevaluator.impl.DefaultInterpreter;
import org.gvsig.expressionevaluator.impl.DefaultOptimizer;
import org.gvsig.expressionevaluator.impl.DefaultSymbolTable;
import org.gvsig.fmap.crs.CRSFactory;
import org.gvsig.fmap.geom.Geometry;
import org.gvsig.fmap.geom.GeometryLocator;
import org.gvsig.fmap.geom.GeometryManager;
import org.gvsig.fmap.geom.exception.CreateGeometryException;
import org.gvsig.fmap.geom.primitive.Point;
import org.gvsig.fmap.geom.primitive.Polygon;
import org.gvsig.tools.library.impl.DefaultLibrariesInitializer;

public class TestOptimizer extends TestCase {

    public TestOptimizer(String testName) {
        super(testName);
    }

    @Override
    protected void setUp() throws Exception {
        super.setUp();
        new DefaultLibrariesInitializer().fullInitialize();
    }

    @Override
    protected void tearDown() throws Exception {
        super.tearDown();
    }

    // TODO add test methods here. The name must begin with 'test'. For example:
    // public void testHello() {}
    protected LexicalAnalyzer createLexicalAnalyzer() {
        SQLLexicalAnalyzer lexer = new SQLLexicalAnalyzer();
        return lexer;
    }

    protected Compiler createCompiler() {
        Compiler compiler = new DefaultCompiler();
        compiler.setLexicalAnalyzer(createLexicalAnalyzer());
        return compiler;
    }

    protected Interpreter createInterpreter(SymbolTable symbolTable) {
        Interpreter interpreter = new DefaultInterpreter();
        interpreter.setSymbolTable(symbolTable);
        return interpreter;
    }

    protected Optimizer createOptimizer(SymbolTable symbolTable) {
        Optimizer optimizer = new DefaultOptimizer(symbolTable);
        return optimizer;
    }

    protected MutableSymbolTable createSymbolTable() {
        DefaultSymbolTable symbolTable = new DefaultSymbolTable();
        symbolTable.setVar("precio", 200);
        symbolTable.setVar("1990", 0.168873933773767);
        return symbolTable;
    }

    private String createExpr1() {
        ExpressionBuilder builder = new DefaultExpressionBuilder();

        ExpressionBuilder.Value expr1 = builder.gt(
                builder.column("campo1"),
                builder.constant(10)
        );
        System.out.println(expr1.toString());
        assertEquals(
                "( (\"campo1\") > (10) )",
                expr1.toString()
        );
        return expr1.toString();
    }

    private Polygon createPolygon() throws CreateGeometryException {
        GeometryManager geometryManager = GeometryLocator.getGeometryManager();
        IProjection proj = CRSFactory.getCRS("EPSG:4326");

        Polygon polygon = geometryManager.createPolygon(Geometry.SUBTYPES.GEOM2D);
        polygon.addVertex(0, 0);
        polygon.addVertex(0, 100);
        polygon.addVertex(100, 100);
        polygon.addVertex(100, 0);
        polygon.addVertex(0, 0);

        polygon.setProjection(proj);
        return polygon;
    }

    private Geometry createPoint(double x, double y) throws CreateGeometryException {
        GeometryManager geometryManager = GeometryLocator.getGeometryManager();
        IProjection proj = CRSFactory.getCRS("EPSG:4326");
        Point point = geometryManager.createPoint(x, y, Geometry.SUBTYPES.GEOM2D);
        point.setProjection(proj);
        return point;
    }

    public void testIdentifier1() {
        String source = "precio";

        Compiler compiler = createCompiler();
        Code code = compiler.compileExpression(source);
        assertEquals("\"precio\"", code.toString());

        SymbolTable symbolTable = createSymbolTable();
        Optimizer optimizer = createOptimizer(symbolTable);
        Code code2 = optimizer.optimize(code);
        assertEquals("\"precio\"", code2.toString());
    }

    public void testIdentifier2() {
        Compiler compiler = createCompiler();
        compiler.getLexicalAnalyzer().setUseBracketsForIdentifiers(true);

        String source = "[precio]";

        Code code = compiler.compileExpression(source);
        assertEquals("\"precio\"", code.toString());

        SymbolTable symbolTable = createSymbolTable();
        Optimizer optimizer = createOptimizer(symbolTable);
        Code code2 = optimizer.optimize(code);
        assertEquals("\"precio\"", code2.toString());
    }

    public void testIdentifier3() {
        String source = "\"precio\"";

        Compiler compiler = createCompiler();
        Code code = compiler.compileExpression(source);
        assertEquals("\"precio\"", code.toString());

        SymbolTable symbolTable = createSymbolTable();
        Optimizer optimizer = createOptimizer(symbolTable);
        Code code2 = optimizer.optimize(code);
        assertEquals("\"precio\"", code2.toString());
    }

    public void testTrue() {
        String source = "true";

        Compiler compiler = createCompiler();
        Code code = compiler.compileExpression(source);
        assertEquals("TRUE", code.toString());

        SymbolTable symbolTable = createSymbolTable();
        Optimizer optimizer = createOptimizer(symbolTable);
        Code code2 = optimizer.optimize(code);
        assertEquals("TRUE", code2.toString());
    }

    public void testFalse() {
        String source = "false";

        Compiler compiler = createCompiler();
        Code code = compiler.compileExpression(source);
        assertEquals("FALSE", code.toString());

        SymbolTable symbolTable = createSymbolTable();
        Optimizer optimizer = createOptimizer(symbolTable);
        Code code2 = optimizer.optimize(code);
        assertEquals("FALSE", code2.toString());
    }

    public void testNull() {
        String source = "null";

        Compiler compiler = createCompiler();
        Code code = compiler.compileExpression(source);
        assertEquals("NULL", code.toString());

        SymbolTable symbolTable = createSymbolTable();
        Optimizer optimizer = createOptimizer(symbolTable);
        Code code2 = optimizer.optimize(code);
        assertEquals("NULL", code2.toString());
    }

    public void testNotTrue() {
        String source = "not true";

        Compiler compiler = createCompiler();
        Code code = compiler.compileExpression(source);
        assertEquals("NOT(TRUE)", code.toString());

        SymbolTable symbolTable = createSymbolTable();
        Optimizer optimizer = createOptimizer(symbolTable);
        Code code2 = optimizer.optimize(code);
        assertEquals("FALSE", code2.toString());
    }

    public void testInteger() {
        String source = "23";

        Compiler compiler = createCompiler();
        Code code = compiler.compileExpression(source);
        assertEquals("23", code.toString());

        SymbolTable symbolTable = createSymbolTable();
        Optimizer optimizer = createOptimizer(symbolTable);
        Code code2 = optimizer.optimize(code);
        assertEquals("23", code2.toString());
    }

    public void operator(String operatorName) {
        String source = "precio " + operatorName + " 23";

        Compiler compiler = createCompiler();
        Code code = compiler.compileExpression(source);
        assertEquals( "(\"precio\" "+operatorName+" 23)", code.toString());

        SymbolTable symbolTable = createSymbolTable();
        Optimizer optimizer = createOptimizer(symbolTable);
        Code code2 = optimizer.optimize(code);
        assertEquals( "(\"precio\" "+operatorName+" 23)", code2.toString());
    }

    public void testOperators() {

        operator("=");
        operator("<>");
        operator(">");
        operator(">=");
        operator("<");
        operator("<=");
        operator("LIKE");
        operator("ILIKE");
        operator("||");
        operator("+");
        operator("-");
        operator("*");
//        operator("^");
        operator("OR");
        operator("AND");
        operator("%");
        operator("IS");
    }

    
    public void testAddMul() {
        String source = "precio + 10 * 2 + 20 + 30";

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("(((\"precio\" + (10 * 2)) + 20) + 30)", code.toString());

        SymbolTable symbolTable = createSymbolTable();
        Optimizer optimizer = createOptimizer(symbolTable);
        Code code2 = optimizer.optimize(code);
        assertEquals("(((\"precio\" + 20) + 20) + 30)", code2.toString());
    }
    
    public void testMul0() {
        String source = "precio + 10 * 0 + 20 + 30";

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("(((\"precio\" + (10 * 0)) + 20) + 30)", code.toString());

        SymbolTable symbolTable = createSymbolTable();
        Optimizer optimizer = createOptimizer(symbolTable);
        Code code2 = optimizer.optimize(code);
        assertEquals("((\"precio\" + 20) + 30)", code2.toString());
    }

    public void testMul1() {
        String source = "precio + 10 * 1 + 20 + 30";

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("(((\"precio\" + (10 * 1)) + 20) + 30)", code.toString());

        SymbolTable symbolTable = createSymbolTable();
        Optimizer optimizer = createOptimizer(symbolTable);
        Code code2 = optimizer.optimize(code);
        assertEquals("(((\"precio\" + 10) + 20) + 30)", code2.toString());
    }

    public void testDiv1() {
        String source = "precio + 10 / 1 + 20 + 30";

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("(((\"precio\" + (10 / 1)) + 20) + 30)", code.toString());

        SymbolTable symbolTable = createSymbolTable();
        Optimizer optimizer = createOptimizer(symbolTable);
        Code code2 = optimizer.optimize(code);
        assertEquals("(((\"precio\" + 10) + 20) + 30)", code2.toString());
    }

    public void testDiv0() {
        String source = "precio + 0 / 10 + 20 + 30";

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("(((\"precio\" + (0 / 10)) + 20) + 30)", code.toString());

        SymbolTable symbolTable = createSymbolTable();
        Optimizer optimizer = createOptimizer(symbolTable);
        Code code2 = optimizer.optimize(code);
        assertEquals("((\"precio\" + 20) + 30)", code2.toString());
    }

    public void testAddMulPar() {
        String source = "(precio + 10) * 2 + 20 + 30";

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("((((\"precio\" + 10) * 2) + 20) + 30)", code.toString());

        SymbolTable symbolTable = createSymbolTable();
        Optimizer optimizer = createOptimizer(symbolTable);
        Code code2 = optimizer.optimize(code);
        assertEquals("((((\"precio\" + 10) * 2) + 20) + 30)", code2.toString());
    }

    public void testAbs() {
        String source = "precio + abs(10)";

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("(\"precio\" + abs(10))", code.toString());

        SymbolTable symbolTable = createSymbolTable();
        Optimizer optimizer = createOptimizer(symbolTable);
        Code code2 = optimizer.optimize(code);
        assertEquals("(\"precio\" + 10)", code2.toString());
    }

    public void testAbs2() {
        String source = "precio + abs(-10)";

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("(\"precio\" + abs(-10))", code.toString());

        SymbolTable symbolTable = createSymbolTable();
        Optimizer optimizer = createOptimizer(symbolTable);
        Code code2 = optimizer.optimize(code);
        assertEquals("(\"precio\" + 10)", code2.toString());
    }

    public void testPI() {
        String source = "precio + PI()";

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("(\"precio\" + PI())", code.toString());

        SymbolTable symbolTable = createSymbolTable();
        Optimizer optimizer = createOptimizer(symbolTable);
        Code code2 = optimizer.optimize(code);
        assertEquals("(\"precio\" + 3.141592653589793)", code2.toString());
    }

    public void testCeil() {
        String source = "precio + CEIL(PI())";

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("(\"precio\" + CEIL(PI()))", code.toString());

        SymbolTable symbolTable = createSymbolTable();
        Optimizer optimizer = createOptimizer(symbolTable);
        Code code2 = optimizer.optimize(code);
        assertEquals("(\"precio\" + 4.0)", code2.toString());
    }

    public void testConcat() {
        String source = "CONCAT(precio,' euros')";

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("CONCAT(\"precio\", ' euros')", code.toString());

        SymbolTable symbolTable = createSymbolTable();
        Optimizer optimizer = createOptimizer(symbolTable);
        Code code2 = optimizer.optimize(code);
        assertEquals("CONCAT(\"precio\", ' euros')", code2.toString());
    }

    public void testConcat2() {
        String source = "CONCAT(precio,' euros')";

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("CONCAT(\"precio\", ' euros')", code.toString());

        SymbolTable symbolTable = createSymbolTable();
        Optimizer optimizer = createOptimizer(symbolTable);
        Code code2 = optimizer.optimize(code);
        assertEquals("CONCAT(\"precio\", ' euros')", code2.toString());
    }

    public void testInvokeFunction1() {
        String source = "ST_Area(GEOMETRY)";

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("ST_Area(\"GEOMETRY\")", code.toString());

        SymbolTable symbolTable = createSymbolTable();
        Optimizer optimizer = createOptimizer(symbolTable);
        Code code2 = optimizer.optimize(code);
        assertEquals("ST_Area(\"GEOMETRY\")", code2.toString());
    }

    public void testInvokeMethod1() {
        String source = "'hola'.length()";

        Compiler compiler = createCompiler();

        Code code = compiler.compileExpression(source);
        assertEquals("'hola'->length()", code.toString());

        SymbolTable symbolTable = createSymbolTable();
        Optimizer optimizer = createOptimizer(symbolTable);
        Code code2 = optimizer.optimize(code);
        assertEquals("'hola'->length()", code2.toString());
    }

    public void testSpatialExpression1() throws Exception {
        IProjection proj = CRSFactory.getCRS("EPSG:4326");
        ExpressionBuilder builder = new DefaultExpressionBuilder();
        builder.geometry_support_type(ExpressionBuilder.GeometrySupportType.WKB);

        String source = builder.and(
                builder.custom(this.createExpr1()),
                builder.ST_Intersects(
                        builder.geometry(createPolygon()),
                        builder.column("GEOMETRY")
                )
        ).toString();

        MutableSymbolTable symbolTable = createSymbolTable();
        symbolTable.setVar("campo1", 11);
        symbolTable.setVar("GEOMETRY", createPoint(50, 50));

        Interpreter interpreter = createInterpreter(symbolTable);
        Compiler compiler = createCompiler();
        Code code = compiler.compileExpression(source);
        System.out.println(code.toString());
        assertEquals(
                "((\"campo1\" > 10) AND ST_Intersects(ST_GeomFromWKB(DECODE('000000000300000001000000050000000000000000000000000000000000000000000000004059000000000000405900000000000040590000000000004059000000000000000000000000000000000000000000000000000000000000', 'hex'), 4326), \"GEOMETRY\"))",
                code.toString()
        );

//        long t1 = System.currentTimeMillis();
//        for (int i = 0; i < 1000; i++) {
//            interpreter.run(code);
//        }
//        System.out.println("Sin optimizar: " + (System.currentTimeMillis() - t1));

//        t1 = System.currentTimeMillis();
        Optimizer optimizer = createOptimizer(symbolTable);
        Code code2 = optimizer.optimize(code);
//        System.out.println("Tiempo de optimizacion: " + (System.currentTimeMillis() - t1));
        System.out.println(code2.toString());
        assertEquals(
                "((\"campo1\" > 10) AND ST_Intersects('POLYGON ((0 0, 0 100, 100 100, 100 0, 0 0))'::geometry, \"GEOMETRY\"))",
                code2.toString()
        );

//        t1 = System.currentTimeMillis();
//        for (int i = 0; i < 1000; i++) {
//            interpreter.run(code2);
//        }
//        System.out.println("Optimizado: " + (System.currentTimeMillis() - t1));

    }

}
