package org.gvsig.expressionevaluator.swing.impl;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import org.gvsig.expressionevaluator.ExpressionEvaluatorLocator;
import org.gvsig.expressionevaluator.ExpressionEvaluatorManager;
import org.gvsig.expressionevaluator.SymbolTable;
import org.gvsig.expressionevaluator.SymbolTableFactory;
import org.gvsig.expressionevaluator.swing.Element;
import org.gvsig.expressionevaluator.swing.ElementFactory;
import org.gvsig.expressionevaluator.swing.ExpressionBuilderConfig;
import org.gvsig.expressionevaluator.swing.ExpressionEvaluatorSwingManager;
import org.gvsig.tools.util.PropertiesSupportHelper;

/**
 *
 * @author jjdelcerro
 */
public class DefaultExpressionBuilderConfig
        implements ExpressionBuilderConfig, ExpressionBuilderConfig.Preferences {

    private List<Element> elements;
    private int limit;
    private HashMap<String, SymbolTable> symbolTables;
    private final ExpressionEvaluatorSwingManager manager;
    private boolean scriptsEnabled = true;
    private SymbolTable previewSymbolTable;
    private boolean SQLCompatible;
    private final PropertiesSupportHelper propertiesSupportHelper;
    private boolean showLabelsOfAttributes;
    
    public DefaultExpressionBuilderConfig(ExpressionEvaluatorSwingManager manager) {
        this.manager = manager;
        this.limit = 6;
        this.SQLCompatible = false;
        this.propertiesSupportHelper = new PropertiesSupportHelper();
        this.showLabelsOfAttributes = true;
        initSymbolTables();
    }

    @Override
    public Preferences getPreferences() {
        return this;
    }

    private void initSymbolTables() {
        ExpressionEvaluatorManager theManager = ExpressionEvaluatorLocator.getManager();
        for (SymbolTableFactory factory : theManager.getSymbolTableFactories()) {
            if (factory.isAutoload()) {
                this.addSymbolTable(factory.create());
            }
        }
    }

    @Override
    public List<Element> getElements() {
        if (this.elements == null) {
            this.elements = manager.createElements(this.getSymbolTables());
            for (ElementFactory factory : manager.getElementFactories()) {
                if (factory.isAutoload()) {
                    Element e = factory.createElement();
                    this.elements.add(e);
                    e.setConfig(this);
                }
            }
        }
        return Collections.unmodifiableList(elements);
    }

    @Override
    public void addElement(Element element) {
        if( element==null ) {
            return;
        }
        this.getElements(); // Forzamos a crear la lista de elementos
        for (Element theElement : this.elements) {
            if (theElement.getName().equalsIgnoreCase(element.getName())) {
                return;
            }
        }
        this.elements.add(element);
        element.setConfig(this);
    }

    @Override
    public Collection<SymbolTable> getSymbolTables() {
        return Collections.unmodifiableCollection(this.symbolTables.values());
    }

    @Override
    public void addSymbolTable(String name) {
        ExpressionEvaluatorManager theManager = ExpressionEvaluatorLocator.getManager();
        SymbolTable symbolTable = theManager.getSymbolTable(name);
        if (symbolTable == null) {
            throw new IllegalArgumentException("Can't locate symbol table '" + name + "'.");
        }
        addSymbolTable(symbolTable);
    }

    @Override
    public void addSymbolTable(SymbolTable symbolTable) {
        if (symbolTable == null) {
            return;
        }
        if (this.elements != null) {
            throw new IllegalStateException("Can not add symbol tables once the elements are built.");
        }
        if (this.symbolTables == null) {
            this.symbolTables = new HashMap<>();
        }
        this.symbolTables.put(symbolTable.getName(), symbolTable);
    }

    @Override
    public void removeAllSymbolTables() {
        this.symbolTables.clear();
    }

    @Override
    public int getSimpleElementsLimit() {
        return this.limit;
    }

    @Override
    public void setSimpleElementsLimit(int limit) {
        this.limit = limit;
    }

    @Override
    public void removeSymbolTable(String name) {
        this.symbolTables.remove(name);
    }

//    @Override
//    public void setScriptsEnabled(boolean enabled) {
//        this.scriptsEnabled = enabled;
//    }
//
//    @Override
//    public boolean getScriptsEnabled() {
//        return this.scriptsEnabled;
//    }
//
    @Override
    public SymbolTable getPreviewSymbolTable() {
        return this.previewSymbolTable;
    }

    @Override
    public void setPreviewSymbolTable(SymbolTable symbolTable) {
        this.previewSymbolTable = symbolTable;
    }

    @Override
    public void setSQLCompatible(boolean SQLCompatible) {
        this.SQLCompatible = SQLCompatible;
    }

    @Override
    public boolean isSQLCompatible() {
        return SQLCompatible;
    }

    @Override
    public Object getProperty(String string) {
        return this.propertiesSupportHelper.getProperty(string);
    }

    @Override
    public void setProperty(String string, Object o) {
        this.propertiesSupportHelper.setProperty(string, o);
    }

    @Override
    public Map<String, Object> getProperties() {
        return this.propertiesSupportHelper.getProperties();
    }

    @Override
    public boolean getShowLabelsOfElements() {
        return this.showLabelsOfAttributes;
    }

    @Override
    public void setShowLabelsOfElements(boolean showLabels) {
        this.showLabelsOfAttributes = showLabels;
    }

    
}
