/**
 * gvSIG. Desktop Geographic Information System.
 *
 * Copyright (C) 2007-2013 gvSIG Association.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 *
 * For any additional information, do not hesitate to contact us
 * at info AT gvsig.com, or visit our website www.gvsig.com.
 */
package org.gvsig.fmap.mapcontrol.tools.Behavior;

import java.awt.Color;
import java.awt.Point;
import java.awt.event.MouseEvent;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.gvsig.fmap.geom.Geometry;
import org.gvsig.fmap.geom.GeometryException;
import org.gvsig.fmap.geom.aggregate.MultiLine;
import org.gvsig.fmap.geom.primitive.Arc;
import org.gvsig.fmap.geom.primitive.Ellipse;

import org.gvsig.fmap.geom.primitive.Line;
import org.gvsig.fmap.mapcontext.ViewPort;
import org.gvsig.fmap.mapcontrol.MapControlDrawer;
import org.gvsig.fmap.mapcontrol.tools.BehaviorException;
import org.gvsig.fmap.mapcontrol.tools.Events.MeasureEvent;
import org.gvsig.fmap.mapcontrol.tools.Listeners.EllipseListener;
import org.gvsig.fmap.mapcontrol.tools.Listeners.ToolListener;

/**
 * <p>
 * Behavior that allows user to draw a polyline by its vertexes on the image of
 * the associated <code>MapControl</code> using a
 * {@link EllipseListener EllipseListener}.</p>
 *
 * @author Vicente Caballero Navarro
 * @author Pablo Piqueras Bartolom
 */
public class EllipseBehavior extends Behavior {

    /**
     * The abscissa coordinate of all vertexes of the polyline.
     */
    protected List<Double> arrayX = new ArrayList<>();

    /**
     * The ordinate coordinate of all vertexes of the polyline.
     */
    protected List<Double> arrayY = new ArrayList<>();

    /**
     * Determines if user is setting the vertexes (with one click of the button
     * 1 of the mouse), or not.
     */
    protected boolean isClicked = false;

    /**
     * Tool listener used to work with the <code>MapControl</code> object.
     *
     * @see #getListener()
     * @see #setListener(ToolListener)
     */
    protected EllipseListener listener;
    private Point2D m_FirstPoint;
    private Point2D m_SecondPoint;
    private Point2D m_ThirdPoint;

    /**
     * <p>
     * Creates a new behavior for drawing a polyline by its vertexes.</p>
     *
     * @param mli listener used to permit this object to work with the
     * associated <code>MapControl</code>
     */
    public EllipseBehavior(EllipseListener mli) {
        super(Behavior.BUTTON_LEFT);
        listener = mli;
    }

    /*
	 * (non-Javadoc)
	 * @see com.iver.cit.gvsig.fmap.tools.Behavior.Behavior#paintComponent(java.awt.Graphics)
     */
    @Override
    public void paintComponent(MapControlDrawer mapControlDrawer) {
        mapControlDrawer.setColor(Color.black);
        if (arrayX.size() == 1) {
            drawCircle(mapControlDrawer);
        } else if (arrayX.size()  > 1) {
            drawEllipseLine(mapControlDrawer);
        }

    }

    /*
	 * (non-Javadoc)
	 * @see com.iver.cit.gvsig.fmap.tools.Behavior.Behavior#mousePressed(java.awt.event.MouseEvent)
     */
    @Override
    public void mousePressed(MouseEvent e) throws BehaviorException {

        if (this.isMyButton(e)) {
            if (arrayX.isEmpty()) {
                Point2D point = getMapControl().getViewPort().toMapPoint(e.getPoint());
                addPoint(point);
                m_FirstPoint = point;
            } else if (arrayX.size() == 1) {
                Point2D point = getMapControl().getViewPort().toMapPoint(e.getPoint());
                addPoint(point);
                m_SecondPoint = point;
            } else {
                Point2D point = getMapControl().getViewPort().toMapPoint(e.getPoint());
                addPoint(point);
                m_ThirdPoint = point;
//                Double[] x = new Double[3];
//                Double[] y = new Double[3];
//
//                x[0] = m_FirstPoint.getX();
//                x[1] = m_SecondPoint.getX();
//                x[2] = m_ThirdPoint.getX();
//
//                y[0] = m_FirstPoint.getY();
//                y[1] = m_SecondPoint.getY();
//                y[2] = m_ThirdPoint.getY();

		MeasureEvent event = new MeasureEvent((Double[]) arrayX.toArray(new Double[arrayX.size()]), (Double[]) arrayY.toArray(new Double[arrayY.size()]), e);

                listener.ellipseFinished(event);

                arrayX.clear();
                arrayY.clear();
                m_FirstPoint = null;
                m_SecondPoint = null;
                m_ThirdPoint = null;
                isClicked = false;
                getMapControl().repaint();
                return;
            }

            isClicked = true;

            getMapControl().repaint();
            if (listener.cancelDrawing()) {
                getMapControl().cancelDrawing();
                isClicked = false;
                getMapControl().repaint();
            }
        }
    }

    /*
	 * (non-Javadoc)
	 * @see com.iver.cit.gvsig.fmap.tools.Behavior.Behavior#mouseDragged(java.awt.event.MouseEvent)
     */
    @Override
    public void mouseDragged(MouseEvent e) throws BehaviorException {
        mouseMoved(e);
    }

    /**
     * <p>
     * Changes the last point added of the polyline.</p>
     *
     * @param p a point which will replace the last added to the polyline
     */
    protected void changeLastPoint(Point2D p) {
        if (arrayX.size() > 0) {
            arrayX.set(arrayX.size() - 1, p.getX());
            arrayY.set(arrayY.size() - 1, p.getY());
        }
    }

    /*
	 * (non-Javadoc)
	 * @see com.iver.cit.gvsig.fmap.tools.Behavior.Behavior#mouseMoved(java.awt.event.MouseEvent)
     */
    @Override
    public void mouseMoved(MouseEvent e) throws BehaviorException {
        if (isClicked) {
            //System.err.println("moved despues de click");
//            changeLastPoint(getMapControl().getViewPort().toMapPoint(e.getPoint()));

//            MeasureEvent event = new MeasureEvent((Double[]) arrayX.toArray(new Double[arrayX.size()]), (Double[]) arrayY.toArray(new Double[arrayY.size()]), e);
            if (arrayX.size() == 1) {
                Point2D point = getMapControl().getViewPort().toMapPoint(e.getPoint());
                m_SecondPoint = point;
                ViewPort vp = getMapControl().getMapContext().getViewPort();
                Double[] x = new Double[2];
                Double[] y = new Double[2];
//                x[0] = vp.toMapPoint(m_FirstPoint).getX();
//                x[1] = vp.toMapPoint(m_SecondPoint).getX();
//                y[0] = vp.toMapPoint(m_FirstPoint).getY();
//                y[1] = vp.toMapPoint(m_SecondPoint).getY();
                x[0] = m_FirstPoint.getX();
                x[1] = m_SecondPoint.getX();
                y[0] = m_FirstPoint.getY();
                y[1] = m_SecondPoint.getY();
                MeasureEvent event = new MeasureEvent(x, y, e);
                listener.circle(event);
            } else if (arrayX.size() == 2) {
                Point2D point = getMapControl().getViewPort().toMapPoint(e.getPoint());
                m_ThirdPoint = point;
                ViewPort vp = getMapControl().getMapContext().getViewPort();
            
                Double[] x = new Double[3];
                Double[] y = new Double[3];
//                x[0] = vp.toMapPoint(m_FirstPoint).getX();
//                x[1] = vp.toMapPoint(m_SecondPoint).getX();
//                x[2] = vp.toMapPoint(m_ThirdPoint).getX();
//
//                y[0] = vp.toMapPoint(m_FirstPoint).getY();
//                y[1] = vp.toMapPoint(m_SecondPoint).getY();
//                y[2] = vp.toMapPoint(m_ThirdPoint).getY();
                
                x[0] = m_FirstPoint.getX();
                x[1] = m_SecondPoint.getX();
                x[2] = m_ThirdPoint.getX();

                y[0] = m_FirstPoint.getY();
                y[1] = m_SecondPoint.getY();
                y[2] = m_ThirdPoint.getY();
                MeasureEvent event = new MeasureEvent(x, y, e);

                listener.ellipse(event);
            }

            getMapControl().repaint();
        }
    }

    /**
     * <p>
     * Draws the polyline in the <code>Graphics2D</code> of the associated
     * <code>MapControl</code>.</p>
     *
     * @param mapControlDrawer
     */
    protected void drawEllipseLine(MapControlDrawer mapControlDrawer) {

        double radio;
        mapControlDrawer.setColor(Color.black);

        if ((m_FirstPoint != null) && (m_SecondPoint != null) && (m_ThirdPoint != null)) {
            ViewPort vp = getMapControl().getMapContext().getViewPort();
//            Point2D p1 = vp.toMapPoint(m_FirstPoint);
//            Point2D p2 = vp.toMapPoint(m_SecondPoint);
//            Point2D p3 = vp.toMapPoint(m_ThirdPoint);
            
            Point2D p1 = m_FirstPoint;
            Point2D p2 = m_SecondPoint;
            Point2D p3 = m_ThirdPoint;
            radio = p2.distance(p3);
            if (radio > 0.0) {
                MultiLine ellipse = null;
                try {
                    ellipse = createEllipse(p1, p2, p3).toLines();
                } catch (GeometryException ex) {
                    Logger.getLogger(EllipseBehavior.class.getName()).log(Level.SEVERE, null, ex);
                }
                if (ellipse != null) {
                    mapControlDrawer.draw(ellipse);
                }
            }

        }
    }

    protected void drawCircle(MapControlDrawer mapControlDrawer) {
        double radio;
        mapControlDrawer.setColor(Color.black);

        if ((m_FirstPoint != null) && (m_SecondPoint != null)) {
            ViewPort vp = getMapControl().getMapContext().getViewPort();
//            Point2D p1 = vp.toMapPoint(m_FirstPoint);
//            Point2D p2 = vp.toMapPoint(m_SecondPoint);
            Point2D p1 = m_FirstPoint;
            Point2D p2 = m_SecondPoint;
            radio = p1.distance(p2)/2;
            if (radio > 0.0) {
                Arc arc = null;
                arc = createArc((p1.getX()+p2.getX())/2, (p1.getY()+p2.getY())/2,
                        radio, 0, Math.PI * 2);
                if (arc != null) {
                    mapControlDrawer.draw(arc);
                }
            }

        }
    }

    /**
     * <p>
     * Adds a new point to the polyline.</p>
     *
     * @param p a new point to the polyline
     */
    protected void addPoint(Point2D p) {
        arrayX.add(p.getX());
        arrayY.add(p.getY());
    }

    /**
     * <p>
     * Sets a tool listener to work with the <code>MapControl</code> using this
     * behavior.</p>
     *
     * @param listener a <code>EllipseListener</code> object for this behavior
     */
    public void setListener(ToolListener listener) {
        this.listener = (EllipseListener) listener;
    }

    /*
	 * (non-Javadoc)
	 * @see com.iver.cit.gvsig.fmap.tools.Behavior.Behavior#getListener()
     */
    @Override
    public ToolListener getListener() {
        return listener;
    }
}
