/**
 * gvSIG. Desktop Geographic Information System.
 *
 * Copyright (C) 2007-2013 gvSIG Association.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * For any additional information, do not hesitate to contact us at info AT
 * gvsig.com, or visit our website www.gvsig.com.
 */
/*
 * AUTHORS (In addition to CIT):
 * 2010 {Prodevelop}   {Task}
 */
package org.gvsig.installer.lib.impl.execution;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import org.apache.commons.io.FileUtils;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.gvsig.installer.lib.api.InstallerManager;
import org.gvsig.installer.lib.api.PackageInfo;
import org.gvsig.installer.lib.api.Version;
import org.gvsig.installer.lib.api.creation.MakePluginPackageServiceException;
import org.gvsig.installer.lib.api.execution.InstallPackageService;
import org.gvsig.installer.lib.api.execution.InstallPackageServiceException;
import org.gvsig.installer.lib.impl.DefaultInstallerManager;
import org.gvsig.installer.lib.impl.utils.Download;
import org.gvsig.installer.lib.spi.InstallPackageProviderServices;
import org.gvsig.installer.lib.spi.InstallerProviderLocator;
import org.gvsig.installer.lib.spi.InstallerProviderManager;
import org.gvsig.installer.lib.spi.execution.InstallPackageProvider;
import org.gvsig.tools.service.Manager;
import org.gvsig.tools.service.ServiceException;
import org.gvsig.tools.task.SimpleTaskStatus;

/**
 * @author <a href="mailto:jpiera@gvsig.org">Jorge Piera Llodr&aacute;</a>
 */
public class DefaultInstallPackageService extends Thread implements
        InstallPackageService {

    private static final Logger LOG = LoggerFactory
            .getLogger(DefaultInstallPackageService.class);

    private static final String PACKAGE_FILE_NAME = "packages.gvspki";

    private PackagesCache packagesCache = null;
//    private Map<PackageInfo, String> zipEntriesMap = null;
//    private List<PackageInfo> packageInfos = null;
    private InstallerManager manager;
    private InstallPackageProviderServices installerProviderServices = null;

    public DefaultInstallPackageService(DefaultInstallerManager manager) {
        super();
        this.manager = manager;
        this.reset();
    }

    public void reset() {
//        packageInfos = new ArrayList<PackageInfo>();
//        zipEntriesMap = new HashMap<PackageInfo, String>();
        installerProviderServices = InstallerProviderLocator
                .getProviderManager().createInstallerProviderServices();
        packagesCache = new PackagesCache(installerProviderServices);
    }

    public class InstallerApplicationDirectoryNotFoundException extends
            InstallPackageServiceException {

        private static final long serialVersionUID = -1130408094135962456L;

        private static final String message = "Aplication directory '%(directory)s' not found";

        private static final String KEY = "_aplication_directory_XdirectoryX_not_found";

        public InstallerApplicationDirectoryNotFoundException(File file) {
            super(message, KEY, serialVersionUID);
            setValue("directory", file.toString());
        }

    }

    public class InstallerNoDirectoryException extends
            InstallPackageServiceException {

        private static final long serialVersionUID = -8685263049644983769L;

        private static final String message = "'%(directory)s' is not a directory";

        private static final String KEY = "_XdirectoryX_is_not_a_directory";

        public InstallerNoDirectoryException(File file) {
            super(message, KEY, serialVersionUID);
            setValue("directory", file.toString());
        }

    }

    public class InstallerFileNotFoundException extends
            InstallPackageServiceException {

        private static final long serialVersionUID = 556517830330132149L;

        private static final String message = "File '%(file)s' not found";

        private static final String KEY = "_file_XfileX_not_found";

        public InstallerFileNotFoundException(File file) {
            super(message, KEY, serialVersionUID);
            setValue("file", file.toString());
        }

    }

    public class InstallerBundleNotFoundException extends
            InstallPackageServiceException {

        private static final long serialVersionUID = 5065410511582625301L;

        private static final String message = "File '%(file)s' not found";

        private static final String KEY = "_file_XfileX_not_found";

        public InstallerBundleNotFoundException(File file,
                FileNotFoundException e) {
            super(message, e, KEY, serialVersionUID);
            setValue("file", file.toString());
        }

    }

    public class InstallerIOException extends InstallPackageServiceException {

        private static final long serialVersionUID = 3153613550157712363L;

        private static final String message = "IO error installing the file '%(file)s'";

        private static final String KEY = "_IO_error installing_file_XfileX_";

        public InstallerIOException(File file, IOException e) {
            super(message, e, KEY, serialVersionUID);
            setValue("file", file.toString());
        }

    }

    public class InstallerFileDownloadException extends
            InstallPackageServiceException {

        private static final long serialVersionUID = 8640183295766490512L;

        private static final String message = "File '%(url)s' download error";

        private static final String KEY = "_File_XurlX_download_error";

        public InstallerFileDownloadException(URL url, IOException e) {
            super(message, e, KEY, serialVersionUID);
            setValue("url", url.toString());
        }

    }

    public class InstallerPackageNotFoundException extends
            InstallPackageServiceException {

        private static final long serialVersionUID = 1726608498886963868L;

        private static final String message = "Package not found";

        private static final String KEY = "_package_not_found";

        public InstallerPackageNotFoundException() {
            super(message, KEY, serialVersionUID);
        }

    }

    public class InstallerNoPackageException extends
            InstallPackageServiceException {

        private static final long serialVersionUID = -2292735515704746966L;

        private static final String message = "Package does not exist";

        private static final String KEY = "_package__does_not_exist";

        public InstallerNoPackageException() {
            super(message, KEY, serialVersionUID);
        }

    }

    public class InstallerProviderCreationException extends
            InstallPackageServiceException {

        private static final long serialVersionUID = -7985786807492393584L;

        private static final String message = "Error creating the provider";

        private static final String KEY = "_Error_creating_the_provider";

        public InstallerProviderCreationException(ServiceException e) {
            super(message, e, KEY, serialVersionUID);
        }

    }

    @Override
    public void installPackage(File applicationDirectory,
            PackageInfo packageInfo) throws InstallPackageServiceException {
        if (!applicationDirectory.exists()) {
            LOG.warn("Can install package '" + packageInfo.getCode() + "', application folder '" + applicationDirectory.toString() + "' does not exits.");
            throw new InstallerApplicationDirectoryNotFoundException(
                    applicationDirectory);
        }
        if (!packagesCache.contains(packageInfo)) {
            LOG.warn("Can install package '" + packageInfo.getCode() + "', package not found.");
            throw new InstallerPackageNotFoundException();
        }

        InstallPackageProvider installerExecutionProvider = createProvider(packageInfo);

        File file = packagesCache.getFile(packageInfo);
        try {
            InputStream packageStream = this.packagesCache.getInputStream(packageInfo);
            try {
                installerExecutionProvider.install(applicationDirectory,
                        packageStream, packageInfo);
            } catch (InstallPackageServiceException e) {
                IOUtils.closeQuietly(packageStream);
                packageStream = this.packagesCache.getInputStream(packageInfo);
                installerExecutionProvider.installLater(applicationDirectory,
                        packageStream, packageInfo);
            } finally {
                IOUtils.closeQuietly(packageStream);
            }

        } catch (FileNotFoundException e) {
            throw new InstallerFileNotFoundException(file);
        } catch (IOException e) {
            throw new InstallerIOException(file, e);
        }

    }

    @Override
    public void installPackage(File applicationDirectory, String packageCode)
            throws InstallPackageServiceException {
        PackageInfo packageInfo = getPackageInfo(packageCode);
        if (packageInfo == null) {
            throw new InstallerNoPackageException();
        }
        installPackage(applicationDirectory, packageInfo);
    }

    private InstallPackageProvider createProvider(PackageInfo packageInfo)
            throws InstallPackageServiceException {
        InstallerProviderManager installerProviderManager = (InstallerProviderManager) ((DefaultInstallerManager) manager)
                .getProviderManager();

        try {
            return installerProviderManager.createExecutionProvider(packageInfo
                    .getType());
        } catch (ServiceException e) {
            throw new InstallerProviderCreationException(e);
        }
    }

    @Override
    public PackageInfo getPackageInfo(int index) {
        if (index >= packagesCache.size()) {
            return null;
        }
        return packagesCache.get(index);
    }

    @Override
    public PackageInfo getPackageInfo(String packageCode) {
        for( PackageInfo packageInfo : packagesCache ) {
            if (packageInfo.getCode().equals(packageCode)) {
                return packageInfo;
            }
        }
        return null;
    }

    @Override
    public void addBundle(File bundle, SimpleTaskStatus taskStatus) throws InstallPackageServiceException {
        Map<PackageInfo, String> zipEntriesNames = new HashMap<>();
        List<PackageInfo> packages = new ArrayList<>();
        
        if (!bundle.exists()) {
            throw new InstallPackageServiceException();
        }

        BufferedInputStream bis = null;
        FileInputStream fis = null;
        try {
            fis = new FileInputStream(bundle);
            bis = new BufferedInputStream(fis);
            installerProviderServices.readPackageInfo(
                bis, packages, zipEntriesNames, bundle.getName(), taskStatus
            );
        } catch (FileNotFoundException e) {
            throw new InstallerBundleNotFoundException(bundle, e);
        } finally {
            IOUtils.closeQuietly(bis);
            IOUtils.closeQuietly(fis);
            bis = null;
            fis = null;
        }
        
        if( packages.size()==1 ) {
            PackageInfo pkg = packages.get(0);
            String entryName = zipEntriesNames.get(pkg);
            if( entryName.endsWith("."+manager.getDefaultPackageFileExtension()) ) {
                // Es un xxx.gvspkg
                packagesCache.addPackage(pkg, bundle);
                return;
            } else if( entryName.endsWith("."+manager.getDefaultIndexSetFileExtension()) ) {
                // Es un xxx.gvspki
                packagesCache.addPackageIndex(pkg);
            } else {
                throw new IllegalArgumentException("Can't recognize the file '"+bundle.getAbsolutePath()+"'.");
            }
        }
        
        packages.clear();
        zipEntriesNames.clear();
        try {
            fis = new FileInputStream(bundle);
            bis = new BufferedInputStream(fis);
            installerProviderServices.readPackageSetInfo(
                bis, packages, zipEntriesNames, taskStatus
            );
        } catch (FileNotFoundException e) {
            throw new InstallerBundleNotFoundException(bundle, e);
        } finally {
            IOUtils.closeQuietly(bis);
            IOUtils.closeQuietly(fis);
            bis = null;
            fis = null;
        }
        for( PackageInfo pkg : packages ) {
            String entryName = zipEntriesNames.get(pkg);
            if( entryName.endsWith("."+manager.getDefaultPackageFileExtension()) ) {
                // Es un xxx.gvspkg
                packagesCache.addBundledPackage(pkg, bundle, entryName);
            } else if( entryName.endsWith("."+manager.getDefaultIndexSetFileExtension()) ) {
                // Es un xxx.gvspki
                packagesCache.addBundledPackageIndex(pkg, bundle, entryName);
            } else {
                throw new IllegalArgumentException("Can't recognize the file '"+bundle.getAbsolutePath()+"'.");
            }
        }
    }

    @Override
    public void addBundle(URL bundleURL, SimpleTaskStatus taskStatus) throws InstallPackageServiceException {
        File bundle;
        String urlString = bundleURL.toString();
        if (urlString.endsWith(InstallerManager.PACKAGE_EXTENSION) || urlString.endsWith(InstallerManager.PACKAGE_INDEX_EXTENSION)) {
            manager.setDownloadBaseURL(bundleURL);
            bundle = downloadFile(bundleURL, PACKAGE_FILE_NAME);
            addBundle(bundle,taskStatus);
        } else {
            if (!urlString.endsWith("/")) {
                urlString += "/";
            }

            Version version = manager.getVersionEx();
            urlString += ("dists/"
                    + version.getMayor() + "." + version.getMinor() + "."
                    + version.getRevision() + "/" + PACKAGE_FILE_NAME);

            URL completeURL;
            try {
                completeURL = new URL(urlString);
            } catch (MalformedURLException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
                return;
            }

            manager.setDownloadBaseURL(completeURL);
            bundle = downloadFile(completeURL, PACKAGE_FILE_NAME);
            addBundle(bundle,taskStatus);
        }

    }

    private File downloadFile(URL bundleURL, String defaultFileName)
            throws InstallPackageServiceException {
        try {
            Download download = new Download();
            return download.downloadFile(bundleURL, defaultFileName);
        } catch (IOException e) {
            throw new InstallerFileDownloadException(bundleURL, e);
        }
    }

    @Override
    public void addBundlesFromDirectory(File directory, SimpleTaskStatus taskStatus)
            throws InstallPackageServiceException {
        if (!directory.isDirectory()) {
            throw new InstallerNoDirectoryException(directory);
        }
        Collection<File> files = FileUtils.listFiles(
            directory, 
            new String[] {
                manager.getDefaultPackageFileExtension(),
                manager.getDefaultPackageSetFileExtension(),
                manager.getDefaultIndexSetFileExtension()
            },
            true
        );
        for( File file : files ) {
            addBundle(file,taskStatus);
        }
    }

    @Override
    public int getPackageCount() {
        if (packagesCache == null) {
            return 0;
        }
        return packagesCache.size();
    }

    @Override
    public Manager getManager() {
        return this.manager;
    }

    public void downloadPackage(PackageInfo packageInfo)
            throws InstallPackageServiceException {
        this.downloadPackage(packageInfo, null);
    }

    @Override
    public void downloadPackage(PackageInfo packageInfo,
            SimpleTaskStatus taskStatus) throws InstallPackageServiceException {
        packagesCache.download(packageInfo, taskStatus);
    }

    @Override
    public List<String> getDefaultSelectedPackagesIDs() {
        return installerProviderServices.getDefaultSelectedPackagesIDs();
    }

    @Override
    public Properties getProperties() {
        return installerProviderServices.getProperties();
    }

    @Override
    public List<String> getCategories() {
        Set<String> categories = new HashSet<>();
        for( PackageInfo pkginfo : packagesCache ) {
            List<String> pkgcategories = pkginfo.getCategories();
            categories.addAll(pkgcategories);
        }
        try {
            PackageInfo[] pkgs = manager.getInstalledPackages();
            for( PackageInfo pkginfo : pkgs ) {
                List<String> pkgcategories = pkginfo.getCategories();
                categories.addAll(pkgcategories);
            }

        } catch (MakePluginPackageServiceException e) {
            // Ignore exceptions
        }
        ArrayList<String> l = new ArrayList<>(categories);
        Collections.sort(l);
        return l;
    }

    @Override
    public List<String> getTypes() {
        Set<String> types = new HashSet<>();

        for( PackageInfo pkginfo : packagesCache ) {
            types.add(pkginfo.getType());
        }
        try {
            PackageInfo[] pkgs = manager.getInstalledPackages();
            for( PackageInfo pkginfo : pkgs ) {
                types.add(pkginfo.getType());
            }

        } catch (MakePluginPackageServiceException e) {
            // Ignore exceptions
        }
        List<String> l = new ArrayList<>(types);
        Collections.sort(l);
        return l;
    }

    @Override
    public boolean needInstallPackageProviders() {
        boolean need = false;
        String ss = this.getProperties().getProperty("installers");
        if (ss != null) {
            String[] installerNames = StringUtils.split(ss);
            for (String installerName : installerNames) {
                InstallerProviderInstallers installer = new InstallerProviderInstallers(this, installerName);
                if( !installer.isAlreadyRegistered() ) {
                    need = true;
                    break;
                }
            }
        }
        return need;
    }

    @Override
    public void installPackageProviders(SimpleTaskStatus taskStatus) {
        Properties props = this.getProperties();
        String ss = props.getProperty("installers");
        if (ss == null) {
            return;
        }

        try {
            String[] installerNames = StringUtils.split(ss);
            taskStatus.setRangeOfValues(0, installerNames.length);
            int count=0;
            for (String installerName : installerNames) {
                taskStatus.message(installerName);
                taskStatus.setCurValue(count++);
                InstallerProviderInstallers installer = new InstallerProviderInstallers(this, installerName);
                installer.install();
            }
            taskStatus.setCurValue(count++);
        } finally {
            taskStatus.terminate();
        }
    }

}
