
package org.gvsig.installer.lib.impl.execution;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import org.gvsig.installer.lib.api.PackageInfo;
import org.gvsig.installer.lib.spi.InstallPackageProviderServices;
import org.gvsig.tools.task.SimpleTaskStatus;


public class PackagesCache implements Iterable<PackageInfo> {

    private interface CacheEntry {
        public PackageInfo getPackage();
        public InputStream getInputStream();
        public File getFile();
        public void download(SimpleTaskStatus taskStatus);
        public boolean isDownloaded();
    }
    
    private class PackageCacheEntry implements CacheEntry {
        private final PackageInfo pkg;
        private final File file;
        
        PackageCacheEntry(PackageInfo pkg, File file) {
            this.pkg = pkg;
            this.file = file;
        }

        @Override
        public PackageInfo getPackage() {
            return this.pkg;
        }
        
        @Override
        public InputStream getInputStream() {
            InputStream bis;
            try {
                FileInputStream fis = new FileInputStream(file);
                bis = new BufferedInputStream(fis);
            } catch (FileNotFoundException ex) {
                throw new RuntimeException("Can't open package of '"+pkg.getCode()+"'.", ex);
            }
            return bis;
        }

        @Override
        public File getFile() {
            return this.file;
        }

        @Override
        public void download(SimpleTaskStatus taskStatus) {
        }

        @Override
        public boolean isDownloaded() {
            return true;
        }
    }
    
    private class IndexCacheEntry implements CacheEntry {
        private final PackageInfo pkg;
        private File file;
        
        IndexCacheEntry(PackageInfo pkg) {
            this.pkg = pkg;
            this.file = null;
        }

        @Override
        public PackageInfo getPackage() {
            return this.pkg;
        }
        
        @Override
        public void download(SimpleTaskStatus taskStatus) {
            if( file == null ) {
                try {
                    file = pkg.downloadFile(taskStatus);
                } catch (Exception ex) {
                    throw new RuntimeException("Can't download package of '"+pkg.getCode()+"'.", ex);
                }
            }
        }

        @Override
        public File getFile() {
            this.download(null);
            return this.file;
        }
        
       @Override
        public InputStream getInputStream() {
            InputStream bis;
            try {
                if( file == null ) {
                    this.download(null);
                }
                FileInputStream fis = new FileInputStream(this.file);
                bis = new BufferedInputStream(fis);
            } catch (FileNotFoundException ex) {
                throw new RuntimeException("Can't open package of '"+pkg.getCode()+"'.", ex);
            }
            return bis;
        }

        @Override
        public boolean isDownloaded() {
            return this.file!=null;
        }        
    }

    private class BundledPackageCacheEntry implements CacheEntry {
        private final PackageInfo pkg;
        private final File bundleFile;
        private final String bundleEntryName;
        
        BundledPackageCacheEntry(PackageInfo pkg, File bundleFile, String bundleEntryName) {
            this.pkg = pkg;
            this.bundleFile = bundleFile;
            this.bundleEntryName = bundleEntryName;
        }

        @Override
        public PackageInfo getPackage() {
            return this.pkg;
        }
        
        @Override
        public File getFile() {
            return this.bundleFile;
        }

        @Override
        public void download(SimpleTaskStatus taskStatus) {
        }

        @Override
        public InputStream getInputStream() {
            InputStream is;
            try {
                FileInputStream fis = new FileInputStream(bundleFile);
                BufferedInputStream bis = new BufferedInputStream(fis);
                is = installerProviderServices.searchPackage(bis,bundleEntryName);
            } catch (Exception ex) {
                throw new RuntimeException("Can't open package of '"+pkg.getCode()+"'.", ex);
            }
            return is;
        }

        @Override
        public boolean isDownloaded() {
            return true;
        }        
    }

    private class BundledIndexCacheEntry extends IndexCacheEntry {

        private final File bundleFile;
        
        BundledIndexCacheEntry(PackageInfo pkg, File bundleFile, String bundleEntryName) {
            super(pkg);
            this.bundleFile = bundleFile;
        }

        @Override
        public File getFile() {
            return this.bundleFile;
        }
    }

    private List<CacheEntry> cacheEntries = null;
    private Map<PackageInfo, CacheEntry> package2cacheEntry = null;
    private InstallPackageProviderServices installerProviderServices;
    
    public PackagesCache(InstallPackageProviderServices installerProviderServices) {
        this.installerProviderServices = installerProviderServices;
        this.package2cacheEntry = new HashMap<>();
        this.cacheEntries = new ArrayList<>();
    }
    
    public void addPackage(PackageInfo pkg, File f) {
        CacheEntry cacheEntry = new PackageCacheEntry(pkg, f);
        this.cacheEntries.add(cacheEntry);
        this.package2cacheEntry.put(pkg, cacheEntry);
    }

    public void addPackageIndex(PackageInfo pkg) {
        CacheEntry cacheEntry = new IndexCacheEntry(pkg);
        this.cacheEntries.add(cacheEntry);
        this.package2cacheEntry.put(pkg, cacheEntry);
    }

    public void addBundledPackage(PackageInfo pkg, File bundleFile, String bundleEntryName) {
        CacheEntry cacheEntry = new BundledPackageCacheEntry(pkg, bundleFile, bundleEntryName);
        this.cacheEntries.add(cacheEntry);
        this.package2cacheEntry.put(pkg, cacheEntry );
    }

    public void addBundledPackageIndex(PackageInfo pkg, File bundleFile, String bundleEntryName) {
        CacheEntry cacheEntry = new BundledIndexCacheEntry(pkg, bundleFile, bundleEntryName);
        this.cacheEntries.add(cacheEntry);
        this.package2cacheEntry.put(pkg, cacheEntry );
    }

    /**
     * Return the file associated to this package.
     * If the package is a standalone package it return the file of package.
     * If the package is a standalone package index, return the file of the 
     * associated package, downloading it if need.
     * If is a package in a bundle, return the file of the bundle.
     * 
     * @param packageInfo
     * @return 
     */
    public File getFile(PackageInfo packageInfo) {
        CacheEntry entry = this.package2cacheEntry.get(packageInfo);
        return entry.getFile();
    }

    public boolean contains(PackageInfo packageInfo) {
        return this.package2cacheEntry.containsKey(packageInfo);
    }

    public boolean isDownloaded(PackageInfo packageInfo) {
        CacheEntry entry = this.package2cacheEntry.get(packageInfo);
        return entry.isDownloaded();
    }
    
    public void download(PackageInfo packageInfo, SimpleTaskStatus taskStatus) {
        CacheEntry entry = this.package2cacheEntry.get(packageInfo);
        entry.download(taskStatus);
    }

    /**
     * Return the InputStream of the package.
     * If is a package index, download if need the package and return 
     * a InputStream of the package associated to the index.
     * If is a package in a bundle, return a InputStream without extract 
     * the package of the bundle.
     * 
     * @param packageInfo
     * @return 
     */
    public InputStream getInputStream(PackageInfo packageInfo) {
        CacheEntry entry = this.package2cacheEntry.get(packageInfo);
        return entry.getInputStream();
    }

    @Override
    public Iterator<PackageInfo> iterator() {
        return new Iterator<PackageInfo>() {
            Iterator<CacheEntry> it = cacheEntries.iterator();
            @Override
            public boolean hasNext() {
                return it.hasNext();
            }

            @Override
            public PackageInfo next() {
                CacheEntry x = it.next();
                return x.getPackage();
            }

            @Override
            public void remove() {
                throw new UnsupportedOperationException();
            }
        };
    }

    public int size() {
        return this.package2cacheEntry.size();
    }
    
    public PackageInfo get(int n) {
        return this.cacheEntries.get(n).getPackage();
    }
}
