/**
 * gvSIG. Desktop Geographic Information System.
 *
 * Copyright (C) 2007-2013 gvSIG Association.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * For any additional information, do not hesitate to contact us at info AT
 * gvsig.com, or visit our website www.gvsig.com.
 */
package org.gvsig.installer.swing.impl.execution.wizard;

import java.util.ArrayList;
import java.util.List;

import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.SwingUtilities;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.gvsig.gui.beans.wizard.WizardPanelWithLogo;
import org.gvsig.gui.beans.wizard.panel.NotContinueWizardException;
import org.gvsig.gui.beans.wizard.panel.OptionPanel;
import org.gvsig.i18n.Messages;
import org.gvsig.installer.lib.api.PackageInfo;
import org.gvsig.installer.lib.api.execution.InstallPackageService;
import org.gvsig.installer.swing.api.SwingInstallerLocator;
import org.gvsig.installer.swing.api.SwingInstallerManager;
import org.gvsig.installer.swing.impl.DefaultSwingInstallerManager;
import org.gvsig.installer.swing.impl.execution.DefaultInstallWizardPanel;
import org.gvsig.installer.swing.impl.execution.InstallWizardPanel_ext;
import org.gvsig.installer.swing.impl.panel.DefaultProgressPanel;
import org.gvsig.tools.swing.api.ToolsSwingLocator;
import org.gvsig.tools.task.AbstractMonitorableTask;
import org.gvsig.tools.task.SimpleTaskStatus;

/**
 * @author gvSIG Team
 * @version $Id$
 *
 */
public class DownloadProgressWizardPage extends DefaultProgressPanel implements
        OptionPanel {

    private static final long serialVersionUID = -7064334977678611609L;
    private static final Logger logger = LoggerFactory.getLogger(DownloadProgressWizardPage.class);

    private final InstallWizardPanel_ext wizardPanel;

    public DownloadProgressWizardPage(InstallWizardPanel_ext wizardPanel) {
        super();
        this.wizardPanel = wizardPanel;
    }

    @Override
    public JPanel getJPanel() {
        return this;
    }

    @Override
    public String getPanelTitle() {
        return wizardPanel.getTranslation("_download_progress");
    }

    @Override
    public void lastPanel() {
        wizardPanel.setFinishButtonEnabled(false);
        wizardPanel.setCancelButtonEnabled(true);
    }

    @Override
    public void nextPanel() throws NotContinueWizardException {
    }

    @Override
    public void updatePanel() {
        if( wizardPanel.getDirection() == WizardPanelWithLogo.ACTION_PREVIOUS ) {
            wizardPanel.skip();
            return;
        }
        boolean AreTherePackagesToDownload = false;

        this.wizardPanel.setNextButtonEnabled(false);
        this.wizardPanel.setBackButtonEnabled(false);
        this.wizardPanel.setFinishButtonEnabled(false);

        InstallPackageService service = wizardPanel.getInstallerExecutionService();
        service.getPackageCount();

        List<PackageInfo> packagesToInstall = wizardPanel.getPackagesToInstall();
        for (PackageInfo packageToInstall : packagesToInstall) {
            if( packageToInstall.getDownloadURL() != null ) {
                AreTherePackagesToDownload = true;
                break;
            }
        }

        if (!AreTherePackagesToDownload) {
            wizardPanel.skip();
        }
        
        UpdatePanel task = new UpdatePanel(this);
        wizardPanel.addCancellableTask(task);
        bind(task.getTaskStatus());
        task.setDaemon(true);
        task.start();
        
    }

    private String getText(String key) {
        return this.wizardPanel.getTranslation(key);
    }

    private static class UpdatePanel extends AbstractMonitorableTask {

        private final DownloadProgressWizardPage panel;

        public UpdatePanel(DownloadProgressWizardPage panel) {
            super(panel.getText("_downloading_files"));
            this.panel = panel;
        }

        @Override
        public synchronized void run() {
            SimpleTaskStatus taskStatus = (SimpleTaskStatus) this.getTaskStatus();
            final List<String> not_downloaded = new ArrayList<>();
            try {

                List<PackageInfo> packagesToInstall = this.panel.wizardPanel.getPackagesToInstall();
                InstallPackageService installerService = this.panel.wizardPanel.getInstallerExecutionService();

                taskStatus.setRangeOfValues(0, packagesToInstall.size());
                for (int i = 0; i < packagesToInstall.size(); i++) {
                    taskStatus.setCurValue(i);
                    PackageInfo packageInfo = packagesToInstall.get(i);
                    taskStatus.message(packageInfo.getName());
                    if (packageInfo.getDownloadURL() != null) {
                        try {
                            installerService.downloadPackage(
                                    packageInfo,
                                    taskStatus
                            );
                        } catch (Exception e) {
                            not_downloaded.add(packageInfo.getName());
                            logger.info("Can't download package '"+packageInfo.getName()+"'.", e);
                        }
                    }

                    if (not_downloaded.size() > 0) {
                        break;
                    }
                }
                SwingUtilities.invokeAndWait(new Runnable() {

                    @Override
                    public void run() {
                        if (not_downloaded.isEmpty()) {
                            panel.wizardPanel.setNextButtonEnabled(true);
                            panel.wizardPanel.doAction(WizardPanelWithLogo.ACTION_NEXT);
                        } else {
                            panel.wizardPanel.setNextButtonEnabled(false);
                        }
                        panel.wizardPanel.setBackButtonEnabled(true);
                    }
                });                

            } catch(Exception ex) {
                logger.warn("Problems downloading packages.", ex);
                ToolsSwingLocator.getThreadSafeDialogsManager().messageDialog(
                        panel.getText("_Cant_download_packages"), 
                        panel.getText("_Warning"), 
                        JOptionPane.WARNING_MESSAGE
                );
            } finally {
                if (not_downloaded.isEmpty()) {
                    taskStatus.message(panel.getText("_finished"));
                } else {
                    taskStatus.message(panel.getText("_finished_Some_files_not_downloaded"));
                    showNotDownloaded(not_downloaded);
                }
                taskStatus.terminate();
                taskStatus.remove();
            }
        }

    }

    private static synchronized void showNotDownloaded(final List<String> names) {

        int sz = names.size();
        String list_str = "";
        for (int i = 0; i < sz; i++) {
            list_str = list_str + "\n - " + names.get(i);
        }

        list_str = Messages.getText("_These_packages_not_downloaded")
                + ":\n" + list_str + "\n";
        ToolsSwingLocator.getThreadSafeDialogsManager().messageDialog(
                list_str, 
                Messages.getText("_Warning"), 
                JOptionPane.WARNING_MESSAGE
        );
    }


}
