/**
 * gvSIG. Desktop Geographic Information System.
 *
 * Copyright (C) 2007-2013 gvSIG Association.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 *
 * For any additional information, do not hesitate to contact us
 * at info AT gvsig.com, or visit our website www.gvsig.com.
 */
package org.gvsig.symbology.fmap.mapcontext.rendering.symbol.style;

import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.geom.Rectangle2D;
import java.io.IOException;
import java.net.URL;


import org.gvsig.fmap.mapcontext.rendering.symbols.ISymbol;
import org.gvsig.fmap.mapcontext.rendering.symbols.SymbolDrawingException;
import org.gvsig.svgsupport.SVGRenderer;
import org.gvsig.tools.ToolsLocator;
import org.gvsig.tools.dynobject.DynStruct;
import org.gvsig.tools.persistence.PersistenceManager;
import org.gvsig.tools.persistence.PersistentState;
import org.gvsig.tools.persistence.exception.PersistenceException;
import org.gvsig.tools.util.Callable;
import org.gvsig.tools.util.ToolsUtilLocator;

/**
 * Style for a SVG files.This is a XML specification and file format for
 * describing two-dimensional vector graphics, both static and animated.
 * SVG can be purely declarative or may include scripting. Images can contain
 * hyperlinks using outbound simple XLinks.It is an open standard created
 * by the World Wide Web Consortium..
 * 
 */
public class SVGStyle extends BackgroundFileStyle {

    public static final String SVG_STYLE_PERSISTENCE_DEF_NAME = "SVGStyle";
    private static final String SOURCE = "source";
    private final SVGRenderer renderer;

    /**
     * Constructor method
     * 
     */
    public SVGStyle() {
        this.renderer = ToolsUtilLocator.getSVGSupportManager().createRenderer();
    }

    @Override
    public void drawInsideRectangle(Graphics2D g,
        Rectangle rect,
        boolean keepAspectRatio) throws SymbolDrawingException {
        this.renderer.drawInsideRectangle(g, rect, keepAspectRatio);
    }

    @Override
    public boolean isSuitableFor(ISymbol symbol) {
        return true;
    }

    @Override
    public void setSource(URL url) throws IOException {

    	source = url;
        this.renderer.setSource(url);
    }

    @Override
    public Rectangle getBounds() {
        try {
            Rectangle2D r = this.renderer.getBounds();
            return new Rectangle((int) r.getX(),
                (int) r.getY(),
                (int) r.getWidth(),
                (int) r.getHeight());
        } catch (Exception e) {
            return new Rectangle();
        }
    }

    @Override
    public void drawOutline(Graphics2D g, Rectangle r) throws SymbolDrawingException {
        drawInsideRectangle(g, r);
    }

    
    @Override
    public void loadFromState(PersistentState state) throws PersistenceException {
    	try {
    		String sourceSymbolInLibrary = state.getString(SOURCE_SYMBOL_IN_LIBRARY);
    		if (sourceSymbolInLibrary != null){
    			setSource(new URL(getSymbolLibraryURL().toString()+sourceSymbolInLibrary));
    		} else {
    			setSource(state.getURL(SOURCE));
    		}
    	} catch (Exception e) {
    		throw new PersistenceCantSetSourceException(e);
    	}
    }

    @Override
    public void saveToState(PersistentState state) throws PersistenceException {
    	if (isLibrarySymbol()){
    		state.set(SOURCE_SYMBOL_IN_LIBRARY, getSourceSymbolInLibrary());
    	} else {
    		state.setNull(SOURCE_SYMBOL_IN_LIBRARY);
    	}
        state.set(SOURCE, source);
    }

    public static class RegisterPersistence implements Callable {

        @Override
        public Object call() throws Exception {
            PersistenceManager manager = ToolsLocator.getPersistenceManager();
            if (manager.getDefinition(SVG_STYLE_PERSISTENCE_DEF_NAME) == null) {
                DynStruct definition =
                    manager.addDefinition(SVGStyle.class,
                        SVG_STYLE_PERSISTENCE_DEF_NAME,
                        SVG_STYLE_PERSISTENCE_DEF_NAME
                            + " Persistence definition",
                        null,
                        null);

                // Extend the Style base definition
                definition.extend(manager.getDefinition(BACKGROUND_FILE_STYLE_PERSISTENCE_DEFINITION_NAME));

                definition.addDynFieldURL(SOURCE).setMandatory(true);
                definition.addDynFieldString(SOURCE_SYMBOL_IN_LIBRARY).setMandatory(false);
            }
            return Boolean.TRUE;
        }

    }    
}
