/**
 * gvSIG. Desktop Geographic Information System.
 *
 * Copyright (C) 2007-2013 gvSIG Association.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 *
 * For any additional information, do not hesitate to contact us
 * at info AT gvsig.com, or visit our website www.gvsig.com.
 */
package org.gvsig.app.gui.styling;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.GradientPaint;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.List;

import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JList;
import javax.swing.ListCellRenderer;

import org.gvsig.gui.ColorTablePainter;
import org.gvsig.i18n.Messages;
import org.gvsig.symbology.swing.SymbologySwingLocator;

public class ColorSchemeSelector {

    private List<ColorTablePainter> colorTables = new ArrayList<>();

    private final JComboBox wrappedComboBox;

    /**
     * Constructor method
     *
     * @param wrappedComboBox
     * @param interpolateValues
     */
    public ColorSchemeSelector(final JComboBox wrappedComboBox, boolean interpolateValues) {
        this.wrappedComboBox = wrappedComboBox;
        this.wrappedComboBox.removeAllItems();
        this.wrappedComboBox.setPreferredSize(new Dimension(150, 20));
        
        // Remove listeners to prevent strange behavior when you call this class
        // repeatedly with the same combo.
        for( ActionListener l : this.wrappedComboBox.getActionListeners()) {
            this.wrappedComboBox.removeActionListener(l);
        }   
        this.colorTables = SymbologySwingLocator.getSwingManager().createColorTables();
        if( colorTables != null ) {
            for( int i = 0; i < colorTables.size(); i++ ) {
                ColorTablePainter colorTable = colorTables.get(i);
                wrappedComboBox.addItem(colorTable);
            }
            this.wrappedComboBox.addActionListener(new ActionListener() {
                @Override
                public void actionPerformed(ActionEvent e) {
                    doUpdateToolTip();
                }
            });
            this.wrappedComboBox.setRenderer(new ListCellRenderer() {
                @Override
                public Component getListCellRendererComponent(JList list, Object value, int index, boolean isSelected, boolean cellHasFocus) {
                    ColorTablePainter colorTable = (ColorTablePainter) value;
                    ColorSchemeItemPainter paintItem = new ColorSchemeItemPainter(colorTable.getTableName(), colorTable, isSelected);
                    paintItem.setPreferredSize(wrappedComboBox.getPreferredSize());
                    return paintItem;
                }
            });
        }

    }

    private void doUpdateToolTip() {
        ColorTablePainter colorTable = (ColorTablePainter) this.wrappedComboBox.getSelectedItem();
        if( colorTable != null ) {
            this.wrappedComboBox.setToolTipText(colorTable.getTableName());
        } else {
            this.wrappedComboBox.setToolTipText(Messages.getText("select_a_color_scheme"));
        }
    }

    /**
     * Returns an array composed with the selected colors
     *
     * @return
     */
    public Color[] getSelectedColors() {
        ColorTablePainter colorTable = (ColorTablePainter) this.wrappedComboBox.getSelectedItem();
        if( colorTable == null ) {
            return null;
        }
        return colorTable.getColors();
    }

    public void setSelectedColors(Color[] colors) {

        if( colors == null ) {
            this.wrappedComboBox.setSelectedIndex(0);
            return;
        }
        colorTables.clear();
        for( int i = 0; i < this.wrappedComboBox.getItemCount(); i++ ) {
            colorTables.add((ColorTablePainter) this.wrappedComboBox.getItemAt(i));
        }

        int n = 0;
        int colorsCode = getCode(colors);
        for( ColorTablePainter colorTable : colorTables ) {
            Color[] curcolors = colorTable.getColors();
            if( getCode(curcolors) == colorsCode ) {
                this.wrappedComboBox.setSelectedIndex(n);
                this.wrappedComboBox.repaint();
                return;
            }
            n++;
        }
        if( this.wrappedComboBox.getItemCount() > 0 ) {
            this.wrappedComboBox.setSelectedItem(0);
        }

    }
    
    private int getCode(Color[] table) {
        StringBuilder builder = new StringBuilder();
        for( Color color : table ) {
            builder.append(color.getRed());
            builder.append(color.getGreen());
            builder.append(color.getBlue());
            builder.append(color.getAlpha());
        }
        return builder.toString().hashCode();
    }

    private class ColorSchemeItemPainter extends JComponent {

        private static final long serialVersionUID = -6448740563809113949L;
        private boolean isSelected = false;
        private ColorTablePainter colorTablePaint = null;

        /**
         * Constructor method
         *
         * @param name
         * @param colorTablePaint
         * @param isSelected
         */
        public ColorSchemeItemPainter(String name, ColorTablePainter colorTablePaint, boolean isSelected) {
            super();
            this.colorTablePaint = colorTablePaint;
            this.isSelected = isSelected;
            setToolTipText(name);
        }

        @Override
        public void paintComponent(Graphics g) {
            Graphics2D g2 = (Graphics2D) g;
            if( isSelected ) {
                Color color1 = new Color(89, 153, 229);
                Color color2 = new Color(31, 92, 207);
                g2.setPaint(new GradientPaint(0, 1, color1, 0, getHeight() - 1, color2, false));
                g2.fillRect(0, 1, getWidth(), getHeight() - 1);
                g2.setColor(new Color(61, 123, 218));
                g2.drawLine(0, 0, getWidth(), 0);
                g2.setColor(Color.white);
            } else {
                g2.setColor(Color.white);
                g2.fillRect(0, 0, getWidth(), getHeight());
                g2.setColor(Color.black);
            }

            colorTablePaint.paint(g2, isSelected);
        }
    }

}
