/* gvSIG. Geographic Information System of the Valencian Government
 *
 * Copyright (C) 2007-2008 Infrastructures and Transports Department
 * of the Valencian Government (CIT)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 *
 */
package org.gvsig.timesupport.impl.animation;

import java.util.Date;
import org.gvsig.timesupport.AbsoluteInterval;
import org.gvsig.timesupport.Duration;
import org.gvsig.timesupport.Interval;
import org.gvsig.timesupport.RelativeInstant;
import org.gvsig.timesupport.RelativeInterval;
import org.gvsig.timesupport.TimeSupportLocator;
import org.gvsig.timesupport.TimeSupportManager;
import org.gvsig.timesupport.animation.TimeAnimation;
import org.gvsig.tools.logger.FilteredLogger;
import org.gvsig.tools.observer.ObservableHelper;
import org.gvsig.tools.observer.Observer;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class TimeAnimationImpl implements TimeAnimation, Runnable {

    private static final Logger logger = LoggerFactory.getLogger(TimeAnimationImpl.class);
    
    private int status = TimeAnimation.STOPED;

    private RelativeInterval interval = null;

    private long stepInMillis;

    private int currentPosition;
    private int currentIteration;

    private int iterationCount;

    private Thread thread;

    private final TimeAnimationRenderer renderer;

    private final ObservableHelper observers;
    private double framesPerSecond;
    private long pauseTime;
    private int stepScale;

    public TimeAnimationImpl(TimeAnimationRenderer renderer) {
        this.observers = new ObservableHelper();
        this.renderer = renderer;
        this.pauseTime = 10;
        this.iterationCount = 1;
        this.framesPerSecond = 0.3;
        this.stepScale = TimeAnimation.MILLISECOND;
    }

    @Override
    public void addObserver(Observer obsrvr) {
        this.observers.addObserver(obsrvr);
    }

    @Override
    public void deleteObserver(Observer obsrvr) {
        this.observers.deleteObserver(obsrvr);
    }

    @Override
    public void deleteObservers() {
        this.observers.deleteObservers();
    }

    @Override
    public RelativeInterval getInterval() {
        return this.interval;
    }

    @Override
    public void setInterval(RelativeInterval interval) {
        if( this.status != TimeAnimation.STOPED ) {
            if( interval == null ) {
                throw new IllegalArgumentException("Can't set interval (null), Illegal state (" + this.getStatusLabel() + ").");
            }
            throw new IllegalArgumentException("Can't set interval (" + interval.toString() + "), Illegal state (" + this.getStatusLabel() + ").");
        }
        this.interval = interval;
        this.inferStepScale();
        this.notifyObservers();
    }

    @Override
    public long getStep() {
        return this.stepInMillis;
    }

    @Override
    public void setStep(long step) {
        if( this.status != TimeAnimation.STOPED ) {
            throw new IllegalArgumentException("Can't set step (" + step + "), Illegal state (" + this.getStatusLabel() + ").");
        }
        this.stepScale = TimeAnimation.MILLISECOND;
        this.stepInMillis = step;
        this.notifyObservers();
    }

    @Override
    public long getStep(int scale) {
        switch( scale ) {
        case TimeAnimation.MILLISECOND:
            return this.stepInMillis;
        case TimeAnimation.SECOND:
            return this.stepInMillis / 1000;
        case TimeAnimation.MINUTE:
            return this.stepInMillis / (60 * 1000);
        case TimeAnimation.HOUR:
            return this.stepInMillis / (60 * 60 * 1000);
        case TimeAnimation.DAY:
            return this.stepInMillis / (24 * 60 * 60 * 1000);
        case TimeAnimation.WEEK:
            return this.stepInMillis / (7 * 24 * 60 * 60 * 1000);
        case TimeAnimation.MONTH:
            // Steps in months are not accurate
            return this.stepInMillis / (30 * 24 * 60 * 60 * 1000);
        case TimeAnimation.YEAR:
        default:
            return this.stepInMillis * (365 * 24 * 60 * 60 * 1000);
        }
    }

    @Override
    public int getStepScale() {
        return this.stepScale;
    }

    @Override
    public void setStep(long step, int scale) {
        if( this.status != TimeAnimation.STOPED ) {
            throw new IllegalArgumentException("Can't set step (" + step + "," + this.getScaleLabel(scale) + "), Illegal state (" + this.getStatusLabel() + ").");
        }
        this.stepScale = scale;
        switch( scale ) {
        case TimeAnimation.MILLISECOND:
            this.stepInMillis = step;
            break;
        case TimeAnimation.SECOND:
            this.stepInMillis = step * 1000;
            break;
        case TimeAnimation.MINUTE:
            this.stepInMillis = step * 60 * 1000;
            break;
        case TimeAnimation.HOUR:
            this.stepInMillis = step * 60 * 60 * 1000;
            break;
        case TimeAnimation.DAY:
            this.stepInMillis = step * 24 * 60 * 60 * 1000;
            break;
        case TimeAnimation.WEEK:
            this.stepInMillis = step * 7 * 24 * 60 * 60 * 1000;
            break;
        case TimeAnimation.MONTH:
            // Steps in months are not accurate
            this.stepInMillis = step * 30 * 24 * 60 * 60 * 1000;
            break;
        case TimeAnimation.YEAR:
        default:
            this.stepInMillis = step * 365 * 24 * 60 * 60 * 1000;
            break;
        }
        this.notifyObservers();
    }

    @Override
    public void setFramesPerSecond(double fps) {
        this.framesPerSecond = fps;
        this.notifyObservers();
    }

    @Override
    public double getFramesPerSecond() {
        return this.framesPerSecond;
    }

    @Override
    public void setPauseTime(long time) {
        this.pauseTime = time;
    }

    @Override
    public long getPauseTime() {
        return this.pauseTime;
    }

    @Override
    public int getIterations() {
        return this.iterationCount;
    }

    @Override
    public void setIterations(int iterations) {
        if( this.status != TimeAnimation.STOPED ) {
            throw new IllegalArgumentException("Can't set num iterations (" + iterations + "), Illegal state (" + this.getStatusLabel() + ").");
        }
        this.iterationCount = iterations;
        this.notifyObservers();
    }

    @Override
    public int getCurrentPosition() {
        return this.currentPosition;
    }

    @Override
    public void setCurrentPosition(int position) {
        int stepCount = this.getStepCount();
        if( position < 0 || position > stepCount ) {
            throw new IllegalArgumentException("Can't set position " + position + ", out of range (0," + stepCount + ").");
        }
        this.currentPosition = position;
//        if( !this.isPlaying() ) {
            Interval currentTime = this.getCurrentTime();
            this.renderer.render(currentTime);
//        }
        this.notifyObservers();
    }

    @Override
    public int getCurrentIteration() {
        return this.currentIteration;
    }

    @Override
    public int getStepCount() {
        Duration duration = this.interval.toStandardDuration();
        long stepCount = duration.getMillis() / this.stepInMillis;
        return (int) stepCount;
    }

    @Override
    public RelativeInterval getCurrentTime() {
        TimeSupportManager manager = TimeSupportLocator.getManager();
        if( this.interval instanceof RelativeInterval ) {
            RelativeInterval interval = (RelativeInterval) this.interval;
            RelativeInterval time = manager.createRelativeInterval(
                interval.getStart().toMillis() + (this.currentPosition * this.stepInMillis),
                interval.getStart().toMillis() + (this.currentPosition * this.stepInMillis) + this.stepInMillis
            );
            return time;
        }
        AbsoluteInterval interval = (AbsoluteInterval) this.interval;
        RelativeInterval time = manager.createRelativeInterval(
            interval.getStart().getMillis() + (this.currentPosition * this.stepInMillis),
            interval.getStart().getMillis() + (this.currentPosition * this.stepInMillis) + this.stepInMillis
        );
        return time;
    }

    @Override
    public int getStatus() {
        return this.status;
    }

    @Override
    public boolean isPlaying() {
        return this.status == TimeAnimation.PLAYING;
    }

    @Override
    public synchronized void stop() {
        if( this.status!=TimeAnimation.STOPED ) {
            try {
                this.renderer.end();
            } catch(Throwable th) {
                logger.warn("Problems do render.end",th);
            }
        }
        this.status = TimeAnimation.STOPED;
        currentIteration = 0;
        currentPosition = 0;
        this.notifyObservers();
    }

    @Override
    public synchronized void pause() {
        this.status = TimeAnimation.PAUSED;
        this.notifyObservers();
    }

    @Override
    public synchronized void play() {
        switch(this.status) {
        case TimeAnimation.PLAYING:
            return;
        case TimeAnimation.STOPED:
            try {
                this.renderer.begin();
            } catch(Throwable th) {
                logger.warn("Problems do render.begin",th);
            }
            break;
        case TimeAnimation.PAUSED:
            break;
        }
        this.status = TimeAnimation.PLAYING;
        if( this.thread == null || !this.thread.isAlive() ) {
            this.thread = new Thread(this, "TimeAnimation"+Integer.toHexString(this.hashCode()));
            this.thread.start();
        }
        this.notifyObservers();
    }

    @Override
    public void run() {
        int stepCount = this.getStepCount();
        int frameDuration = (int)(1000.0 / this.framesPerSecond);

        FilteredLogger logger = new FilteredLogger(this.logger, "TimeAnimation", 10);
        
        for( ; currentIteration < this.iterationCount; currentIteration++ ) {
            for( ; currentPosition < stepCount; currentPosition++ ) {
                while( this.status == TimeAnimation.PAUSED ) {
                    try {
                        Thread.sleep(this.pauseTime);
                    } catch (InterruptedException ex) {
                    }
                }
                if( this.status == TimeAnimation.STOPED ) {
                    return;
                }
                long renderStartTime = new Date().getTime();
                this.setCurrentPosition(currentPosition);                
//                Interval currentTime = this.getCurrentTime();
//                try {
//                    this.renderer.render(currentTime);
//                } catch(Throwable th) {
//                    logger.warn("Problems do renderer.render ("+currentTime+").",th);
//                }
//                notifyObservers();
                long timeRemaining = frameDuration - (new Date().getTime() - renderStartTime);
                if( timeRemaining > 0 ) {
                    try {
                        Thread.sleep(timeRemaining);
                    } catch (InterruptedException ex) {
                    }
                }
            }
        }
        this.stop();
    }

    private void notifyObservers() {
        Thread th = new Thread(new Runnable() {
            @Override
            public void run() {
                observers.notifyObservers(TimeAnimationImpl.this, null);
            }
        }, "TimeAnimationNotify"+Integer.toHexString(this.hashCode()));
        th.start();
    }

    @Override
    public void inferStepScale() {
        RelativeInstant start = interval.getStart();
        RelativeInstant end = interval.getEnd();
        int years = end.getYear() - start.getYear();
        int scale = TimeAnimation.YEAR;
        if( years == 1 ) {
            scale = TimeAnimation.MONTH;
        }
        if( years < 1 ) {
            int months = end.getMonthOfYear() - start.getMonthOfYear();
            scale = TimeAnimation.MONTH;
            if( months == 1 ) {
                scale = TimeAnimation.DAY;
            }
            if( months < 1 ) {
                int day = end.getDayOfMonth() - start.getDayOfMonth();
                scale = TimeAnimation.DAY;
                if( day == 1 ) {
                    scale = TimeAnimation.HOUR;
                }
                if( day < 1 ) {
                    int hour = end.getHourOfDay() - start.getHourOfDay();
                    scale = TimeAnimation.HOUR;
                    if( hour == 1 ) {
                        scale = TimeAnimation.SECOND;
                    }
                    if( hour < 1 ) {
                        int second = end.getSecondOfDay() - start.getSecondOfDay();
                        scale = TimeAnimation.SECOND;
                        if( second == 1 ) {
                            scale = TimeAnimation.MILLISECOND;
                        }
                        if( second < 1 ) {
                            scale = TimeAnimation.MILLISECOND;
                        }
                    }
                }
            }
        }
        this.setStep(1, scale);
    }

    private String getStatusLabel() {
        return this.getStatusLabel(this.status);
    }

    @Override
    public String getStatusLabel(int status) {
        switch( status ) {
        case TimeAnimation.PLAYING:
            return "Playing";
        case TimeAnimation.PAUSED:
            return "Paused";
        case TimeAnimation.STOPED:
            return "Stoped";
        default:
            return "Unknow(" + this.status + ")";
        }
    }

    @Override
    public String getScaleLabel(int scale) {
        switch( scale ) {
        case TimeAnimation.MILLISECOND:
            return "Milliseconds";
        case TimeAnimation.SECOND:
            return "Seconds";
        case TimeAnimation.MINUTE:
            return "Minutes";
        case TimeAnimation.HOUR:
            return "Hours";
        case TimeAnimation.DAY:
            return "Days";
        case TimeAnimation.WEEK:
            return "Weeks";
        case TimeAnimation.MONTH:
            return "Months";
        case TimeAnimation.YEAR:
            return "Years";
        default:
            return "Unknow(" + scale + ")";
        }
    }
}
