/**
 * gvSIG. Desktop Geographic Information System.
 *
 * Copyright (C) 2007-2017 gvSIG Association.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * For any additional information, do not hesitate to contact us at info AT
 * gvsig.com, or visit our website www.gvsig.com.
 */
package org.gvsig.timesupport.swing.impl.panel;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.swing.BoxLayout;
import javax.swing.ButtonGroup;
import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.JRadioButton;

import org.gvsig.i18n.Messages;
import org.gvsig.timesupport.Instant;
import org.gvsig.timesupport.Interval;
import org.gvsig.timesupport.swing.api.panel.TimeSelectorPanel;
import org.gvsig.timesupport.swing.impl.components.InstantComparator;
import org.gvsig.timesupport.swing.impl.panel.TimePanel.TIME_MODE;
import org.gvsig.timesupport.swing.impl.rdv.TimeAdjustmentListener;
import org.gvsig.timesupport.swing.impl.rdv.TimeEvent;
import org.gvsig.tools.observer.Observer;

/**
 * Creates a panel with slider and time control calendars
 */
public class DefaultTimeSliderPanel extends JPanel implements
    TimeSelectorPanel, TimeAdjustmentListener {

    private static final long serialVersionUID = 213502960867760481L;

    private JRadioButton rdbNone = null;
    private JRadioButton rdbInterval = null;
    private JRadioButton rdbInstant = null;
    private TimePanel timePanel = null;
    private List<Observer> observers = null;
    private Interval interval;

    public DefaultTimeSliderPanel() {
        super();
        observers = new ArrayList<>();
    }

    @Override
    public void setIntervals(Interval interval) {
        this.interval = interval;
        initialize(this.interval.getStart(), this.interval.getEnd());
    }

    /**
     * Intializes in x, y coordinates and start,end dates
     *
     * @param x
     * @param y
     * @param start
     * @param end
     */
    public void initialize(Instant start, Instant end) {
        removeAll();
        revalidate();
        repaint();
        createUI();
        if( start.isAbsolute() ) {
            timePanel.setTimeMode(TIME_MODE.ABSOLUTE, start, end);
        } else {
            timePanel.setTimeMode(TIME_MODE.RELATIVE, start, end);
        }
    }

    @Override
    public void timeChanged(TimeEvent event) {
        if( getInstantRadioButton().isSelected() ) {
            Instant instante = timePanel.getInstant();
            if( !this.observers.isEmpty() ) {
                for( Observer o : this.observers ) {
                    o.update(this, new UpdateTimeNotification(instante));
                }
            }
        }
    }

    @Override
    public void rangeChanged(TimeEvent event) {
        if( getIntervalRadioButton().isSelected() ) {
            if( !this.observers.isEmpty() ) {
                for( Observer o : this.observers ) {
                    o.update(this, new UpdateTimeNotification(timePanel.getInterval()));
                }
            }
        }
    }

    @Override
    public void boundsChanged(TimeEvent event) {

    }

    @Override
    public void setBounds(Rectangle r) {
        super.setBounds(new Rectangle(r.x, r.y, WIDTH, HEIGHT));
    }

    @Override
    public boolean isInterval() {
        return getIntervalRadioButton().isSelected();
    }

    @Override
    public boolean isInstant() {
        return getInstantRadioButton().isSelected();
    }

    @Override
    public boolean isNone() {
        return getNoneRadioButton().isSelected();
    }

    @Override
    public void addObserver(Observer o) {
        observers.add(o);
    }

    @Override
    public void deleteObserver(Observer o) {
        observers.remove(o);
    }

    @Override
    public void deleteObservers() {
        observers.clear();
    }

    public void setInstants(Instant start, Instant end) {
        timePanel.setInstants(start, end);
    }

    private void createUI() {
        setLayout(new BorderLayout());
        add(createRadioButtons(),BorderLayout.NORTH);
        timePanel = new TimePanel();
        timePanel.setListener(this);
        add(timePanel,BorderLayout.CENTER);
        
        Dimension size = this.getPreferredSize();
        if( size.width<400 ) {
            size.width = 400;
        }
        if( size.height<140 ) {
            size.height = 140;
        }
        this.setPreferredSize(size);
        timePanel.fireRangeInterval();
    }

    private JPanel createRadioButtons() {
        ButtonGroup group = new ButtonGroup();
        getNoneRadioButton().addActionListener(new ActionNoneButton(this));
        group.add(getNoneRadioButton());
        getIntervalRadioButton().addActionListener(new ActionListener() {

            @Override
            public void actionPerformed(ActionEvent e) {
                timePanel.setEnabled(true);
                timePanel.resetValues();
                timePanel.setValueChangeableSlider(false);
                timePanel.fireRangeInterval();
            }
        });
        getIntervalRadioButton().setSelected(true);
        group.add(getIntervalRadioButton());
        getInstantRadioButton().addActionListener(new ActionListener() {

            @Override
            public void actionPerformed(ActionEvent e) {
                timePanel.setEnabled(true);
                timePanel.setEnabledEndCalendar(false);
                timePanel.setValueChangeableSlider(true);

                timePanel.resetValues();
                timePanel.fireValueInstant();
            }
        });
        group.add(getInstantRadioButton());
        // return group;
        JPanel radio = new JPanel();
        radio.setLayout(new FlowLayout());
        radio.add(getNoneRadioButton());
        radio.add(getIntervalRadioButton());
        radio.add(getInstantRadioButton());

        return radio;
    }

    private JRadioButton getNoneRadioButton() {
        if( rdbNone == null ) {
            rdbNone = new JRadioButton(Messages.getText("rdButton_none"));
        }
        return rdbNone;
    }

    private JRadioButton getIntervalRadioButton() {
        if( rdbInterval == null ) {
            rdbInterval = new JRadioButton(
                Messages.getText("rdButton_interval"));
        }
        return rdbInterval;
    }

    private JRadioButton getInstantRadioButton() {
        if( rdbInstant == null ) {
            rdbInstant = new JRadioButton(Messages.getText("rdButton_instant"));
        }
        return rdbInstant;
    }

//    private JButton getBtnClose() {
//        if( btnClose == null ) {
//            btnClose = new JButton(Messages.getText("temporal_btn_cancel"));
//        }
//        return btnClose;
//    }

    void noneButtonAction() {
        if( !this.observers.isEmpty() ) {
            for( Observer o : this.observers ) {
                o.update(this, new UpdateTimeNotification(null));
            }
        }
    }

    @Override
    public JComponent asJComponent() {
        return this;
    }

    private class ActionNoneButton implements ActionListener {

        private final DefaultTimeSliderPanel panel;

        public ActionNoneButton(DefaultTimeSliderPanel panel) {
            this.panel = panel;
        }

        @Override
        public void actionPerformed(ActionEvent e) {
            timePanel.setEnabled(false);

            this.panel.noneButtonAction();
        }

    }

    @Override
    public void setInstants(List<Instant> instants) {
        java.util.Collections.sort(instants, new InstantComparator());
        //Delete duplicates
        Iterator<Instant> iterador = instants.iterator();
        List<Instant> sortedList = new ArrayList<>();
        Instant obj;
        while( iterador.hasNext() ) {
            obj = iterador.next();
            if( !sortedList.contains(obj) ) {
                sortedList.add(obj);
            }
        }
        timePanel.setInstants(sortedList);
    }

}
