/**
 * gvSIG. Desktop Geographic Information System.
 *
 * Copyright (C) 2007-2017 gvSIG Association.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * For any additional information, do not hesitate to contact us at info AT
 * gvsig.com, or visit our website www.gvsig.com.
 */
package org.gvsig.timesupport.swing.impl.panel.animation;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.net.URL;

import javax.swing.ImageIcon;
import javax.swing.JComponent;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import org.gvsig.timesupport.Interval;

import org.gvsig.timesupport.RelativeInterval;
import org.gvsig.timesupport.animation.TimeAnimation;
import org.gvsig.timesupport.swing.api.panel.AnimationBarPanel;
import org.gvsig.tools.ToolsLocator;
import org.gvsig.tools.i18n.I18nManager;
import org.gvsig.tools.observer.Observable;
import org.gvsig.tools.observer.Observer;

/**
 * Main panel for animation bar
 *
 */
public class AnimationBarPanelImpl extends AnimationBarPanelImplLayout implements AnimationBarPanel, Observer {

    private static final long serialVersionUID = 1L;
    private final String pattern = "dd/MM/yyyy HH:mm:ss:SSS";
    private final TimeAnimation animation;

    public AnimationBarPanelImpl(TimeAnimation animation) {
        this.animation = animation;
        this.initComponents();
        this.updateComponents(animation);
        animation.addObserver(this);
    }

    private void initComponents() {
        sliderSpeed.setMinimum(0);
        sliderSpeed.setMaximum(100);
        sliderSpeed.setValue(0);
        sliderSpeed.addChangeListener(new ChangeListener() {
            @Override
            public void stateChanged(ChangeEvent e) {
                doSliderChanged();
            }
        });
        
        URL playUrl = getClass().getClassLoader().getResource("org/gvsig/timesupport/swing/impl/images/play.png");
        if( playUrl != null ) {
            ImageIcon playIcon = new ImageIcon(playUrl);
            buttonPlay.setIcon(playIcon);
        }
        buttonPlay.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                animation.play();
            }
        });
        buttonPlay.setEnabled(true);

        URL stopUrl = getClass().getClassLoader().getResource("org/gvsig/timesupport/swing/impl/images/stop.png");
        if( stopUrl != null ) {
            ImageIcon stopIcon = new ImageIcon(stopUrl);
            buttonStop.setIcon(stopIcon);
        }
        buttonStop.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                animation.stop();
            }
        });
        buttonStop.setEnabled(false);

        URL pauseUrl = getClass().getClassLoader().getResource("org/gvsig/timesupport/swing/impl/images/pause.png");
        if( pauseUrl != null ) {
            ImageIcon pauseIcon = new ImageIcon(pauseUrl);
            buttonPause.setIcon(pauseIcon);
        }
        buttonPause.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                animation.pause();
            }
        });
        buttonPause.setEnabled(false);
        
        I18nManager i18nManager = ToolsLocator.getI18nManager();
        timeStep.setText(i18nManager.getTranslation("timestep"));
        speedOfProgress.setText(i18nManager.getTranslation("speedofprogress"));
        
    }

    private void updateComponents(TimeAnimation animation) {
        I18nManager i18n = ToolsLocator.getI18nManager();

        sliderSpeed.setMinimum(0);
        sliderSpeed.setMaximum(animation.getStepCount());

        Interval window = animation.getCurrentTime();
        textWindowTime.setText(
            window.getStart().toString(pattern) + 
            " to " + 
            window.getEnd().toString(pattern)
        );
        RelativeInterval interval = animation.getInterval();
        labelMinInterval.setText(interval.getStart().toString(pattern));
        labelMaxInterval.setText(interval.getEnd().toString(pattern));
        textTimeStep.setText(
            animation.getStep(animation.getStepScale()) + 
            " " + 
            i18n.getTranslation(this.animation.getScaleLabel(animation.getStepScale()))
        );        
        textTimeSpeed.setText(String.format("%.3f fps", animation.getFramesPerSecond()));

        switch(animation.getStatus() ) {
        case TimeAnimation.STOPED:
            buttonPlay.setEnabled(true);
            buttonStop.setEnabled(false);
            buttonPause.setEnabled(false);
            break;
        case TimeAnimation.PLAYING:
            buttonPlay.setEnabled(false);
            buttonStop.setEnabled(true);
            buttonPause.setEnabled(true);
            break;
        case TimeAnimation.PAUSED:
            buttonPlay.setEnabled(true);
            buttonStop.setEnabled(true);
            buttonPause.setEnabled(false);
            break;
        }

        sliderSpeed.setValue(animation.getCurrentPosition());
    }
    
    private void doSliderChanged() {
        this.animation.setCurrentPosition(this.sliderSpeed.getValue());
    }
    
    @Override
    public void update(Observable o, Object animation) {
        if( o instanceof TimeAnimation ) {
            this.updateComponents((TimeAnimation) o);
        }
    }

    @Override
    public JComponent asJComponent() {
        return this;
    }

    @Override
    public TimeAnimation getAnimation() {
        return this.animation;
    }
}
