/**
 * gvSIG. Desktop Geographic Information System.
 *
 * Copyright (C) 2007-2017 gvSIG Association.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * For any additional information, do not hesitate to contact us at info AT
 * gvsig.com, or visit our website www.gvsig.com.
 */
package org.gvsig.timesupport.swing.impl.panel.animation;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import javax.swing.JComponent;
import javax.swing.SpinnerNumberModel;
import org.apache.commons.lang3.math.NumberUtils;
import org.gvsig.timesupport.RelativeInterval;
import org.gvsig.timesupport.animation.TimeAnimation;
import org.gvsig.timesupport.swing.api.panel.AnimationPreferencesPanel;
import org.gvsig.timesupport.swing.impl.panel.TimePanel;
import org.gvsig.tools.ToolsLocator;
import org.gvsig.tools.i18n.I18nManager;
import org.gvsig.tools.swing.api.ListElement;

/**
 * Main panel for animation preferences
 */
public class AnimationPreferencesPanelImpl extends AnimationPreferencesPanelImplLayout implements AnimationPreferencesPanel {

    private static final long serialVersionUID = -2490295024250846833L;

    private final TimeAnimation animation;
    private TimePanel timePanel;

    private final int[] scales = {
        TimeAnimation.MILLISECOND,
        TimeAnimation.SECOND,
        TimeAnimation.MINUTE,
        TimeAnimation.HOUR,
        TimeAnimation.DAY,
        TimeAnimation.WEEK,
        TimeAnimation.MONTH,
        TimeAnimation.YEAR
    };

    public AnimationPreferencesPanelImpl(TimeAnimation animation) {
        this.animation = animation;
        this.initComponents();
        this.updateComponents(animation);
    }

    private void initComponents() {
        I18nManager i18n = ToolsLocator.getI18nManager();

        this.timePanel = new TimePanel();
        this.pnlTimeSelector.setLayout(new BorderLayout());
        this.pnlTimeSelector.add(this.timePanel, BorderLayout.CENTER);

        SpinnerNumberModel fpsModel = new SpinnerNumberModel(
            new Float(0.3), 
            null, 
            null, 
            new Float(1.0)
        ); 
        this.spnFps.setModel(fpsModel);

        this.cboStepScale.removeAllItems();
        for( int scale : scales ) {
            this.cboStepScale.addItem(new ListElement(
                this.animation.getScaleLabel(scale),
                scale
            ));
        }

        this.cboRepeatMode.removeAllItems();
        this.cboRepeatMode.addItem(i18n.getTranslation("_Without_repetition"));
        this.cboRepeatMode.addItem(i18n.getTranslation("_Infinite"));
        this.cboRepeatMode.addItem(i18n.getTranslation("_repeat_N_times"));
        this.cboRepeatMode.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                doRepeatModeChanged();
            }
        });
        this.setPreferredSize(new Dimension(430,170));
    }

    private void updateComponents(TimeAnimation animation) {
        this.timePanel.initialize(
            this.animation.getInterval().getStart(),
            this.animation.getInterval().getEnd()
        );

        this.txtStep.setText(String.valueOf(this.animation.getStep(animation.getStepScale())));

        for( int i = 0; i < scales.length; i++ ) {
            ListElement<Integer> element = (ListElement<Integer>) this.cboStepScale.getItemAt(i);
            if( element!=null && element.getValue() == animation.getStepScale() ) {
                this.cboStepScale.setSelectedIndex(i);
                break;
            }
        }

        this.spnFps.setValue(animation.getFramesPerSecond());
        
        switch( animation.getIterations() ) {
        case 1:
            this.cboRepeatMode.setSelectedIndex(0);
            this.txtIterations.setText("1");
            break;
        case Integer.MAX_VALUE:
            this.cboRepeatMode.setSelectedIndex(1);
            this.txtIterations.setText(String.valueOf(Integer.MAX_VALUE));
            break;
        default:
            this.cboRepeatMode.setSelectedIndex(2);
            this.txtIterations.setText(String.valueOf(animation.getIterations()));
            break;
        }
        
    }
    
    private void updateAnimation(TimeAnimation animation) {
        animation.setInterval((RelativeInterval) timePanel.getInterval());
        
        animation.setStep(
            NumberUtils.toLong(this.txtStep.getText(), animation.getStep()),
            ((ListElement<Integer>)(this.cboStepScale.getSelectedItem())).getValue()
        );
        
        animation.setFramesPerSecond(((Number) this.spnFps.getValue()).doubleValue());
        
        switch(this.cboRepeatMode.getSelectedIndex()) {
        case 0:
            animation.setIterations(1);
            break;
        case 1:
            animation.setIterations(Integer.MAX_VALUE);
            break;
        case 2:
            animation.setIterations(NumberUtils.toInt(this.txtIterations.getText(), 1));
            break;
        }
    }

    @Override
    public TimeAnimation getAnimation() {
        this.updateAnimation(animation);
        return this.animation;
    }

    @Override
    public JComponent asJComponent() {
        return this;
    }

    private void doRepeatModeChanged() {
        switch( this.cboRepeatMode.getSelectedIndex() ) {
        case -1:
            return;
        case 0:
        case 1:
            this.txtIterations.setEnabled(false);
        case 2:
            this.txtIterations.setEnabled(true);
        }
    }
}
