package org.gvsig.webbrowser.jfx;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.net.MalformedURLException;
import java.net.URL;
import java.text.MessageFormat;
import org.gvsig.webbrowser.WebBrowserPanel;
import javafx.event.EventHandler;
import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.application.Platform;
import javafx.concurrent.Worker;
import javafx.embed.swing.JFXPanel;
import javafx.scene.web.WebView;
import javafx.scene.Scene;
import javafx.scene.web.WebEngine;
import javafx.scene.web.WebEvent;
import javax.swing.ImageIcon;
import javax.swing.JComponent;
import javax.swing.JTextField;

public class JFXWebBrowserPanel extends JFXWebBrowserPanelView implements WebBrowserPanel {

    private static final long serialVersionUID = -5048314340699267249L;

    private JFXWebBrowser browser;

    private class JFXWebBrowser extends JFXPanel {

        private static final long serialVersionUID = 3021277085276506035L;

        private WebView view;
        private WebEngine engine;
        private Scene scene;

        private class MyEventHandler implements EventHandler<WebEvent<String>> {

            @Override
            public void handle(WebEvent<String> event) {
                if( event.getEventType() == WebEvent.STATUS_CHANGED ) {
                    String s = event.getData();
                    if( s == null ) {
                        lblStatus.setText("");
                    } else {
                        lblStatus.setText(s);
                    }

                }
            }
        }

        private class MyChangeListener implements ChangeListener<Worker.State> {

            @Override
            public void changed(ObservableValue<? extends Worker.State> ov, Worker.State oldState, Worker.State newState) {
                if( newState == Worker.State.SUCCEEDED ) {
                    txtURL.setText(engine.getLocation());
                }
            }
        }

        public JFXWebBrowser() {
            this.view = null;
            this.engine = null;
            this.scene = null;

            Platform.runLater(new Runnable() {
                @Override
                public void run() {
                    createScene();
                }
            });
            Platform.setImplicitExit(false);
        }

        private void createScene() {
            this.view = new WebView();
            this.engine = this.view.getEngine();
            this.engine.setOnStatusChanged(new MyEventHandler());
            this.engine.getLoadWorker().stateProperty().addListener(new MyChangeListener());
            this.engine.setJavaScriptEnabled(true);
            this.scene = new Scene(this.view);
            this.setScene(this.scene);
        }

        private void load(URL url) {
            this.engine.load(url.toString());
            this.revalidate();
        }

        private void loadContent(String content, String contentType) {
            this.engine.loadContent(content, contentType);
            this.revalidate();
        }

        private void goHistory(int index) {
            this.engine.getHistory().go(index);
        }

        public void setPage(final URL url) {
            Platform.runLater(new Runnable() {
                @Override
                public void run() {
                    load(url);
                }
            });
        }

        public void setContent(final String content, final String contentType) {
            Platform.runLater(new Runnable() {
                @Override
                public void run() {
                    loadContent(content, contentType);
                }
            });
        }

        public void goPrevious() {
            Platform.runLater(new Runnable() {
                @Override
                public void run() {
                    goHistory(-1);
                }
            });
        }

        public void goNext() {
            Platform.runLater(new Runnable() {
                @Override
                public void run() {
                    goHistory(1);
                }
            });
        }

        public String getPage() {
            return this.engine.getLocation();
        }

        public String getTitle() {
            return this.engine.getTitle();
        }

        public Object executeScript(String js) {
            Object r = this.engine.executeScript(js);
            return r;
        }

        public void find(final String text, final boolean backwards) {
            try {
                Platform.runLater(new Runnable() {
                    @Override
                    public void run() {
                        String script = MessageFormat.format(
                            "window.find(\"{0}\",false,{1},true)",
                            text,
                            backwards ? "true" : "false"
                        );

                    }
                });
            } catch (Throwable th) {

            }
        }

    }

    public JFXWebBrowserPanel() {
        this.initComponents();
    }

    private void initComponents() {
        this.browser = new JFXWebBrowser();
        this.containerBrowser.setLayout(new BorderLayout());
        this.containerBrowser.add(this.browser, BorderLayout.CENTER);

        this.btnPrevious.setIcon(imageIcon("images/arrow_left.png"));
        this.btnNext.setIcon(imageIcon("images/arrow_right.png"));
        this.btnRefresh.setIcon(imageIcon("images/arrow_refresh.png"));
        this.btnConfig.setIcon(imageIcon("images/add.png"));
        this.btnSearchPrevious.setIcon(imageIcon("images/arrow_up.png"));
        this.btnSearchNext.setIcon(imageIcon("images/arrow_down.png"));
        this.imgSearch.setIcon(imageIcon("images/search.png"));

        this.txtURL.addKeyListener(new KeyAdapter() {
            @Override
            public void keyPressed(KeyEvent keyEvent) {
                if( keyEvent.getKeyChar() == '\n' ) {
                    setPage(txtURL.getText());
                }
            }
        });
        this.txtSearch.addKeyListener(new KeyAdapter() {
            @Override
            public void keyPressed(KeyEvent event) {
                JTextField jtext = txtSearch;
                if( event.getKeyChar() == 0x1b ) { // ESC
                  jtext.setText("");
                  browser.find("", true);
                } else if( event.getKeyChar() == '\n' ) {
                  browser.find(jtext.getText(), true);
                }
            }
        });
        this.btnSearchPrevious.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                browser.find(txtSearch.getText(),false);
            }
        });
        this.btnSearchNext.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                browser.find(txtSearch.getText(),true);
            }
        });
        this.btnPrevious.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                browser.goPrevious();
            }
        });
        this.btnNext.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                browser.goNext();
            }
        });
        this.setPreferredSize(new Dimension(600, 500));
    }

    private ImageIcon imageIcon(String res) {
        URL url = this.getClass().getResource(res);
        if( url == null ) {
            return new ImageIcon();
        }
        return new ImageIcon(url);
    }

    @Override
    public void setPage(String url) {
        try {
            this.setPage(new URL(url));
        } catch (MalformedURLException ex) {
            throw new RuntimeException(ex);
        }
    }

    @Override
    public void setPage(URL url) {
        this.txtURL.setText(url.toString());
        this.browser.setPage(url);
    }

    @Override
    public String getPage() {
        return this.browser.getPage();
    }

    @Override
    public void setContent(final String content, final String contentType) {
        this.browser.setContent(content, contentType);
    }

    @Override
    public String getTitle() {
        return this.browser.getTitle();
    }

    @Override
    public JComponent asJComponent() {
        return this;
    }
}
