/**
 * gvSIG. Desktop Geographic Information System.
 *
 * Copyright (C) 2007-2013 gvSIG Association.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * For any additional information, do not hesitate to contact us at info AT
 * gvsig.com, or visit our website www.gvsig.com.
 */
package org.gvsig.app.gui.preferencespage;

import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;

import javax.swing.DefaultComboBoxModel;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JColorChooser;
import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.JTextField;
import javax.swing.SpinnerNumberModel;
import javax.swing.SwingUtilities;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import org.apache.commons.lang3.BooleanUtils;

import org.cresques.cts.IProjection;
import org.gvsig.andami.IconThemeHelper;
import org.gvsig.andami.PluginServices;
import org.gvsig.andami.PluginsLocator;
import org.gvsig.andami.preferences.IPreference;
import org.gvsig.andami.preferences.StoreException;
import org.gvsig.app.ApplicationLocator;
import org.gvsig.app.gui.panels.CRSSelectPanel;
import org.gvsig.app.gui.panels.crs.ISelectCrsPanel;
import org.gvsig.app.project.ProjectPreferences;
import org.gvsig.fmap.crs.CRSFactory;
import org.gvsig.fmap.mapcontext.MapContext;
import org.gvsig.fmap.mapcontext.MapContextLocator;
import org.gvsig.i18n.Messages;

import org.gvsig.tools.ToolsLocator;
import org.gvsig.tools.dynobject.DynObject;
import org.gvsig.tools.i18n.I18nManager;
import org.gvsig.tools.swing.api.ToolsSwingLocator;
import org.gvsig.tools.swing.api.windowmanager.WindowManager;
import org.gvsig.utils.XMLEntity;

/**
 * View document configuration page.
 *
 * Here the user can establish what settings wants to use by default regarding
 * to the document View.
 *
 */
public class ViewPage extends ViewPageView implements ChangeListener, ItemListener, IPreference {
    //
    // The class ViewPageView can extend from AbstractPreferencePage, not JPanel,
    // if rereate it from the GUI builder change te class declaration to:
    //    public abstract class ViewPageView extends AbstractPreferencePage

    private static final long serialVersionUID = 1204688910771374514L;

    public static String id = ViewPage.class.getName();

    private static final IProjection FACTORY_DEFAULT_PROJECTION
            = MapContextLocator.getMapContextManager().getDefaultCRS();

    private static String[] unitsNames = null;
    private static String[] areaUnitsNames = null;

    private static final double DEFAULT_ZOOM_IN_FACTOR = 2.0;
    private static final double DEFAULT_ZOOM_OUT_FACTOR = 0.5;
    private static final Color FACTORY_DEFAULT_SELECTION_COLOR = Color.YELLOW;
    private static final Color FACTORY_DEFAULT_VIEW_BACK_COLOR = Color.WHITE;
    private static int FACTORY_DEFAULT_MAP_UNITS;
    private static int FACTORY_DEFAULT_DISTANCE_UNITS;
    private static int FACTORY_DEFAULT_DISTANCE_AREA;

    private final ImageIcon icon;

    private ProjectionController defaultProjection = null;
    private ColorChooserController defaultViewBackColor = null;
    private ColorChooserController defaultSelectionColor = null;

    private boolean hasChanges = false;

    private static class ProjectionController implements ActionListener {

        private static final long serialVersionUID = -1442364274073324390L;

        private final JTextField txtLabel;
        private final JButton btnShowDialog;
        private IProjection proj = null;
        private ChangeListener changeListener;

        public ProjectionController(JTextField txtLabel, JButton btnShowDialog) {
            this.txtLabel = txtLabel;
            this.btnShowDialog = btnShowDialog;

            this.btnShowDialog.addActionListener(this);
            this.set(null);
        }

        public void addChangeListener(ChangeListener changeListener) {
            this.changeListener = changeListener;
        }

        public void fireChanged() {
            if (this.changeListener != null) {
                this.changeListener.stateChanged(null);
            }
        }

        public void set(IProjection proj) {
            this.proj = proj;
            if (this.proj == null) {
                this.txtLabel.setText("");
            } else {
                this.txtLabel.setText(proj.getAbrev());
            }
        }

        public IProjection get() {
            return this.proj;
        }

        @Override
        public void actionPerformed(ActionEvent e) {
            ISelectCrsPanel panel = CRSSelectPanel.getUIFactory().getSelectCrsPanel(
                    CRSFactory.getCRS(this.txtLabel.getText()), false);
            ToolsSwingLocator.getWindowManager().showWindow(
                    panel.asJComponent(), 
                    ToolsLocator.getI18nManager().getTranslation("selecciona_sistema_de_referencia"), 
                    WindowManager.MODE.DIALOG
            );
            if (panel.isOkPressed()) {
                this.set(panel.getProjection());
                this.fireChanged();
            }
        }

    }

    private static class ColorChooserController implements ActionListener, ChangeListener {

        private final JTextField txtLabel;
        private final JButton btnShowDialog;
        private final JSlider sldAlpha;
        private Color color;
        private boolean allowNull;
        private ChangeListener changeListener;

        public ColorChooserController(JTextField txtLabel, JButton btnShowDialog, JSlider sldAlpha, boolean allowNull) {
            this.txtLabel = txtLabel;
            this.btnShowDialog = btnShowDialog;
            this.sldAlpha = sldAlpha;
            this.btnShowDialog.addActionListener(this);
            if (this.sldAlpha != null) {
                this.sldAlpha.addChangeListener(this);
            }
            this.allowNull = allowNull;
            if (!this.allowNull) {
                this.color = Color.BLACK;
            }
        }

        public ColorChooserController(JTextField txtLabel, JButton btnShowDialog) {
            this(txtLabel, btnShowDialog, null, false);
        }

        public ColorChooserController(JTextField txtLabel, JButton btnShowDialog, JSlider sldAlpha) {
            this(txtLabel, btnShowDialog, sldAlpha, false);
        }

        @Override
        public void stateChanged(ChangeEvent e) {
            if (this.color == null) {
                set(Color.BLACK);
            }
            if (this.sldAlpha == null) {
                this.color = new Color(
                        color.getRed(),
                        color.getGreen(),
                        color.getBlue()
                );
            } else {
                this.color = new Color(
                        color.getRed(),
                        color.getGreen(),
                        color.getBlue(),
                        this.sldAlpha.getValue()
                );
            }
            this.txtLabel.setBackground(this.color);
        }

        public void addChangeListener(ChangeListener changeListener) {
            this.changeListener = changeListener;
        }

        public void fireChanged() {
            if (this.changeListener != null) {
                this.changeListener.stateChanged(null);
            }
        }

        @Override
        public void actionPerformed(ActionEvent e) {
            Color c = JColorChooser.showDialog(
                    SwingUtilities.windowForComponent(this.btnShowDialog),
                    Messages.getText("choose_color"),
                    this.get()
            );
            if (c == null) {
                return;
            }
            set(c);
        }

        private void set(Color color) {
            if (color == null) {
                if (allowNull) {
                    this.color = null;
                    return;
                }
                color = Color.BLACK;
            }
            this.color = color;
            this.txtLabel.setBackground(this.color);
            if (this.sldAlpha != null) {
                this.sldAlpha.setValue(this.color.getAlpha());
            }
        }

        public Color get() {
            return this.color;
        }

    }

    /**
     * Creates a new panel containing View preferences settings.
     *
     */
    public ViewPage() {
        super();

        initUnitsNames();

        icon = IconThemeHelper.getImageIcon("document-view-icon");

        this.chkKeepScaleOnResizing.setEnabled(false);
        
        this.defaultProjection = new ProjectionController(
                this.txtDefaultProjection,
                this.btnChangeProjection
        );
        this.defaultProjection.addChangeListener(this);

        this.defaultSelectionColor = new ColorChooserController(
                this.txtDefaultSelectionColor,
                this.btnDefaultSelectionColor,
                this.sldDefaultSelectionColor
        );
        this.defaultSelectionColor.addChangeListener(this);

        this.defaultViewBackColor = new ColorChooserController(
                this.txtDefaultViewBackColor,
                this.btnDefaultViewBackColor
        );
        this.defaultViewBackColor.addChangeListener(this);

        this.chkInvisibleNewLayers.addChangeListener(this);
        this.chkKeepScaleOnResizing.addChangeListener(this);

        this.jCmbDistanceArea.addItemListener(this);
        this.jCmbDistanceUnits.addItemListener(this);
        this.jCmbMapUnits.addItemListener(this);

        this.translateAll();
        this.initializeValues();
    }

    private void translateAll() {
        I18nManager i18n = ToolsLocator.getI18nManager();

        this.lblDefaultProjection.setText(i18n.getTranslation("default_projection"));
        this.chkInvisibleNewLayers.setText(i18n.getTranslation("options.view.invisible_new_layers"));
        this.chkKeepScaleOnResizing.setText(i18n.getTranslation("options.view.keep_scale_on_resizing"));
        this.lblZoomInFactor.setText(i18n.getTranslation("zoom_in_factor"));
        this.lblZoomOutFactor.setText(i18n.getTranslation("zoom_out_factor"));
        this.lblDefaultViewBackColor.setText(i18n.getTranslation("options.view.default_view_back_color"));
        this.lblDefaultSelectionColor.setText(i18n.getTranslation("options.view.default_selection_color"));
        this.lblMapUnits.setText(i18n.getTranslation("map_units"));
        this.lblDistanceUnits.setText(i18n.getTranslation("distance_units"));
        this.lblDistanceArea.setText(i18n.getTranslation("distance_area"));
        this.chkHideLegendOfNonVisibleLayers.setText(i18n.getTranslation("_Hide_legend_of_non_visible_layers"));
        this.chkOpenWithANewMaximizedView.setText(i18n.getTranslation("_Open_with_a_new_maximized_view"));
        this.btnChangeProjection.setText(i18n.getTranslation("_Change"));
        this.lblAlpha.setText(i18n.getTranslation("Alpha"));
    }

    @Override
    public boolean isResizeable() {
        return true;
    }

    @Override
    public void initializeValues() {
        Double zif;
        Double zof;

        ProjectPreferences projectPreferences = ApplicationLocator.getProjectManager().getProjectPreferences();
        PluginServices plugin = PluginsLocator.getManager().getPlugin(this);
        DynObject props = plugin.getPluginProperties();
        XMLEntity xml = plugin.getPersistentXML();

        this.defaultProjection.set(projectPreferences.getDefaultProjection());

        this.chkInvisibleNewLayers.setSelected(projectPreferences.getAddNewLayersInInvisibleMode());
        this.chkKeepScaleOnResizing.setSelected(projectPreferences.getKeepScaleOnResizing());
        this.chkOpenWithANewMaximizedView.setSelected(projectPreferences.getOpenWithANewMaximizedView());
        this.chkHideLegendOfNonVisibleLayers.setSelected(projectPreferences.getHideLegendInToCOfNonVisibleLayers());
        this.txtZoomInFactor.setModel(
                new SpinnerNumberModel(
                        projectPreferences.getZoomInFactor(), //initial value
                        0.01, //min
                        1000, //max
                        0.25 //step
                )
        );
        this.txtZoomOutFactor.setModel(
                new SpinnerNumberModel(
                        projectPreferences.getZoomOutFactor(), //initial value
                        0.01, //min
                        1000, //max
                        0.25 //step
                )
        );
        MapContext.ZOOMINFACTOR = projectPreferences.getZoomInFactor();
        MapContext.ZOOMOUTFACTOR = projectPreferences.getZoomOutFactor();

        this.defaultViewBackColor.set(projectPreferences.getDefaultViewBackColor());
        this.defaultSelectionColor.set(projectPreferences.getDefaultSelectionColor());

        this.jCmbMapUnits.setModel(new DefaultComboBoxModel(unitsNames));
        this.jCmbMapUnits.setSelectedIndex(projectPreferences.getDefaultMapUnits());

        this.jCmbDistanceUnits.setModel(new DefaultComboBoxModel(unitsNames));
        this.jCmbDistanceUnits.setSelectedIndex(projectPreferences.getDefaultDistanceUnits());

        this.jCmbDistanceArea.setModel(new DefaultComboBoxModel(areaUnitsNames));
        this.jCmbDistanceArea.setSelectedIndex(projectPreferences.getDefaultDistanceArea());
    }

    @Override
    public void stateChanged(ChangeEvent e) {
        this.hasChanges = true;
    }

    @Override
    public void itemStateChanged(ItemEvent e) {
        this.hasChanges = true;
    }

    @Override
    public String getID() {
        return id;
    }

    @Override
    public String getTitle() {
        I18nManager i18nManager = ToolsLocator.getI18nManager();
        return i18nManager.getTranslation("Vista");
    }

    @Override
    public JPanel getPanel() {
        return this;
    }

    @Override
    public void storeValues() throws StoreException {

        ProjectPreferences projectPreferences = ApplicationLocator.getProjectManager().getProjectPreferences();

        try {
            double zif = (double) txtZoomInFactor.getValue();
            double zof = (double) txtZoomOutFactor.getValue();
            if (zif == 0 || zof == 0) {
                throw new NumberFormatException();
            }

            projectPreferences.setZoomInFactor(zif);
            projectPreferences.setZoomOutFactor(zof);
            MapContext.ZOOMINFACTOR = zif;
            MapContext.ZOOMOUTFACTOR = zof;

            projectPreferences.setDefaultProjection(this.defaultProjection.get());
            projectPreferences.setDefaultSelectionColor(defaultSelectionColor.get());
            projectPreferences.setDefaultViewBackColor(defaultViewBackColor.get());

            projectPreferences.setDefaultMapUnits(jCmbMapUnits.getSelectedIndex());
            projectPreferences.setDefaultDistanceUnits(jCmbDistanceUnits.getSelectedIndex());
            projectPreferences.setDefaultDistanceArea(jCmbDistanceArea.getSelectedIndex());

            projectPreferences.setAddNewLayersInInvisibleMode(chkInvisibleNewLayers.isSelected());
            projectPreferences.setKeepScaleOnResizing(chkKeepScaleOnResizing.isSelected());
            projectPreferences.setOpenWithANewMaximizedView(chkOpenWithANewMaximizedView.isSelected());
            projectPreferences.setHideLegendInToCOfNonVisibleLayers(this.chkHideLegendOfNonVisibleLayers.isSelected());
            
        } catch (Exception e) {
            throw new StoreException(PluginServices.getText(this, "factor_zoom_incorrecto"));
        }
    }

    @Override
    public void initializeDefaults() {
        defaultProjection.set(FACTORY_DEFAULT_PROJECTION);
        txtZoomInFactor.setModel(
                new SpinnerNumberModel(
                        DEFAULT_ZOOM_IN_FACTOR, //initial value
                        1, //min
                        1000, //max
                        0.25 //step
                )
        );
        txtZoomOutFactor.setModel(
                new SpinnerNumberModel(
                        DEFAULT_ZOOM_OUT_FACTOR, //initial value
                        1, //min
                        1000, //max
                        0.25 //step
                )
        );        
//        txtZoomInFactor.setText(String.valueOf(DEFAULT_ZOOM_IN_FACTOR));
//        txtZoomOutFactor.setText(String.valueOf(DEFAULT_ZOOM_OUT_FACTOR));
        this.chkInvisibleNewLayers.setSelected(false);
        this.chkKeepScaleOnResizing.setSelected(false);
        this.chkHideLegendOfNonVisibleLayers.setSelected(true);
        this.defaultViewBackColor.set(FACTORY_DEFAULT_VIEW_BACK_COLOR);
        this.defaultSelectionColor.set(FACTORY_DEFAULT_SELECTION_COLOR);
        this.jCmbMapUnits.setSelectedIndex(FACTORY_DEFAULT_MAP_UNITS);
        this.jCmbDistanceUnits.setSelectedIndex(FACTORY_DEFAULT_DISTANCE_UNITS);
        this.jCmbDistanceArea.setSelectedIndex(FACTORY_DEFAULT_DISTANCE_AREA);

    }

    @Override
    public ImageIcon getIcon() {
        return icon;
    }

    @Override
    public boolean isValueChanged() {
        return this.hasChanges;
    }

    @Override
    public void setChangesApplied() {
        this.hasChanges = false;
    }

    private void initUnitsNames() {
        I18nManager i18n = ToolsLocator.getI18nManager();
        if (unitsNames == null) {
            String[] names = MapContext.getDistanceNames();
            unitsNames = new String[names.length];
            for (int i = 0; i < names.length; i++) {
                unitsNames[i] = i18n.getTranslation(names[i]);
                if (names[i].equals("Metros")) {
                    FACTORY_DEFAULT_DISTANCE_UNITS = i;
                    FACTORY_DEFAULT_DISTANCE_AREA = i;
                    FACTORY_DEFAULT_MAP_UNITS = i;
                }
            }
        }
        if( areaUnitsNames == null ) {
            String[] names = MapContext.getAreaNames();
            areaUnitsNames = new String[names.length];
            for (int i = 0; i < names.length; i++) {
                areaUnitsNames[i] = i18n.getTranslation(names[i])+MapContext.getOfLinear(i);
            }
        }
    }
}
