/**
 * gvSIG. Desktop Geographic Information System.
 *
 * Copyright (C) 2007-2013 gvSIG Association.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 *
 * For any additional information, do not hesitate to contact us
 * at info AT gvsig.com, or visit our website www.gvsig.com.
 */
package org.gvsig.app.gui.selectionByTheme;

import org.cresques.cts.IProjection;
import org.gvsig.andami.messages.NotificationManager;
import org.gvsig.fmap.dal.exception.DataException;
import org.gvsig.tools.dispose.DisposableIterator;
import org.gvsig.fmap.dal.feature.Feature;
import org.gvsig.fmap.dal.feature.FeatureQuery;
import org.gvsig.fmap.dal.feature.FeatureSelection;
import org.gvsig.fmap.dal.feature.FeatureSet;
import org.gvsig.fmap.dal.feature.FeatureStore;
import org.gvsig.fmap.geom.Geometry;
import org.gvsig.fmap.mapcontext.layers.FLayer;
import org.gvsig.fmap.mapcontext.layers.vectorial.FLyrVect;
import org.gvsig.fmap.mapcontext.layers.vectorial.SpatialEvaluatorsFactory;
import org.gvsig.tools.dispose.DisposeUtils;
import org.gvsig.tools.evaluator.Evaluator;
import org.gvsig.tools.evaluator.EvaluatorData;
import org.gvsig.tools.evaluator.EvaluatorException;
import org.gvsig.tools.evaluator.EvaluatorFieldsInfo;

public class MySelectionByThemeListener implements SelectionByThemeListener {

    @Override
    public void newSet(FLayer[] layersToSelect, FLayer selectionLayer, int action) {
        if (selectionLayer instanceof FLyrVect) {
            try {
                FeatureSelection selection = ((FLyrVect) selectionLayer).getFeatureStore().getFeatureSelection();
                SelectByTheme filter = new SelectByTheme(
                        selection,
                        selectionLayer.getMapContext().getProjection(),
                        action
                );

                for (FLayer layerToSelect : layersToSelect) {
                    if (layerToSelect instanceof FLyrVect) {
                        FeatureStore storeToQuery = ((FLyrVect) layerToSelect).getFeatureStore();
                        FeatureQuery fQuery = storeToQuery.createFeatureQuery();
                        filter.setData(storeToQuery, null);
                        fQuery.setFilter(filter);
                        FeatureSet featureSet = null;
                        try {
                            featureSet = storeToQuery.getFeatureSet(fQuery);
                            storeToQuery.getFeatureSelection().deselectAll();
                            storeToQuery.setSelection(featureSet);
                        } finally {
                            DisposeUtils.disposeQuietly(featureSet);
                        }
                    }
                }
            } catch (Exception e) {
                NotificationManager.addError("error_selection_by_theme", e);
            }
        }
    }

    @Override
    public void addToSet(FLayer[] layersToSelect, FLayer selectionLayer, int action) {
        if (selectionLayer instanceof FLyrVect) {
            try {
                FeatureSelection selection = ((FLyrVect) selectionLayer).getFeatureStore().getFeatureSelection();
                SelectByTheme filter = new SelectByTheme(
                        selection,
                        selectionLayer.getMapContext().getProjection(),
                        action
                );

                for (FLayer layerToSelect : layersToSelect) {
                    if (layerToSelect instanceof FLyrVect) {
                        if (!selectionLayer.getFullEnvelope().intersects(layerToSelect.getFullEnvelope())) {
                            continue;
                        }
                        FeatureStore storeToQuery = ((FLyrVect) layerToSelect).getFeatureStore();
                        FeatureQuery fQuery = storeToQuery.createFeatureQuery();
                        filter.setData(storeToQuery, null);
                        fQuery.setFilter(filter);
                        FeatureSet newFeatureSet = null;
                        FeatureSelection previousFeatureSet = null;
                        try {
                            newFeatureSet = storeToQuery.getFeatureSet(fQuery);
                            previousFeatureSet = storeToQuery.getFeatureSelection();
                            previousFeatureSet.select(newFeatureSet);
                            storeToQuery.setSelection(previousFeatureSet);
                        } finally {
                            DisposeUtils.disposeQuietly(newFeatureSet);
                            DisposeUtils.disposeQuietly(previousFeatureSet);
                        }
                    }
                }
            } catch (Exception e) {
                NotificationManager.addError("error_selection_by_theme", e);
            }
        }
    }

    @Override
    public void fromSet(FLayer[] layersToSelect, FLayer selectionLayer, int action) {
        if (selectionLayer instanceof FLyrVect) {
            try {
                FeatureSelection selection = ((FLyrVect) selectionLayer).getFeatureStore().getFeatureSelection();
                SelectByTheme filter = new SelectByTheme(
                        selection,
                        selectionLayer.getMapContext().getProjection(),
                        action
                );

                for (FLayer layerToSelect : layersToSelect) {
                    if (layerToSelect instanceof FLyrVect) {
                        if (!selectionLayer.getFullEnvelope().intersects(layerToSelect.getFullEnvelope())) {
                            continue;
                        }
                        FeatureStore storeToQuery = ((FLyrVect) layerToSelect).getFeatureStore();
                        FeatureQuery fQuery = storeToQuery.createFeatureQuery();
                        filter.setData(
                                storeToQuery,
                                (FeatureSelection) storeToQuery.getSelection()
                        );
                        fQuery.setFilter(filter);
                        FeatureSet featureSet = null;
                        try {
                            featureSet = storeToQuery.getFeatureSet(fQuery);
                            storeToQuery.getFeatureSelection().deselectAll();
                            storeToQuery.setSelection(featureSet);
                        } finally {
                            DisposeUtils.disposeQuietly(featureSet);
                        }
                    }
                }
            } catch (Exception e) {
                NotificationManager.addError("error_selection_by_theme", e);
            }
        }
    }

    class SelectByTheme implements Evaluator {

        private FeatureSelection selection = null;
        private final IProjection projection;
        private FeatureStore store;
        private String geomName;
        private EvaluatorFieldsInfo info;
        private FeatureSelection featureSelection;
        private final int action;

        
        public SelectByTheme(
                FeatureSelection selection,
                IProjection projection,
                int action
        ) {
            this.selection = selection;
            this.projection = projection;
            this.action = action;
        }

        public void setData(
                FeatureStore store,
                FeatureSelection featureSelection
            ) throws DataException {
            this.store = store;
            this.geomName = store.getDefaultFeatureType().getDefaultGeometryAttributeName();
            this.featureSelection = featureSelection;
            this.info = new EvaluatorFieldsInfo();
            this.info.addFieldValue(this.geomName);
        }

        @Override
        public Object evaluate(EvaluatorData data) throws EvaluatorException {
            if (featureSelection != null && !featureSelection.isSelected(((Feature) data.getContextValue("feature")))) {
                return false;
            }
            DisposableIterator features = null;
            try {
                features = selection.fastIterator();

                while (features.hasNext()) {
                    Feature feature = (Feature) features.next();
                    Geometry geometry = feature.getGeometry(geomName);
                    Evaluator eval;
                    switch (action) {
                        case SelectionByTheme.INTERSECTS:
                            eval = SpatialEvaluatorsFactory.getInstance().intersects(
                                    geometry,
                                    projection,
                                    store
                            );
                            break;
                        case SelectionByTheme.CONTAINS:
                            eval = SpatialEvaluatorsFactory.getInstance().contains(
                                    geometry,
                                    projection,
                                    store
                            );
                            break;
                        case SelectionByTheme.CROSSES:
                            eval = SpatialEvaluatorsFactory.getInstance().crosses(
                                    geometry,
                                    projection,
                                    store
                            );
                            break;
                        case SelectionByTheme.DISJOINT:
                            eval = SpatialEvaluatorsFactory.getInstance().disjoint(
                                    geometry,
                                    projection,
                                    store
                            );
                            break;
                        case SelectionByTheme.EQUALS:
                            eval = SpatialEvaluatorsFactory.getInstance().equals(
                                    geometry,
                                    projection,
                                    store
                            );
                            break;
                        case SelectionByTheme.OVERLAPS:
                            eval = SpatialEvaluatorsFactory.getInstance().overlaps(
                                    geometry,
                                    projection,
                                    store
                            );
                            break;
                        case SelectionByTheme.TOUCHES:
                            eval = SpatialEvaluatorsFactory.getInstance().touches(
                                    geometry,
                                    projection,
                                    store
                            );
                            break;
                        case SelectionByTheme.WITHIN:
                            eval = SpatialEvaluatorsFactory.getInstance().within(
                                    geometry,
                                    projection,
                                    store
                            );
                            break;

                        default:
                            eval = SpatialEvaluatorsFactory.getInstance().intersects(
                                    geometry,
                                    projection,
                                    store
                            );
                            break;
                    }

                    Boolean evaluate = (Boolean) eval.evaluate(data);
                    if ( evaluate) {
                        return evaluate;
                    }
                }
            } catch (DataException e) {
                NotificationManager.addError("error_selection_by_theme", e);
            } finally {
                DisposeUtils.dispose(features);
            }
            return false;
        }

        @Override
        public String getName() {
            return "select by theme";
        }

        @Override
        public String getSQL() {
            return null;
        }

        @Override
        public String getDescription() {
            return getName();
        }

        @Override
        public EvaluatorFieldsInfo getFieldsInfo() {
            return info;
        }
    }
}
