/*
 * Copyright (C) 2017 gvSIG Association
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.gvsig.app.project.documents.view.gui;

import java.awt.BorderLayout;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import javax.swing.Icon;
import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.JTabbedPane;
import org.gvsig.tools.swing.api.Component;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 *
 * @author jjdelcerro
 */
public class DefaultViewInformationArea implements ViewInformationArea {

    protected final Logger logger = LoggerFactory.getLogger(DefaultViewInformationArea.class);

    private static class Element implements ViewInformationAreaElement {
        
        private final Component component;
        private final String id;
        private String label;
        private Icon icon;
        private int priority;
        private String tip;
        
        public Element(Component component, String id, int priority, String label, Icon icon, String tip) {
            this.component = component;
            this.id = id;
            this.priority = priority;
            this.label = label;
            this.icon = icon;
            this.tip = tip;
        }
        
        @Override
        public int getPriority() {
            return this.priority;
        }

        @Override
        public Component getComponent() {
            return this.component;
        }
        
        @Override
        public JComponent asJComponent() {
            return this.component.asJComponent();
        }
        
        @Override
        public String getLabel() {
            return this.label;
        }
        
        @Override
        public void setLabel(String label) {
            this.label = label;
        }
        
        @Override
        public String getTip() {
            return this.tip;
        }
        
        @Override
        public Icon getIcon() {
            return this.icon;
        }

        @Override
        public void setIcon(Icon icon) {
            this.icon = icon;
        }

        @Override
        public void setPriority(int priority) {
            this.priority = priority;
        }

        @Override
        public void setTip(String tip) {
            this.tip = tip;
        }
    }
    
    private final Map<String,ViewInformationAreaElement> elements;
    private JPanel panel;
    private JTabbedPane tab;
    
    public DefaultViewInformationArea() {
        this.elements = new HashMap<>();
        this.panel = null;
        this.tab = null;
    }
    
    @Override
    public void add(final JComponent component, String id, int priority, String label, Icon icon, String tip) {
        this.add(
            new Component() {
                @Override
                public JComponent asJComponent() {
                    return component;
                }
            }, 
            id, priority, label, icon, tip
        );
    }

    @Override
    public void add(Component component, String id, int priority, String label, Icon icon, String tip) {
        Element element = new Element(component, id, priority, label, icon, tip);
        this.elements.put(id, element);
        this.update();
    }

    @Override
    public ViewInformationAreaElement get(String id) {
        ViewInformationAreaElement x = this.elements.get(id);
        return x;
    }

    @Override
    public Iterator<ViewInformationAreaElement> iterator() {
        Iterator<ViewInformationAreaElement> x = this.elements.values().iterator();
        return x;
    }
    
    @Override
    public void remove(String id) {
        this.elements.remove(id);
        this.update();
    }

    @Override
    public void update() {
        if( this.panel == null ) {
            return;
        }
        this.panel.removeAll();
        if( this.elements.isEmpty() ) {
            return;
        }
        List<Element> elements = new ArrayList(this.elements.values());
        if( elements.size()==1 ) {
            Element element = elements.get(0);
            this.panel.add(element.asJComponent(), BorderLayout.CENTER);
            return;
        }
        Collections.sort(elements, new Comparator<Element>() {
            @Override
            public int compare(Element o1, Element o2) {
                return Integer.compare(o1.getPriority(),o2.getPriority());
            }
        });
        JTabbedPane tab = new JTabbedPane();
        tab.setTabPlacement(JTabbedPane.BOTTOM);
        tab.setTabLayoutPolicy(JTabbedPane.SCROLL_TAB_LAYOUT);
        for( Element element : elements ) {
            tab.addTab(
                element.getLabel(), 
                element.getIcon(), 
                element.asJComponent(), 
                element.getTip()
            );
        }
        int index = 0;
        if( this.tab != null ) { 
            index = this.tab.getSelectedIndex();
        }
        if( index < 0 ) {
            index = 0;
        } else if( index >= tab.getTabCount() ) {
            index = tab.getTabCount()-1;
        }
        try {
            tab.setSelectedIndex(index);
        } catch(Exception ex) {
            logger.warn("Can't restore the selected tab", ex);
        }
        this.tab = tab;
        this.panel.add(this.tab, BorderLayout.CENTER);
    }
    
    @Override
    public JComponent asJComponent() {
        if( this.panel == null ) {
            this.panel = new JPanel();
            this.panel.setLayout(new BorderLayout());
            this.update();
        }
        return this.panel;
    }
    
}
