/**
 * gvSIG. Desktop Geographic Information System.
 *
 * Copyright (C) 2007-2013 gvSIG Association.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 *
 * For any additional information, do not hesitate to contact us
 * at info AT gvsig.com, or visit our website www.gvsig.com.
 */
package org.gvsig.app.project.documents.view.legend.gui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Collection;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.prefs.Preferences;
import javax.swing.Action;

import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JTextArea;
import javax.swing.JTree;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.MutableTreeNode;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.gvsig.andami.PluginServices;
import org.gvsig.andami.messages.NotificationManager;
import org.gvsig.fmap.dal.serverexplorer.filesystem.FilesystemServerExplorer;
import org.gvsig.fmap.dal.serverexplorer.filesystem.FilesystemServerExplorerParameters;
import static org.gvsig.fmap.dal.serverexplorer.filesystem.swing.FilesystemExplorerWizardPanel.OPEN_LAYER_FILE_CHOOSER_ID;
import org.gvsig.fmap.mapcontext.exceptions.LegendLayerException;
import org.gvsig.fmap.mapcontext.layers.FLayer;
import org.gvsig.fmap.mapcontext.layers.FLayers;
import org.gvsig.fmap.mapcontext.layers.operations.Classifiable;
import org.gvsig.fmap.mapcontext.layers.operations.ClassifiableVectorial;
import org.gvsig.fmap.mapcontext.layers.vectorial.FLyrVect;
import org.gvsig.fmap.mapcontext.rendering.legend.IClassifiedVectorLegend;
import org.gvsig.fmap.mapcontext.rendering.legend.ILegend;
import org.gvsig.fmap.mapcontext.rendering.legend.IVectorLegend;
import org.gvsig.symbology.SymbologyLocator;
import org.gvsig.symbology.swing.SymbologySwingLocator;
import org.gvsig.symbology.swing.SymbologySwingManager;
import org.gvsig.tools.ToolsLocator;
import org.gvsig.tools.folders.FoldersManager;

/**
 * Implements the panel which allows the user to control all the information
 * about the
 * legends of a layer in order to improve the information that it offers to the
 * user.
 * There are options to create, save or load an existing legend.
 *
 * @author jaume dominguez faus - jaume.dominguez@iver.es
 */
public class LegendManager extends AbstractThemeManagerPage implements LegendsPanel {

    private static final Logger logger =
        LoggerFactory.getLogger(LegendManager.class);

    private static final long serialVersionUID = 7989057553773181019L;

    public static class Pages implements Iterable<ILegendPanel> {

        private Map<Class<? extends ILegendPanel>, ILegendPanel> pages;

        public Pages() {
            this.pages = new HashMap<>();
        }
        
        public void add(ILegendPanel page) {
            this.pages.put(page.getClass(), page);
        }

        public Iterator<ILegendPanel> iterator() {
            return this.pages.values().iterator();
        }

        public boolean contains(Class<? extends ILegendPanel> pageClass) {
            return this.pages.containsKey(pageClass);
        }

        public ILegendPanel get(Class<? extends ILegendPanel> pageClass) {
            return this.pages.get(pageClass);
        }

        public Collection<ILegendPanel> asCollection() {
            return this.pages.values();
        }
    }

    private FLayer layer;
    private ILegend legend; 
    private Pages pages = new Pages();
    
    private JPanel topPanel = null;
    private JTextArea titleArea = null;
    private JPanel preview = null;
    private JScrollPane jTitleScrollPane = null;
    private JTree jTreeLegends;
    private ILegendPanel activePanel;
    private JScrollPane legendTreeScrollPane;
//    private boolean dirtyTree_;
    private final DefaultMutableTreeNode root = new DefaultMutableTreeNode();
    private DefaultTreeModel treeModel;
    private JScrollPane jPanelContainer;
    private JPanel jCentralPanel;
    private JSplitPane jSplitPane;
    private boolean isTreeListenerDisabled;
    private JButton btnOptionalActions;
    private JPopupMenu menuOptionalActions;
    private Hashtable<FLayer, ILegend> table = null;
    private boolean empty = true;
    private JLabel iconLabel;

    public static String defaultLegendFolderPath;
    {
        FoldersManager folderManager = ToolsLocator.getFoldersManager();
        defaultLegendFolderPath = folderManager.getLastPath(
                "LegendsFolder", 
                folderManager.getHome()
        ).getAbsolutePath();
    }


    public LegendManager() {
        initialize();
    }

    private void initialize() {
        setLayout(new BorderLayout());
        add(getTopPanel(), BorderLayout.NORTH);
        add(getSplitPane(), BorderLayout.CENTER);
        setSize(500, 360);
        treeModel = new DefaultTreeModel(root);
    }

    private JSplitPane getSplitPane() {
        if (jSplitPane == null) {
            jSplitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT);
            JPanel aux = new JPanel(new BorderLayout(0, 5));
            aux.add(getLegendTreeScrollPane(), BorderLayout.CENTER);
            aux.add(getPreviewPanel(), BorderLayout.SOUTH);
            jSplitPane.setLeftComponent(aux);
            jSplitPane.setRightComponent(getCentralPanel());
            jSplitPane.setDividerLocation(150);
        }
        return jSplitPane;
    }

    private JPanel getCentralPanel() {
        if (jCentralPanel == null) {
            jCentralPanel = new JPanel(new BorderLayout(0, 10));
            jCentralPanel.add(getTitleScroll(), BorderLayout.NORTH);
            jCentralPanel.add(getJPanelContainer(), BorderLayout.CENTER);
        }
        return jCentralPanel;
    }

    private JScrollPane getJPanelContainer() {
        if (jPanelContainer == null) {
            jPanelContainer = new JScrollPane();
        }
        return jPanelContainer;
    }

    /**
     * This method initializes jPanel
     *
     * @return javax.swing.JPanel
     */
    private JPanel getTopPanel() {
        if (topPanel == null) {
            topPanel = new JPanel(new FlowLayout(FlowLayout.RIGHT, 5, 0));
            topPanel.setPreferredSize(new Dimension(638, 31));
            topPanel.setBorder(javax.swing.BorderFactory.createTitledBorder(
                null, "",
                javax.swing.border.TitledBorder.DEFAULT_JUSTIFICATION,
                javax.swing.border.TitledBorder.DEFAULT_POSITION, null, null));
            topPanel.add(getBtnOptionalActions(), null);
        }
        return topPanel;
    }

    private JButton getBtnOptionalActions() {
        if (btnOptionalActions == null) {
            btnOptionalActions =
                new JButton(PluginServices.getText(this, "_Mas_opciones")
                    + "...");
            btnOptionalActions.addActionListener(new ActionListener() {
                @Override
                public void actionPerformed(ActionEvent e) {
                    Point p = btnOptionalActions.getLocationOnScreen();
                    menuOptionalActions.show(asJComponent(),0,0);
                    menuOptionalActions.setLocation(p.x,p.y+btnOptionalActions.getHeight());
                }
            });
        }
        return btnOptionalActions;
    }

    /**
     * This method initializes jTextArea
     *
     * @return javax.swing.JTextArea
     */
    private JTextArea getTitleArea() {
        if (titleArea == null) {
            titleArea = new JTextArea();
            titleArea.setBackground(java.awt.SystemColor.control);
            titleArea.setLineWrap(true);
            titleArea.setRows(0);
            titleArea.setWrapStyleWord(false);
            titleArea.setEditable(false);
            titleArea.setPreferredSize(new java.awt.Dimension(495, 40));
        }
        return titleArea;
    }

    /**
     * This method initializes jPanel1
     *
     * @return javax.swing.JPanel
     */
    private JPanel getPreviewPanel() {
        if (preview == null) {
            preview = new JPanel();
            preview.setBorder(javax.swing.BorderFactory
                .createBevelBorder(javax.swing.border.BevelBorder.LOWERED));
            preview.setBackground(java.awt.SystemColor.text);
            preview.setLayout(new BorderLayout(5, 5));
            preview.add(getIconLabel());
            preview.setPreferredSize(new Dimension(getSplitPane()
                .getDividerLocation(), 130));
            preview.setBackground(Color.white);
        }
        return preview;
    }

    private JLabel getIconLabel() {
        if (iconLabel == null) {
            iconLabel = new JLabel();
            iconLabel.setVerticalAlignment(JLabel.CENTER);
            iconLabel.setHorizontalAlignment(JLabel.CENTER);
        }

        return iconLabel;
    }

    /**
     * This method initializes jScrollPane
     *
     * @return javax.swing.JScrollPane
     */
    private JScrollPane getLegendTreeScrollPane() {
        if (legendTreeScrollPane == null) {
            legendTreeScrollPane = new JScrollPane();
            legendTreeScrollPane.setViewportView(getJTreeLegends());
        }
        return legendTreeScrollPane;
    }

    /**
     * <p>
     * Adds a new fully-featured legend panel to the LegendManager.<br>
     * </p>
     *
     * <p>
     * <b>CAUTION:</b> Trying to add a child page whose parent hasn't been added
     * yet causes the application to fall in an infinite loop. This is a known
     * bug, sorry. Just avoid this case or try to fix it (lol).<br>
     * </p>
     *
     * @deprecated use SymbologySwingManager.registerLegendEditor
     */
    public static void addLegendPage(Class<? extends ILegendPanel> iLegendPanelClass) {
        SymbologySwingManager manager = SymbologySwingLocator.getSwingManager();
        manager.registerLegendEditor(iLegendPanelClass);
    }

    /**
     * Causes the component to be autofilled with the legend pages that
     * were added through the static method addLegendPage(ILegendPanel page)
     */
    private void fillDialog() {
        if (empty) {
            SymbologySwingManager manager = SymbologySwingLocator.getSwingManager();

            Iterator<ILegendPanel> it = manager.getLegendEditors(layer).iterator();
            while( it.hasNext() ) {
                ILegendPanel page = it.next();
                pages.add(page);
            }
            addOptionalActions();
            for(ILegendPanel page : this.pages ) {
                doInsertNode(treeModel, page);
            }
            getJTreeLegends().setModel(treeModel);
            getJTreeLegends().repaint();
            empty = false;
        }
    }

    private void addOptionalActions() {
        this.menuOptionalActions = new JPopupMenu();
        Iterable<Action> actions = SymbologySwingLocator.getSwingManager().getOptionalActionOfLegendsPanel();
        for( Action action : actions ) {
            JMenuItemForOptionalAction item = new JMenuItemForOptionalAction(action);
            this.menuOptionalActions.add(item);
        }
    }
    
    private class JMenuItemForOptionalAction extends JMenuItem implements ActionListener {

        private static final long serialVersionUID = 1656264978338543368L;
        Action action;
        
        JMenuItemForOptionalAction(Action action) {
            this.action = action;
            this.configurePropertiesFromAction(action);
            this.addActionListener(this);
        }
        
        @Override
        public void actionPerformed(ActionEvent e) {
            e.setSource(LegendManager.this);
            this.action.actionPerformed(e);
        }
    }
    
    @SuppressWarnings("unchecked")
    private DefaultMutableTreeNode findNode(Class searchID) {
        String title;
        try {
            title =
                ((ILegendPanel) Class.forName(searchID.getName()).newInstance())
                    .getTitle();
        } catch (Exception e) {
            // this should be impossible, but anyway this case will be treat as
            // the node does not
            // exist.
            return null;
        }

        Enumeration e = root.breadthFirstEnumeration();
        while (e.hasMoreElements()) {
            DefaultMutableTreeNode nodeAux =
                (DefaultMutableTreeNode) e.nextElement();
            if (nodeAux != null) {
                ILegendPanel legend = (ILegendPanel) nodeAux.getUserObject();
                if (legend == null)
                    continue; // Root node
                if (legend.getTitle().equals(title)) {
                    return nodeAux;
                }
            }
        }
        return null;
    }

    /**
     * If parent_node == null, add to root.
     * Returns added node
     *
     * @param tm
     * @param parent_node
     * @param item
     */
    private MutableTreeNode insertNodeHere(
        DefaultTreeModel tm,
        MutableTreeNode parent_node,
        ILegendPanel item) {

        MutableTreeNode pn = null;

        if (parent_node == null) {
            pn = root;
        } else {
            pn = parent_node;
        }

        DefaultMutableTreeNode nodeValue = new DefaultMutableTreeNode(item);
        int children = pn.getChildCount();
        int pos = 0;
        String pTitle = null;

        for (int i = 0; i < children; i++) {
            DefaultMutableTreeNode node =
                (DefaultMutableTreeNode) tm.getChild(pn, i);
            if (node.getUserObject() instanceof ILegendPanel) {
                pTitle = ((ILegendPanel) node.getUserObject()).getTitle();
                if (pTitle.compareTo(item.getTitle()) < 0) {
                    pos++;
                }
            }
        }
        tm.insertNodeInto(nodeValue, pn, pos);
        return nodeValue;
    }

    /**
     * Returns inserted node
     *
     * @param tm
     * @param page
     * @return
     */
    private MutableTreeNode doInsertNode(DefaultTreeModel tm, ILegendPanel page) {

        if (tm == null || page == null) {
            return null;
        }

        MutableTreeNode aux = findNode(page.getClass());
        if (aux != null) {
            return aux;
        }

        Class parent_class = page.getParentClass();

        if (parent_class != null) {
//            if (pages.containsKey(parent_class)) {
            if (pages.contains(parent_class)) {
                ILegendPanel parent = (ILegendPanel) pages.get(parent_class);

                aux = doInsertNode(tm, parent);
                if (aux != null) {
                    return insertNodeHere(tm, aux, page);
                } else {
                    return null;
                }

            } else {
                return null;
            }
        } else {
            // add to root
            return insertNodeHere(tm, null, page);
        }


    }





    private JScrollPane getTitleScroll() {
        if (jTitleScrollPane == null) {
            jTitleScrollPane = new JScrollPane();
            jTitleScrollPane.setBounds(2, 2, 498, 42);
            jTitleScrollPane.setViewportView(getTitleArea());
        }
        return jTitleScrollPane;
    }

    private JTree getJTreeLegends() {
        if (jTreeLegends == null) {
            jTreeLegends = new JTree();
            jTreeLegends.setRootVisible(false);
            MyTreeCellRenderer treeCellRenderer = new MyTreeCellRenderer();
            treeCellRenderer.setOpenIcon(null);
            treeCellRenderer.setClosedIcon(null);
            treeCellRenderer.setLeafIcon(null);

            jTreeLegends.setCellRenderer(treeCellRenderer);
            jTreeLegends.setShowsRootHandles(true);
            jTreeLegends
                .addTreeSelectionListener(new javax.swing.event.TreeSelectionListener() {

                    public void valueChanged(
                        javax.swing.event.TreeSelectionEvent e) {
                        if (isTreeListenerDisabled)
                            return;
                        DefaultMutableTreeNode node =
                            (DefaultMutableTreeNode) jTreeLegends
                                .getLastSelectedPathComponent();

                        if (node == null)
                            return;
                        setActivePage((ILegendPanel) node.getUserObject());
                    }
                });
            jTreeLegends.putClientProperty("JTree.linestyle", "Angled");
            jTreeLegends.getSelectionModel().setSelectionMode(
                TreeSelectionModel.SINGLE_TREE_SELECTION);
        }
        return jTreeLegends;
    }

    @Override
    public void setActivePage(ILegendPanel page) {
        if (page.getPanel() == null) {
            // this is what happens when the user clicked in a parent node
            // which only acts as a folder, and does not manage any legend
            // then it expands and selects the first child
            DefaultMutableTreeNode node = findNode(page.getClass());
            if (treeModel.getChildCount(node) > 0) {
                DefaultMutableTreeNode dmn =
                    (DefaultMutableTreeNode) treeModel.getChild(node, 0);
                page = (ILegendPanel) dmn.getUserObject();
                setActivePage(page);
                expandAndSelect(page);
            }
        } else {
            // show the page
            activePanel = page;
            setIcon(activePanel.getIcon());

            activePanel.setData(layer, legend);
            getTitleArea().setText(activePanel.getDescription());
            jPanelContainer.setViewportView(activePanel.getPanel());
        }
    }

    @Override
    public ILegendPanel getActivePage() {
        throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
    }

    private void setIcon(ImageIcon icon) {
        getIconLabel().setIcon(icon);
    }

    private class MyTreeCellRenderer extends DefaultTreeCellRenderer {

        private static final long serialVersionUID = -6013698992263578041L;

        public MyTreeCellRenderer() {
        }

        public Component getTreeCellRendererComponent(JTree tree, Object value,
            boolean sel, boolean expanded, boolean leaf, int row,
            boolean hasFocus) {

            super.getTreeCellRendererComponent(tree, value, sel, expanded,
                leaf, row, hasFocus);
            if (value instanceof DefaultMutableTreeNode) {
                DefaultMutableTreeNode node = (DefaultMutableTreeNode) value;
                if (node.getUserObject() instanceof ILegendPanel) {
                    ILegendPanel legend = (ILegendPanel) node.getUserObject();
                    this
                        .setText(legend.getPanel() == null ? "<html><b>"
                            + legend.getTitle() + "</b></html>" : legend
                            .getTitle());
                }
            }
            return this;
        }

    }

    private void expandAndSelect(Object node) {
        isTreeListenerDisabled = true;
        // will expand the tree and select the node
        int i = 0;
        boolean exit = false;

        TreePath tp = null;
        // find the page in the tree
        while (i < jTreeLegends.getRowCount() && !exit) {
            // see if this row is the node that we are looking for

            tp = jTreeLegends.getPathForRow(i);
            Object[] obj = tp.getPath();
            for (int j = 0; j < obj.length && !exit; j++) {
                Object o = ((DefaultMutableTreeNode) obj[j]).getUserObject();

                if (o != null && o.getClass().equals(node.getClass())
                    && o.equals(node)) {
                    // found it! collapse the tree
                    while (i >= 0) {
                        jTreeLegends.collapseRow(i);
                        i--;
                    }
                    exit = true;
                }
            }
            jTreeLegends.expandRow(i);
            i++;
        }

        // expand the tree and set the selection
        if (tp != null) {
            jTreeLegends.expandPath(tp);
            jTreeLegends.setSelectionPath(tp);
        }
        isTreeListenerDisabled = false;
    }

    public String getName() {
        return PluginServices.getText(this, "Simbologia");
    }

    public void acceptAction() {
        applyAction();
        // automatically handled by the ThemeManagerWindow
    }

    public void cancelAction() {
        // does nothing
    }

    public void applyAction() {
        legend = activePanel.getLegend();

        if (table != null && table.size() > 1)
            applyRestOfLegends(table, layer.getMapContext().getLayers());

        /*
         * try to apply the legend to all the active layers that
         * can accept it
         */
        FLayer[] activeLyrs = layer.getMapContext().getLayers().getActives();
        for (int i = 0; i < activeLyrs.length; i++) {
            FLayer laux = activeLyrs[i];

            if (activeLyrs[i] instanceof FLayers) {
                laux = getFirstActiveLayerVect((FLayers) activeLyrs[i]);
            }

            if (laux instanceof ClassifiableVectorial) {
                ClassifiableVectorial aux2 = (ClassifiableVectorial) laux;
                try {
                    if (legend instanceof IClassifiedVectorLegend) {
                        // Es una leyenda que necesita un recordset con un
                        // nombre de campo. Comprobamos que ese recordset
                        // tiene ese nombre de campo y es del tipo esperado
                        IClassifiedVectorLegend cl =
                            (IClassifiedVectorLegend) legend;

                        if (aux2 instanceof FLyrVect) {

                            if (cl.getValues().length == 0) {
                                JOptionPane.showMessageDialog(
                                    (Component) PluginServices.getMainFrame(),
                                    PluginServices.getText(this,
                                        "no_es_posible_aplicar_leyenda_vacia"));
                                return;
                            }

                            aux2.setLegend((IVectorLegend) legend);
                        }
                    } else
                        if (legend instanceof IVectorLegend) {
                            aux2.setLegend((IVectorLegend) legend);
                        }
                } catch (LegendLayerException e) {
                    NotificationManager.addError(PluginServices.getText(this,
                        "legend_exception"), e);
                }
            }
        }
    }

    private void applyRestOfLegends(Hashtable<FLayer, ILegend> table2,
        FLayers layers) {

        for (int i = 0; i < layers.getLayersCount(); i++) {
            FLayer my_layer = layers.getLayer(i);

            if (!(my_layer instanceof FLayers)) {
                if (my_layer instanceof ClassifiableVectorial) {
                    try {
                        if (table.containsKey(my_layer)) {
                            ClassifiableVectorial lyr =
                                (ClassifiableVectorial) my_layer;
                            lyr.setLegend((IVectorLegend) table.get(my_layer));
                        }

                    } catch (LegendLayerException e) {
                        // TODO Auto-generated catch block
                        e.printStackTrace();
                    }
                }
            } else
                applyRestOfLegends(table, (FLayers) my_layer);
        }
    }

    @Override
    public void setModel(FLayer layer) {
        this.layer = layer;
        applyLegend(((Classifiable) layer).getLegend());
    }

    /**
     * Applies the legend to the layer.
     *
     * @param aLegend
     *            , legend that the user wants to apply
     */
    private void applyLegend(ILegend aLegend) {
        this.legend = aLegend;
        fillDialog();
        for(ILegendPanel page : this.pages ) {
//        Enumeration<Class<? extends ILegendPanel>> en = pages.keys();
//        while (en.hasMoreElements()) {
//            ILegendPanel page = (ILegendPanel) pages.get(en.nextElement());
            if (legend.getClass().equals(page.getLegendClass())) {
                setActivePage(page);
                expandAndSelect(page);
                return;
            }
        }
        NotificationManager.addWarning(PluginServices.getText(this,
            "caution_no_registered_panel_associated_to_"
                + "loaded_legend_the_legend_wont_be_applied"));
    }

    @Override
    public int getPriority() {
    	return 800;
    }

    @Override
    public ILegend getLegend() {
        return this.legend;
    }

    @Override
    public void setLegend(ILegend legend) {
        applyLegend(legend);
    }

    @Override
    public FLayer getLayer() {
        return this.layer;
    }

    @Override
    public void setLayer(FLayer layer) {
        this.layer = layer;
        applyLegend(((Classifiable) layer).getLegend());
    }

    @Override
    public ILegendPanel getPage(Class<? extends ILegendPanel> pageClass) {
        return this.pages.get(pageClass);
    }

    @Override
    public Collection<ILegendPanel> getPages() {
        return this.pages.asCollection();
    }
    
}
