/**
 * gvSIG. Desktop Geographic Information System.
 *
 * Copyright (C) 2007-2013 gvSIG Association.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * For any additional information, do not hesitate to contact us at info AT
 * gvsig.com, or visit our website www.gvsig.com.
 */
package org.gvsig.coreplugin;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.prefs.Preferences;
import org.apache.commons.io.FileUtils;
import org.apache.commons.lang3.StringUtils;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.gvsig.andami.IconThemeHelper;
import org.gvsig.andami.Launcher;
import org.gvsig.andami.PluginServices;
import org.gvsig.andami.PluginsLocator;
import org.gvsig.andami.PluginsManager;
import org.gvsig.andami.plugins.Extension;
import org.gvsig.andami.plugins.config.generate.Extensions;
import org.gvsig.andami.plugins.config.generate.PluginConfig;
import org.gvsig.andami.preferences.DlgPreferences;
import org.gvsig.andami.preferences.IPreference;
import org.gvsig.andami.preferences.IPreferenceExtension;
import org.gvsig.coreplugin.preferences.general.BrowserControlPage;
import org.gvsig.coreplugin.preferences.general.ExtensionPage;
import org.gvsig.coreplugin.preferences.general.ExtensionsPage;
import org.gvsig.coreplugin.preferences.general.FolderingPage;
import org.gvsig.coreplugin.preferences.general.GeneralPage;
import org.gvsig.coreplugin.preferences.general.IconThemePage;
import org.gvsig.coreplugin.preferences.general.NotificationsPage;
import org.gvsig.coreplugin.preferences.general.ScreenSettingsPage;
import org.gvsig.coreplugin.preferences.general.SkinPreferences;
import org.gvsig.coreplugin.preferences.general.appearance.AppearancePage;
import org.gvsig.coreplugin.preferences.network.FirewallPage;
import org.gvsig.coreplugin.preferences.network.NetworkPage;
import org.gvsig.tools.ToolsLocator;
import org.gvsig.tools.dynobject.DynObject;
import org.gvsig.tools.extensionpoint.ExtensionPoint;
import org.gvsig.tools.extensionpoint.ExtensionPointManager;
import org.gvsig.tools.folders.FoldersManager;
import org.gvsig.tools.i18n.I18nManager;
import org.gvsig.tools.util.ToolsUtilLocator;
import org.gvsig.utils.XMLEntity;

/**
 *
 * <p>
 * Extension that provides support for visual application configuration through
 * a Preferences dialog where the user can specify its own settings for general
 * purpose aspects.
 * </p>
 * <p>
 * Adding new preference pages is made through ExtensionPoints by invoking
 * <b>
 * this.extensionPoints.add("AplicationPreferences","YourIPreferencesClassName",
 * yourIPreferencesPage);
 * </b>
 * and then call <b>DlgPreferences.refreshExtensionPoints();</b>
 * </p>
 *
 *
 */
public class PreferencesExtension extends Extension {

	public static final String PROJECTS_FOLDER_PROPERTY_NAME = "ProjectsFolder";
	public static final String DATA_FOLDER_PROPERTY_NAME = "DataFolder";
	public static final String TEMPLATES_FOLDER_PROPERTY_NAME = "TemplatesFolder";
	public static final String SYMBOL_LIBRARY_FOLDER_PROPERTY_NAME = "SymbolLibraryFolder";
    
    private static final Logger logger = LoggerFactory.getLogger(PreferencesExtension.class);
    private final ExtensionPointManager extensionPoints = ToolsLocator.getExtensionPointManager();

    private boolean initilizedExtensions = false;

    @Override
    public void initialize() {
        initializeUseNativeFileChooser();
        initializeFoldersManager();
        registerIcons();
        initializeCoreExtensions();
    }

    private void initializeUseNativeFileChooser() {
        PluginsManager pluginsManager = PluginsLocator.getManager();
        PluginServices plugin = pluginsManager.getPlugin(PreferencesExtension.class);

        DynObject pluginProperties = plugin.getPluginProperties();

        boolean useNativeFileChooser = (Boolean) pluginProperties.getDynValue("useNativeFileChooser");
        ToolsUtilLocator.getFileDialogChooserManager().setUseNativeFileChooser(useNativeFileChooser);
    }

    private void initializeFoldersManager() {
        try {
            Preferences prefs = Preferences.userRoot().node( "gvsig.foldering" );
            FoldersManager foldersManager = ToolsLocator.getFoldersManager();
            for( String name : new String[] {
                    PROJECTS_FOLDER_PROPERTY_NAME,
                    DATA_FOLDER_PROPERTY_NAME,
                    TEMPLATES_FOLDER_PROPERTY_NAME,
                    SYMBOL_LIBRARY_FOLDER_PROPERTY_NAME
                }) {
                String fname = prefs.get(name, null);
                if( !StringUtils.isEmpty(fname) ) {
                    foldersManager.set(name, new File(fname).getAbsoluteFile());
                }
            }
        } catch(Exception ex) {
            logger.warn("Can't initialize default folders in FolderManager.",ex);
        }
    }
    
    private void registerIcons() {
        IconThemeHelper.registerIcon("action", "edit-setup", this);
        IconThemeHelper.registerIcon("action", "application-exit", this);

        // AppearancePage.java
        IconThemeHelper.registerIcon("preferences", "edit-setup-appearance", this);
        // BrowserControlPage.java
        IconThemeHelper.registerIcon("preferences", "edit-setup-browsercontrol", this);
        // DirExtensionPage.java
        IconThemeHelper.registerIcon("preferences", "edit-setup-dirextension", this);
        // ExtensionPage.java
        IconThemeHelper.registerIcon("preferences", "edit-setup-extension", this);
        // ExtensionsPage.java
        IconThemeHelper.registerIcon("preferences", "edit-setup-extensions", this);
        // GeneralPage.java
        IconThemeHelper.registerIcon("preferences", "edit-setup-general", this);
        // FolderingPage.java
        IconThemeHelper.registerIcon("preferences", "edit-setup-foldering", this);
        // LanguagePage.java
        IconThemeHelper.registerIcon("preferences", "edit-setup-language", this);
        // ScreenSettingsPage.java
        IconThemeHelper.registerIcon("preferences", "edit-setup-screensetting", this);
        // FirewallPage.java
        IconThemeHelper.registerIcon("preferences", "edit-setup-firewall", this);
        // NetworkPage.java
        IconThemeHelper.registerIcon("preferences", "edit-setup-network", this);
        // SkinPreferences.java
        IconThemeHelper.registerIcon("preferences", "edit-setup-skin", this);
        IconThemeHelper.registerIcon("preferences", "edit-setup-icontheme", this);
    }

    @Override
    public void execute(String actionCommand) {
        if (!this.initilizedExtensions) {
            initializeExtensions();
            initializeExtensionsConfig();
            this.initilizedExtensions = true;
        }
        if ("edit-setup".equalsIgnoreCase(actionCommand)) {
            DlgPreferences dlgPreferences = PluginServices.getDlgPreferences();
            dlgPreferences.refreshExtensionPoints();
            PluginServices.getMDIManager().addWindow(dlgPreferences);
        } else if ("application-exit".equalsIgnoreCase(actionCommand)) {
            Launcher.closeApplication();
        }
    }

    @Override
    public boolean isEnabled() {
        return true;
    }

    @Override
    public boolean isVisible() {
        return true;
    }

    private void initializeCoreExtensions() {
        addPreferencesPage(new GeneralPage());
        addPreferencesPage(new NetworkPage());
        addPreferencesPage(new FirewallPage());
        addPreferencesPage(new ExtensionsPage());
        addPreferencesPage(new AppearancePage());
        addPreferencesPage(new FolderingPage());
        addPreferencesPage(new ScreenSettingsPage());
        addPreferencesPage(new SkinPreferences());
        addPreferencesPage(new IconThemePage());
        addPreferencesPage(new NotificationsPage());
        String os = System.getProperty("os.name").toLowerCase();
        if (os.contains("linux") || os.contains("unix")) {
            addPreferencesPage(new BrowserControlPage());
        }
    }

    private void initializeExtensionsConfig() {
        HashMap pc = Launcher.getPluginConfig();
        ArrayList array = new ArrayList();
        Iterator iter = pc.values().iterator();

        while (iter.hasNext()) {
            array.add(((PluginConfig) iter.next()).getExtensions());
        }
        Extensions[] exts = (Extensions[]) array.toArray(new Extensions[0]);
        for (Extensions ext1 : exts) {
            for (int j = 0; j < ext1.getExtensionCount(); j++) {
                org.gvsig.andami.plugins.config.generate.Extension ext = ext1.getExtension(j);
                addPreferencesPage(new ExtensionPage(ext));
            }
        }
    }

    private void initializeExtensions() {
        Iterator i = Launcher.getExtensionIterator();
        I18nManager i18nManager = ToolsLocator.getI18nManager();

        while (i.hasNext()) {
            Object extension = i.next();

            if (extension instanceof IPreferenceExtension) {
                IPreferenceExtension pe = (IPreferenceExtension) extension;
                IPreference[] pp;
                try {
                    pp = pe.getPreferencesPages();
                } catch (Throwable th) {
                    logger.warn("Can't get preference pages from '"+extension.getClass().getName()+"'.",th);
                    continue;
                }
                if (pp == null) {
                    String msg = "Preferences page is NULL. Extension "
                            + extension.getClass().getName()
                            + " implements IPreferenceExtension but does not provide pref. pages. ";
                    logger.info(msg);
                } else {
                    for (IPreference pp1 : pp) {
                        try {
                            addPreferencesPage(pp1);
                        } catch (Throwable th) {
                            String msg = "cant_initialize_values_for_{0}_page_in_{1}";
                            String[] params = {pp1.getTitle(), extension.getClass().getName()};
                            logger.warn(i18nManager.getTranslation(msg, params), th);
                        }
                    }
                }
            }
        }
    }

    public void addPreferencesPage(IPreference pp) {
        ExtensionPoint ep = this.extensionPoints.add("AplicationPreferences", "");
        pp.initializeValues();
        ep.append(pp.getID(), "", pp);
    }

    @Override
    public void postInitialize() {
        super.postInitialize();
        DlgPreferences dlgPreferences = PluginServices.getDlgPreferences();
        dlgPreferences.refreshExtensionPoints();
        dlgPreferences.storeValues();
    }
}
