package org.gvsig.geodb.databaseworkspace;

import org.gvsig.fmap.dal.DatabaseWorkspaceManager;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import org.apache.commons.lang.StringUtils;
import org.gvsig.featureform.swing.JFeaturesForm;
import org.gvsig.fmap.dal.DALLocator;
import org.gvsig.fmap.dal.DataManager;
import org.gvsig.fmap.dal.DataStoreParameters;
import org.gvsig.fmap.dal.StoresRepository;
import org.gvsig.fmap.dal.feature.FeatureStore;
import org.gvsig.fmap.dal.store.jdbc.JDBCServerExplorerParameters;
import org.gvsig.fmap.dal.swing.DALSwingLocator;
import org.gvsig.fmap.dal.swing.DataSwingManager;
import org.gvsig.fmap.dal.swing.searchpanel.FeatureStoreSearchPanel;
import static org.gvsig.fmap.dal.DatabaseWorkspaceManager.CONFIG_NAME_STORESREPOSITORYID;
import static org.gvsig.fmap.dal.DatabaseWorkspaceManager.CONFIG_NAME_STORESREPOSITORYLABEL;
import org.gvsig.tools.swing.api.Component;
import org.gvsig.tools.swing.api.ToolsSwingLocator;
import org.gvsig.tools.swing.api.pickercontroller.PickerController;
import org.gvsig.tools.swing.api.windowmanager.Dialog;
import org.gvsig.tools.swing.api.windowmanager.WindowManager;
import org.gvsig.tools.swing.api.windowmanager.WindowManager_v2;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 *
 * @author jjdelcerro
 */
@SuppressWarnings("UseSpecificCatch")
public class WorkspaceConfigurationPanel
        extends WorkspaceConfigurationPanelView
        implements Component
    {

    private static final Logger LOGGER = LoggerFactory.getLogger(WorkspaceConfigurationPanel.class);

    private class TableController {

        private final JButton btnCreate;
        private final JButton btnDrop;
        private final JButton btnSearch;
        private final JButton btnShowForm;
        private final int tableid;

        public TableController(
                int tableid,
                JButton btnCreate,
                JButton btnDrop,
                JButton btnSearch,
                JButton btnShowForm
        ) {
            this.tableid = tableid;
            this.btnCreate = btnCreate;
            this.btnDrop = btnDrop;
            this.btnSearch = btnSearch;
            this.btnShowForm = btnShowForm;
            this.btnCreate.addActionListener(new ActionListener() {
                @Override
                public void actionPerformed(ActionEvent e) {
                    doCreate();
                }
            });
            this.btnDrop.addActionListener(new ActionListener() {
                @Override
                public void actionPerformed(ActionEvent e) {
                    doDrop();
                }
            });
            this.btnShowForm.addActionListener(new ActionListener() {
                @Override
                public void actionPerformed(ActionEvent e) {
                    doShowForm();
                }
            });
            this.btnSearch.addActionListener(new ActionListener() {
                @Override
                public void actionPerformed(ActionEvent e) {
                    doSearch();
                }
            });
        }

        private void doCreate() {
            DataManager dataManager = DALLocator.getDataManager();
            DatabaseWorkspaceManager workspace = dataManager.createDatabaseWorkspaceManager(getConnection());
            if (workspace.existsTable(this.tableid)) {
                return;
            }
            workspace.createTable(this.tableid);
        }

        private void doShowForm() {
            try {
                DataManager dataManager = DALLocator.getDataManager();
                DatabaseWorkspaceManager workspace = dataManager.createDatabaseWorkspaceManager(getConnection());
                if (!workspace.existsTable(this.tableid)) {
                    return;
                }
                FeatureStore store = workspace.getTable(this.tableid);
                DataSwingManager dalSwingManager = DALSwingLocator.getSwingManager();
                JFeaturesForm form = dalSwingManager.createJFeaturesForm(store);
                WindowManager winManager = ToolsSwingLocator.getWindowManager();
                winManager.showWindow(
                        form.asJComponent(),
                        "Resources",
                        WindowManager.MODE.WINDOW
                );
            } catch (Exception ex) {
                LOGGER.warn("Can't show form of resources table.", ex);
            }
        }

        private void doDrop() {
            DataManager dataManager = DALLocator.getDataManager();
            DatabaseWorkspaceManager workspace = dataManager.createDatabaseWorkspaceManager(getConnection());
            if (!workspace.existsTable(this.tableid)) {
                return;
            }
            workspace.dropTable(this.tableid);
        }

        private void doSearch() {
            try {
                DataManager dataManager = DALLocator.getDataManager();
                DatabaseWorkspaceManager workspace = dataManager.createDatabaseWorkspaceManager(getConnection());
                if (!workspace.existsTable(this.tableid)) {
                    return;
                }
                FeatureStore store = workspace.getTable(this.tableid);
                DataSwingManager dalSwingManager = DALSwingLocator.getSwingManager();
                FeatureStoreSearchPanel search = dalSwingManager.createFeatureStoreSearchPanel(store);
                WindowManager winManager = ToolsSwingLocator.getWindowManager();
                winManager.showWindow(
                        search.asJComponent(),
                        "Resources",
                        WindowManager.MODE.WINDOW
                );
            } catch (Exception ex) {
                LOGGER.warn("Can't show search panel for resources table.", ex);
            }
        }

        public void setEnabled(boolean enabled) {
            this.btnCreate.setEnabled(enabled);
            this.btnDrop.setEnabled(enabled);
            this.btnSearch.setEnabled(enabled);
            this.btnShowForm.setEnabled(enabled);
        }
    }

    private TableController resourcesTableController;
    private TableController repositoryTableController;
    private TableController configurationTableController;
    private PickerController<JDBCServerExplorerParameters> connectionPicker;

    public WorkspaceConfigurationPanel() {
        this.initComponents();
    }

    private void initComponents() {
        this.connectionPicker = DALSwingLocator.getSwingManager().createJDBCConnectionPickerController(
                this.cboConnections,
                this.btnAddConnection
        );
        this.connectionPicker.addChangeListener(new ChangeListener() {
            @Override
            public void stateChanged(ChangeEvent e) {
                doUpdateEnableComponents();
                doUpdateStoresRepositoryInfo();
            }
        });
        this.resourcesTableController = new TableController(
                DatabaseWorkspaceManager.TABLE_RESOURCES, 
                btnCreateResourceTable, 
                btnDropResourceTable, 
                btnSearchResourceTable, 
                btnShowResourceTable
        );
        this.repositoryTableController = new TableController(
                DatabaseWorkspaceManager.TABLE_REPOSITORY, 
                btnCreateRepositoryTable, 
                btnDropRepositoryTable, 
                btnSearchRepositoryTable, 
                btnShowRepositoryTable
        );
        this.configurationTableController = new TableController(
                DatabaseWorkspaceManager.TABLE_CONFIGURATION, 
                btnCreateConfigurationTable, 
                btnDropConfigurationTable, 
                btnSearchConfigurationTable, 
                btnShowConfigurationTable
        );

        this.btnRepositoryAddTable.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                doAddTableToRepository();
            }
        });
        
        this.btnConnectToRepository.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                doConnectToRepository();
            }
        });
        
        doUpdateEnableComponents();
    }

    @Override
    public JComponent asJComponent() {
        return this;
    }

    public JDBCServerExplorerParameters getConnection() {
        JDBCServerExplorerParameters conn = this.connectionPicker.get();
        return conn;
    }

    private void doUpdateEnableComponents() {
        boolean theEnabled = this.getConnection() != null;
        this.repositoryTableController.setEnabled(theEnabled);
        this.resourcesTableController.setEnabled(theEnabled);
        this.configurationTableController.setEnabled(theEnabled);
        this.btnRepositoryAddTable.setEnabled(theEnabled);
        this.btnConnectToRepository.setEnabled(theEnabled);
    }
    
    private void doAddTableToRepository() {
        final RepositoryAddTablePanel panel = new RepositoryAddTablePanel();
        WindowManager_v2 winManager = (WindowManager_v2) ToolsSwingLocator.getWindowManager();
        final Dialog dialog = winManager.createDialog(
                panel,
                "Add table to repository",
                null, 
                WindowManager_v2.BUTTONS_OK_CANCEL
        );
        dialog.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                if( dialog.getAction()==WindowManager_v2.BUTTON_OK ) {
                    DataManager dataManager = DALLocator.getDataManager();
                    DatabaseWorkspaceManager workspace = dataManager.createDatabaseWorkspaceManager(getConnection());
                    String tableName = panel.getTableName();
                    DataStoreParameters parameters = panel.getDataStoreParameters();
                    if( !workspace.writeStoresRepositoryEntry(tableName, parameters) ) {
                        // TODO: mensaje al usuario,no se ha podido crear
                    }
                }
            }
        });
        dialog.show(WindowManager.MODE.WINDOW);
    }

    private void doUpdateStoresRepositoryInfo() {
        DataManager dataManager = DALLocator.getDataManager();
        DatabaseWorkspaceManager workspace = dataManager.createDatabaseWorkspaceManager(getConnection());
        String id = workspace.get(CONFIG_NAME_STORESREPOSITORYID);
        String label = workspace.get(CONFIG_NAME_STORESREPOSITORYLABEL);
        this.txtStoresRepositoryId.setText(StringUtils.defaultIfBlank(id, ""));
        this.txtStoresRepositoryLabel.setText(StringUtils.defaultIfBlank(label, ""));
        
        this.btnConnectToRepository.setEnabled(!StringUtils.isBlank(id));
    }
    
    private void doConnectToRepository() {
        DataManager dataManager = DALLocator.getDataManager();
        DatabaseWorkspaceManager workspace = dataManager.createDatabaseWorkspaceManager(getConnection());
        StoresRepository repo = workspace.getStoresRepository();
        dataManager.getStoresRepository().addRepository(repo);
    }
}
