/**
 * gvSIG. Desktop Geographic Information System.
 *
 * Copyright (C) 2007-2013 gvSIG Association.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 *
 * For any additional information, do not hesitate to contact us
 * at info AT gvsig.com, or visit our website www.gvsig.com.
 */
package org.gvsig.labeling.placements;

import java.util.ArrayList;

import org.gvsig.fmap.geom.Geometry;
import org.gvsig.fmap.geom.Geometry.TYPES;
import org.gvsig.fmap.geom.type.GeometryType;
import org.gvsig.fmap.mapcontext.ViewPort;
import org.gvsig.fmap.mapcontext.rendering.legend.styling.ILabelClass;
import org.gvsig.fmap.mapcontext.rendering.legend.styling.IPlacementConstraints;
import org.gvsig.symbology.fmap.mapcontext.rendering.legend.styling.LabelLocationMetrics;
import org.gvsig.tools.task.Cancellable;

public class MultiShapePlacement implements ILabelPlacement {
	private final ILabelPlacement pointPlacement;
	private final ILabelPlacement linePlacement;
	private final ILabelPlacement polygonPlacement;


	/**
	 * Creates a new instance of MultiShapePlacement initializing the respective
	 * placements to those passed as parameters. Null values are allowed for
	 * the parameters and will cause that no label will be placed when the
	 * geometry belongs to such null values.
	 *
	 * @param pointPlacement, the placement for points
	 * @param linePlacement, the placement for lines
	 * @param polygonPlacement, the placement for polygons
	 */
	public MultiShapePlacement(
			ILabelPlacement pointPlacement,
			ILabelPlacement linePlacement,
			ILabelPlacement polygonPlacement) {
		this.pointPlacement = pointPlacement;
		this.linePlacement = linePlacement;
		this.polygonPlacement = polygonPlacement;
	}


    @Override
	public ArrayList<LabelLocationMetrics> guess(
			ILabelClass lc,
			Geometry geom,
			IPlacementConstraints placementConstraints,
			double cartographicSymbolSize,
			Cancellable cancel, ViewPort vp) {
		
		MultiShapePlacementConstraints pc =
				(MultiShapePlacementConstraints) placementConstraints;

		GeometryType gt = geom.getGeometryType();
        if( gt.isTypeOf(TYPES.POINT) || gt.isTypeOf(TYPES.MULTIPOINT) ) {
            if( pointPlacement != null ) {
                return pointPlacement.guess(lc, geom, pc.getPointConstraints(), cartographicSymbolSize, cancel, vp);
            }
        } else if( gt.isTypeOf(TYPES.CURVE) || gt.isTypeOf(TYPES.MULTICURVE) ) {
            if( linePlacement != null ) {
                return linePlacement.guess(lc, geom, pc.getLineConstraints(), cartographicSymbolSize, cancel, vp);
            }

        } else if( gt.isTypeOf(TYPES.SURFACE) || gt.isTypeOf(TYPES.MULTISURFACE) ) {
            if( polygonPlacement != null ) {
                return polygonPlacement.guess(lc, geom, pc.getPolygonConstraints(), cartographicSymbolSize, cancel, vp);
            }

        }
        return CannotPlaceLabel.NO_PLACES;
	}

    @Override
	public boolean isSuitableFor(
			IPlacementConstraints placementConstraints,
			int shapeType) {
		// TODO shoud I also ask to each placement if it is suitable
		// for these constraints????
		return (shapeType == TYPES.GEOMETRY || shapeType == TYPES.AGGREGATE);
	}




}
