/* gvSIG. Geographic Information System of the Valencian Government
 *
 * Copyright (C) 2007-2008 Infrastructures and Transports Department
 * of the Valencian Government (CIT)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 *
 */
package org.gvsig.timesupport.app.animation.extension;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Iterator;
import javax.swing.SwingUtilities;
import org.gvsig.andami.plugins.Extension;
import org.gvsig.app.ApplicationLocator;
import org.gvsig.app.ApplicationManager;
import org.gvsig.app.project.ProjectManager;
import org.gvsig.app.project.documents.view.ViewDocument;
import org.gvsig.app.project.documents.view.ViewManager;
import org.gvsig.app.project.documents.view.gui.AbstractViewPanel;
import org.gvsig.app.project.documents.view.gui.IView;
import org.gvsig.fmap.dal.DataStore;
import org.gvsig.fmap.mapcontext.layers.FLayer;
import org.gvsig.fmap.mapcontext.layers.FLayers;
import org.gvsig.fmap.mapcontext.layers.operations.SingleLayer;
import org.gvsig.fmap.mapcontrol.MapControl;
import org.gvsig.timesupport.Interval;
import org.gvsig.timesupport.RelativeInterval;
import org.gvsig.timesupport.Time;
import org.gvsig.timesupport.TimeSupportLocator;
import org.gvsig.timesupport.animation.TimeAnimation;
import org.gvsig.timesupport.animation.TimeAnimation.TimeAnimationRenderer;
import org.gvsig.timesupport.swing.api.TimeSupportSwingLocator;
import org.gvsig.timesupport.swing.api.TimeSupportSwingManager;
import org.gvsig.timesupport.swing.api.panel.AnimationBarPanel;
import org.gvsig.timesupport.swing.api.panel.AnimationPreferencesPanel;
import org.gvsig.tools.ToolsLocator;
import org.gvsig.tools.i18n.I18nManager;
import org.gvsig.tools.swing.api.ToolsSwingLocator;
import org.gvsig.tools.swing.api.windowmanager.Dialog;
import org.gvsig.tools.swing.api.windowmanager.WindowManager;
import org.gvsig.tools.swing.api.windowmanager.WindowManager_v2;

public class AnimationSupportExtension extends Extension {

    private TimeSupportSwingManager swingManager = null;

    @Override
    public void initialize() {
        // Do nothing
    }

    @Override
    public void postInitialize() {
        super.postInitialize();
        swingManager = TimeSupportSwingLocator.getSwingManager();
    }

    @Override
    public void execute(String actionCommand) {
        if( actionCommand.equalsIgnoreCase("AnimationSupport") ) {
            final ViewDocument viewDocument = getView();
            if( viewDocument == null ) {
                return;
            }
            ProjectManager projectManager = ApplicationLocator.getProjectManager();
            ViewManager viewManager = (ViewManager) projectManager.getDocumentManager(ViewManager.TYPENAME);
            final IView viewPanel = (IView) viewManager.getMainWindow(viewDocument);
            final MapControl mapControl = viewPanel.getMapControl();
            RelativeInterval interval = (RelativeInterval) viewDocument.getMapContext().getTimeContext().getInterval();
            if( interval != null ) {
                final int savedPaintMode = mapControl.getPaintMode();
                final Time savedTimeFilter = mapControl.getMapContext().getViewPort().getTime();
                I18nManager i18n = ToolsLocator.getI18nManager();
                final TimeAnimation animation = TimeSupportLocator.getManager().
                    createTimeAnimation(new TimeAnimationRenderer() {
                        @Override
                        public void begin() {
                            mapControl.setPaintMode(MapControl.PAINT_WHEN_DRAW_LAYERS_IS_COMPLETED);
                        }

                        @Override
                        public void render(final Interval time) {
                            if( !SwingUtilities.isEventDispatchThread() ) {
                                SwingUtilities.invokeLater(new Runnable() {
                                    @Override
                                    public void run() {
                                        render(time);
                                    }
                                });
                            }
                            viewDocument.getMapContext().getViewPort().setTime(time);
                            viewDocument.getMapContext().invalidate();
                        }

                        @Override
                        public void end() {
                            mapControl.setPaintMode(savedPaintMode);
                            mapControl.getMapContext().getViewPort().setTime(savedTimeFilter);
                        }
                    }
                    );
                animation.setInterval(interval);

                final AnimationPreferencesPanel prefs = swingManager.createAnimationPreferencesPanel(animation);

                WindowManager_v2 winManager = (WindowManager_v2) ToolsSwingLocator.getWindowManager();
                final Dialog dialog = winManager.createDialog(
                    prefs.asJComponent(),
                    i18n.getTranslation("animation_preferences"),
                    null,
                    WindowManager_v2.BUTTONS_OK_CANCEL
                );
                dialog.addActionListener(new ActionListener() {
                    @Override
                    public void actionPerformed(ActionEvent e) {
                        if( dialog.getAction() == WindowManager_v2.BUTTON_OK ) {
                            doAnimation(viewDocument, prefs.getAnimation());
                        }
                        prefs.asJComponent().setVisible(false);
                    }
                });
                dialog.show(WindowManager.MODE.TOOL);
            }
        }
    }

    private void doAnimation(ViewDocument viewDocument, TimeAnimation animation) {
        I18nManager i18n = ToolsLocator.getI18nManager();
        WindowManager_v2 winManager = (WindowManager_v2) ToolsSwingLocator.getWindowManager();
        AnimationBarPanel panel = swingManager.createAnimationBarPanel(animation);
        winManager.showWindow(
            panel.asJComponent(),
            i18n.getTranslation("animation_bar") + " - " + viewDocument.getName(),
            WindowManager.MODE.TOOL
        );
    }

    @Override
    public boolean isEnabled() {
        ViewDocument viewDocument = getView();
        if( viewDocument == null ) {
            return false;
        }
        FLayers layers = viewDocument.getMapContext().getLayers();
        for( Iterator<FLayer> iterator = layers.deepiterator(); iterator.hasNext(); ) {
            FLayer layer = iterator.next();
            if( layer instanceof SingleLayer ) {
                DataStore dataStore = ((SingleLayer) layer).getDataStore();
                if( dataStore.getInterval() != null ) {
                    return true;
                }
            }
        }
        return false;
    }

    /**
     * Gets the active view
     *
     * @return {@link AbstractViewPanel}
     */
    private ViewDocument getView() {
        ApplicationManager application = ApplicationLocator.getManager();
        ViewDocument viewDocument = (ViewDocument) application.getActiveDocument(ViewManager.TYPENAME);
        return viewDocument;
    }

    @Override
    public boolean isVisible() {
        return getView()!=null;
    }

}
